#include "EMVtk.h"
#include <cstring>
#include <cassert>
#include <cstdlib>
#include <cctype>

EMVtk::EMVtk(const char* filename)
    : fname(filename)
{
    in = 0;
    filetype = VTK_UNKNOWN;
    datatype = DATA_UNKNOWN;
    nx = 0;
    ny = 0;
    nz = 0;
    originx = 0;
    originy = 0;
    originz = 0;
    spacingx = 0;
    spacingy = 0;
    spacingz = 0;
}

EMVtk::EMVtk(FILE* infile)
    : in(infile)
{
    fname = 0;
    filetype = VTK_UNKNOWN;
    datatype = DATA_UNKNOWN;
    nx = 0;
    ny = 0;
    nz = 0;
    originx = 0;
    originy = 0;
    originz = 0;
    spacingx = 0;
    spacingy = 0;
    spacingz = 0;
}

EMVtk::~EMVtk()
{
    if (in) {
	fclose(in);
	in = 0;
    }
}

static int is_empty_line(const char* line)
{
    if (line) {
	const char* tmp = line;
	for (; !tmp; tmp++) {
	    if(!isspace(tmp[0])) {
		return 0;
	    }
	}
    }
    return 1;
}

static int samestr(const char* s1, const char* s2)
{
    return (strncmp(s1, s2, strlen(s2)) == 0);
}

	
int EMVtk::init()
{
    if(fname) {
	in = fopen(fname, "rb");
	if (!in) {
	    fprintf(stderr, "Error: cannot open file '%s'\n", fname);
	    return 1;
	}
    }
    
    char buf[1024];
    int bufsz = sizeof(buf);
    const char* magic = "# vtk DataFile Version";
    
    if (fgets(buf, bufsz, in)) {
	if (strncmp(buf, magic, strlen(magic)) != 0) {
	    return 1;
	}
    }
    else {
	return 1;
    }

    while(fgets(buf, bufsz, in)) {
	if (samestr(buf, "LOOKUP_TABLE default")) {
	    break;
	}
	else if (samestr(buf, "ASCII")) {
	    filetype = VTK_ASCII;
	}
	else if (samestr(buf, "BINARY")) {
	    filetype = VTK_BINARY;
	}
	else if (samestr(buf, "DIMENSIONS")) {
	    sscanf(buf, "DIMENSIONS %d %d %d", &nx, &ny, &nz);
	}
	else if (samestr(buf, "ORIGIN")) {
	    sscanf(buf, "ORIGIN %f %f %f", &originx, &originy, &originz);
	}
	else if (samestr(buf, "SPACING")) {
	    sscanf(buf, "SPACING %f %f %f", &spacingx, &spacingy, &spacingz);
	    if (spacingx != spacingy || spacingx != spacingz || spacingy != spacingz) {
		fprintf(stderr, "Error: cannot handle non-uniform spacing VTK file\n");
		return 1;
	    }
	}
	else if (samestr(buf, "SCALARS")) {
	    char datatypestr[32];
	    char scalartype[32];
	    sscanf(buf, "SCALARS %s %s", &scalartype, &datatypestr);

	    if (samestr(datatypestr, "unsigned_short")) {
		datatype = UNSIGNED_SHORT;
	    }
	    else if (samestr(datatypestr, "float")) {
		datatype = FLOAT;
	    }
	    else {
		fprintf(stderr, "Error: unknown data type: %s", datatypestr);
		return 1;
	    }
	}
	
    }
#ifdef DEBUG
    printf("vtk nx = %d, ny = %d, nz = %d, type = %d\n", nx, ny, nz, datatype);
#endif
    return 0;
}


int EMVtk::read_data(float** p_data)
{
    if (!p_data || !(*p_data)) {
	fprintf(stderr, "Error: null pointer in read_data()\n");
	return 1;
    }

    if (filetype == VTK_ASCII) {
	int bufsz = nx*16;
	char* buf = new char[bufsz];
	int i = 0;
    
	while(fgets(buf, bufsz, in)) {
	    int bufslen = strlen(buf) - 1;
	    char numstr[32];
	    int k = 0;
	    for (int j = 0; j < bufslen; j++) {
		if (!isspace(buf[j])) {
		    numstr[k++] = buf[j];
		}
		else {
		    numstr[k] = '\0';
		    (*p_data)[i++] = atoi(numstr);
		    k = 0;
		}
	    }
	}
	delete [] buf;
	buf = 0;
    }
    else if (filetype == VTK_BINARY) {
	fprintf(stderr, "Error: binary VTK is not supported\n");
	return 1;
    }
    
    return 0;
}

