#include <string.h>

#include "PyEMData.h"
#include "PyEuler.h"
#include "util.h"
#include <boost/python/detail/api_placeholder.hpp>
#include <sys/types.h>
#ifdef WIN32
	#include <UnixToWin.h>
#else
	#include <unistd.h>
#endif
#include <stdlib.h>
#include <float.h>

static void pylist2emlist(const PyList& pylist, List* emlist)
{
    if (!emlist) {
	return;
    }
    
    
    for (int i = 0; i < python::len(pylist); i++) {
	PyEMData* item = python::extract<PyEMData*>(pylist[i]);
	emlist->addObject(item);
    }
}
 
static ImageType get_imgtype(const char* filename)
{
    const char* exts[] = {".mrc", ".tnf", ".spi", ".dm2", ".img", ".sal",
			  ".pgm", ".map", ".emim", ".pif", ".png", ".vtk",
			  ".dm3", ".tif", ".tiff", ".ssw", ".hed", ".hdf", ".df3", ".xplor"};
    
    ImageType types[] = {IMAGE_MRC, IMAGE_MRC, IMAGE_SPIDER, IMAGE_GATAN,
			 IMAGE_IMAGIC, IMAGE_SAL, IMAGE_PGM, IMAGE_ICOS, 
			 IMAGE_EMIM, IMAGE_PIF, IMAGE_PNG, IMAGE_VTK, IMAGE_DM3, 
			 IMAGE_TIF, IMAGE_TIF, IMAGE_SPIDER_SWAP, IMAGE_IMAGIC, IMAGE_HDF, IMAGE_DF3, IMAGE_XPLOR};

    int flen = strlen(filename);
    int num_ext = sizeof(types) / sizeof(ImageType);
    
    for (int i = 0; i < num_ext; i++) {
	int elen = strlen(exts[i]);
	if (flen > elen && strcasecmp(&filename[flen - elen], exts[i]) == 0) {
	    return types[i];
	}
    }
    
    return IMAGE_ANY;
}

PyEMData::PyEMData() : EMData()
{

}


PyEMData::~PyEMData()
{   
}

python::tuple PyEMData::get_loc(int ismax)
{
    int flags = getFlags();
    int di = 1;
    
    if ((flags&EMDATA_COMPLEX)&& !(flags&EMDATA_RI)) {
	di=2;
    }
    
    float* mydata = getDataRO();
    int nz = zSize();
    int ny = ySize();
    int nx = xSize();
    int minj = 0;
    int mink = 0;
    int minl = 0;
    int maxj = 0;
    int maxk = 0;
    int maxl = 0;
    float max = -FLT_MAX;
    float min = FLT_MAX;
    
    for (int j = 0; j < nz; j++) {
	for (int k = 0; k < ny; k++) {
	    for (int l = 0; l < nx; l += di) {
		float v = mydata[j*ny*nx + k*nx + l];
		
		if (v > max) {
		    max = v;
		    minj = j;
		    mink = k;
		    minl = l;
		}

		if (v < min) {
		    min = v;
		    maxj = j;
		    maxk = k;
		    maxl = l;
		}
	    }
	}
    }
    doneData();

    if (ismax) {
	return python::make_tuple(maxl, maxk, maxj);
    }
    
    return python::make_tuple(minl, mink, minj);
}

python::tuple PyEMData::py_MinLoc()
{
    return get_loc(0);
}

python::tuple PyEMData::py_MaxLoc()
{
    return get_loc(1);
}

int PyEMData::display()
{
    char fname[32];
    sprintf(fname, "XXXXXX");
    mkstemp(fname);
    unlink(fname);
    sprintf(fname, "%s.mrc", fname);
    writeMRC(fname);
    char cmd[64];
    sprintf(cmd, "v2 %s", fname);
    int err = system(cmd);
    unlink(fname);
    return err;
}

PyArray PyEMData::get_num_array() 
{
    float* d = getDataRO();
    int nx = xSize();
    int ny = ySize();
    int nz = zSize();
    int n = nx * ny * nz;

    PyList datalist = PyList();
    
    for (int i = 0; i < nz; i++) {
	PyList ll = PyList();
	for (int j = 0; j < ny; j++) {
	    PyList l = PyList();
	    array2list(&d[i* nx * ny + j * nx], l, nx);
	    ll.append(l);
	}
	datalist.append(ll);
    }
    doneData();
    
    return PyArray(datalist);
}


void PyEMData::set_num_array(PyArray& array)
{
    int nz = python::len(array);
    int ny = python::len(array[0]);
    int nx = python::len(array[0][0]);

    update();
    setSize(nx, ny, nz);
    
    float* rdata = getData();
    
    for (int i = 0; i < nz; i++) {
	for (int j = 0; j < ny; j++) {
	    PyArray array2 = python::extract<PyArray>(array[i][j]);
	    PyList l = PyList(array2);
	    list2array(l, &rdata[i * nx * ny + j * nx]);
	}
    }
    doneData();
}


// return 0 if OK; 
int PyEMData::readImage(const char* filename, int n, int nodata, ImageType img_type)
{
    if (!filename) {
	printf("Error: no file is input\n");
	return -1;
    }

    switch(img_type) {
    case IMAGE_ANY:
	return EMData::readImage(filename, n, nodata);
    case IMAGE_SAL:
	return readSAL(filename, nodata);
    case IMAGE_PGM:
	return readPGM(filename, nodata);
    case IMAGE_ICOS:
	return readICOS(filename, nodata);
    case IMAGE_MRC:
	return readMRC(filename, nodata, n);
    case IMAGE_IMAGIC:
	return readIMAGIC(filename, n, nodata);
    case IMAGE_IMAGIC3D:
	return readIMAGIC3d(filename, nodata);
    case IMAGE_EMIM:
	return readEMIM(filename, n, nodata);
    case IMAGE_SPIDER:
	return readSPIDER(filename, n, nodata);
    case IMAGE_GATAN:
	return readGatan(filename, nodata);
//    case IMAGE_GATAN3:
//	return readGatan3(filename, nodata);
    case IMAGE_PIF:
	return readPIF(filename, n, nodata);
    case IMAGE_PNG:
	return readPNG(filename);
    case IMAGE_VTK:
	return readVTK(filename, nodata);
    case IMAGE_DM3:
	return readDM3(filename, nodata);
    case IMAGE_TIF:
	return readTIF(filename, nodata);
    default:
	printf("Error: invalid file type '%d' for file '%s'\n",
	       img_type, filename);
	return -1;
    }
    
    return 0;
}


// return 0 if OK; 
int PyEMData::writeImage(const char* filename, int n, ImageType img_type,
			 int nolock, int swap, int mode)
{    
    if (!filename) {
	printf("Error: no file is input\n");
	return -1;
    }

    if (img_type == IMAGE_ANY) {
	img_type = get_imgtype(filename);
    }
    
    switch(img_type) {
    case IMAGE_ANY:
	printf("Error: please specify the image type\n");
	return -1;    
    case IMAGE_IMAGIC3D:
	return writeIMAGIC3D(filename, nolock);
    case IMAGE_IMAGIC:
	return writeIMAGIC(filename, n, nolock);
    case IMAGE_ICOS:
	return writeICOS(filename);
    case IMAGE_MRC:
	return writeMRC(filename, mode, n);
    case IMAGE_SPIDER:
	return writeSPIDER(filename, n, swap);
    case IMAGE_SINGLE_SPIDER:
	return writeSingleSPIDER(filename, swap);
    case IMAGE_EMIM:
	return writeEMIM(filename, n);
    case IMAGE_VTK:
	return writeVTK(filename);
    case IMAGE_AMIRA:
	return writeAMIRA(filename);
    case IMAGE_PIF:
	return writePIF(filename);
	case IMAGE_TIF:
	return writeTIF(filename);
    case IMAGE_PNG:
	return writePNG(filename);
	case IMAGE_PGM:
	return writePGM(filename);
    case IMAGE_HDF:
	return writeHDF2(filename, n, nolock);
	case IMAGE_DF3:
	return writeDF3(filename);
    case IMAGE_XPLOR:
	return writeXPLOR(filename);
    case IMAGE_SPIDER_SWAP:
	return writeSPIDER(filename, n, 1);
    default:
	printf("Error: invalid file type '%d' for file '%s'\n",
	       img_type, filename);
	return -1;
    }
    
    return 0;
}

    
void PyEMData::py_calcHist(PyList hist_list, int n, float hmin,
			   float hmax, int add)
{
    float* hist_array = new float[n];
    calcHist(hist_array, n, hmin, hmax, add);
    array2list(hist_array, hist_list, n);
    delete [] hist_array;
    hist_array = 0;
}


python::tuple PyEMData::py_transAlign(EMData* with, int useparent,
				      int intonly,int maxshift)
{
    float peak = 0;
    float align = transAlign(with, useparent, intonly, maxshift, &peak);
    return python::make_tuple(align, peak);
}


void PyEMData::py_calcRCF(EMData* with, PyList sum_list, int NS)
{
    float* sum_array = new float[NS];  
    calcRCF(with, sum_array, NS);
    array2list(sum_array, sum_list, NS);
    delete [] sum_array;
    sum_array = 0;
}

void PyEMData::py_applyRadFn(int n, float x0, float dx, PyList y_list, int interp)
{
    float* y_array = new float[n];
    list2array(y_list, y_array);
    applyRadFn(n, x0, dx, y_array, interp);
    delete [] y_array;
    y_array = 0;
}

void PyEMData::py_addRadNoise(int n, float x0, float dx, PyList y_list, int interp)	   
{
    float* y_array = new float[n];
    list2array(y_list, y_array);
    addRadNoise(n, x0, dx, y_array, interp);
    delete [] y_array;
    y_array = 0;
}

python::tuple PyEMData::py_normalizeTo(EMData* noisy, int keepzero, int invert)
{
    float mult = 0;
    float add = 0;
    normalizeTo(noisy, keepzero, invert, &mult, &add);
    return python::make_tuple(mult, add);
}

python::tuple PyEMData::py_leastSquareNormalizeTo(EMData* to, float low_thresh,
						  float high_thresh)
{
    float scale = 0;
    float shift = 0;
    leastSquareNormalizeTo(to, low_thresh, high_thresh, &scale, &shift);
    return python::make_tuple(scale, shift);
}


void PyEMData::py_calcRadDist(int n, float x0, float dx, PyList y_list)
{
    float* y_array = new float[n];
    calcRadDist(n, x0, dx, y_array);
    array2list(y_array, y_list, n);
    delete [] y_array;
    y_array = 0;
}

void PyEMData::py_calcRadDist(int n, float x0, float dx, PyList y_list,
			      float acen, float amwid)
{
    float* y_array = new float[n];
    calcRadDist(n, x0, dx, y_array, acen, amwid);
    array2list(y_array, y_list, n);
    delete [] y_array;
    y_array = 0;
}

void PyEMData::py_calcAzDist(int n, float a0, float da, PyList d_list,
			     float rmin, float rmax)
{
    float* d_array = new float[n];
    calcAzDist(n, a0, da, d_array, rmin, rmax);
    array2list(d_array, d_list, n);
    delete [] d_array;
    d_array = 0;
}

python::tuple PyEMData::py_get_bounding_box(float thresh)
{
    int xmin, ymin, zmin, xmax, ymax, zmax;
    get_bounding_box(thresh, &xmin, &ymin, &zmin, &xmax, &ymax, &zmax);
    return python::make_tuple(xmin, ymin, zmin, xmax, ymax, zmax);
}

python::tuple PyEMData::py_lcmp(EMData* noisy, int keepzero, int invert)
{
    float scale = 0;
    float shift = 0;
    float result = lcmp(noisy, keepzero, invert, &scale, &shift);
    return python::make_tuple(result, scale, shift);
}

void PyEMData::py_makeAverageWT(PyList in_pylist, EMData* curves, XYData* sf)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeAverageWT(in_emlist, curves, sf);

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_makeAverageCTFCW(PyList in_pylist, XYData* sf, PyList SNR_list)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);
    
    int sz = python::len(SNR_list);
    float* SNR_array = new float[sz];
    list2array(SNR_list, SNR_array);

    makeAverageCTFCW(in_emlist, sf, SNR_array);

    array2list(SNR_array, SNR_list);
    delete [] SNR_array;
    SNR_array = 0;

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_makeAverage(PyList in_pylist, EMData* sigma, int ignore0)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeAverage(in_emlist, sigma, ignore0);

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_makeAverageIter(PyList in_pylist)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);
    makeAverageIter(in_emlist);

    delete in_emlist;
    in_emlist = 0;
}


void PyEMData::py_makeAverageCTFC(PyList in_pylist, float filter)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeAverageCTFC(in_emlist, filter);

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_makeAverageCTFCWauto(PyList in_pylist, const char *outfile)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeAverageCTFCWauto(in_emlist, const_cast<char*>(outfile));
    
    delete in_emlist;
    in_emlist = 0;
}


void PyEMData::py_makeMedian(PyList in_pylist)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeMedian(in_emlist);

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_interpFT3d(float x, float y, float z, PyList ret_list, int mode)
{
    // only 2 elements are used in ret_list
    const int nitems = 2;
    float* ret_array =  new float[nitems];
    interpFT3d(x, y, z, ret_array, mode);
    array2list(ret_array, ret_list, nitems);
    delete [] ret_array;
    ret_array = 0;
}


python::tuple PyEMData::py_normSlice(EMData* slice, float alt, float az, float phi)
{
    float phaseresid = 0;
    float result = normSlice(slice, alt, az, phi, &phaseresid);
    return python::make_tuple(result, phaseresid);
}


PyList PyEMData::py_MaxLocs(int n)
{
    int* sf_array = MaxLocs(n);
    PyList sf_list = PyList();
    array2list(sf_array, sf_list, n);
    return sf_list;
}


PyList PyEMData::py_ctfCurve(int type, XYData* sf)
{
    float* sf_array = ctfCurve(type, sf);
    int array_size = ceil(CTFOS * sqrt(2.0) * ySize() / 2) + 2;
    PyList sf_list = PyList();
    array2list(sf_array, sf_list, array_size);
    return sf_list;
}
#if 1

PyList PyEMData::py_fsc(EMData* with)
{
    int array_size = ySize() / 2;
    float* array1 = fsc(with);
    PyList list1 = PyList();
    array2list(array1, list1, array_size);
    return list1;    
}
#endif

PyList PyEMData::py_getCTF()
{
    const int SZ = 11;
    float* ctf_array = getCTF();
    PyList ctf_list = PyList();
    array2list(ctf_array, ctf_list, SZ);
    return ctf_list;
}

void PyEMData::py_setCTF(PyList ctf_list)
{
    const int SZ = 11;
#ifdef DEBUG
    assert(python::len(ctf_list) == SZ);
#endif
    float* ctf_array = new float[SZ];
    list2array(ctf_list, ctf_array);
    setCTF(ctf_array);
    delete [] ctf_array;
    ctf_array = 0;
}

float PyEMData::py_fscmp(EMData* with, PyList snr)
{
	float score = 0;
	if(python::len(snr)) {
	    int array_size = ceil(CTFOS * sqrt(2.0) * ySize() / 2) + 2;
		#ifdef DEBUG
		assert(python::len(snr) == array_size);
		#endif
		float *snr_array = new float[array_size];
		list2array(snr, snr_array);
		score = fscmp(with, snr_array);
		delete [] snr_array;
		snr_array = 0;
	}
	else {
		score = fscmp(with);
	}
	return score;
}


void PyEMData::py_setSFFile(const char* filename){
	if (!filename){
		fprintf(stderr,"EMData::setSFFile(const char* filename): blank string for \"filename\"\n");
		return;
	}
	mrcH *mrch= getMRCHed();
	if(mrch){
		mrch->nlabl=2;
		mrch->labels[1][79]=0;
		sprintf(mrch->labels[1],"%s",filename);
	}
}
		
PyEMData* PyEMData::py_doFFT()
{
    EMData* e = doFFT();
    gimmeFFT();
    return static_cast<PyEMData*>(e);
}
	

PyEMData& PyEMData::operator+=(float n)
{
    addConst(n);
    return *this;
}

PyEMData& PyEMData::operator-=(float n)
{
    addConst(-n);
    return *this;
}

PyEMData& PyEMData::operator*=(float n)
{
    multConst(n);
    return *this;
}

PyEMData& PyEMData::operator/=(float n)
{
    if (n != 0) {
	multConst(1.0/n);
    }
    else {
	printf("Error: divided by 0\n");
    }
    return *this;
}

void PyEMData::apply_obj(const PyEMData& other_obj, EMOper op)
{
    PyEMData* other = const_cast<PyEMData*>(&other_obj);

    if (other->xSize() != xSize() ||
	other->ySize() != ySize() ||
	other->zSize() != zSize()) {
	printf("Error: the two EMData objects have different sizes\n");
    }
    else {
	if (isComplex()) {
	    ap2ri();
	}
	
	if (other->isComplex()) {
	    other->ap2ri();
	}
	
	float* mydata = getData();
	float* otherdata = 0;
	
	if (other != this) {
	    otherdata = other->getDataRO();
	}
	else {
	    otherdata = mydata;
	}
	
	int size = xSize() * ySize() * zSize();

	for (int i = 0; i < size; i++) {
	    if (op == EM_ADD) {
		mydata[i] += otherdata[i];
	    }
	    else if (op == EM_SUB) {
		mydata[i] -= otherdata[i];
	    }
	    else if (op == EM_MUL) {
		mydata[i] *= otherdata[i];
	    }
	    else if (op == EM_DIV) {
		if (otherdata[i] != 0) {
		    mydata[i] /= otherdata[i];
		}
		else {
		    mydata[i] = 0;
		}
	    }
	}
	if (other != this) {
	    other->doneData();
	}
	doneData();
    }
}

PyEMData* PyEMData::py_matchFilter(EMData* to) {
    return static_cast<PyEMData*>(matchFilter(to));
}


PyEMData* PyEMData::my_copy() const
{
    PyEMData* e = const_cast<PyEMData*>(this);
    return static_cast<PyEMData*>(e->copy());
}

PyEMData* PyEMData::py_copy(int withfft, int withparent)
{
    return static_cast<PyEMData*>(copy(withfft, withparent));
}


PyEMData* PyEMData::py_copyHead()
{
    return static_cast<PyEMData*>(copyHead());
}

    
PyEMData* PyEMData::py_clip(int x0, int y0, int w, int h)
{
    return static_cast<PyEMData*>(clip( x0,  y0,  w,  h));
}

PyEMData* PyEMData::py_clip(int x0, int y0, int z0, int w, int h, int d)
{
    return static_cast<PyEMData*>(clip( x0, y0, z0, w, h, d));
}
PyEMData* PyEMData::py_Parent()
{
    return static_cast<PyEMData*>(Parent());
}

PyEuler* PyEMData::py_getEuler() 
{
    return static_cast<PyEuler*>(getEuler());
}


PyEMData* PyEMData::py_doIFT()
{
    EMData* e = doIFT();
    gimmeFFT();
    return static_cast<PyEMData*>(e);
}   

PyEMData* PyEMData::py_doRadon()
{
    return static_cast<PyEMData*>(doRadon());
}

PyEMData* PyEMData::py_vertACF(int maxdy)
{
    return static_cast<PyEMData*>(vertACF(maxdy));
}


PyEMData* PyEMData::py_convolute(EMData* with)
{
    return static_cast<PyEMData*>(convolute(with));
}

PyEMData* PyEMData::py_calcCCF(EMData* with, int tocorner, EMData* filter)
{
    return static_cast<PyEMData*>(calcCCF(with, tocorner, filter));
}

PyEMData* PyEMData::py_calcFLCF(EMData* with, int r)
{
    return static_cast<PyEMData*>(calcFLCF(with, r, 4));
}

PyEMData* PyEMData::py_calcCCFX(EMData* with, int y0, int y1, int nosum)
{
    return static_cast<PyEMData*>(calcCCFX(with, y0, y1, nosum));
}

PyEMData* PyEMData::py_calcMCF(EMData* with, int tocorner, EMData* filter)
{
    return static_cast<PyEMData*>(calcMCF(with, tocorner, filter));
}


PyEMData* PyEMData::py_littleBigDot(EMData* with, int sigma)
{
    return static_cast<PyEMData*>(littleBigDot(with, sigma));
}

    
PyEMData* PyEMData::py_RTFAlignRadon(EMData* with, int maxshift,
				     EMData* thisf, EMData* radonwith,
				     EMData* radonthis, EMData* radonthisf)
{
    return static_cast<PyEMData*>(RTFAlignRadon(with, maxshift,
						thisf, radonwith,
						radonthis, radonthisf));
}


PyEMData* PyEMData::py_RTAlignRadon(EMData* with, int maxshift,
				    EMData* radonwith, EMData* radonthis)
{
    return static_cast<PyEMData*>(RTAlignRadon(with, maxshift, radonwith, radonthis));
}


PyEMData* PyEMData::py_RTFAlignSlow(EMData* with,EMData* flip, int maxshift)
{
    return static_cast<PyEMData*>(RTFAlignSlow(with, flip, maxshift));
}


PyEMData* PyEMData::py_RTFAlignSlowest(EMData* with, EMData* flip, int maxshift)
{
    return static_cast<PyEMData*>(RTFAlignSlowest(with, flip, maxshift));
}

PyEMData* PyEMData::py_RTFAlignBest(EMData* with, EMData* flip,
				    int maxshift, float *snr)
{
    return static_cast<PyEMData*>(RTFAlignBest(with, flip, maxshift, snr));
}

PyEMData* PyEMData::py_RTFAlign(EMData* with, EMData* flip,
				int usedot, int maxshift)
{
    return static_cast<PyEMData*>(RTFAlign(with, flip, usedot, maxshift));
}


PyEMData* PyEMData::py_RFAlign(EMData* with,
			       EMData* flip,
			       int imask)
{
    return static_cast<PyEMData*>(RFAlign(with, flip, imask));
}

PyEMData* PyEMData::py_RTAlignBest(EMData* with, int maxshift, float *snr)
{
    return static_cast<PyEMData*>(RTAlignBest(with, maxshift, snr));
}


    
PyEMData* PyEMData::py_RTAlign(EMData* with, int usedot, int maxshift)
{
    return static_cast<PyEMData*>(RTAlign(with, usedot, maxshift));
}


PyEMData* PyEMData::py_makeRFP(int premasked, int unwrap)
{
    return static_cast<PyEMData*>(makeRFP(premasked, unwrap));
}

PyEMData* PyEMData::py_unwrap(int r1, int r2, int xs,
			      int dx, int dy, int do360)
{
    return static_cast<PyEMData*>(unwrap(r1, r2, xs, dx, dy, do360));
}

    
				
PyEMData* PyEMData::py_project3d(float alt, float az,  float phi, 
				 int mode,  float thr, float ri)
{
    return static_cast<PyEMData*>(project3d(alt, az, phi, mode, thr, ri));
}

PyEMData* PyEMData::py_fftSlice(float alt, float az, float phi, int mode)
{
    return static_cast<PyEMData*>(fftSlice(alt, az, phi, mode));
}


PyEMData* PyEMData::py_iftSlice()
{
    return static_cast<PyEMData*>(iftSlice());
}

PyEMData* PyEMData::py_row(int n)
{
    return static_cast<PyEMData*>(row(n));
}

PyEMData* PyEMData::py_col(int n)
{
    return static_cast<PyEMData*>(col(n));
}

PyEMData* PyEMData::py_mxMult(EMData* second)
{
    return static_cast<PyEMData*>(mxMult(second));
}

    
PyEMData& PyEMData::operator+=(const PyEMData& other)
{
    apply_obj(other, EM_ADD);
    return *this;
}

PyEMData& PyEMData::operator-=(const PyEMData& other)
{
    apply_obj(other, EM_SUB);
    return *this;
}

PyEMData& PyEMData::operator*=(const PyEMData& other)
{
    apply_obj(other, EM_MUL);
    return *this;
}

PyEMData& PyEMData::operator/=(const PyEMData& other)
{
    apply_obj(other, EM_DIV);
    return *this;
}


PyEMData& operator+(const PyEMData& em, float n)
{
    PyEMData* em1 = em.my_copy();
    return (*em1) += n;
}

PyEMData& operator-(const PyEMData& em, float n)
{
    PyEMData* em1 = em.my_copy();
    return (*em1) -= n;
}

PyEMData& operator*(const PyEMData& em, float n)
{
    PyEMData* em1 = em.my_copy();
    return (*em1) *= n;
}

PyEMData& operator/(const PyEMData& em, float n)
{
    PyEMData* em1 = em.my_copy();
    return (*em1) /= n;
}

    
PyEMData& operator+(float n, const PyEMData& em)
{
    return operator+(em, n);
}

PyEMData& operator-(float n, const PyEMData& em)
{
    return operator-(em, n);
}

PyEMData& operator*(float n, const PyEMData& em)
{
    return operator*(em, n);
}

PyEMData& operator/(float n, const PyEMData& em)
{
    return operator/(em, n);
}

   
PyEMData& operator+(const PyEMData& a, const PyEMData& b)
{
    PyEMData* em = a.my_copy();
    return (*em) += b;
}

PyEMData& operator-(const PyEMData& a, const PyEMData& b)
{
    PyEMData* em = a.my_copy();
    return (*em) -= b;
}

PyEMData& operator*(const PyEMData& a, const PyEMData& b)
{
    PyEMData* em = a.my_copy();
    return (*em) *= b;
}

PyEMData& operator/(const PyEMData& a, const PyEMData& b)
{
    PyEMData* em = a.my_copy();
    return (*em) /= b;
}

python::tuple fileCount(const char* filespec)
{
    int type = 0;
    int nimg = fileCount(filespec, &type);
    ImageType imgtype = static_cast<ImageType>(type);
    const char* imgname = get_imagetype_name(imgtype);
    return python::make_tuple(nimg, imgname);
}

PyList py_readImages(const char *filespec, int n0, int n1, int nodata)
{
    List* images = readImages(filespec, n0, n1, nodata);
    int nimg = images->count();
    PyList images_list = PyList();

    for (int i = 0; i < nimg; i++) {
	images_list.append(static_cast<PyEMData*>(images->objectAt(i)));
    }

    return images_list;
}

PyList py_readImagesExt(const char *filespec, const char *ext,
			int n0, int n1, int nodata)
{
    List* images = readImagesExt(filespec, ext, n0, n1, nodata);
    int nimg = images->count();
    PyList images_list = PyList();

    for (int i = 0; i < nimg; i++) {
	images_list.append(static_cast<PyEMData*>(images->objectAt(i)));
    }

    return images_list;
}


