// cenalignint.C 1/98 - Steve Ludtke
// Copyright 1998 Baylor College of Medicine
// This iteratively centers a set of particles, using only integer shifts
// to prevent 

//N cenalignint
//F This iteratively centers a set of particles, using only integer shifts to avoid interpolation artifacts.
//T LA
//1
//P <input file>	File containing images to center
//P [frac=<0 to d-1>/<d>]	Processes a fraction of the input file.
//P [mask=<rad>]	Applies a mask to the centered images
//P [maxshift=<pixels>]	Maximum shift in pixels
//P [usefilt]	Use .filt file (filtered particles) for centering, but write original images in output
//U cenalignint init.img frac=0,4
//U cenalignint init.img frac=1,4
//U cenalignint init.img frac=2,4
//U cenalignint init.img frac=3,4
//D Cenalignint is used to take a set of boxed particles and move each
//E particle to the center of its box (+- 1 pixel). This is accomplished by
//E iteratively aligning each particle to a rotationally averaged image which
//E is the average of all of the particles. This process may not work well for
//E all particles. It typically works very well on objects with high symmetry
//E (icosahedral viruses), and seems to work well on many objects with lower
//E symmetry. Note that this program does only integer pixel translations,
//E so no interpolation is done.

//D This program is not used by the standard refinement proceedure, but can
//E be used manually before the refinement loop for several purposes: 1)
//E After centering the particles it may be possible to reduce the size (in
//E pixels) of the entire particle set. This will reduce the processing time
//E in the refinement process.  2) Some unusable particles can be eliminated
//E so processor time isn't wasted on them later.  3) The refinement loop
//E may work a little better/converge faster if the particles start out near
//E the center of their boxes.

//D If, during the iterative loop, a particle does not seem to align
//E to a consistent center, that particle is discarded and written to
//E 'bad.hed'. All sucessfully aligned images are written to 'ali.hed'. The
//E reference image used for alignment after each iteration is stored in
//E 'avg.hed'. This file can be examined to see if convergence really has
//E been reached.

//D This process generally works better if the references, and eventually
//E the particles, can be masked during each step. The mask radius should
//E be specified so the mask is just larger than the largest axis of your
//E particle.

#include <cstdio>
#include <cmath>
#include "EMData.h"
#include "List.h"
#include "util.h"

int main(int argc, char *argv[]) {
int i,j,k,nx,nm=0,dmn=1,n,NS,usefilt=0;
List *data,*data2,*bad;
EMData *d,*d2;
char s[80], *alifile="ali.img", *badfile="bad.img", *avgfile="avg.img";
float f,mean,mask=-1,max=-1;

appinit(argc,argv);

if (argc<2) appusage(argv[0]);

for (i=2; i<argc; i++) {
	if (strncmp("num=",argv[i],4)==0) nm=atoi(&argv[i][4]);
	else if (strncmp("denom=",argv[i],6)==0) dmn=atoi(&argv[i][6]);
	else if (strncmp(argv[i],"frac=",5)==0) sscanf(&argv[i][5],"%d/%d",&nm,&dmn);
	else if (strncmp("mask=",argv[i],5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp("maxshift=",argv[i],9)==0) max=atof(&argv[i][9]);
	else if (strncmp("avgfile=",argv[i],8)==0) avgfile=&argv[i][8];
	else if (strncmp("alifile=",argv[i],8)==0) alifile=&argv[i][8];
	else if (strncmp("badfile=",argv[i],8)==0) badfile=&argv[i][8];
	else if (strncmp("usefilt",argv[i],7)==0) usefilt=1;
}

n=fileCount(argv[1]);
NS=n/dmn+1;
if (usefilt) data=readImagesExt(argv[1],".filt",nm*NS,(nm+1)*NS-1);
else data=readImages(argv[1],nm*NS,(nm+1)*NS-1);
if (data==NULL) error("Cannot read input",ERR_EXIT);
printf("Read %d images (%d - %d).\n",data->count(),nm*NS,(nm+1)*NS-1);
data2=new List;
for (i=0; i<data->count(); i++) {
	d2=(EMData *)data->objectAt(i);
	d2->normalize();
	d2->unum=i+nm*NS;
//	d2->applyMask(-1,1);		// noise mask
//	d2->applyMask(-1,0);		// mean mask
	data2->addObject(d2->copy());
}
nx=d2->xSize();
if (max<1) max=nx/4.0;
bad=new List;

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

printf("Images copied and normalized.\n");

/*unlink("avg.img");
unlink("avg.hed");
unlink("ali.img");
unlink("ali.hed");*/

for (i=0; i<12; i++) {
	printf("Loop %d\n",i);
	d=new EMData;
	d->makeAverage(data2);
	d->radialAverage();
	d->applyMask(mask,4);
	
//	if (i==7) 
		d->writeImage(avgfile,-1);

//	sprintf(s,"/tmp/tst%d.img",i);
//	d2=(EMData *)data2->objectAt(0);
//	d2->writeMRC(s);

	mean=0;
	for (j=0; j<data2->count(); j++) {
		d2=(EMData *)data2->objectAt(j);
		f=d2->transAlign(d,1,1,max);
		mean+=f;
		printf("%d. %1.3f\t%1.3f\t%1.2f\n",j,d2->Dx(),d2->Dy(),f);

// delete ones that need too much alignment
		if ((i>4 && f>nx/15) || (i>1 && hypot(d2->Dx(),d2->Dy())>max)){
			bad->addObject(d2->Parent());
			delete d2;
			data->removeObjectAt(j);
			data2->removeObjectAt(j);
			printf("DELETE %d, %f\n",j,f);
			j--;
			continue;
		}
		
		((EMData *)data2->objectAt(j))->rotateAndTranslate();
	}
	delete(d);
	mean/=data2->count();
	if (mean<.8 && i>1) break;
}
printf("%d loops completed\n",i);

//unlink("bad.img");
//unlink("bad.hed");
for (i=0; i<data2->count(); i++) {
	if (usefilt) {
		EMData tmp,*t2;
		t2=((EMData *)data2->objectAt(i));
		tmp.readImage(argv[1],t2->unum);
		tmp.edgeNormalize();
		tmp.setTAlign(t2->Dx(),t2->Dy(),0);
		tmp.setRAlign(0,0,0);
		tmp.rotateAndTranslate();
		tmp.applyMask((int)ceil(mask),0);
		tmp.edgeNormalize();
		tmp.writeImage(alifile,-1);
	}
	else {
		((EMData *)data2->objectAt(i))->normalize();
		((EMData *)data2->objectAt(i))->applyMask((int)ceil(mask),0);
		((EMData *)data2->objectAt(i))->writeImage(alifile,-1);
	}
}
for (i=0; i<bad->count(); i++) ((EMData *)bad->objectAt(i))->writeImage(badfile,-1);

LOG(Ref,"avg.hed",LOG_OUTFILE,NULL);
LOG(Ref,"ali.hed",LOG_OUTFILE,NULL);
LOG(Ref,"bad.hed",LOG_OUTFILE,NULL);
LOG(Ref,NULL,LOG_END,NULL);

printf("%d good, %d bad\n",data2->count(),bad->count());

return 0;
}
