// classesbybasis.C 5/2002  Steve Ludtke

//N classesbybasis
//F This will classify and align a set of particles based on an orthogonal basis as produced by svdcmp or geodcmp
//T LM
//1
//P <input>	Input images to classify
//P <ref set>	Reference data set 
//P <basis set>	A set of orthogonal basis images as generated by svdcmp
//P [mask=<rad>]	Mask radius
//P [frac=<num>/<denom>]	Operate only on a fraction of the raw images
//P [nbasis=<n>]	Number of basis images to use from the basis set. Defaults to all.
//P [saveali=<aligned out>]	File to write aligned images to
//P [savealiparm]	This will save alignment parameters for each image to a text file INSTEAD of writing to the 'saveali' file. 
//P [logit=<label>]	This specifies a label to use in the particle.log file. This records particle number classification throughout a refinement.
//P [groups=<ngrp>]	This will classify into multiple sets. Used by svdmultirefine.py
//P [maxshift=<dxy>]	Maximum radial translation
//P [verbose[=<n>]]	More verbose output
//P [lcmptoo]	This will also generate an lcmp quality factor for each classification
//P [clsbylcmp[=<filt rad>]]	This will classify based on lcmp comparison of the aligned images
//P [quiet]	No logging
//P [noisecomp]	Noise compensation
//P [skip1]	Skips the first basis vector
//P [shrink=<n>]	Scales the images down for initial orientation search, may be risky but gives a big speedup
//P [svweight]	Weight each basis vector by its singular value (rarely a good idea)
//U 

#include <cstdio>
#include <cmath>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "XYData.h"

//#define ANGMAP

struct BEST { float pwr,x,y,ang; int cls,flip; };

// comparison routine for qsort
int bstcmp(const void *d1, const void *d2) {
float p1=((BEST *)d1)->pwr;
float p2=((BEST *)d2)->pwr;
if (p1>p2) return -1;
if (p1<p2) return 1;
return 0;
}

#define BESTKP	20		// length of ordered list

// a little class for maintaining an ordered list
class Olist {
public:
BEST *best;
int N;

Olist(int ntokeep) {
	N=ntokeep;
	best=(BEST *)calloc(N,sizeof(BEST));
	for (int i=0; i<N; i++) {
		best[i].pwr=-1.0e23;
	}
}

~Olist() {
	free(best);
}

// this is used to re-sort the list after improper modifications have been made
void sort() {
	qsort(best,N,sizeof(BEST),bstcmp);
}

// note that this can mess up the sorting, sort() should probably be called after using this
void set(int i,float pwr,int cls,float x,float y,float ang,int flip) {
	best[i].pwr=pwr;
	best[i].cls=cls;
	best[i].x=x;
	best[i].y=y;
	best[i].ang=ang;
	best[i].flip=flip;
}

// this will insert a new element, maintaining proper ordering
void insert(float pwr,int cls,float x,float y,float ang,int flip) {
	if (pwr<best[N-1].pwr) return;

	int i,j;

	for (i=0; i<N; i++) if (pwr>best[i].pwr) break;

	for (j=N-1; j>i; j--) best[j]=best[j-1];
	
	best[i].pwr=pwr;
	best[i].cls=cls;
	best[i].x=x;
	best[i].y=y;
	best[i].ang=ang;
	best[i].flip=flip;
}

};

inline int negmod(int x,int y) {
if (x>=0) return x%y;
return y+x%y;
}

// This will calculate a set of CCFs for a rotated image, 1 for each basis vector, then
// compute the dot product of the resultant projection vector at each pixel location with
// a set of reference vectors, finally storing the strength of the best match in each pixel location
// in the returned image
EMData *mapbasis(EMData *orig,int nbasis,List *basis,int nclsref,float **clsvec,int maxshift,int weight=0) {
	int i,j,k,l;
	int xs=orig->xSize();
	if (maxshift>xs/4) maxshift=xs/4;
	EMData *ret=orig->copyHead();
	ret->zero();
	
	EMData **ccf = (EMData **)malloc(nbasis*sizeof(EMData *));
	for (i=0; i<nbasis; i++) ccf[i]=orig->calcCCF((EMData *)basis->objectAt(i));
	if (weight==1) {
		for (i=0; i<nbasis; i++) 
			ccf[i]->multConst(((EMData *)basis->objectAt(i))->ufloat/1000.0);
	}
	for (i=-maxshift; i<=maxshift; i++) {
		for (j=-maxshift; j<=maxshift; j++) {
			float bdot=-1.0e20;
			int bcmp=0;
			for (k=0; k<nclsref; k++) {
				float dot=0;
				for (l=0; l<nbasis; l++) 
					dot+=clsvec[l][k]*ccf[l]->valueAt(negmod(i,xs),negmod(j,xs));
				if (dot>bdot) { bdot=dot; bcmp=-k; }
			}
			ret->setValueAt(negmod(i,xs),negmod(j,xs),0,bdot);
			ret->setValueAt(i+xs/2,j+xs/2,0,bcmp);		
				// hack. We store the best matching reference number in the unused 1/2 of the map
		}
	}
	
//	ccf[0]->writeIMAGIC("z.hed",0);
//	ccf[1]->writeIMAGIC("z.hed",1);
	for (i=0; i<nbasis; i++) delete ccf[i];
	free(ccf);
	
	ret->getData();
	ret->doneData();
//	ret->writeIMAGIC("z.hed",2);
	return ret;
}

// Alternative version of the above that simply calculates the power rather than comparing to a set of
// possible basis vectors
EMData *mapbasis2(EMData *orig,int nbasis,List *basis,int nclsref,float **clsvec,int maxshift,int weight=0) {
	int i,j,k,l;
	int xs=orig->xSize();
	if (maxshift>xs/4) maxshift=xs/4;
	EMData *ret=orig->copyHead();
	ret->zero();
	
	EMData **ccf = (EMData **)malloc(nbasis*sizeof(EMData *));
	for (i=0; i<nbasis; i++) ccf[i]=orig->calcCCF((EMData *)basis->objectAt(i));
	if (weight==1) {
		for (i=0; i<nbasis; i++) 
			ccf[i]->multConst(((EMData *)basis->objectAt(i))->ufloat/1000.0);
	}
	for (i=-maxshift; i<=maxshift; i++) {
		for (j=-maxshift; j<=maxshift; j++) {
			float bdot=0;
			for (k=0; k<nbasis; k++) bdot+=SQR(ccf[k]->valueAt(negmod(i,xs),negmod(j,xs)));
			ret->setValueAt(negmod(i,xs),negmod(j,xs),0,bdot);
		}
	}
	
//	ccf[0]->writeIMAGIC("z.hed",0);
//	ccf[1]->writeIMAGIC("z.hed",1);
	for (i=0; i<nbasis; i++) delete ccf[i];
	free(ccf);
	
	ret->getData();
	ret->doneData();
//	ret->writeIMAGIC("z.hed",2);
	return ret;
}

// This will determine the projection of image 'orig' into 'basis' space, return is in vec
void project(EMData *orig, int nbasis, List *basis,float *vec,int weight=0) {
	float sum=0;
	int i,j;
	for (i=0; i<nbasis; i++) {
//		vec[i]=orig->dot((EMData *)basis->objectAt(i))/(orig->xSize()*orig->ySize());
		vec[i]=0;
		for (j=0; j<orig->xSize()*orig->ySize(); j++) vec[i]+=((EMData *)basis->objectAt(i))->valueAt(j,0)*orig->valueAt(j,0);
		vec[i]/=(orig->xSize()*orig->ySize());
		
		if (weight==1) vec[i]*=((EMData *)basis->objectAt(i))->ufloat/1000.0;
		sum+=SQR(vec[i]);
	}
	sum=sqrt(sum);
//	for (i=0; i<nbasis; i++) vec[i]/=sum;
}

// This does a dot product between a vector and a single reference from a table of reference vectors
inline float cmpbasis(int nbasis, int clsrefn, float *vec, float **clsvec) {
	float ret=0;
	for (int i=0; i<nbasis; i++) ret+=vec[i]*clsvec[i][clsrefn];
	return ret;
}

int main(int argc, char *argv[]) {
int i,j,k,num=0,denom=1,quiet=0,maxshift=-1,precen=0,lcmptoo=0,skip1=0,shrink=1,savealiparm=0;
int alignn=1,origrt=0,nbasis=-1,maskr=0,verbose=0,groups=0,weight=0,noisecomp=0,mode=1;
float dang=PI,lcmpfilt=0;
char *saveali="aligned.hed",*logit=NULL,*clsref;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=4; i<argc; i++) {
	if (strncmp(argv[i],"multiali=",9)==0) alignn=atoi(&argv[i][9]);
	else if (strncmp(argv[i],"frac=",5)==0) sscanf(&argv[i][5],"%d/%d",&num,&denom);
	else if (strncmp(argv[i],"nbasis=",7)==0) nbasis=atoi(&argv[i][7]);
	else if (strncmp(argv[i],"maxshift=",9)==0) maxshift=atoi(&argv[i][9]);
	else if (strncmp(argv[i],"lcmptoo",7)==0) {
		lcmptoo=1;
		if (argv[i][7]=='=') sscanf(argv[i]+8,"%f",&lcmpfilt);
	}
	else if (strncmp(argv[i],"clsbylcmp",9)==0)	{
		lcmptoo=2;
		if (argv[i][9]=='=') sscanf(argv[i]+10,"%f",&lcmpfilt);
	}
	else if (strncmp(argv[i],"saveali=",8)==0) saveali=argv[i]+8;
	else if (strncmp(argv[i],"savealiparm",11)==0) savealiparm=1;
	else if (strncmp(argv[i],"logit=",6)==0) logit=&argv[i][6];
	else if (strncmp(argv[i],"shrink=",7)==0) shrink=atoi(&argv[i][7]);
	else if (strncmp(argv[i],"mask=",5)==0) maskr=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"usefs=",6)==0) ;
	else if (strncmp(argv[i],"skip1",5)==0) skip1=1;
	else if (strncmp(argv[i],"svweight",8)==0) weight=1;
	else if (strncmp(argv[i],"noisecomp",9)==0) noisecomp=1;
	else if (strncmp(argv[i],"quiet",5)==0) quiet=1;
	else if (strncmp(argv[i],"vecmatchall",8)==0) mode=3;	// This is very slow. Instead of total power, based on match with best-matching projection
	else if (strncmp(argv[i],"vecmatch",8)==0) mode=2;		// not quite as slow, same principle, but total power still used in the initial coarse search
	else if (strncmp(argv[i],"groups=",7)==0) groups=atoi(&argv[i][7]);
	else if (strncmp(argv[i],"verbose",7)==0) {
		if (argv[i][7]!='=') verbose=1;
		else verbose=atoi(&argv[i][8]);
	}
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

if (shrink<1||shrink>8) error(ERR_EXIT,"Shrink must be between 1 and 8");

int bc=fileCount(argv[3]);
if (bc<2) error(ERR_EXIT,"Need at least 2 basis images!");
if (nbasis<0) nbasis=bc-1;
if (nbasis>bc) { error(ERR_WARN,"nbasis > # images in basis file"); nbasis=bc-1; }
if (maxshift==0) precen=1;
clsref=argv[2];

// Normal EMAN logging
int Ref;
if (!quiet) {
	LOGbegin(argc,argv);
	LOG(Ref,argv[1],LOG_INFILE,NULL);
	LOG(Ref,argv[2],LOG_REFFILE,NULL);
}

// read basis set and normalize
List *basis,*sbasis;
if (skip1) basis=readImages(argv[3],1,nbasis);
else {
	basis=readImages(argv[3],0,nbasis-1);
		// hack that makes the weight for the first vector equal to the weight of the second
//	((EMData *)basis->objectAt(0))->setNImg(((EMData *)basis->objectAt(1))->NImg());
}
for (i=0; i<nbasis; i++) {
	EMData *tmp=((EMData *)basis->objectAt(i));
//	tmp->edgeNormalize();
	tmp->multConst(1.0/sqrt(tmp->SumSq()));	// a 'true' normalization
}

if (shrink>1) {
	sbasis=new List;
	for (i=0; i<nbasis; i++) {
		EMData *tmp=((EMData *)basis->objectAt(i));
		sbasis->addObject(tmp->copy(0,0));
		tmp=((EMData *)sbasis->objectAt(i));
		tmp->meanShrink(shrink);
	}
}

if (maxshift<0) maxshift=((EMData *)basis->objectAt(0))->ySize()/8.0;
dang=atan2(1.0,((EMData *)basis->objectAt(0))->ySize()/2.0);

// read classification refs if present, and determine basis projections
int nclsref=fileCount(argv[2]);
if (nclsref<2) error(ERR_EXIT,"Could not read 2 or more references from classify file");

// clsvec is a set of nclsref vectors representing projections of the reference set into basis space
float **clsvec;
clsvec=(float **)malloc(nbasis*sizeof(float*));
for (i=0; i<nbasis; i++) clsvec[i]=(float *)malloc(nclsref*sizeof(float));

// refs is a list of all classification refs when needed
//QList<EMData> refs;
int nimg=fileCount(argv[1]);

EMData *mask=((EMData *)basis->objectAt(0))->copyHead();
mask->one();
if (maskr>3) mask->applyMask(maskr,0);

List refs;

// Noise compensation precalibration
float *ncmp = NULL;
if (noisecomp) {
        if (verbose) printf("Noise compensation:\n");
        ncmp=(float *)calloc(nbasis,sizeof(float));
        for (i=0; i<nimg; i++) {
                EMData in;
                in.readImage(argv[1],i);
                in.edgeNormalize();
                for (j=0; j<nbasis; j++) {
                        float dot=in.dot((EMData *)basis->objectAt(j))/(in.xSize()*in.ySize());
                        ncmp[j]+=dot*dot;
                }
        }
        for (i=0; i<nbasis; i++) {
                ncmp[i]=sqrt(ncmp[i]/(float)nimg);
                if (verbose) printf("%d. %1.3g\n",i,ncmp[i]);
        }
}

for (i=0; i<nbasis; i++) {
        EMData *d=(EMData *)basis->objectAt(i);
        if (weight) d->ufloat=d->NImg();
        else d->ufloat=1.0;
        if (noisecomp) d->ufloat/=ncmp[i];
}
if (noisecomp) weight=1;



if (clsref) {
	if (verbose) printf("Read classification refs (%s)\n",clsref);

	// outer loop over classification refs
	for (i=0; i<nclsref; i++) {
		// read reference
		EMData in;
		in.readImage(argv[2],i);
		in.multConst(1.0/sqrt(in.SumSq()));	// normalize this image too
//		if (lcmptoo>1) refs.append(in.copy(0,0));
		// determine basis projection for each ref
		float norm=0;
		for (j=0; j<nbasis; j++) {
			clsvec[j][i]=in.dot((EMData *)basis->objectAt(j));
			if (weight==1) clsvec[j][i]*=((EMData *)basis->objectAt(j))->ufloat;
			norm+=SQR(clsvec[j][i]);
		}
		norm=sqrt(norm);
		for (j=0; j<nbasis; j++) clsvec[j][i]/=norm;
		if (lcmpfilt>0) in.filter(1.0,lcmpfilt,1);
		if (lcmptoo>1) refs.addObject(in.copy(0,0));

		if (verbose>2) {
			printf("%d.\t",i);
			for (j=0; j<nbasis; j++) printf("%1.4f\t",clsvec[j][i]);
			printf("\n");
		}
	}
}

if (verbose) printf("Begin analysis:\n");
// outer loop over images being filtered
int start=nimg*num/denom;
int stop=nimg*(num+1)/denom;
//benchmark_start(1);
for (i=start; i<stop; i++) {
	float dts=0;

//	float *vec = (float*)malloc(nbasis*sizeof(float));
	float *vec = (float *)malloc(nbasis*sizeof(float));

#ifdef ANGMAP
	char S[80];
	sprintf(S,"angmap%03d.txt",i);
	FILE *amap = fopen(S,"wb");
#endif
	

	if (verbose>1) printf("\n");

	EMData in;
	in.readImage(argv[1],i);
	in.edgeNormalize();
	Euler eul(*in.getEuler());
	if (shrink>1) in.meanShrink(shrink);	// shrink now, undo it later
	// rot/trans alignment my maximizing basis projection pwrer
	EMData *cop=in.copy();
	float x,y,bx2,by2;
	int nx=in.xSize(),ny=in.ySize();
	int fl,bfl=0;
	float ang=0,bang2;
	Olist ol(BESTKP);
	BEST bst;
	bst.pwr=0;
	if (!precen) {
benchmark_start(1);
		for (fl=0; fl<2; fl++) {
			for (ang=0; ang<PI*2; ang+=dang*4.0) {	//4
				cop->setRAlign(ang);
				cop->setTAlign(0,0);
				cop->rotateAndTranslate();
				if (fl) cop->vFlip();
				EMData *ccfs;
				if (shrink>1) {
					if (mode <3) ccfs = mapbasis2(cop,nbasis,sbasis,nclsref,clsvec,maxshift/shrink,weight);
					else ccfs = mapbasis(cop,nbasis,sbasis,nclsref,clsvec,maxshift/shrink,weight);
					ol.insert(ccfs->Max(),-ccfs->valueAt((ccfs->MaxLoc()+ny*ny/2+nx/2)%(nx*ny),0),(ccfs->MaxLoc()%nx),(ccfs->MaxLoc()/nx),ang,fl);
				}
				else {
					if (mode<3) ccfs = mapbasis2(cop,nbasis,basis,nclsref,clsvec,maxshift,weight);
					else ccfs = mapbasis(cop,nbasis,basis,nclsref,clsvec,maxshift,weight);
					ol.insert(ccfs->Max(),-ccfs->valueAt((ccfs->MaxLoc()+ny*ny/2+nx/2)%(nx*ny),0),ccfs->MaxLoc()%nx,ccfs->MaxLoc()/nx,ang,fl);
				}
				if (verbose>2) printf("%f\t%f\n",ang*180./PI,ccfs->Max());
#ifdef ANGMAP
				fprintf(amap,"%g\t%g\n",ang*180./PI,ccfs->Max());
#endif
				delete ccfs;
			}
		}
benchmark_stop(1);

benchmark_start(2);
		// this is an ordered list of the best angles/positions we found
		for (j=0; j<BESTKP; j++) {
			if (ol.best[j].x>nx/2) ol.best[j].x=nx-ol.best[j].x;
			else ol.best[j].x*=-1;
			if (ol.best[j].y>ny/2) ol.best[j].y=ny-ol.best[j].y;
			else ol.best[j].y*=-1;
			ol.best[j].x*=shrink;
			ol.best[j].y*=shrink;
			if (ol.best[j].flip) ol.best[j].y*=-1;
			x= ol.best[j].x*cos(ol.best[j].ang)+ol.best[j].y*sin(ol.best[j].ang);
			y=-ol.best[j].x*sin(ol.best[j].ang)+ol.best[j].y*cos(ol.best[j].ang);
			ol.best[j].x=x;
			ol.best[j].y=y;
			if (verbose>2) printf("%d. %d\t%1.1f,%1.1f  %f\n",j,ol.best[j].cls,ol.best[j].x,ol.best[j].y,ol.best[j].ang*180./PI);
		}
benchmark_stop(2);
	}
	else {
		x=y=0;

		for (ang=0; ang<2.0*PI; ang+=dang*2.) {
			cop->setRAlign(ang);
			cop->setTAlign(0,0,0);
			cop->rotateAndTranslate();
			project(cop,nbasis,basis,vec,weight);

			//TODO insert cmpbasis() loop here and set 'ol' elements
		}
	}


	if (verbose>1) {
		printf("Best initial possibilities:\n");
		for (j=0; j<BESTKP && ol.best[j].pwr>.75*ol.best[0].pwr; j++) printf("%2d. %c %1.1f\t%1.1f,%1.1f \t%1.8f\t%d\n",j,ol.best[j].flip?'F':' ',
			ol.best[j].ang*180./PI,ol.best[j].x,ol.best[j].y,ol.best[j].pwr,ol.best[j].cls);
	}

	// if we did the presearch in shrink mode, we need to undo the effects here
	if (shrink>1) {
		delete cop;
		in.readImage(argv[1],i);
		in.edgeNormalize();
		cop=in.copy();
		nx=in.xSize();
		ny=in.ySize();
	}

	// Count the number of the best positions we wish to do a more detailed
	// search on before we zero the power values
	int ntodo;
	for (ntodo=1; ntodo<BESTKP && ol.best[ntodo].pwr>ol.best[0].pwr*.95; ntodo++);

	for (j=0; j<BESTKP; j++) ol.best[j].pwr=0;		// the CCF and dot products aren't normalized the same

benchmark_start(3);
	// now refine rotation for each of the best possibilities found
	// in the global search above. Only look at peaks at least
	// 75% of the largest amplitude. Should be safe ???
	// If we're in shrink mode, we need to refine x and y a little too
	k=0;
	while (k<ntodo) {
		float bx=ol.best[k].x;
		float by=ol.best[k].y;
//		float sm=shrink-1.0;
		float sm=0;
		bang2=ol.best[k].ang;
		bfl=ol.best[k].flip;
		for (x=bx-sm; x<=bx+sm; x+=1.0) {
			for (y=by-sm; y<=by; y+=1.0) {
				for (ang=bang2-dang*4.0; ang<bang2+dang*4.0; ang+=dang/2.0) {
					cop->setRAlign(ang);
					cop->setTAlign(x,y,0);
					cop->rotateAndTranslate();
					if (bfl) cop->vFlip();
					cop->mult(mask);
					project(cop,nbasis,basis,vec,weight);
					float pwr;
					if (mode<2) {
						pwr=0;
						for (j=0; j<nbasis; j++) pwr+=SQR(vec[j]);
					}
					else pwr=cmpbasis(nbasis,ol.best[k].cls,vec,clsvec);
					if (verbose>3 && pwr>=ol.best[k].pwr) cop->writeIMAGIC("a.hed",k);
					if (pwr>ol.best[k].pwr) ol.set(k,pwr,ol.best[k].cls,x,y,ang,bfl);
//					if (pwr>bpwr) { bpwr=pwr; bang=ang; bx=x; by=y; }
					if (verbose>2) printf("%d. %f\t%f\n",k,ang*180./PI,pwr);
#ifdef ANGMAP
					fprintf(amap,"%g\t%g\n",ang*180./PI,pwr);
#endif
				}
			}
		}
		k++;
	}
	ol.sort();
benchmark_stop(3);

	if (verbose>1) {
		printf("Best possibilities after rot refine:\n");
		for (j=0; j<BESTKP && ol.best[j].pwr>.5*ol.best[0].pwr; j++) printf("%2d. %c %1.1f\t%1.1f,%1.1f \t%1.8f\t%d\n",j,ol.best[j].flip?'F':' ',
			ol.best[j].ang*180./PI,ol.best[j].x,ol.best[j].y,ol.best[j].pwr,ol.best[j].cls);
	}

benchmark_start(4);
	// Now we do x,y and ang. We keep trying a few of the best possibilities
	int nbest=0;
	if (!precen) {
		k=0;
		while (k<BESTKP && ol.best[k].pwr>ol.best[0].pwr*.95) {
			if (verbose>2) printf("try best %d\n",k);
			bx2=ol.best[k].x;
			by2=ol.best[k].y;
			bang2=ol.best[k].ang;
			bfl=ol.best[i].flip;
			// now refine the dx/dy and rotation
			for (x=bx2-.75; x<=bx2+.76; x+=.25) {
				for (y=by2-.75; y<=by2+.76; y+=.25) {
					for (ang=bang2-dang; ang<bang2+dang; ang+=dang/4.) {
						cop->setRAlign(ang);
						cop->setTAlign(x,y,0);
						cop->rotateAndTranslate();
						if (bfl) cop->vFlip();
						cop->mult(mask);
						project(cop,nbasis,basis,vec,weight);
						float pwr=0;
						if (mode<2) {
							pwr=0;
							for (j=0; j<nbasis; j++) pwr+=SQR(vec[j]);
						}
						else pwr=cmpbasis(nbasis,ol.best[k].cls,vec,clsvec);
						if (verbose>3 && pwr>=ol.best[k].pwr) cop->writeIMAGIC("b.hed",k);
						if (pwr>ol.best[k].pwr) ol.set(k,pwr,ol.best[k].cls,x,y,ang,bfl);
#ifdef ANGMAP
						fprintf(amap,"%g\t%g\n",ang*180./PI,pwr);
#endif
					}
				}
			}
			k++;
			nbest=k;
		}
	}
	ol.sort();
benchmark_stop(4);

#ifdef ANGMAP
	fclose(amap);
#endif

	// now we find the best class match for each one

	if (lcmptoo>1) {
		for (j=0; j<nbest; j++) {
			cop->setRAlign(ol.best[j].ang);
			cop->setTAlign(ol.best[j].x,ol.best[j].y,0);
			cop->rotateAndTranslate();
			if (ol.best[j].flip) cop->vFlip();
			cop->mult(mask);
//cop->writeIMAGIC("zal.hed",-1);
			if (lcmpfilt>0) cop->filter(1.0,lcmpfilt,1);
			int bk=0;
			float bpwr=-1.0e30;
			for (k=0; k<nclsref; k++) {			
				EMData *ref=(EMData *)refs.objectAt(k);
				float pwr=4.0-ref->lcmp(cop);
				if (pwr>bpwr) { bpwr=pwr; bk=k; }
				if (verbose>2) printf("%1.3f%c\t",pwr,(pwr==bpwr)?'*':' ');
			}
			ol.set(j,bpwr,bk,ol.best[j].x,ol.best[j].y,ol.best[j].ang,ol.best[j].flip);
		}
	}
	else {
		for (j=0; j<nbest; j++) {
			cop->setRAlign(ol.best[j].ang);
			cop->setTAlign(ol.best[j].x,ol.best[j].y,0);
			cop->rotateAndTranslate();
			if (ol.best[j].flip) cop->vFlip();
			cop->mult(mask);
			project(cop,nbasis,basis,vec,weight);
			if (verbose>2) {
				for (k=0; k<nbasis; k++) printf("%1.3f\t",vec[k]);
				printf("\n");		
			}
			int bk=0;
			float bpwr=-1.0e30;
			for (k=0; k<nclsref; k++) {
				float pwr=cmpbasis(nbasis,k,vec,clsvec);
				if (pwr>bpwr) { bpwr=pwr; bk=k; }
				if (verbose>2) printf("%1.3f%c\t",pwr,(pwr==bpwr)?'*':' ');
			}
			printf("\n");
			ol.set(j,bpwr,bk,ol.best[j].x,ol.best[j].y,ol.best[j].ang,ol.best[j].flip);
		}
	}	
	ol.sort();

	if (verbose>1) {
		printf("Best possibilities after final refine:\n");
		for (j=0; j<nbest; j++) printf("%2d. %c %1.1f\t%1.1f,%1.1f \t%1.5g\t%d\n",j,ol.best[j].flip?'F':' ',
			ol.best[j].ang*180./PI,ol.best[j].x,ol.best[j].y,ol.best[j].pwr,ol.best[j].cls);
	}

	delete cop;
	in.setRAlign(ol.best[0].ang);
	in.setTAlign(ol.best[0].x,ol.best[0].y,0);
	in.rotateAndTranslate();
	if (ol.best[0].flip) in.vFlip();
	in.setRAlign(eul.alt(),eul.az(),eul.phi()-ol.best[0].ang);

	if (savealiparm) {
		char s[256];
		sprintf(s,"%s %d %s %1.8g %1.2f %1.2f\n",saveali,i,argv[1],ol.best[0].ang,ol.best[0].x,ol.best[0].y);
		FSsend(FS_TEXT,strlen(s)+1,"aliparm.txt",s);
	}
	else in.writeIMAGIC(saveali,i);

	float lcmpqual=0;
	if (lcmptoo) {
		EMData cm;
		cm.readImage(argv[2],ol.best[0].cls);
		if (lcmpfilt>0) {
			EMData *cp=in.copy(0,0);
			cp->filter(2.0,lcmpfilt,1);
			cm.filter(2.0,lcmpfilt,1);
			lcmpqual=cm.lcmp(cp,1,0);
			delete cp;
		}
		else lcmpqual=cm.lcmp(&in,1,0);
	}

	int maxj=ol.best[0].cls;
	float max =ol.best[0].pwr;
	if (groups) {
		int gsize=1+(bc-1)/groups;	// this should be exactly an integer anyway
		int grp=maxj/gsize;
		char s[200],t[200],u[512];
		sprintf(s,"%0d/cls%04d.lst",grp,maxj%gsize);
		sprintf(t,"%f",max);
		sprintf(u,"../%s",saveali);
		in.writeLST(s,-1,u,i,t);
	}
	else {
		char s[200],t[200];
		sprintf(s,"cls%04d.lst",maxj);
		sprintf(t,"%f",max);
		if (saveali) in.writeLST(s,-1,saveali,i,t);
		else in.writeLST(s,-1,argv[1],i,t);
	}
			
	if (logit) {
		// order: program tag, particle number, file, ntk rank, quality, defocus, class, iteration tag 
		char s[120];
		float *ctf=in.getCTF();
		if (!in.hasCTF()) ctf[0]=0;
		if (groups) {
			int gsize=bc/groups;	// this should be exactly an integer anyway
			int grp=maxj/gsize;
			char fsp[180];
			char fsp2[1024];
			sprintf(fsp,"%0d/particle.log",grp);
			sprintf(fsp2,"../%s",saveali);
			sprintf(s,"B\t%d\t%s\t0\t%1.5g\t%1.3f\t%d\t%s\t%1.5g\n",i,fsp2,max,-ctf[0],maxj%gsize,logit,lcmpqual);
			FSsend(FS_TEXT,strlen(s)+1,fsp,s);
		}
		else {
			sprintf(s,"B\t%d\t%s\t0\t%1.5g\t%1.3f\t%d\t%s\t%1.5g\n",i,saveali,max,-ctf[0],maxj,logit,lcmpqual);
			FSsend(FS_TEXT,strlen(s)+1,"particle.log",s);
		}
	}
	if (verbose==1||verbose==2) { printf("%d -> %d  (%1.4g)\n",i,maxj,max); fflush(stdout); }
	free(vec);
}
//benchmark_stop(1);
benchmark_report();

if (!quiet) LOG(Ref,0,LOG_END,NULL);

return 0;
}
