// classesbycl.C 3/24/01  Steve Ludtke
// Copyright 2001 Baylor College of Medicine

//N classesbycl
//F This classifies a set of centered particles using self common lines.
//T LC
//1
//P <particles>	Raw particle filespec (must have a .cl. version)
//P <refrences>	Reference image filespec (must have a .cl. version)
//P [assigneuler]	Assigns Eulers to each raw image
//P [eulerinref]	Euler angles should be taken from <references> not common-lines file
//P [frac=<num>/<denom>]	Operate only on a fraction of the raw images
//P [verbose[=<n>]]	More verbose output
//P [matrix]	Generates similarity matrix for entire data set in matrix.dat
//P [logit=<label>]	This specifies a label to use in the particle.log file. This records particle number classification throughout a refinement.
//U  classesbycl start.hed proj.hed
//D This classifies a set of raw particles based on a set of reference projections.
//E The output is a set of cls*.img files, one file for each reference image. This
//E program differs from classesbymra in the use of self common-lines maps for
//E performing the actual comparisons. 

#include <cstdio>
#include <cmath>
//#include <unistd.h>
#include "EMData.h"
#include "List.h"
#include "util.h"

// a little class for maintaining an ordered list
class Olist {
public:
float *dot;
int *loc;
int *n;
EMData **ali;
int N;

Olist(int ntokeep) {
	N=ntokeep;
	dot=(float *)malloc(N*sizeof(float));
	loc=(int *)malloc(N*sizeof(int));
	n=(int *)malloc(N*sizeof(int));
	ali=(EMData **)malloc(N*sizeof(EMData *));
	for (int i=0; i<N; i++) {
		dot[i]=-1.0e20;
		n[i]=-1;
		ali[i]=NULL;
	}
}

~Olist() {
	free(dot);
	free(n);
	free(loc);
	for (int i=0; i<N; i++) delete ali[i];
	free(ali);
}

void insert(float vdot, int vn, int vloc, EMData *vali) {
	int i,j;
	
	for (i=0; i<N; i++) if (dot[i]<vdot) break;

	if (i==N) { if (vali) delete vali; return; }

	if (ali[N-1]) delete ali[N-1];
	for (j=N-1; j>i; j--) {
		dot[j]=dot[j-1];
		loc[j]=loc[j-1];
		n[j]=n[j-1];
		ali[j]=ali[j-1];
	}
	dot[i]=vdot;
	loc[i]=vloc;
	n[i]=vn;
	ali[i]=vali;
}

};

int main(int argc, char *argv[]) {
int i,j,k,*rn,x,y,nr,ns,ntk=1,split=0,num=0,denom=1,ver=0;
int n0,n1,quiet=0,matrix=0,aeuler=0,eir=0;
List *data,*dataref,*datarefnc,*newref=NULL;
EMData *d,*d2,*d2f,*df,*dr,*da,*msk,*msk2;
char s[80],t[160],*f2r = argv[1];
float dot,classes=-999.0,norm,norm2;
float nullctf[11] = { 0,0,0,0,0,0,0,0,0,0,0 };
float *ctf;
char *tree1=NULL,*tree2=NULL,*logit=NULL;
FILE *out;
Olist *best;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=3; i<argc; i++) {        
	if (strncmp(argv[i],"logit=",6)==0) logit=&argv[i][6];
	else if (strncmp(argv[i],"frac=",5)==0) sscanf(&argv[i][5],"%d/%d",&num,&denom);
	else if (strncmp(argv[i],"quiet",5)==0) quiet=1;
	else if (strncmp(argv[i],"assigneuler",11)==0) aeuler=1;
	else if (strncmp(argv[i],"matrix",6)==0) matrix=1;
	else if (strncmp(argv[i],"eulerinref",10)==0) eir=1;
	else if (strncmp(argv[i],"verbose",7)==0) {
		if (strncmp(argv[i],"verbose=",8)==0) ver=atoi(&argv[i][8]);
		else ver=1;
	}
	else if (strncmp(argv[i],"usefs=",6)==0) ;
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

if (num>=denom || num<0) { printf("Invalid frac= specifiation\n"); exit(1); }

f2r=(char *)malloc(strlen(argv[1])+6);
strcpy(f2r,argv[1]);
sprintf(f2r+strlen(argv[1])-4,".cl%s",argv[1]+strlen(argv[1])-4);

// READ Reference Images
dataref=readImagesExt(argv[2],".cl",0,-1);		// read all input images
if (dataref==NULL) error("Cannot read input",ERR_EXIT);
nr=dataref->count();
for (i=0; i<nr; i++) {
	d=(EMData *)dataref->objectAt(i);
//	d->xNormalize();	// sum of each line -> 1
//	d->multConst(1.0/d->Sigma()); //now normalize overall
	if (d->Mean()<=0) error(ERR_EXIT,"Reference image normalization error.");
	d->multConst(1.0/d->Mean());	// Yes. This is a strange normalization
	d->unum=i;
	if (eir) {
		EMData tr;
		tr.readImage(argv[2],i,1);
		d->setRAlign(tr.alt(),tr.az(),tr.phi());
	}
}

ns=fileCount(f2r);
printf("%d images to classify.\n",ns);

// Read the file containing the list of images to exclude
FILE *za = fopen("start.exclude","rb");
char *zapped=NULL;
if ((strcmp(argv[1],"start.hed")==0 || strcmp(argv[1],"start.img")==0) && za) {
	printf("Using image exclusion file\n");
	zapped=(char *)calloc(1,ns);
	while (fgets(s,79,za)) zapped[atoi(s)]=1;
	fclose(za);
}

if (matrix) {
	if (!num) { 
		int tt[3];
		tt[0]=-2;	// truncate file
		tt[1]=nr;
		tt[2]=ns;
		FSsend(FS_BIN,12,"matrix.dat",(char *)tt);
	}
	else sleep(1); // stupid hack to avoid possible file contention
}

printf("Read %d references.\n",nr);

rn=(int *)malloc(nr*sizeof(int));
for (i=0; i<nr; i++) rn[i]=((EMData *)dataref->objectAt(i))->unum;

int Ref=0;
if (!quiet) {
	Ref=LOGbegin(argc,argv);
	LOG(Ref,argv[2],LOG_INFILE,NULL);
	LOG(Ref,argv[1],LOG_INFILE,NULL);
}

x=d->xSize();
y=d->ySize();

d=new EMData;

n0=num*ns/denom;
n1=(num+1)*ns/denom;

printf("processing %d - %d of %d\n",n0,n1,ns);
fflush(stdout);

if (!quiet) LOG(Ref,"cls0000.lst",LOG_OUTFILE,NULL);

float *sim = (float *)malloc(sizeof(float)*nr);

// align and classify
for (i=n0; i<n1; i++) {
	best=new Olist(ntk);
	if (zapped && zapped[i]) continue;	// This particle has been marked as bad

	d->readImage(f2r,i);
	d->xNormalize(1);
	
	for (j=0; j<nr; j++) {
		int loc;
		dr=(EMData *)dataref->objectAt(j);

		da=d->calcCCFX(dr,0,y);
//		dot=da->Max()> -da->Min()?da->Max():-da->Min();
		dot=da->Max();
		loc=da->MaxLoc();
		dot/=dr->xSize()*dr->ySize();

//		dot=(da->Sigma()+SQR(da->Mean()))*da->xSize()*da->ySize()+(dr->Sigma()+SQR(dr->Mean()))*dr->xSize()*dr->ySize()-2*dot;
//		dot/=da->xSize()*da->ySize();
		
		if (da) sim[j]=dot;
		else sim[j]=0;
		if (sim[j]<0) sim[j]=0;
		if (ver) printf("%d. %1.2f\t%1.2f\t%1.2f\n",
			j,dot,da->Min()/(dr->xSize()*dr->ySize()),da->Max()/(dr->xSize()*dr->ySize()));

		if (ver>=9) { 
			d2->makeRFP()->writeIMAGIC("align.hed",-1);
			df->makeRFP()->writeIMAGIC("align.hed",-1);
		}

		best->insert(dot,j,loc,da);	// note that this may delete da
		da=NULL;
		
	}
	
	if (matrix) {
		int *tt = (int *)malloc((2+nr)*sizeof(int));
		tt[0]=-1;
		tt[1]=i;
		memcpy(tt+2,sim,nr*sizeof(int));
		FSsend(FS_BIN,(2+nr)*sizeof(int),"matrix.dat",(char *)tt);
		free(tt);
	}
	

	if (best->dot[0]<0) { 
		printf("No class for %d\n",i);  
		delete d2;
		delete df;
		continue;
	}

	printf("%d -> ",i);
	for (j=0; j<ntk; j++) {
		printf("\t%d (%1.2f)",rn[best->n[j]],best->dot[j]);
		sprintf(s,"%scls%04d.lst",tree1?"t":"",rn[best->n[j]]);
		sprintf(t,"%1.2f",best->dot[j]);
		d->writeLST(s,-1,argv[1],i,t);
			
		// order: program tag, particle number, ntk rank, quality, class, iteration tag 
		if (logit) {
			char s[120];
			sprintf(s,"C\t%d\t%d\t%1.5g\t%d\t%s\n",i,j,best->dot[j],rn[best->n[j]],logit);
			FSsend(FS_TEXT,strlen(s)+1,"particle.log",s);
		}
	}
	
	if (aeuler) {
		dr=(EMData *)dataref->objectAt(best->n[0]);
		EMData tmp;
		tmp.readImage(argv[1],i);
		tmp.setRAlign(dr->alt(),dr->az(),PI*2.0-best->loc[0]*(360.0/dr->xSize())*(PI/180.0));
		tmp.setNImg(1);
		printf("\t%1.0f\t%1.0f\t%1.0f",dr->alt()*180.0/PI,dr->az()*180.0/PI,360.0-(float)best->loc[0]*(360.0/dr->xSize()));
		tmp.writeIMAGIC(argv[1],i);
	}

	printf("\n");
	fflush(stdout);
	
	delete best;
}


if (!quiet) LOG(Ref,NULL,LOG_END,NULL);

return 0;
}



