// eulerproj.C Copyright 1999 Baylor College of Medicine
// 2/6/99 Steven Ludtke
// This program sets euler angles on the images in a file by projection matching
// note that the images to be 'classified' should already be CTF corrected
// (typically with classalign2) if CTF correction is being performed

//N eulerproj
//F This program reassigns Euler angles to a set of class averages by projection matching
//T LT
//1
//P <classes>	The file containing the class averages to be oriented
//P <refrences>	Projections to be used as references
//P [sigfilt]	Apply a real-space filter to use only intense data in the classification
//P [frac=<num>/<denom>]	Operate only on a fraction of the raw images
//P [mask=<radius>]	Circular mask
//P [imask=<radius>]	Circular innner mask
//P [phase]	Use phase error as a similarity criteria
//P [writeali]	Write the aligned images over the original images
//P [verbose]	More verbose output during classification
//U eulerproj classes.img proj.img mask=42
//D When raw images are classified, there may be classification errors. Sometimes the
//E misclassified images may outnumber the correctly classified images in specific classes.
//E This means the class average generated from these images may not have the correct
//E Euler angles. Since the class average is much less noisy, the Euler angle assignment
//E is much more accurate for class averages. This program is used to redetermine the
//E Euler angles for lower noise class averages. It could potentially be used to assign
//E Euler angles to individual particles as well, although this practice is not part
//E of the standard reconstruction procedure in EMAN.

#include <cstdio>
#include <cmath>
#include "EMData.h"
#include "List.h"
#include "util.h"

int main(int argc, char *argv[]) {
int nr,mask=0,imask=0,num=0,denom=1,precen=0,writeali=0,verbose=0,sigfilt=0,phase=0,matrix=0;
int i,j,k,ns,n0,n1,x,y,prevref=0,maxshift,best;
char s[80];
float norm,bestdot,dot,bang[3];
List *dataref;
EMData *msk,*d,*dbest,*d2,*df,*dr,*da;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=3; i<argc; i++) {
	if (strncmp(argv[i],"mask=",5)==0) mask=atoi(&argv[i][5]);
 	else if (strncmp(argv[i],"imask=",6)==0) imask=atoi(&argv[i][6]);
	else if (strncmp(argv[i],"num=",4)==0) num=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"denom=",6)==0) denom=atoi(&argv[i][6]);
	else if (strncmp(argv[i],"frac=",5)==0) sscanf(&argv[i][5],"%d/%d",&num,&denom);
	else if (strncmp(argv[i],"precen",6)==0) precen=1;
	else if (strncmp(argv[i],"sigfilt",7)==0) sigfilt=1;
	else if (strncmp(argv[i],"phase",5)==0) phase=1;
	else if (strncmp(argv[i],"matrix",6)==0) matrix=1;
	else if (strncmp(argv[i],"writeali",8)==0) writeali=1;
	else if (strncmp(argv[i],"verbose",7)==0) verbose=1;
	else if (strncmp(argv[i],"usefs=",6)==0) ;
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

if (num>=denom || num<0) { printf("0 <= num < denom !!!\n"); exit(1); }
if (imask>1 && !precen) { printf("imask may only be used in conjunction with precen\n"); exit(0); }
if (imask==0) imask=8;

// READ Reference Images
nr=fileCount(argv[2]);
dataref=readImages(argv[2],0,nr);		// read all input images
if (dataref==NULL) error("Cannot read input",ERR_EXIT);
printf("Read %d references.\n",dataref->count());

int Ref=0;

if (num==0) {
	Ref=LOGbegin(argc,argv);
	LOG(Ref,argv[2],LOG_INFILE,NULL);
	LOG(Ref,argv[1],LOG_INFILE,NULL);
}


// prepare mask
msk=((EMData *)dataref->objectAt(0))->copy();
msk->setParent(NULL);
msk->one();
msk->applyMask(mask,4);
x=msk->xSize();
y=msk->ySize();

for (j=0; j<nr; j++) {
	dr=(EMData *)dataref->objectAt(j);
	dr->mult(msk);
	if (sigfilt) dr->realFilter(3,dr->Sigma(),-dr->Sigma());
//	else dr->normalize();
}

ns=fileCount(argv[1]);
printf("%d images to classify.\n",ns);


d=new EMData;

norm=x*y/1000.0;
maxshift=y/2-mask;

n0=num*ns/denom;
n1=(num+1)*ns/denom;

EMData mx;
if (matrix) mx.setSize(nr,n1-n0,1);

printf("processing %d - %d of %d\n",n0,n1,ns);
if (num==0) LOG(Ref,"cls000.hed",LOG_OUTFILE,NULL);

// align and classify
for (i=n0; i<n1; i++) {
	d->readImage(argv[1],i);
	if (d->NImg()<0) { prevref=1; continue; }
	d->edgeNormalize();
//	d->normalize();
	if (d->xSize()!=x ||d->ySize()!=y) { printf("Bad image size %d,%d != %d,%d\n",d->xSize(),d->ySize(),x,y); exit(1); }

	d2=d->copy();
	df=d->copy(0,0);
	df->vFlip();
	df=df->copy();

	best=0;
	bestdot=-1000.0;
	dbest=NULL;
	for (j=0; j<nr; j++) {
		if (verbose) {printf(" %d\r",j); fflush(stdout); }
		dr=(EMData *)dataref->objectAt(j);
		if (precen) da=d2->RFAlign(dr,df,imask);
		else {
			da=d2->RTFAlign(dr,df,1,maxshift);
//			if (fabs(da->Dx())>maxshift || fabs(da->Dy())>maxshift) { delete da; continue; }
		}
		da->refineAlign(dr);
		if (!da) continue;
		da->mult(msk);
//		da->normalize();
//		dot=da->dot(dr);
#ifndef ALTCMP
		if (phase==1) dot=(2.0-da->pcmp(dr,NULL))*500.0;
		else dot=(2.0-da->lcmp(dr,1))*500.0;
#else
		dot=(da->fscmp(d2))/(float)da->ySize();
#endif
		if (matrix) mx.setValueAt(j,i-n0,0,dot);

		if (dot>bestdot) {
			bestdot=dot;
			best=j;
			bang[0]=dr->alt();
			bang[1]=dr->az();
			bang[2]=dr->phi();
			if (dbest) delete dbest;
			dbest=da;
		}
		else delete da;
	}

	if (!dbest) {
		printf("No class for %d\n",i);
		delete d2;
		delete df;
		continue;
	}

	if (writeali) {
		if (verbose) printf("%3d. (%3d) %1.1f,%1.1f,%1.1f  ->  ",i,best,
			d->alt()*180.0/PI,d->az()*180.0/PI,d->phi()*180.0/PI);
		dbest->setRAlign(bang[0],bang[1],bang[2]);
		if (verbose) printf("%1.1f,%1.1f,%1.1f  %s  (%f)\n",bang[0]*180.0/PI,bang[1]*180.0/PI,bang[2]*180.0/PI,
			(dbest->getFlags()&EMDATA_FLIP)?"flip":"",bestdot);
		dbest->writeIMAGIC(argv[1],i);
		if (verbose) { printf("wrote aligned\n"); fflush(stdout); }
	} else {
		if (verbose) printf("%3d. (%3d) %1.1f,%1.1f,%1.1f  ->  ",i,best,
			d->alt()*180.0/PI,d->az()*180.0/PI,d->phi()*180.0/PI);
		if (dbest->getFlags()&&EMDATA_FLIP) d->vFlip();	// flip isn't stored in Euler, so must be done
		d->setRAlign(bang[0],bang[1],bang[2]+dbest->alt());
		if (verbose) printf("%1.1f,%1.1f,%1.1f %s\n",bang[0]*180.0/PI,bang[1]*180.0/PI,(bang[2]+dbest->alt())*180.0/PI,
			(dbest->getFlags()&EMDATA_FLIP)?"flip":"");
		d->writeIMAGIC(argv[1],i);
		if (verbose) { printf("wrote unaligned\n"); fflush(stdout); }
	}


	delete d2;
	delete df->Parent();
	delete df;
	delete dbest;

	//replace the reference image in the file
	if (prevref) {
		dr=(EMData *)dataref->objectAt(best);
		dr->setNImg(-1);
		dr->writeIMAGIC(argv[1],i-1);
		if (verbose) { printf("wrote reference\n"); fflush(stdout); }
	}

	prevref=0;
}


if (matrix) mx.writeMRC("matrix.mrc");
if (num==0) LOG(Ref,NULL,LOG_END,NULL);

return 0;
}
