// foldhunterP.C	June 1999 Wen Jiang
// Copyright 1999 Baylor College of Medicine

//N foldhunterP
//F This program is used to quantitatively fit a subunit/domain into a large structure density map at intermediate resolution.
//T X
//1
//P <map>	input file, single 3D map in MRC format, usually the reconstruction map
//P <model>	input file, single 3D map in MRC format, the probe map e.g. from PDB atomic structure blurred to similar resolution to <map>
//P <outputmap>	the output density map in MRC format, is the rotated and then translated probe map according to the best match parameters
//P [cpu=<number of CPU>]	number of processors to use, default to 1
//P [da=<angle>]	angular search step size in degree
//P [startangle=<alt,az,phi>]	starting angles, search will be around it (+-range/2), in degree
//P [range=<angle range>|<r1,r2,r3>]	range of the search will cover, single value will set three ranges the same, or three different ranges could be set, in degree
//P [smart]	if smart is set, the search will go deep around the best orientation with finer angular step size until angular step size is smaller than 1 degree or finestda
//P [mcf]	use MCF instead the default CCF for correlation function
//P [flcf=<radius>]	STILL EXPERIMETAL: use fast local correlation instead the default CCF for correlation function, input radius of the mask
//P [sphere_region=<x0,y0,z0,r>]	only consider solutions in given spheric region, in pixel, relative to center of the map
//P [cyl_region=<x0,y0,z0,rmin,rmax,h>]	only consider solutions in given cylinder shell region, in pixel, relative to center of the map
//P [keep=<solution_num>]	number of the top matches will be outputed, default to 20
//P [mask=<x0,y0,z0,r>]	won't consdier solutions inside the masked spheric region, in pixel, relative to center of the map
//P [apix=<angstrom_per_pixel>]	Angstrom/voxel, if set, the output for translation will be in Angstrom, otherwise in pixel
//P [res=<angstrom>]	the resolution of map, only effective when da is not set and used to calculate the maximal angular step size. default to 8 Angstrom
//P [peakwidth=<pixels>]	pixels assumed for each match peak width, to avoid too close peaks being considered, default to 5 pixels
//P [finestda=<finest angle stepsize>]	smallest angular step size that 'smart' feature will use,d efault to 1 degree
//P [int=<file>]	filename for intermediate data output
//P [relorig=<x,y,z>]	the position of the small boxed density(model) in the large map, required for [explorer=< >] option, default to 0,0,0
//P [explorer=<binary file>]	output a file to read by an IRIS Explorer module ReadFoldhunter to apply the solutions, only work when [relorig=<x,y,z>] is set
//P [log=<filename>]	the file to send the intermediate outputs, if not set, stdout (screen) will be used
//P [bat=<batch file>]	output a csh script to transform the probe to match the ref map according to the top solutions
//P [genpdb=<pdb file>]	This will generate a new PDB file in the determined orientations
//P [verbose]	output the intermediate information and the time spent in the orientation search will be monitored
//U foldhunterP	ref.mrc probe.mrc out.mrc cpu=8 explorer=out.exp relorig=123,234,345 bat=out.bat
//U foldhunterP	ref.mrc probe.mrc out.mrc da=20 smart
//U foldhunterP	ref.mrc  probe.mrc  out.mrc  da=1 startangle=35.7,99.3,234.7 range=10.0
//D This program is classified as experimental at this point. Expect
//E a forthcoming publication with more details. We do not suggest
//E using it until it has been better characterized.
//D This program is parellelized using pthreads to run on shared memory computers.
//D This program tries to find the best matches for a probe density map (such as a known structure
//E from PDB) to be fitted into the reference map which is usually a lower resolution
//E reconstruction map by cryoEM, this will hopefully give more information about how the domains
//E arrangement in the context of large assemblies .
//D The requirements for the two input maps are that they are cubic, same size, similar resolution.
//D The methodology is simply cross correlation. The correlation coefficients are used as the
//E criterion to find the best matches. The final outputs are just simplely the top ranked matches, there
//E is no intention to build intelligence to determine if the matches are correct. Make sure to
//E visually inspect the top matches. Experience showed that the correct match needs not
//E to be the best match.
//D The program outputs the Euler angles and the translations for all matches, but only actually
//E transforms and outputs the probe map according to the best match. When viewed together with the
//E reference map, this output map will overlap the corresponding domain in reference map if the
//E solution is correct.
//D There are three ways to check for other matches.
//D First, one could use the <b>[explorer=< >]</b> option which will output a binary file to be read in
//E by an IRIS Explorer module developed in <a href="http://ncmi.bcm.tmc.edu">NCMI</a>. This module provides interface for
//E interactively inspect each of the matchs.
//D Second, the <b>[bat=<batch file>]</b> option will output a CSH script which would handle the correct transformations
//E for all the matches. Just run the csh script to get all the transformed probe structures.
//D Third, one could manually transform to match the equivalent
//E domain in reference map. Be sure to transform in correct order: <b>rotate first, then translate</b>.
//E <a href="http://ncmi.bcm.tmc.edu/~stevel/EMAN/doc/progs/proc3d.html"><b>proc3d</b></a>
//E could be used to do these transforms, but unfortunately <b>proc3d</b> does it in reversed order
//E (translate first, then rotate), be be sure to <b>transform in two steps</b> instead of one step if
//E <b>proc3d</b> is used.
//D More specifically, to transform the model map to match the corresponding part in the reference map:
//D     1:      rotate the model by the 3 rotation angles
//D                    e.g.    proc3d model out1 rot=30,60,80
//D
//D     2:      translate the output of step 1 by the 3 translations
//D                    e.g.    proc3d out1 out trans=3,-4,5
//D Please visit web page <a href="http://ncmi.bcm.tmc.edu/~wjiang/hhfh">Helixhunter and Foldhunter</a>
//E for more detailed information.
#include <cstdio>
#include <cmath>
#include "EMData.h"
#include "Vect.h"
#include "Cylinder.h"
#include <cassert>
#include <cfloat>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <pthread.h>
#include <vector>
#include <cstring>
using namespace std;

struct Para {
	int	ithread;
	int	nz;
	float	da1;
	float	da2;
	float	da3;
	float	alt0, az0, phi0;
	float	range1, range2, range3;
	bool timing;
	ostream * log;
  float radius;
};

EMData *model, *map;
vector < EMData* > best, balt, baz, bphi;
bool usemcf = false;
bool useflcf=false;
float	gmax = (-FLT_MAX);

void *foldhunter_thread	(void *arg)
{
	Para para = *(Para * )arg;
	int	ithread = para.ithread;
	int	nz = para.nz;
	float	DA1 = para.da1;
	float	DA2 = para.da2;
	float	DA3 = para.da3;
	float	alt0 = para.alt0,   az0 = para.az0,   phi0 = para.phi0;
	float	range1 = para.range1, 	range2 = para.range2, range3 = para.range3;
	bool timing = para.timing;
	ostream * log = para.log;
	EMData * tmp = model->copy();
	tmp->setTAlign(0, 0, 0);

	float	*dbest = best[ithread]->getData();
	float	*dalt = balt[ithread]->getData();
	float	*daz = baz[ithread]->getData();
	float	*dphi = bphi[ithread]->getData();

	float	r10, r11, r20, r21, r30, r31;
	float radius;

	if (2 * range1 < DA1) {
		r10 = 0; 
		r11 = DA1 / 8; 
	} else {
		r10 = range1; 
		r11 = range1 - DA1 / 8.0;
	}
	if (2 * range2 < DA2) {
		r20 = 0; 
		r21 = DA2 / 8; 
	} else {
		r20 = range2; 
		r21 = range2 - DA2 / 8.0;
	}
	if (2 * range3 < DA3) {
		r30 = 0; 
		r31 = DA3 / 8; 
	} else {
		r30 = range3; 
		r31 = range3 - DA3 / 8.0;
	}

	/*
for (float alt=alt0-range1; alt<alt0+range1-DA/8.0; alt+=DA) {
	for (float az=az0-range2; az<az0+range2-DA/8.0; az+=DA) {
		for (float phi=phi0-range3;phi<phi0+range3-DA/8.0;phi+=DA) {
*/
	for (float alt = alt0 - r10; alt < alt0 + r11; alt += DA1) {
		for (float az = az0 - r20; az < az0 + r21; az += DA2) {
			for (float phi = phi0 - r30; phi < phi0 + r31; phi += DA3) {
				tmp->setRAlign(alt, az, phi);
				tmp->rotateAndTranslate();
				// now do a CCF
				EMData * ccf;
				if (usemcf){	
					ccf = map->calcMCF(tmp,1);
					ccf->multConst(1.0e9);
//					ccf->toCorner();
					}

				if (useflcf){
				  tmp->applyMask(radius,4);
				  ccf=map->calcFLCF(tmp,radius,4);	       
				  ccf->toCorner();
				}
				else{	
				  ccf = map->calcCCF(tmp,1);
//				  ccf->toCorner();
				  }


				// Now keep track of the best orientation at each voxel
				float	*d = ccf->getDataRO();
				int	index = 0;
				float	cmax = (-FLT_MAX);
				for (int k = 0; k < nz; k++) {
					for (int j = 0; j < nz; j++) {
						for (int i = 0; i < nz; i++) {
							if (d[index] > cmax)
								cmax = d[index];
							if (cmax > gmax)
								gmax = cmax;
							if (d[index] > dbest[index]) {
								dbest[index] = d[index];
								dalt[index] = alt;
								daz[index] = az;
								dphi[index] = phi;
							}
							index++;
						}
					}
				}
				ccf->doneData();

				if (!timing)
					(*log) << "thread " << ithread << ":\talt= " << alt * 180.0 / PI << "\taz= " << az
					    *180.0 / PI << "\tphi= " << phi * 180.0 / PI << "\tccf coeff= " << cmax << "\tcurrent max coeff= " <<
					    gmax << endl;
				delete ccf;
			}
		}
	}
	best[ithread]->doneData();
	balt[ithread]->doneData();
	baz[ithread]->doneData();
	bphi[ithread]->doneData();

	return 0;
}



int	main(int argc, char *argv[]) 
{
	int	nz, i, j, k, l, keepnum = 20, HLFW = 2, numcpu = 1;
	float	dx0, dy0, dz0;

	float	alt0 = PI / 2.0, az0 = PI, phi0 = PI, range1 = PI / 2.0, range2 = PI, range3 = PI;
	float	alt, az, phi, x, y, z, s, bm, bs;
	float	DA1 = -1, DA2 = -1, DA3 = -1;	//(10.0*PI/180.0);
	float	finestda1 = 1.0*PI / 180.0;
	float	finestda2 = 1.0*PI / 180.0;
	float	finestda3 = 1.0*PI / 180.0;
	float	mx0 = 0, my0 = 0, mz0 = 0, mr = 0, apix = 0;
	float	cx0 = 0, cy0 = 0, cz0 = 0, crmin = 0, crmax=0,  ch = 0;
	float	sx0 = 0, sy0 = 0, sz0 = 0, sr = 0, div = 5;
	float	res = -1;	// default to 8angstrom, won't be effective if da is set
	float rox=0, roy=0, roz=0;
	float radius;
	float dtrans=0;
	char *genpdb=NULL;

	char	*intfile = NULL;
	char	*logfile = NULL;
	char	*batfile = NULL;
	char	*expfile = NULL;	// if output binary file for an explorer module
	bool timing = true, smart = false, relorig=false;
	ostream * log = &cout;
	ofstream dummy;
	
	appinit(argc,argv);
	if (argc < 4) {
		appusage(argv[0]);
		exit(1);
	}
	
	if (argc > 4) {
		for (i = 4; i < argc; i++) {
			if (strncmp(argv[i], "int=", 4) == 0) 
				intfile = &argv[i][4];
			else if (strncmp(argv[i], "cpu=", 4) == 0) 
				numcpu = atoi(&argv[i][4]);
/*
			else if (strncmp(argv[i], "da=", 3) == 0) 
				assert((DA = atof(&argv[i][3]) * PI / 180.0) > 0.0);
*/
			else if (strncmp(argv[i], "da=", 3) == 0) {
				int	numda = sscanf(&argv[i][3], "%f,%f,%f", &DA1, &DA2, &DA3);
				DA1 = DA1 * PI / 180.0;	
				DA2 = DA2 * PI / 180.0;	
				DA3 = DA3 * PI / 180.0;
				if (numda != 3 && numda >= 1) {
					DA3 = DA2 = DA1;
				}
			} 				
			else if (strncmp(argv[i], "res=", 4) == 0) 
				res = atof(&argv[i][4]);
			else if (strncmp(argv[i], "keep=", 5) == 0) 
				keepnum = atoi(&argv[i][5]);
			else if (strncmp(argv[i], "peakwidth=", 10) == 0) 
				HLFW = atoi(&argv[i][10]) / 2;
			else if (strncmp(argv[i], "genpdb=", 6) == 0) genpdb=argv[i]+7;
			else if (strncmp(argv[i], "log=", 4) == 0) {
				logfile = &argv[i][4];
				dummy.open(logfile, ios::out);
				log = &dummy;
			} 
			else if (strncmp(argv[i], "startangle=", 11) == 0) {
				sscanf(&argv[i][11], "%f,%f,%f", &alt0, &az0, &phi0);
				alt0 *= PI / 180.0;	
				az0 *= PI / 180.0;	
				phi0 *= PI / 180.0;
			} 
			else if (strncmp(argv[i], "mask=", 5) == 0) {
				sscanf(&argv[i][5], "%f,%f,%f,%f", &mx0, &my0, &mz0, &mr);
			} 
			else if (strncmp(argv[i], "cyl_region=", 11) == 0) {
				sscanf(&argv[i][11], "%f,%f,%f,%f,%f,%f", &cx0, &cy0, &cz0, &crmin, &crmax, &ch);
			} 
			else if (strncmp(argv[i], "sphere_region=", 14) == 0) {
				sscanf(&argv[i][14], "%f,%f,%f,%f", &sx0, &sy0, &sz0, &sr);
			} 
			else if (strncmp(argv[i], "range=", 6) == 0) {
				int	num = sscanf(&argv[i][6], "%f,%f,%f", &range1, &range2, &range3);
				range1 = (range1 * PI / 180.0 / 2);	
				range2 = range2 * PI / 180.0 / 2;	
				range3 = range3 * PI / 180.0 / 2;
				if (num != 3 && num >= 1) {
					range3 = range2 = range1;
				}
			} 
			else if (strncmp(argv[i], "apix=", 5) == 0) 
				apix = atof(&argv[i][5]);
			else if (strncmp(argv[i], "relorig=", 8) == 0) {
				sscanf(&argv[i][8],"%f,%f,%f",&rox,&roy,&roz);
				relorig=true;
			}
			else if (strncmp(argv[i], "smart", 5) == 0) {
				smart = true;
			} 
			else if (strncmp(argv[i], "mcf", 5) == 0) {
				usemcf = true;
			}  
			else if (strncmp(argv[i], "flcf", 4) == 0) {
				sscanf(&argv[i][4], "%f", &radius);
				useflcf = true;
			}
/*			
			else if (strncmp(argv[i], "finestda", 8) == 0) {
				finestda = atof(&argv[i][8]) * PI / 180.0;
			} 
*/			
			else if (strncmp(argv[i], "finestda=", 9) == 0) {
				int	numfda = sscanf(&argv[i][6], "%f,%f,%f", &finestda1, &finestda2, &finestda3);
				finestda1 = finestda1 * PI / 180.0;	
				finestda2 = finestda2 * PI / 180.0;	
				finestda3 = finestda3 * PI / 180.0;
				if (numfda != 3 && numfda >= 1) {
					finestda3 = finestda2 = finestda1;
				}
			}			
			else if (strncmp(argv[i], "bat=", 4) == 0) 
				batfile = &argv[i][4];
			else if (strncmp(argv[i], "explorer=", 9) == 0)
				expfile = &argv[i][9];
			else if (strncmp(argv[i], "verbose", 7) == 0) 
				timing = false;
			else if (strncmp(argv[i], "dtrans=", 7) ==0){
				sscanf(&argv[i][7], "%f", &dtrans);
				}
			else { 
				cerr << argv[0] << ":\tunknown parameter " << argv[i] << '\n'; 
				exit(1); 
			}
		}
	}

	if(expfile && !relorig){
		cerr <<argv[0]<<": \"[explorer=<binary file>]\" requires \"[relorig=<x,y,z>]\" specified !\n";
		exit(2);
	} 	
	
	vector < float > fbcoeff(keepnum), fbdx(keepnum), fbdy(keepnum), fbdz(keepnum), fbalt(keepnum), fbaz(keepnum), fbphi(keepnum);
	
	if(apix<=0) {
		cerr<<argv[0]<<": \"apix\" is not set, default to 1 Angstrom/pixel. Input correct value if you know the correct value\n";
		apix=1;
	}
	// started reading in map and model
	
	map = new EMData;
	//if (map[0]->readMRC(argv[1])) { cerr << "Error reading image "<<argv[1]<<endl; exit(1); }
	if (map->readImage(argv[1], -1)) { 
		cerr << argv[0] << ":\terror reading image " << argv[1] << endl; 
		exit(2); 
	}
	map->normalize();
	nz = map->zSize();
	if (nz != map->xSize() || nz != map->ySize()) {
		cerr << argv[0] << ":\tthe map \"" << argv[1] << "\" MUST be in a cubic box (nx=ny=nz)\n";
		exit(3);
	}

	model = new EMData;
	if (model->readImage(argv[2], -1)) { 
		cerr << argv[0] << ":\terror reading image " << argv[2] << endl; 
		exit(4); 
	}
	model->normalize();
	nz = model->zSize();
	if (nz != model->xSize() || nz != model->ySize()) {
		cerr << argv[0] << ":\tthe map \"" << argv[2] << "\" MUST be in a cubic box (nx=ny=nz)\n";
		exit(4);
	}


	if (nz != map->zSize()) {
		cerr << "The map and model MUST be in cubic boxes of the SAME size\n";
		exit(5);
	}

	if (DA1 <= 0) {
		if (res <= 0) 
			res = 8;
		DA1 = 2 * asin(res / nz);
		printf("da is set to %.1f degree for resolution %f\n", DA1 * 180 / PI, res);
	}
	if (DA2 <= 0) {
		if (res <= 0) 
			res = 8;
		DA1 = 2 * asin(res / nz);
		printf("da is set to %.1f degree for resolution %f\n", DA1 * 180 / PI, res);
	}
	if (DA3 <= 0) {
		if (res <= 0) 
			res = 8;
		DA1 = 2 * asin(res / nz);
		printf("da is set to %.1f degree for resolution %f\n", DA1 * 180 / PI, res);
	}
	
	time_t t = time(NULL);

	// finished reading in map and model

	int	Ref = LOGbegin(argc, argv);
	LOG(Ref, argv[1], LOG_INFILE, NULL);

	//map->normalizeTo(model);	//make the signal at the same scale

	best.resize(numcpu);
	balt.resize(numcpu);
	baz.resize(numcpu);
	bphi.resize(numcpu);

	for (i = 0; i < numcpu; i++) {
		best[i] = new EMData;
		best[i]->setSize(nz, nz, nz);
		best[i]->zero();

		balt[i] = new EMData;
		balt[i]->setSize(nz, nz, nz);
		balt[i]->zero();

		baz[i] = new EMData;
		baz[i]->setSize(nz, nz, nz);
		baz[i]->zero();

		bphi[i] = new EMData;
		bphi[i]->setSize(nz, nz, nz);
		bphi[i]->zero();
	}

	int	round = 1;
	pthread_t * thrd = new pthread_t[numcpu];
	Para * para = new Para[numcpu];
	do {
		(*log) << "\nSearch round " << round << ":\torientation range: alt=(" << (alt0 - range1) * 180 / PI << "->"
		    << (alt0 + range1) * 180 / PI << ")\t"
		     << "az=(" << (az0 - range2) * 180 / PI << "->" << (az0 + range2) * 180 / PI << ")\t"
		     << "phi=(" << (phi0 - range3) * 180 / PI << "->" << (phi0 + range3) * 180 / PI << ")\t"
		     << "stepsize= " << DA1 * 180 / PI << "," << DA2 * 180 / PI << "," << DA3 * 180 / PI << " degrees" << endl;


		int	ithread;
		for (ithread = 0; ithread < numcpu; ithread++) {
			para[ithread].ithread = ithread;
			para[ithread].nz = nz;
			para[ithread].da1 = DA1;
			para[ithread].da2 = DA2;
			para[ithread].da3 = DA3;	
			para[ithread].timing = timing;
			para[ithread].log = log;

			if (2 * range1 > numcpu * DA1) {
				float	chunk = 2*range1 / numcpu;
				para[ithread].alt0 = alt0 - range1 + chunk * (ithread + 0.5);
				para[ithread].az0 = az0;
				para[ithread].phi0 = phi0;
				para[ithread].range1 = range1 / numcpu;
				para[ithread].range2 = range2;
				para[ithread].range3 = range3;
			} else if (2 * range2 > numcpu * DA2) {
				float	chunk = 2*range2 / numcpu;
				para[ithread].alt0 = alt0;
				para[ithread].az0 = az0 - range2 + chunk * (ithread + 0.5);
				para[ithread].phi0 = phi0;
				para[ithread].range1 = range1;
				para[ithread].range2 = range2 / numcpu;
				para[ithread].range3 = range3;
			} else if (2 * range3 > numcpu * DA3) {
				float	chunk = 2*range3 / numcpu;
				para[ithread].alt0 = alt0;
				para[ithread].az0 = az0;
				para[ithread].phi0 = phi0 - range3 + chunk * (ithread + 0.5);
				para[ithread].range1 = range1;
				para[ithread].range2 = range2;
				para[ithread].range3 = range3 / numcpu;
			}


			if ((pthread_create(&thrd[ithread], NULL, foldhunter_thread, &para[ithread])) != 0) {
				printf("Cannot creat thread %d\n", ithread);
				exit(-1);
			}
		}

		for (ithread = 0; ithread < numcpu; ithread++) 
			pthread_join(thrd[ithread], NULL); // wait until all threads finish

		// merge all into one best[0]

		if (numcpu > 1) {
			vector < float * > dbest(numcpu), dalt(numcpu), daz(numcpu), dphi(numcpu);
			for (i = 0; i < numcpu; i++) {
				dbest[i] = best[i]->getDataRO();
				dalt[i] = balt[i]->getDataRO();
				daz[i] = baz[i]->getDataRO();
				dphi[i] = bphi[i]->getDataRO();
			}
			for (k = 0; k < nz * nz * nz; k++) {
				for (i = 1; i < numcpu; i++) {
					if (dbest[i][k] > dbest[0][k]) {
						dbest[0][k] = dbest[i][k];
						dalt[0][k] = dalt[i][k];
						daz[0][k] = daz[i][k];
						dphi[0][k] = dphi[i][k];
					}
				}
			}
			for (i = 0; i < numcpu; i++) {
				best[i]->doneData();
				balt[i]->doneData();
				baz[i]->doneData();
				bphi[i]->doneData();
			}
		}

		// copy best[0] to others
		if (numcpu > 1) {
			for (i = 1; i < numcpu; i++) {
				delete best[i];	
				best[i] = best[0]->copy();
				delete balt[i];	
				balt[i] = balt[0]->copy();
				delete baz[i];	
				baz[i] = baz[0]->copy();
				delete bphi[i];	
				bphi[i] = bphi[0]->copy();
			}
		}


		float	*dbest = best[0]->getData();
		float	*dalt = balt[0]->getData();
		float	*daz = baz[0]->getData();
		float	*dphi = bphi[0]->getData();

		if (crmax > 0 && crmin>0 && crmax>crmin) {	// only consider the region in the cylinder shell
			cx0 += nz / 2;	
			cy0 += nz / 2;	
			cz0 += nz / 2;	// all relative to the center, to be consisitent to the translation output
			for (int k = 0; k < nz; k++) {
				float	tk = fabs(k-cz0);
				for (int j = 0; j < nz; j++) {
					float	tj = j-cy0;
					for (int i = 0; i < nz; i++) {
						float	ti = i-cx0;
						float radius2 = ti * ti + tj * tj;
						if ( radius2 < crmin * crmin || radius2 > crmax * crmax || tk > ch / 2.) {
							dbest[k*nz*nz+j*nz+i] = (-FLT_MAX);
						}
					}
				}
			}
		}

		if (sr > 0) {	// only consider the region in the sphere
			sx0 += nz / 2;	
			sy0 += nz / 2;	
			sz0 += nz / 2;	// all relative to the center, to be consisitent to the translation output
			for (int k = 0; k < nz; k++) {
				float	tk = k-sz0;
				for (int j = 0; j < nz; j++) {
					float	tj = j-sy0;
					for (int i = 0; i < nz; i++) {
						float	ti = i-sx0;
						if ((ti * ti + tj * tj + tk * tk) > sr * sr) {
							dbest[k*nz*nz+j*nz+i] = (-FLT_MAX);
						}
					}
				}
			}
		}

		if (dtrans > 0) {	// only consider delta translations relative to center
			for (int k = 0; k < nz; k++) {
				float	tk = k-(nz/2);
				for (int j = 0; j < nz; j++) {
					float	tj = j-(nz/2);
					for (int i = 0; i < nz; i++) {
						float	ti = i-(nz/2);
						if ((ti * ti + tj * tj + tk * tk) > dtrans * dtrans) {
							dbest[k*nz*nz+j*nz+i] = (-FLT_MAX);

						}
					}
				}
			}
		}

		if (mr > 0) {	//mask out regions that will be ignored, this just a shpere right now, might change to ecclipsoid later
			mx0 += nz / 2;	
			my0 += nz / 2;	
			mz0 += nz / 2;	// all relative to the center, to be consisitent to the translation output
			int	r = int(mr) + 1;
			for (int k = -r; k < r; k++) {
				float	tk = int(k + mz0)-mz0;
				int	tz = int(k + mz0) % nz;
				for (int j = -r; j < r; j++) {
					float	tj = int(j + my0)-my0;
					int	ty = int(j + my0) % nz;
					for (int i = -r; i < r; i++) {
						float	ti = int(i + mx0)-mx0;
						if ((ti * ti + tj * tj + tk * tk) <= mr * mr) {
							int	tx = int(i + mx0) % nz;
							dbest[tz*nz*nz+ty*nz+tx] = (-FLT_MAX);
						}
					}
				}
			}
		}

		/*
		if (cr > 0 || sr > 0) {	// limit the peak width inside the included region
			HLFW = HLFW < cr ? HLFW : (int)(cr+0.5);
			HLFW = HLFW < sr ? HLFW : (int)(sr+0.5);
		}
		*/

		int	maxindex;
		float	max;
		for (int num = 0; num < keepnum; num++) {
			max = (-FLT_MAX);
			for (i = 0; i < nz * nz * nz; i++) {
				if (dbest[i] > max) {
					max = dbest[i];
					maxindex = i;
				}
			}
			alt = dalt[maxindex];
			az = daz[maxindex];
			phi = dphi[maxindex];
			int	z, y, x;
			float	dx, dy, dz;
			float	fdx, fdy, fdz, norm;
			fdx = fdy = fdz = norm = 0.0;	// for gravity center

			z = maxindex / (nz * nz);
			y = (maxindex - z * nz * nz) / nz;
			x = maxindex - z * nz * nz - y * nz;
			dz = z - nz / 2;	
			dy = y - nz / 2;	
			dx = x - nz / 2;
			if (num == 0) {
				alt0 = alt;
				az0 = az;
				phi0 = phi;
				dx0 = dx;
				dy0 = dy;
				dz0 = dz;
			} //set the starting angles for next round

			for (int kk = z - HLFW; kk <= z + HLFW; kk++) {
				k = (kk + nz) % nz;
				for (int jj = y - HLFW; jj <= y + HLFW; jj++) {
					j = (jj + nz) % nz;
					for (int ii = x - HLFW; ii <= x + HLFW; ii++) {
						i = (ii + nz) % nz;
						int	index = k*nz*nz + j*nz + i;		// only the pixels around the peak that have the same orientation as the peak will be ignored
						if ( hypot3(ii-x,jj-y,kk-z)<=HLFW*HLFW && (dbest[index]>FLT_MIN && dbest[index]<FLT_MAX) && 
							 (dalt[maxindex] == dalt[index]) && (daz[maxindex] == daz[index]) &&
							 (dphi[maxindex] == dphi[index])) {
							norm += dbest[index];
							fdx += (ii - x) * dbest[index];	
							fdy += (jj - y) * dbest[index];	
							fdz += (kk - z) * dbest[index];	//gravity center of the peak
							dbest[index] = (-FLT_MAX);	// mask out the same peak voxels
						}
					}
				}
			}

			if(goodf(&norm) && norm){
				fdx /= norm;	
				fdy /= norm;	
				fdz /= norm;
				fdx += dx;
				fdy += dy;
				fdz += dz;
			}
			else {
				fdx = dx;	
				fdy = dy;	
				fdz = dz;
			}
			fbcoeff[num]=max;
			fbdx[num] = fdx;	
			fbdy[num] = fdy;	
			fbdz[num] = fdz;	
			fbalt[num] = alt;	
			fbaz[num] = az;	
			fbphi[num] = phi;

			fdx *= apix;	
			fdy *= apix;	
			fdz *= apix;	

			(*log) << "Solution " << num << ":\trotation = ( " << alt * 180.0 / PI << " , " << az * 180.0 / PI
			    << " , " << phi * 180.0 / PI << " )\ttranslation = ( " << fdx << " , " << fdy << " , " << fdz <<
			    " )\tcoeff = " << max << endl;

			if (genpdb) {
				char s[256];
				sprintf(s,"procpdb.py %s fh.%d.ent rot=%1.3f,%1.3f,%1.3f  posttrans=%1.2f,%1.2f,%1.2f",genpdb,num,alt*180.0/PI,az*180.0/PI,phi*180.0/PI,dx*apix,dy*apix,dz*apix);
				system(s);
				printf("%s\n",s);
			}
		}
		best[0]->doneData();
		balt[0]->doneData();
		baz[0]->doneData();
		bphi[0]->doneData();

		// set up for next round run
		if (2 * range1 > DA1)
			range1 = DA1 * 2.0 / 3.0;
		if (2 * range2 > DA2)
			range2 = DA2 * 2.0 / 3.0;
		if (2 * range3 > DA3)
			range3 = DA3 * 2.0 / 3.0;

		div = numcpu > 5 ? numcpu : 5;
		DA1 /= div;
		DA2 /= div;
		DA3 /= div;
		round++;
	} while (smart && (DA1 >= (finestda1 / div)) && (DA2 >= (finestda2 / div)) && (DA3 >= (finestda3 / div))); //make sure the last round has step size less than 1 degree

	t = time(NULL) - t; 	// time spent in the loop

	(*log) << "\nTotal time used:\t" << t << "\tseconds\n";

	if (batfile) {
		FILE * batfp = fopen(batfile, "wb");
		fprintf(batfp, "#!/bin/csh -fx\n");

		fprintf(batfp, "set alt=(");
		for (i = 0; i < fbdx.size(); i++)
			fprintf(batfp, " %.2f", fbalt[i] * 180 / PI);
		fprintf(batfp, " )\n");

		fprintf(batfp, "set az=(");
		for (i = 0; i < fbdx.size(); i++)
			fprintf(batfp, " %.2f", fbaz[i] * 180 / PI);
		fprintf(batfp, " )\n");

		fprintf(batfp, "set phi=(");
		for (i = 0; i < fbdx.size(); i++)
			fprintf(batfp, " %.2f", fbphi[i] * 180 / PI);
		fprintf(batfp, " )\n");

		fprintf(batfp, "set dx=(");
		for (i = 0; i < fbdx.size(); i++)
			fprintf(batfp, " %.2f", fbdx[i]);
		fprintf(batfp, " )\n");

		fprintf(batfp, "set dy=(");
		for (i = 0; i < fbdx.size(); i++)
			fprintf(batfp, " %.2f", fbdy[i]);
		fprintf(batfp, " )\n");

		fprintf(batfp, "set dz=(");
		for (i = 0; i < fbdx.size(); i++)
			fprintf(batfp, " %.2f", fbdz[i]);
		fprintf(batfp, " )\n");

		fprintf(batfp, "foreach i (");
		for (i = 0; i < fbdx.size(); i++)
			fprintf(batfp, " %d", i);
		fprintf(batfp, " )\n");
		fprintf(batfp, "\tproc3d %s tmp.mrc rot=$alt[$i],$az[$i],$phi[$i]\n", argv[2]);
		fprintf(batfp, "\tproc3d tmp.mrc %s.$i trans=$dx[$i],$dy[$i],$dz[$i]\nend\nrm tmp.mrc\n", argv[3]);
		fclose(batfp);
		string tmpstr = string("chmod +x ") + string(batfile);
		run((char *)tmpstr.c_str());
		printf("\nBatch commands saved to file \"%s\" for transforming with all soultions\n", batfile);
	}

	if(expfile){
		Match *solutions=new Match[keepnum];
		memset(solutions,0,sizeof(Match)*keepnum);
		
		Quater quat;
		for(i=0; i<keepnum; i++){
			Euler tmpe(fbalt[i],fbaz[i],fbphi[i]);
			quat.fromEuler(tmpe);
			solutions[i].qx=quat.nx;
			solutions[i].qy=quat.ny;
			solutions[i].qz=quat.nz;
			solutions[i].qa=quat.Q;
			solutions[i].tx=fbdx[i]*apix;
			solutions[i].ty=fbdy[i]*apix;
			solutions[i].tz=fbdz[i]*apix;
			solutions[i].c =fbcoeff[i];
		}
			
		FHH	fhheader;
		memset(&fhheader,0,sizeof(FHH));
		
		strcpy(fhheader.version,"#foldhunter V1 binary\n");        
		fhheader.n=keepnum;          //number of foldhunter matches 

    	if(strstr(argv[1],"/"))	strcpy(fhheader.fileRef,argv[1]); 
		else	sprintf(fhheader.fileRef,"%s/%s",getenv("PWD"),argv[1]);
		
    	if(strstr(argv[2],"/"))	strcpy(fhheader.fileProbe,argv[2]); 
		else	sprintf(fhheader.fileProbe,"%s/%s",getenv("PWD"),argv[2]);
		
    	if(!intfile)	strcpy(fhheader.fileCoeff,"");
		else if(strstr(intfile,"/"))	strcpy(fhheader.fileCoeff,intfile);
		else	sprintf(fhheader.fileCoeff,"%s/%s",getenv("PWD"),intfile);
		
		fhheader.apix=apix;
		
		if(relorig) {
			fhheader.rox=rox;	// should be in Angstrom
			fhheader.roy=roy;
			fhheader.roz=roz;
		}
		else {
			fhheader.rox=0;
			fhheader.roy=0;
			fhheader.roz=0;
		}
		
		fhheader.pcx=nz/2*apix;
		fhheader.pcy=nz/2*apix;
		fhheader.pcz=nz/2*apix;
			
		FILE* expfp = fopen(expfile,"wb");
		fwrite(&fhheader,sizeof(FHH),1,expfp);	//write the fixed header
		fwrite(solutions,sizeof(Match),keepnum,expfp);	// write the variable number of matches
		fclose(expfp);
	} 

	// output the best aligned  map
	EMData * tmp = model->copy();
	tmp->setTAlign(0, 0, 0);
	tmp->setRAlign(alt0, az0, phi0);
	tmp->rotateAndTranslate();
	EMData * out = tmp->copy();
	out->setTAlign(dx0, dy0, dz0);
	out->setRAlign(0.0, 0.0, 0.0);
	out->rotateAndTranslate();
	out->setPixel(map->Pixel());
	out->setXYZOrigin(map->getXorigin(), map->getYorigin(),  map->getZorigin());
	out->writeMRC(argv[3]);
	delete tmp;	
	delete out;

	if (intfile) {
		best[0]->setPixel(map->Pixel());
		best[0]->setXYZOrigin(map->getXorigin(), map->getYorigin(),  map->getZorigin());
		best[0]->writeMRC(intfile);
	}

	(*log) << "\nTo transform the model map to match the corresponding part in the reference map:\n";
	(*log) << "\t1:\trotate the model by the 3 rotation angles\n\t\t\te.g.\tproc3d model out1 rot=30,60,80\n\n";
	(*log) << "\t2:\ttranslate the output of step 1 by the 3 translations\n\t\t\te.g.\tproc3d out1 out trans=3,-4,5\n";


	delete map;	
	delete model;
	for (i = 0; i < numcpu; i++) {
		delete best[i]; 
		delete balt[i]; 
		delete baz[i]; 	
		delete bphi[i];
	}
	LOG(Ref, argv[3], LOG_OUTFILE, NULL);
	LOG(Ref, NULL, LOG_END, NULL);
}


