// make3dicos.C Copyright 7/21/00 Steve Ludtke
// This program takes a set of projections (classes) with euler angles and constructs
// a 3d model in fourier space using a gaussian interpolation algorithm.

//N make3dicos
//F This program builds 3D models from a set of particles with icosahedral symmetry
//T X
//1
//P <input file>	File containing class averages
//P [out=<outfile>]	Output 3D MRC file
//P [pad=<padsize>]	To reduce Fourier artifacts, the model is typically padded by ~25%
//P [proc=<nproc>]	Number of processors to use.
//U make3dicos classes.1.img out=threed.1.mrc pad=128

//D Experimental - works, but not perfectly. make3d is better (but much slower)

//D This program is used to build a 3d model from a set of projections/class
//E averages. The individual projections must already have assigned euler
//E angles.

//D Reconstructions are performed in Fourier space. This causes certain
//E artifacts to appear in the 3d model. These artifacts can be reduced
//E by padding the images in real space. 50% padding produces excellent
//E results in most cases, ie - an 80x80 image would be padded to 120x120.
//E This would be specified with 'pad=120'. 25% padding provides adequate
//E results in most cases. The other artifact is caused by interpolation
//E in Fourier space. 'make3d' supports 6 levels of interpolation. Type
//E 1 is 'nearest neighbor' interpolation and is very fast, but produces
//E significant artifacts. The default in most cases is type 2, which 
//E generally produces the best
//E all-around results. If you have a very small number of slices, you
//E might try mode=5. This effect can be observed qualitatively using the 
//E <a href=v4.html>v4</a> program.
//E This program allows you to interactively select the interpolation
//E mode used for generating forward projections of a 3d model. The artifacts
//E in v4 are more severe, but qualitatively similar to those produced
//E by Fourier reconstruction.

//D Finally, since Fourier reconstruction DOES produce some undesirable
//E artifacts, generally this is just used as a preliminary model. Once
//E this model has been generated, typically this model will be improved
//E using a real-space iterative technique. This is implimented in
//E <a href=make3diter.html>make3diter</a>.

#include <cstdio>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "Euler.h"
#include <cmath>
#include <cctype>
#include "VolFrag.h"

int main(int argc, char *argv[]) {
int pad=-1,proc=1;
int i,n,ng,toclip;
char *file = "threed.mrc";

appinit(argc,argv);

if (argc<2) { appusage(argv[0]); exit(1); }

n=fileCount(argv[1]);
if (n<3) appusage(argv[0]);

for (i=2; i<argc; i++) {	
	if (strncmp(argv[i],"out=",4)==0) file=&argv[i][4];
	else if (strncmp(argv[i],"pad=",4)==0) pad=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"proc=",5)==0) proc=atoi(&argv[i][5]);
	else { printf("Unknown argument '%s'\n",argv[i]); exit(1); }
}

if (proc>1) { error(ERR_WARN,"Multiprocessors disabled in make3dicos\n"); proc=1; }

EMData rx;
rx.readImage(argv[1],0);
toclip=rx.xSize();
if (pad==-1) pad=toclip;

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

const char *tmp=tmpnam2();
FILE *out = fopen(tmp,"wb");
if (!out) error("Cannot open temporary file !!!",ERR_EXIT);

for (i=0; i<proc; i++) {
	fprintf(out,"make3dicos2 %s m3ditmp.%0d pad=%d frac=%0d/%0d\n",argv[1],i,pad,i,proc);
}
fclose(out);

char s[MAXPATHLEN];
sprintf(s,"runpar proc=%0d file=%s",proc,tmp);
run(s);

VolFrag fragr,fragi,fragn,frag;
fragr.setSize(pad);
fragr.setupIcos();
fragr.zero();
fragi.setSize(pad);
fragi.setupIcos();
fragi.zero();
fragn.setSize(pad);
fragn.setupIcos();
fragn.zero();

for (i=0; i<proc; i++) {
	sprintf(s,"m3ditmp.%0d.r.frg",i);
	frag.readFile(s);
	fragr.addFrag(&frag);
	
	sprintf(s,"m3ditmp.%0d.i.frg",i);
	frag.readFile(s);
	fragi.addFrag(&frag);

	sprintf(s,"m3ditmp.%0d.n.frg",i);
	frag.readFile(s);
	fragn.addFrag(&frag);
}


EMData *data = new EMData,*datar;
data->setSize(pad+2,pad,pad);
data->setComplex(1);
data->setRI(1);
data->zero();
float *d = data->getData(),nm;

ng=0;
for (int z=0; z<pad; z++) {
	for (int y=0; y<pad; y++) {
		for (int x=0; x<pad; x++) {
			if (!fragn.validXYZ(x,y,z)) continue;
			nm=fragn.valueAt(x,y,z);
			if (nm==0) nm=1.0;
			d[x*2+y*data->xSize()+z*data->xSize()*data->ySize()]=fragr.valueAt(x,y,z)/nm;
			d[x*2+y*data->xSize()+z*data->xSize()*data->ySize()+1]=fragi.valueAt(x,y,z)/nm;
			ng++;
			//printf("%3d %3d %3d\t%f\t%f\t%f\n",x,y,z,fragr.valueAt(x,y,z),fragi.valueAt(x,y,z),nm);
		}
	}
}			
data->doneData();

printf("ng=%d\n",ng);

// Invert the transform
data->toCorner();
datar=data->doIFT();
if (toclip!=pad) datar=datar->clip((pad-toclip)/2,(pad-toclip)/2,(pad-toclip)/2,toclip,toclip,toclip);
data->gimmeFFT();
delete data;

datar->writeMRC("nosym.mrc");		// write the unsymmetrized file
LOG(Ref,"nosym.mrc",LOG_OUTFILE,NULL);

// apply icosahedral symmetry in real space (this should really be done 
// in Fourier space to avoid overlap problems)
data=new EMData;
data->setSize(datar->xSize(),datar->ySize(),datar->zSize());
data->zero();

int ns=0;
Euler a,e;
a.setAngle(0,0,0);
a.setSym("icos");
data->zero();
for (e=a.FirstSym(); e.valid(); e=a.NextSym(),ns++) {
	printf("Angle %f,%f,%f\n",e.alt()*180/PI,e.az()*180/PI,e.phi()*180/PI);
	datar->readMRC("nosym.mrc");	// This avoids needing a second copy in memory, which might be tight
	datar->setRAlign(e.alt(),e.az(),e.phi());
	datar->rotateAndTranslate();
	data->add(datar);
}
data->multConst(1.0/ns);

data->writeMRC(file);
LOG(Ref,file,LOG_OUTFILE,NULL);

LOG(Ref,NULL,LOG_END,NULL);

return 0;
}
