// make3dreal.C 6/00 Steve Ludtke
// This program takes a set of projections (classes) with euler angles and constructs
// a 3d model in real space with unfiltered back-projection. Intended to be
// combined with make3d results

//N make3dreal
//F This program builds a 3d model from a set of projections (class averages) using unfiltered back-projection
//T LR
//1
//P <input file>	File containing class averages
//P [out=<outfile>]	Output 3D MRC file
//P [noweight]	Normally the class averages are weighted by the number of raw particles used, this disables that
//P [sym=<c4,icos,etc>]	The symmetry to apply to the model 
//P [mask=<rad>]	real-space mask radius
//P [fftmerge=<fftmodel>]	Fourier model to merge with real model
//U make3dreal classes.1.img out=threed.1.mrc sym=c4 mask=42
//D This program is used to build a 3d model from a set of projections/class
//E averages. The individual projections must already have assigned euler
//E angles in their headers. This program builds a model using unfiltered
//E back-projection. It is intended to be combined with a Fourier model,
//E or some other filtering technique.
//D 'sym=<symmetry>' specifies the symmetry of the 3d model. If nothing is
//E specified, the model is assumed to have no symmetry. Currently, only
//E Cn, Dn, oct and icos symmetries are supported.

#include <cstdio>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "Euler.h"
#include <cmath>
#include <cctype>

int main(int argc, char *argv[]) {
int mask=-1,rsym=1,pad=0,nowt=0;
List *data,*dataf,*dataf2;
EMData *d,*d2,*out;
int i,j,k,g,imc=0,n,ny,ny2;
float norm,pr,thr,r;
float prm,prs,prn;
char *file="threed.mrc",*sym="c1",*resmap=NULL,*ffm=NULL;
Euler e,*a;

appinit(argc,argv);

if (argc<2) { appusage(argv[0]); exit(1); }

n=fileCount(argv[1]);
if (n<3) appusage(argv[0]);

for (i=2; i<argc; i++) {	
	if (strncmp(argv[i],"out=",4)==0) file=&argv[i][4];
	else if (strncmp(argv[i],"mask=",5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"noweight",8)==0) nowt=1;
	else if (strncmp(argv[i],"fftmerge=",9)==0) ffm=&argv[i][9];
	else if (strncmp(argv[i],"sym=",4)==0) {
		sym=&argv[i][4];
		Euler e;
		e.setSym(sym);
		if (e.getMaxSymEl()==0) { printf("Unsupported symmetry\n"); exit(1); }
	}
	else { printf("Unknown argument '%s'\n",argv[i]); exit(1); }
}

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

dataf=new List;
dataf2=new List;

// Read slices
printf("Read data\n");
data=readImages(argv[1],0,-1);
if (data->count()<3) {
	printf("Need at least 3 images!\n");
	exit(1);
}
if (nowt) {
	for (i=0; i<data->count(); i++) {
		d=(EMData *)data->objectAt(i);
		if (d->NImg()>=0) d->setNImg(1);
	}
}

// If there are projections in the input file, we can use them to get a
// decent relative normalization for each projection
for (i=1; i<data->count(); i++) {
	d=(EMData *)data->objectAt(i);
	d2=(EMData *)data->objectAt(i-1);
	if (d->NImg()>0 && d2->NImg()<0) d->normalizeTo(d2);
}

out=new EMData;
out->setSize(d->xSize(),d->xSize(),d->xSize());		// use whatever the last image is, all should be the same
out->zero();

for (i=0; i<data->count(); i++) {
	d=(EMData *)data->objectAt(i);
	if (d->NImg()<0) continue;
	out->insertSlice3DReal(d,d->NImg());
	printf("%d. %f\t%f\t%f\t%d\n",i,d->alt()*180.0/PI,d->az()*180.0/PI,d->phi()*180.0/PI,d->NImg());
}

// Apply symmetry if it was specified
if (sym) {
	Euler e,*a;
	EMData *d,*d2;
	int ns=0;

	d=out->copy();
	d->setParent(NULL);
	d2=d->copy();

	printf("Applying %s symmetry\n",sym);
	a=d->getEuler();
	a->setAngle(0,0,0);
	a->setSym(sym);
	out->zero();
	for (e=a->FirstSym(); e.valid(); e=a->NextSym(),ns++) {
		printf("Angle %f,%f,%f\n",e.alt()*180/PI,e.az()*180/PI,e.phi()*180/PI);
		d2->setRAlign(e.alt(),e.az(),e.phi());
		d2->rotateAndTranslate();
		out->add(d2);
	}
	out->multConst(1.0/ns);
	delete d;
	delete d2;
}

out->normalize();

// Apply a s filter to approximately correct for lack of weighted averaging in Fourier space
out->filter(0,0,10);

out->writeMRC(file);

// Now, if a Fourier reconstructed model was provided, merge it in
if (ffm) {
	printf("Merge models\n");
	EMData *out0,*f0,*f1;
	float *d0,*d1,r;
	ny=out->ySize();

	out0=new EMData;
	f0=new EMData;
	f1=new EMData;
	
	out0->readImage(ffm,0);
	f0=out0->doFFT();
	f0->toCorner();
//	f0->ri2ap();
	f1=out->doFFT();
	f1->toCorner();
//	f1->ri2ap();

	d0=f0->getData();
	d1=f1->getData();

	for (k=g=-ny/2; k<ny/2; k++) {
		for (j=-ny/2; j<ny/2; j++) {
			for (i=0; i<ny/2+1; i++,g+=2) {
				r=sqrt(hypot3(k,j,i));
				if (r==0||r>ny/2) { d0[g]=0; continue; }
				d0[g]=d0[g]*(r*2/ny)+d1[g]*(1.0-r*2/ny);
			}
		}
	}
	f0->doneData();
	f1->doneData();
	
	f0->toCorner();
	out=f0->doIFT();
	out->writeMRC("merged.mrc");


}

LOG(Ref,file,LOG_OUTFILE,NULL);

LOG(Ref,NULL,LOG_END,NULL);

return 0;
}
