// make3dnonorm.C 10/03 Steve Ludtke
// This program takes a set of projections (classes) with euler angles and constructs
// a 3d model in fourier space using a gaussian interpolation algorithm.

//N make3dnonorm
//F This program builds 3D models from a set of class averages using Fourier techniques
//T LR
//1
//P <input file>	File containing class averages
//P [out=<outfile>]	Output 3D MRC file
//P [noweight]	Normally the class averages are weighted by the number of raw particles used, this disables that
//P [sym=<c4,icos,etc>]	The symmetry to apply to the model
//P [pad=<padsize>]	To reduce Fourier artifacts, the model is typically padded by ~25%
//P [mode=<1-6>]	Specifies the interpolation size, 2 is the default
//P [mask=<rad>]	real-space mask radius
//P [norm]	Normalize the slices based on associated projections in a 'classes' style file
//P [resmap=<name>]	Generates a 'resolution map' in another 3D MRC file
//P [inorm]	This will use a special weighting scheme to compensate for poor SNR sampling on the unit sphere
//U make3d classes.1.img out=threed.1.mrc hard=25 sym=c4 pad=128 mask=42
//D This program is virtually identical to make3d except for 3 changes.
//E First, it has a reduced memory footprint because it doesn't keep
//E all of the particles/class-averages in memory. Second, it only normalizes
//E slices on request, and does so by comparing to associated class-averages.
//E Finally, there is no mechanism for excluding
//E 'bad' slices.
//D inorm can be used to compensate for cases where strange high resolution radial
//E artifacts occur. This is generally due to nonuniform orientation distributions
//E producing azimuthal variations in signal/noise ratio. It may have some odd noise
//E enchancing effects, though.


#include <cstdio>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "Euler.h"
#include <cmath>
#include <cctype>

int main(int argc, char *argv[]) {
int mode=2,mask=-1,rsym=1,pad=0,nowt=0,norm=0;
List *data,*dataf,*dataf2;
EMData *d,*d2,*dn=NULL,*f,*out,*out2;
int i,j,k,g,imc=0,n,ny,ny2,savenorm=0,inorm=0;
char *file="threed.mrc",*sym="c1",*resmap=NULL,*snrfile=NULL;
Euler e,*a;
float **SNR=NULL;

appinit(argc,argv);

if (argc<2) { appusage(argv[0]); exit(1); }

n=fileCount(argv[1]);
if (n<3) { printf("Input file too small ?!?!\n"); exit(1); }

for (i=2; i<argc; i++) {
	if (strncmp(argv[i],"out=",4)==0) file=&argv[i][4];
	else if (strncmp(argv[i],"mode=",5)==0) mode=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"norm",4)==0) norm=1;
	else if (strncmp(argv[i],"inorm",5)==0) inorm=1;
	else if (strncmp(argv[i],"mask=",5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"resmap=",7)==0) resmap=&argv[i][7];
	else if (strncmp(argv[i],"noweight",8)==0) nowt=1;
	else if (strncmp(argv[i],"sym=",4)==0) {
		sym=&argv[i][4];
		Euler e;
		e.setSym(sym);
		if (e.getMaxSymEl()==0) { printf("Unsupported symmetry\n"); exit(1); }
	}
	else if (strncmp(argv[i],"pad=",4)==0) pad=atoi(&argv[i][4]);
	else { printf("Unknown argument '%s'\n",argv[i]); exit(1); }
}

int nimg=fileCount(argv[1]);
if (nimg<3) error(ERR_EXIT,"At least 3 input images must be provided");

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

// The actual reconstruction loop
	imc=0;
	for (i=0; i<nimg; i++) {
		d=new EMData();
		d->readImage(argv[1],i);
		if (i==0) {
			ny=d->ySize();
			if (pad) ny2=pad; else ny2=ny;
			out=new EMData;
			out->setup4IS(ny2);
		}
		if (d->NImg()<=0) {
			if (dn) delete dn;
			dn=d;
			continue;
		}

		printf("%d/%d\t%d\t%5.1f  %5.1f  %5.1f",i,nimg,
			d->NImg(),d->alt()*180.0/PI,d->az()*180.0/PI,d->phi()*180.0/PI);

		if (norm) {
			if (dn) {
				float a,b;
				d->normalizeTo(dn,1,1,&a,&b);
				printf("\t%1.3f  %1.3f\n",a,b);
			}
			else printf("Can't normalize without projections\n");
		}
		else printf("\n");

		d->applyMask(mask,6);
		d->getEuler()->setSym(sym);

		d2=d->clip((ny-ny2)/2,(ny-ny2)/2,ny2,ny2);
		f=d2->doFFT();
		f->toCorner();
		d2->gimmeFFT();
		f->setRAlign(d->alt(),d->az(),d->phi());
		f->setNImg(d->NImg());

		if (d->NImg()>0) {
			imc+=d->NImg();
			a=d->getEuler();
			for (e=a->FirstSym(); e.valid(); e=a->NextSym()) {
				out->insertSlice(f,e.alt(),e.az(),e.phi(),mode,d->NImg());
			}
		}
		delete d;
		delete d2;
		delete f;
	}

out->doneSlice(inorm);
out->toCorner();
out2=out->doIFT();

if (pad) out2=out2->clip((ny2-ny)/2,(ny2-ny)/2,(ny2-ny)/2,ny,ny,ny);
out2->applyMask(mask,6);

out2->writeMRC(file);

LOG(Ref,NULL,LOG_END,NULL);

return 0;
}
