// matrixembed.C 10/19/2000  Steven Ludtke

//N matrixembed
//F This program will embed a mutual distance matrix generated by the 'matrix' option of classesbymra
//T X
//1
//P <input file>	Source file from classesbymra
//P <output file>	Destination file (ascii)
//P [euler=<path>]	Path to the original image file for retreiving euler info
//P [dim=<n>]	Dimensionality of the embedded space, generally 3
//P [start=<n>]	File containing starting positions
//P [hemisphere]	Restrict the final points to lie on the surface of a hemisphere
//P [fixq=<val>]	Output with a fixed quality factor rather than the calculated one
//P [refpos=<file>]	A set of fixed positions that the distance matrix should be referenced to.
//P [assigneuler]	In conjunction with the hemisphere option, this will assign eulers to each particle
//P [bestref=<n>]	This will find a set of n images which are as different as possible for use as a reference set.
//U matrixembed matrix.dat groel.loc
//D This is an experimental program for visualizing mutual distance matrices generated by
//E the 'matrix' option of classesbymra. It generates an ascii file with the best (almost)
//E possible representation of the distances in the input file in (usually) 3 dimensions.
//E It also outputs a mean stress value for each point in the embedded space.
//D This is basically correspondence analysis.
//D Normally the input matrix file is the result of 'classesbymra proj.hed proj.hed matrix', but
//E it is also possible to use this program to calculate positions referenced to a precalculated
//E embedded space (the result of matrixembed on a square matrix). That is, once an embedded space
//E has been determined, you can determine the embedded positions of individual particles using
//E a matrix produced by 'classesbymra start.hed proj.hed matrix'. In this case, use the 'refpos='
//E option to specify the output coordinates from the initial matrixembed run.

#include <cstdio>
#include <cmath>
//#include <ccmath.h>
#include "util.h"
#include <cstring>
#include "EMData.h"

float stepdown(EMData *mx, EMData **pos,float start=-1.);
void normvec(EMData *pos);
double sigma(EMData *mx, EMData *pos);
EMData *dsdp(EMData *mx, EMData *pos);
double scale(EMData *mx, EMData *pos);
void radjust(EMData *pos);
void distshell(EMData *pos,int flag=0);

int sphere=0;
float spherer=0.5;

int main(int argc,char *argv[]) {
int i,j,k,l,DIM=3,eulerize=0,bestref=0;
int nr,np;	// nr is number of references, np is number of classified points
char *euler=0,*start=0,*refpos=0;
float *eul=NULL;	// These are the per-image euler angles
float fixq=9999.0;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=3; i<argc; i++) {	
	if (strncmp(argv[i],"euler=",6)==0) euler=&argv[i][6];
	else if (strncmp(argv[i],"dim=",4)==0) DIM=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"fixq=",5)==0) fixq=atof(&argv[i][5]);
	else if (strncmp(argv[i],"start=",6)==0) start=&argv[i][6];
	else if (strncmp(argv[i],"refpos=",7)==0) refpos=&argv[i][7];
	else if (strncmp(argv[i],"hemisphere",10)==0) sphere=1;
	else if (strncmp(argv[i],"assigneuler",11)==0) eulerize=1;
	else if (strncmp(argv[i],"bestref=",8)==0) bestref=atoi(&argv[i][8]);
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

if (DIM<2) error(ERR_EXIT, "dim >= 2");

EMData mx,*pos;	// mx is the distance matrix, pos contains DIM x np positions

mx.readImage(argv[1],0);
nr=mx.xSize();
np=mx.ySize();

if (nr!=np && !refpos) error(ERR_EXIT,"refpos must be specified for non-square matrices");
if (bestref && !euler) error(ERR_EXIT,"euler must be specified for bestref to work properly");
if (bestref>np) bestref=np;

pos=new EMData;
pos->setSize(DIM,np);
pos->zero();

// read Eulers if present
if (euler) {
	eul=(float *)malloc(np*3*sizeof(float));
	EMData img;
	
	for (i=0; i<np; i++) {
		img.readImage(euler,i,1);		// read image header
		eul[i*3]=img.alt();
		eul[i*3+1]=img.az();
		eul[i*3+2]=img.phi();
	}
}	

// initialize positions
float *posd = pos->getData();

for (i=0; i<np*DIM; i++) posd[i]=grand(0,1.0);		// random positions with a gaussian distribution
pos->doneData();

// Now optimize
float step=1.0;
i=0;
while (step>.001) {
	step=stepdown(&mx,&pos);
	printf("%d. %f\t%f\n",i++,step,sigma(&mx,pos));
}

// If we're doing a spherical shell, adjust the radius and try again
if (sphere) {
	for (j=0; j<8; j++) {
		radjust(pos);
		distshell(pos,1);
		printf("New radius = %1.3f\n",spherer);
	
		step=stepdown(&mx,&pos,2.0);
	
		float step=1.0;
		i=0;
		while (step>.001) {
			step=stepdown(&mx,&pos);
			printf("%d. %f\t%f\n",i++,step,sigma(&mx,pos));
		}
	}
}

// Find the best (or almost the best) set of references from this data set, the 1st image is always included
short *flag = NULL;
if (bestref>0) {
	flag=(short *)calloc(2,np);
	EMData xf;
	
	unlink("best.hed");
	unlink("best.img");
	
	// First item in the 'best' file is always the first image (we always want the top view)
	// This means we might not always get the optimal set of references, just a good set
	flag[0]=1;
	xf.readImage(euler,0);
	xf.writeIMAGIC("best.hed",-1);
	
	// loop in i is for each output image
	for (i=1; i<bestref; i++) {
		float bdist=-1.0e30;
		int bn=-1;
		// j tests each image to see if it's the best
		for (j=0; j<np; j++) {
			if (flag[j]) continue;
//			float tdist=0;
			float tdist=1.0e30;
			// k generates a quality factor for the current image
			for (k=0; k<np; k++) {
				if (!flag[k]) continue;
//				if (mx.valueAt(k,j,0)<=0) tdist+=1000.0;
//				else tdist+=1.0/mx.valueAt(k,j,0);
//				else tdist+=exp(-mx.valueAt(k,j,0));
//				tdist+=exp(-SQR(mx.valueAt(k,j,0))*100.0);
				if (mx.valueAt(k,j,0)<tdist) tdist=mx.valueAt(k,j,0);
			}
			if (tdist>bdist) { bn=j; bdist=tdist; }
		}
		if (bn<0) error(ERR_EXIT,"No best image found. This shouldn't happen (ever).");
		xf.readImage(euler,bn);
		xf.writeIMAGIC("best.hed",-1);
		flag[bn]=i+1;
	}
	
}

// Write out the optimized results
posd=pos->getData();
FILE *out=fopen(argv[2],"wb");
fprintf(out,"#MTX\t%d\n",DIM);
for (i=0; i<np; i++) {
	for (j=0; j<DIM; j++) {
		fprintf(out,"%1.4f\t",posd[i*DIM+j]);
	}
	
	float te=0;
	for (k=0; k<nr; k++) {
		float s=0;
		for (l=0; l<DIM; l++) s+=SQR(posd[i*DIM+l]-posd[k*DIM+l]);
		s=sqrt(s);
		te+=SQR(s-mx.valueAt(i,k)); 
	}
	if (flag) { te=(float)flag[i]/bestref; if (te==0) te=2.0; }
	else te=sqrt(te);
	
	if (euler) fprintf(out,"%f\t%f\t%f\t%f\n",te,eul[i*3],eul[i*3+1],eul[i*3+2]);
	else fprintf(out,"%f\n",te);
}
fclose(out);

// now assign euler angles to each image and write a new file, designed for use with sphere
if (eulerize) {
	char *used = (char *) calloc(sizeof(char),nr);
	EMData im1,im2,*ima;

	for (i=0; i<np; i++) {
		// find the nearest unused image for alignment
		float d,dist=1.0e6,distu=1.0e6;
		int best= -1,bestu= -1;
		for (j=0; j<np; j++) {
			if (i==j) continue;
			d=sqrt(hypot3(posd[i*DIM]-posd[j*DIM],posd[i*DIM+1]-posd[j*DIM+1],posd[i*DIM+2]-posd[j*DIM+2]));
			if (d<dist && !used[j] && (i<10||j<i)) { dist=d; best=j; }
			if (d<distu) { distu=d; bestu=j; }
		}
		if (best==-1) { printf("alignerr %d\n",i); best=bestu;}
		if (best<i) im2.readImage("euler.hed",best);
		else im2.readImage(euler,best);
		im1.readImage(euler,i);
		ima=im1.RTFAlign(&im2);
		ima->setRAlign((float)atan2((float)hypot(posd[i*DIM],posd[i*DIM+1]),posd[i*DIM+2]),atan2(posd[i*DIM+1],posd[i*DIM]),(float)0);
		ima->writeIMAGIC("euler.hed",i);
		delete ima;
		used[best]|=1;		//1 indicates an image that has been aligned to
	}
}

}

// This forces all points to lie on a hemispherical shell
// if flag=1, then data that doesn't cover the hemisphere will get stretched out so it does
void distshell(EMData *pos,int flag) {
float *pd = pos->getData();
int i,j;

// This will 'collapse' the data to z>0
float m=0;
if (flag) m=spherer-.0001;
for (i=0; i<pos->ySize(); i++) m=pd[i*pos->xSize()+2]<m?pd[i*pos->xSize()+2]:m;

m=spherer/(spherer-m);
for (i=0; i<pos->ySize(); i++) pd[i*pos->xSize()+2]=spherer+(pd[i*pos->xSize()+2]-spherer)*m;

// First, center the data in x/y
// find x,y center of mass
float xs=0,ys=0,t=0;
for (i=0; i<pos->ySize(); i++) {
	xs+=pd[i*pos->xSize()];
	ys+=pd[i*pos->xSize()+1];
	t+=1.0;	
}

xs/=t;
ys/=t;

// adjust center
for (i=0; i<pos->ySize(); i++) {
	pd[i*pos->xSize()]-=xs;
	pd[i*pos->xSize()+1]-=ys;
}

// Now we put each point on a shell
for (i=0; i<pos->ySize(); i++) {
	float l=0;
	if (pd[i*pos->xSize()+2]<0) pd[i*pos->xSize()+2]=0;
	for (j=0; j<pos->xSize(); j++) l+=SQR(pd[i*pos->xSize()+j]); 
	l=sqrt(l);
	for (j=0; j<pos->xSize(); j++) pd[i*pos->xSize()+j]/=(l/spherer);
}

pos->doneData();
}

// This will take one steepest descent step
float stepdown(EMData *mx, EMData **pos,float start) {
float cursig = sigma(mx,*pos);
static float lastst=2.0;

if (start>0) lastst=start;

EMData *pos2;
EMData *ds=dsdp(mx,*pos);
EMData *dsc=ds->copy();

//printf("%1.2f\n",cursig);
float st,sig,bsig=cursig,bst=0;
int l=0;
for (st=lastst*2.0,ds->multConst(lastst*2.0); st>.001; st*=.5,ds->multConst(.5)) {
	pos2=(*pos)->copy(0,0);
	pos2->add(ds);

/*	if (sphere && cursig<100.0) {
		float sr1=spherer,sr2=spherer*.98,sr3=spherer*1.02;
	
		spherer=sr1;
		distshell(pos2);
		sig=sigma(mx,pos2);

		spherer=sr2;
		distshell(pos2);
		float sig2=sigma(mx,pos2);

		spherer=sr3;
		distshell(pos2);
		float sig3=sigma(mx,pos2);
		
		if (sig<sig2 && sig<sig3) { spherer=sr1; distshell(pos2); } 
		else if (sig2<sig3) { spherer=sr2; sig=sig2; distshell(pos2); }
		else { spherer=sr3; sig=sig3; }
	}*/
	if (sphere) {
		distshell(pos2);
		sig=sigma(mx,pos2);
	}
	else sig=sigma(mx,pos2);

//	printf("\t%1.4f\t%1.2f\n",st,sig);
	if (sig<bsig) { bsig=sig; bst=st; l=1; }
	else if (l) break;
	delete pos2;
}
lastst=bst;
if (bsig>=cursig) { delete ds; delete dsc; return 0; }

// Now step the original position by the amount that gives the best sigma
pos2=(*pos)->copy(0,0);
dsc->multConst(bst);
pos2->add(dsc);
if (sphere) distshell(pos2);
delete *pos;
*pos=pos2;

delete ds;
delete dsc;
//scale(mx,pos2);
return bst;
}

// Normalizes a data set so sqrt(sum(xn^2))=1
void normvec(EMData *pos) {
float *d = pos->getData();
float n = 0;
int i;

for (i=0; i<pos->xSize()*pos->ySize(); i++) n+=d[i]*d[i];
n=sqrt(n);

for (i=0; i<pos->xSize()*pos->ySize(); i++) d[i]/=n;
pos->doneData();
}



// This calculates 'error' in the positions that could be corrected by overall scaling
// This routine requires mx to be square
double scale(EMData *mx, EMData *pos) {
int i,j,k;
double sigone,sigtot=0,dtot=0;

for (i=0; i<mx->xSize(); i++) {
	for (j=0; j<mx->xSize(); j++) {
		sigone=0;
		for (k=0; k<pos->xSize(); k++) sigone+=SQR(pos->valueAt(k,i)-pos->valueAt(k,j));
		sigtot+=(sqrt(sigone)-mx->valueAt(i,j));
		dtot+=sqrt(sigone);
	}
}
sigtot/=dtot;
printf("scale error = %lf\n",sigtot);
return sigtot;
}

// This calculates the 'quality' of a given set of positions vs a specified matrix
// This routine requires mx to be square
double sigma(EMData *mx, EMData *pos) {
int i,j,k;
double sigone,sigtot=0;

for (i=0; i<mx->xSize(); i++) {
	for (j=0; j<mx->xSize(); j++) {
		sigone=0;
		for (k=0; k<pos->xSize(); k++) sigone+=SQR(pos->valueAt(k,i)-pos->valueAt(k,j));
		sigtot+=SQR(sqrt(sigone)-mx->valueAt(i,j));
	}
}
sigtot=sqrt(sigtot);
return sigtot;
}

// x/y range based radius
void radjust(EMData *pos) {
float r=0,cr;

for (int i=0; i<pos->ySize(); i++) {
	cr=hypot(pos->valueAt(0,i),pos->valueAt(1,i));
	if (cr>r) r=cr;
}

spherer=r;
}

// This calculates the partial derivatives of each coordinate in a position matrix
// Again, this is for a square matrix only
EMData *dsdp(EMData *mx, EMData *pos) {
int p,i,j,k;
double sigone,sigtot,t,o;
EMData *ret;

ret=pos->copyHead();
ret->zero();

for (p=0; p<pos->ySize(); p++) {
	for (i=0; i<mx->xSize(); i++) {
		sigone=0;
		for (k=0; k<pos->xSize(); k++) sigone+=SQR(pos->valueAt(k,i)-pos->valueAt(k,p));
		if (sigone<=0) continue;
		sigone=-2.0*(mx->valueAt(i,p)/sqrt(sigone)-1.0);
		
//		printf("%1.2f\t",sigone);
		for (k=0; k<pos->xSize(); k++) {
//			if (i==0) o=0; else 
			o=ret->valueAt(k,p);
			ret->setValueAt(k,p,0,o+sigone*(pos->valueAt(k,i)-pos->valueAt(k,p)));
		}
	}
//	printf("\n");
}
normvec(ret);

return ret;
}

