// segment3d.C  4/3/00  Steve Ludtke
//N segment3d
//F Split a 3d map into several pieces.
//T LT
//1
//P <input file>	Source file
//P <output file>	Destination file
//P [watershed[=<lp>[,<hp>]]	Apply simplistic watershed segmentation instead of the default (k-means), with an optional pre-filter
//P [nseg=<n seg>]	Number of segments to split map into
//P [morph=<vector file>]	Takes the result of a previous segmentation and uses it to seed a new segmentation
//P [ampw]	Apply a straight line integral amplitude weight to the distance parameter
//P [thr=<val>]	Only segment voxels above the given threshold, neg values use sigma multipliers
//P [split]	Separate out the individual segments 
//P [maxit=<n>]	Specify maximum iterations before giving up on convergence
//P [itxp=<n>]	After the segmentation, iteratively extend the segments into regions below threshold, n is the number of voxels of expansion
//P [vector=<outfile>]	Write segment centers to a text file  x,y,z,dx,dy,dz
//P [apix=<apix>]	A/pix for chimera output
//P [chimeraout=<outfile.cmm>]	Write segment centers to a chimera marker file
//P [pdb=<outfile>]	Write segment centers to a PDB file
//P [gaussmod=<mrc out>]	Construct a 3d model from Gaussians centered at the group centers.
//P [randomize]	This will randomize the gaussians by rotating them randomly with a fixed radius
//P [stats]	Produce a file containing mean density values in each segment
//P [segmaxv]	Instead of an integer segmap, writes the 'max' value for the segment
//P [resegbystats=<thr>]	Merge segments with mean densities above or below the threshold value
//P [sym=<sym>]	Insure a symmetric segmentation
//U segment3d threed.1a.mrc seg.mrc 12 thr=1
//D This program will take a 3D map and that map into 'segments'.
//E That is, every voxel in the map will be assigned to one of <n seg> segments. The assignment
//E is performed in such a way that high density regions will be grouped together.
//D Note that using the 'watershed' option combined with 'nseg=' will not produce
//E a true 'watershed' segmentation of the image. Instead, it will produce nseg segments
//E then will 'flood fill' the remaining areas above the threshold.

#include <cstdio>
#include <cmath>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include <ctime>

struct SLIST { float v; int l; };

struct Segment { float x,y,z,s,v,sigx,sigy,sigz; };

int cmp(const void *x,const void *y) {
	float v=*((float *)x)-*((float *)y);
	if (v>0) return -1;
	if (v<0) return 1;
	return 0;
}

int main(int argc, char *argv[]) {
int i,j,k,l,x,y,z,rize=0,maxseg=100000,segmaxv=0;
int ampw=0,nseg=2,nseg1s=2,split=0,mit=999999,it=0,itxp=0,mode=0,stats=0,nsym=1;
float thr=-1.0e20,lp=0,hp=-10.0,sstatthr=0;
float h,bh,ra,apix=1.0;
float *ds,*dd;
char *gaussmod = NULL,*vector_file = NULL, *pdb=NULL,*chimout=NULL,*sym="C1";
char *morph=NULL;
EMData *data,*segmap;
Segment *seg,*oseg,*tseg;

appinit(argc,argv);

if (argc<4) appusage(argv[0]);


for (i=3; i<argc; i++) {
	if (strncmp(argv[i],"ampw",4)==0) ampw=1;
	else if (strncmp(argv[i],"nseg=",5)==0) nseg=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"morph=",6)==0) morph=argv[i]+6;
	else if (strncmp(argv[i],"thr=",4)==0) thr=atof(&argv[i][4]);
	else if (strncmp(argv[i],"split",5)==0) split=1;
	else if (strncmp(argv[i],"itxp=",5)==0) itxp=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"maxit=",6)==0) mit=atoi(&argv[i][6]);
	else if (strncmp(argv[i],"vector=",7)==0) vector_file=&argv[i][7];
	else if (strncmp(argv[i],"pdb=",4)==0) pdb=&argv[i][4];
	else if (strncmp(argv[i],"apix=",5)==0) apix=atof(&argv[i][5]);
	else if (strncmp(argv[i],"gaussmod=",9)==0) gaussmod=&argv[i][9];
	else if (strncmp(argv[i],"chimeraout=",11)==0) chimout=&argv[i][11];
	else if (strncmp(argv[i],"resegbystats=",13)==0) { sstatthr=atof(&argv[i][13]); if (!stats) stats=-1; }
	else if (strncmp(argv[i],"randomize",9)==0) rize=1;
	else if (strncmp(argv[i],"stats",5)==0) stats=1;
	else if (strncmp(argv[i],"segmaxv",7)==0) segmaxv=1;
	else if (strncmp(argv[i],"watershed",9)==0) {
		if (argv[i][9]=='=') sscanf(argv[i]+10,"%f , %f",&lp,&hp);
		mode=1;
	}
	else if (strncmp(argv[i],"sym=",4)==0) {
		sym=&argv[i][4];
		Euler e;
		e.setSym(sym);
		if (e.getMaxSymEl()==0) { printf("Unsupported symmetry\n"); exit(1); }
		nsym=e.getMaxSymEl();
	}
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

if (mode==1 && nseg!=2) { printf("Note: with watershed, nseg specifies a maximum value\n"); maxseg=nseg; }

//if (nsym>1 && morph) { printf("Morph incompatible with sym=\n"); exit(1); }

// This is fast enough we can just do it twice
if (morph) {
	FILE *old=fopen(morph,"r");
	if (!old) { printf("Cannot read morph file\n"); exit(1); }
	char s[256];
	float a;
	nseg=0;
	while ( fgets(s,255,old) && sscanf(s," %f %f %f",&a,&a,&a)) nseg++;
	fclose(old);
	nseg/=nsym;
}

nseg*=nsym;
int nsegal=nseg<1000?1000:nseg;
seg=(Segment *)calloc(nseg<1000?1000:nseg,sizeof(Segment));
oseg=(Segment *)calloc(nseg<1000?1000:nseg,sizeof(Segment));
nseg1s=nseg/nsym;

data=new EMData;
data->readImage(argv[1],-1);
if (lp!=0) data->filter(hp,lp,1);
if (thr<0) thr=-data->Sigma()*thr;	// thr < 0 is a sigma multiplier

segmap=new EMData;
segmap=data->copyHead();
segmap->zero();

dd=data->getDataRO();
ds=segmap->getData();

int nx=data->xSize();
int ny=data->ySize();
int nz=data->zSize();

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

// Watershed segmentation
if (mode==1) {
	if (thr<-1.0e19) thr=data->Mean()+data->Sigma()/2.0;
	int n=0;	// # voxels above threshold
	for (i=0; i<nx*ny*nz; i++) {
		if (dd[i]>=thr) n++;		// 2 passes, silly but...
		ds[i]=-1.0;
	}
	printf("%d/%d voxels to process\n",n,nx*ny*nz);

	SLIST *sorted=(SLIST *)malloc(sizeof(SLIST)*n);
	n=0;
	// we won't touch a 1-pixel border around the volume
	// Find all pixels we need to segment (above the threshold)
	for (k=1; k<nz-1; k++) {
		for (j=1; j<ny-1; j++) {
			for (i=1; i<nx-1; i++) {
				l=i+j*nx+k*nx*ny;
				if (dd[l]>=thr) { sorted[n].v=dd[l]; sorted[n].l=l; n++; }
			}
		}
	}

	qsort(sorted,n,sizeof(SLIST),cmp);
	for (i=0; i<3; i++) printf("%d\t%f\n",sorted[i].l,sorted[i].v);

	nseg=0;
	for (i=0; i<n; i++) {
		int l=sorted[i].l;
		for (z=-1; z<=1; z++) {
			for (y=-1; y<=1; y++) {
				for (x=-1; x<=1; x++) {
					if (ds[l+x+y*nx+z*nx*ny]>=0) ds[l]=ds[l+x+y*nx+z*nx*ny];
				}
			}
		}
		if (ds[l]<0) {
			if (nseg==maxseg) break;
			if (nseg>nsegal-2) {
				nsegal*=2;
				seg=(Segment *)realloc(seg,nsegal*sizeof(Segment));
			}
			ds[l]=nseg;
			seg[nseg].x=(l%(nx*ny))%nx;
			seg[nseg].y=(l%(nx*ny))/nx;
			seg[nseg].z=l/(nx*ny);
			seg[nseg].sigx=seg[nseg].sigy=seg[nseg].sigz=0;
			nseg++;
		}
	}

	// This implies that we hit the specified max # of segments, so we need to do some progressive flood-filling
	if (i<n) {
		int ch=1;
		while (ch) {
			ch=0;
			for (j=i; j<n; j++) {
				int l=sorted[j].l;
				if (ds[l]>=0) continue;
				if (ds[l-1]>=0) ds[l]=ds[l-1];
				else if (ds[l+1]>=0) ds[l]=ds[l+1];
				else if (ds[l-nx]>=0) ds[l]=ds[l-nx];
				else if (ds[l+nx]>=0) ds[l]=ds[l+nx];
				else if (ds[l-nx*ny]>=0) ds[l]=ds[l-nx*ny];
				else if (ds[l+nx*ny]>=0) ds[l]=ds[l+nx*ny];
				if (ds[l]>=0) ch=1;
			}
		}
	}
	data->doneData();
	printf("%d segments generated\n",nseg);
}
else {
// This is k-means

// This is fast enough we can just do it twice
if (morph) {
	FILE *old=fopen(morph,"r");
	if (!old) { printf("Cannot read morph file\n"); exit(1); }
	char s[256];
	float a,b,c;
	int i=0;
	while ( fgets(s,255,old) && sscanf(s," %f %f %f",&a,&b,&c)) {
		seg[i].x=a;
		seg[i].y=b;
		seg[i].z=c;
		i++;
	}
	fclose(old);
}
else {
	// Initially we start with random segment positions
	srandom(time(0));
	for (i=0; i<nseg1s; i++) {
		seg[i].x=random()%nx;
		seg[i].y=random()%ny;
		seg[i].z=random()%nz;
	}
}
// symmetrize the initial segments
if (nsym>1) {
	printf("Applying %d-fold symmetry\n",nsym);
	Euler e,a(0,0,0);
	a.setSym(sym);
	for (j=0,e=a.FirstSym(); e.valid(); j++,e=a.NextSym()) {
		float *mat=e.mat();

		for (i=0; i<nseg1s; i++) {
			seg[i+nseg1s*j].x=(seg[i].x-nx/2)*mat[0]+(seg[i].y-ny/2)*mat[1]+(seg[i].z-nz/2)*mat[2]+nx/2;
			seg[i+nseg1s*j].y=(seg[i].x-nx/2)*mat[3]+(seg[i].y-ny/2)*mat[4]+(seg[i].z-nz/2)*mat[5]+ny/2;
			seg[i+nseg1s*j].z=(seg[i].x-nx/2)*mat[6]+(seg[i].y-ny/2)*mat[7]+(seg[i].z-nz/2)*mat[8]+nz/2;
		}
	}
}

while (it<mit) {
	it++;
	printf("Voxel assignment (%d, %d, %d)\n",nseg,nsym,nseg1s);

	// Now we assign each voxel to a segment based on various distance criteria
	for (z=0,k=0; z<nz; z++) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx; x++,k++) {
				if (dd[k]<=thr) { ds[k]=-1.0; continue; }

				bh=nx*ny*nz;	//not important, just big
				// find which segment this voxel goes in
				for (i=0; i<nseg; i++) {
					if (ampw) {
						h=sqrt(hypot3(x-seg[i].x,y-seg[i].y,z-seg[i].z));
						float sum=0;
						for (float rr=0; rr<=1.0; rr+=1.0/h) 
							sum+=data->valueAt(x*rr+seg[i].x*(1.0-rr),y*rr+seg[i].y*(1.0-rr),z*rr+seg[i].z*(1.0-rr)); 
						h=h*h/sum;
						if (h<bh) { bh=h; j=i; }
						
					}
					else {
						h=hypot3(x-seg[i].x,y-seg[i].y,z-seg[i].z);
						if (h<bh) { bh=h; j=i; }
					}
				}
				ds[k]=j;
			}
		}
	}

	// now we redetermine the center locations
	tseg=seg;
	seg=oseg;
	oseg=tseg;
	memset(seg,0,nseg*sizeof(Segment));
	for (z=0,k=0; z<nz; z++) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx; x++,k++) {
				if (dd[k]>thr && ds[k]>=0) {
					seg[(int)ds[k]].x+=x*dd[k];
					seg[(int)ds[k]].y+=y*dd[k];
					seg[(int)ds[k]].z+=z*dd[k];
					seg[(int)ds[k]].s+=dd[k];	// normalization
					seg[(int)ds[k]].v+=1.0;		// count of voxels in segment
				}
			}
		}
	}			

	for (k=0; k<nseg; k++) {
		if (seg[k].s) {
			seg[k].x/=seg[k].s;
			seg[k].y/=seg[k].s;
			seg[k].z/=seg[k].s;
		}
		printf("%d. %f\t%f\t%f\t%d\n",k,seg[k].x,seg[k].y,seg[k].z,(int)seg[k].v);
		
		if (morph) {
			if (seg[k].v==0) {
				seg[k].x=oseg[k].x;
				seg[k].y=oseg[k].y;
				seg[k].z=oseg[k].z;
			}
		}
		else {
			if (seg[k].v==0) {
				seg[k].x=random()%nx;
				seg[k].y=random()%ny;
				seg[k].z=random()%nz;
			}
		}
	}

	// symmetrize based on the first symmetric subset
	if (nsym>1) {
		printf("Applying %d-fold symmetry\n",nsym);
		Euler e,a;
		a.setSym(sym);
		float bestsum=0;
		int bestj=0;
		for (j=0,e=a.FirstSym(); e.valid(); j++,e=a.NextSym()) {
			float *mat=e.mat();
	
			float sum=0;
			for (i=0; i<nseg1s; i++) {
				seg[i+nseg1s*j].x=(seg[i].x-nx/2)*mat[0]+(seg[i].y-ny/2)*mat[1]+(seg[i].z-nz/2)*mat[2]+nx/2;
				seg[i+nseg1s*j].y=(seg[i].x-nx/2)*mat[3]+(seg[i].y-ny/2)*mat[4]+(seg[i].z-nz/2)*mat[5]+ny/2;
				seg[i+nseg1s*j].z=(seg[i].x-nx/2)*mat[6]+(seg[i].y-ny/2)*mat[7]+(seg[i].z-nz/2)*mat[8]+nz/2;
				sum+=seg[i+nseg1s*j].x;
//				sum+=seg[i+nseg1s*j].x+seg[i+nseg1s*j].y+seg[i+nseg1s*j].z;
			}
			printf("sum=%f (%d)\n",sum,j);
			if (sum>bestsum) { bestsum=sum; bestj=j; }
		}
		// This tries to consistently pick as the first segment the one most in 1,1,1 direction
		// for comparison between runs
		if (bestj!=0) {
			for (i=0; i<nseg1s; i++) {
				float t=seg[i].x;
				seg[i].x=seg[i+bestj*nseg1s].x;
				seg[i+bestj*nseg1s].x=t;
				t=seg[i].y;
				seg[i].y=seg[i+bestj*nseg1s].y;
				seg[i+bestj*nseg1s].y=t;
				t=seg[i].z;
				seg[i].z=seg[i+bestj*nseg1s].z;
				seg[i+bestj*nseg1s].z=t;
			}
		}
//		for (k=0; k<nseg; k++) printf("%d. %f\t%f\t%f\t%d\n",k,seg[k].x,seg[k].y,seg[k].z,(int)seg[k].v);
	}

	ra=0;
	for (k=0; k<nseg; k++) {
		ra+=hypot3(seg[k].x-oseg[k].x,seg[k].y-oseg[k].y,seg[k].z-oseg[k].z);
	}
	ra/=(float)nseg;
	printf("%d. --------------   %f\n",it,ra);
	if (ra<.001) break;
}
data->doneData();
}

if (itxp) {
	float *smc=(float *)malloc(nx*ny*nz*sizeof(float));

	for (i=0; i<itxp; i++) {
		memcpy(smc,ds,nx*ny*nz*sizeof(float));
		for (z=1; z<nz-1; z++) {
			for (y=1; y<ny-1; y++) {
				for (x=1; x<nx-1; x++) {
					k=x+y*nx+z*nx*ny;
					if (ds[k]!=-1.0) continue;
					// if a neighbor has a segment number, use it, if ambiguous, use the largest one (arbitrary)
					ds[k]=smc[k-1];
					if (smc[k+1]>ds[k]) ds[k]=smc[k+1];
					if (smc[k-nx]>ds[k]) ds[k]=smc[k-nx];
					if (smc[k+nx]>ds[k]) ds[k]=smc[k+nx];
					if (smc[k-nx*ny]>ds[k]) ds[k]=smc[k-nx*ny];	
					if (smc[k+nx*ny]>ds[k]) ds[k]=smc[k+nx*ny];
				}
			}
		}
	}	
	free(smc);
}

if (stats) {
	float *nsv=(float *)calloc(nseg+1,sizeof(float));		// number of voxels in segment
	float *vval=(float *)calloc(nseg+1,sizeof(float));		// voxel values in segment
	float *max=(float *)calloc(nseg+1,sizeof(float));		// voxel values in segment
	float *map=(float *)calloc(nseg+1,sizeof(float));		// map each segment to a new one

	for (i=0; i<nx*ny*nz; i++) {
		nsv[(int)ds[i]+1]+=1.0;
		vval[(int)ds[i]+1]+=dd[i];
		if (max[(int)ds[i]+1]<dd[i]) max[(int)ds[i]+1]=dd[i];
	}

	if (stats==1) {
		FILE *out=fopen("seg3d.stats","wb");
		for (i=0; i<nseg; i++) {
			if (nsv[i]==0) nsv[i]=.000001;
			printf("%d. %0.0f voxels  mean den -> %1.4g    max den -> %1.4g\n",i,nsv[i],vval[i]/nsv[i],max[i]);
//			fprintf(out,"%1.4g\n",vval[i]/nsv[i]);
			fprintf(out,"%1.4g\n",max[i]);
		}
		fclose(out);
	}

	if (segmaxv) {
		for (i=0; i<nx*ny*nz; i++) {
			ds[i]=max[(int)ds[i]+1];
		}
	}

	// This will split the existing segments by their maximum value into 2 final segments
	if (sstatthr) {
		map[0]=-1.0;
		for (i=1; i<nseg; i++) {
			if (max[i]>sstatthr) map[i]=1.0; else map[i]=0.0;
		}

		for (i=0; i<nx*ny*nz; i++) {
			ds[i]=map[(int)ds[i]+1];
		}
		nseg=2;
	}
}

// This will write the individual segments into individual 3D data files
if (split) {
	int x0,x1,y0,y1,z0,z1;
	char fsp[100];

	for (i=0; i<nseg; i++) {
		EMData *cp = data->copy();
		dd=cp->getData();

		// find the limits of this segment and mask it out
		x0=y0=z0=nx+ny+nz;
		x1=y1=z1=0;
		for (z=0,k=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,k++) {
					if (ds[k]==i) {
						if (x<x0) x0=x;
						if (x>x1) x1=x;
						if (y<y0) y0=y;
						if (y>y1) y1=y;
						if (z<z0) z0=z;
						if (z>z1) z1=z;
					}
					else dd[k]=0;
				}
			}
		}
		cp->doneData();
		EMData *clip=cp->clip(x0,y0,z0,x1-x0+1,y1-y0+1,z1-z0+1);
		sprintf(fsp,"seg.%d.mrc",i);
		clip->writeMRC(fsp);
		delete clip;
		delete cp;

		printf("Wrote %s   %d x %d x %d\n",fsp,x1-x0+1,y1-y0+1,z1-z0+1);
	}

}


// now write the final segment map
segmap->doneData();

segmap->addConst(1.0);
segmap->writeMRC(argv[2]);
segmap->addConst(-1.0);
LOG(Ref,argv[2],LOG_OUTFILE,NULL);

// Write a text file with the vectors if requested
if (vector_file) {
	if (morph) {
		FILE *out = fopen(vector_file,"wb");
		if (out) {
			FILE *old=fopen(morph,"r");
			char s[256];
			float a,b,c;
						
			for (i=0; i<nseg; i++) {
				fgets(s,255,old);
				sscanf(s," %f %f %f",&a,&b,&c);
				fprintf(out,"%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\n",seg[i].x-data->xSize()/2,seg[i].y-data->ySize()/2,seg[i].z-data->zSize()/2,
					a-data->xSize()/2,b-data->ySize()/2,c-data->zSize()/2);
			}
			fclose(out);
		}	
	}
	else {
		FILE *out = fopen(vector_file,"wb");
		if (out) {
			for (i=0; i<nseg; i++) {
				fprintf(out,"%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\n",seg[i].x,seg[i].y,seg[i].z,seg[i].sigx,seg[i].sigy,seg[i].sigz);
			}
			fclose(out);
		}
	}
}

// Chimera output, this mode contains much of the important functionality for
// automatic segmentation of 8ish angstrom resolution models
if (chimout) {
	FILE *out=fopen(chimout,"wb");
	if (out) {

		// calculate a connectivity matrix indicating strength of connections between
		// pairs of centers
		printf("\nCheck connectivity\n");
		float *segdist=(float *)calloc(nseg*nseg,sizeof(float));
		for (z=1; z<segmap->zSize()-1; z++) {
			for (y=1; y<segmap->ySize()-1; y++) {
				for (x=1; x<segmap->xSize()-1; x++) {

					int va=segmap->valueAt(x,y,z);
					if (va>=0) {
						int xx,yy,zz;
						for (zz=z-1; zz<=z+1; zz++) {
							for (yy=y-1; yy<=y+1; yy++) {
								for (xx=x-1; xx<=x+1; xx++) {
									int va2=segmap->valueAt(xx,yy,zz);
									// This original option sums the product of 2 values along the surface where 2 segments meet
//									if (va2>=0 && va2!=va) segdist[va+va2*nseg]+=data->valueAt(x,y,z)*data->valueAt(xx,yy,zz);

									// This option takes the maximum value along the surface of intersection
									if (va2>=0 && va2!=va) {
										float tp=data->valueAt(x,y,z)*data->valueAt(xx,yy,zz);
										if (tp>segdist[va+va2*nseg]) segdist[va+va2*nseg]=tp;
									}
								}
							}
						}
					}

				}
			}
		}
		float maxc=0;
		for (i=0; i<nseg*nseg; i++) if (segdist[i]>maxc) maxc=segdist[i];
		printf("Max conn %f\n",maxc);

		// If symmetry, we want to try to identify 1 connected symmetric unit
		int *blist=(int *)calloc(nseg1s,sizeof(int));
		if (nsym>1 && !morph) {

			blist[0]=0;
			for (i=1; i<nseg1s; i++) {			// We end up with nseg1s good points from any symmetric unit
				float best=0;
				int bestj=0;
				for (j=1; j<nseg; j++) {		// These are the possible segments we may include
					float val=0;
					for (k=0; k<i; k++) {		// sum over all connectivities to point j
						val+=segdist[blist[k]+j*nseg];
						for (l=0; l<nsym; l++) if ((blist[k]+l*nseg1s)%nseg==j) break;		// exclude any already included element in any symmetric unit
						if (l!=nsym) break;
					}
					if (k<i) continue;
					if (val>best) { best=val; bestj=j; }
				}
				blist[i]=bestj;
				printf("%d\t%f\n",bestj,best);
			}
		}
		else {
			for (i=0; i<nseg1s; i++) blist[i]=i;
		}

/*		for (i=0; i<nseg; i++) {
			for (j=i+1; j<nseg; j++) {
				if (segdist[i+j*nseg]) fprintf(out,"<link id1=\"%d\" id2=\"%d\" r=\"1\" g=\"1\" b=\"1\" radius=\"%1.3f\"/>\n",i+1,j+1,segdist[i+j*nseg]/maxc);
			}
		}*/

		if (morph) {
			FILE *old=fopen(morph,"r");
			char s[256];
			float a,b,c;
			
			fprintf(out,"<marker_set name=\"%s\">\n",chimout);
			int ii,jj;
			for (ii=0; ii<nseg; ii++) {
				i=ii;
				fprintf(out,"<marker id=\"%d\" x=\"%1.3f\" y=\"%1.3f\" z=\"%1.3f\" radius=\"%0.2f\"/>\n",
					i*2,apix*(seg[i].x-data->xSize()/2),apix*(seg[i].y-data->ySize()/2),apix*(seg[i].z-data->zSize()/2),apix/3.0);
				fgets(s,255,old);
				sscanf(s," %f %f %f",&a,&b,&c);
				fprintf(out,"<marker id=\"%d\" x=\"%1.3f\" y=\"%1.3f\" z=\"%1.3f\" radius=\"%0.2f\"/>\n",
					i*2+1,apix*(a-data->xSize()/2),apix*(b-data->ySize()/2),apix*(c-data->zSize()/2),apix);
			}
			fclose(old);

			for (i=0; i<nseg; i++) {
				fprintf(out,"<link id1=\"%d\" id2=\"%d\" r=\"1\" g=\"1\" b=\"1\" radius=\"%1.3f\"/>\n",i*2,i*2+1,apix/3.0);
			}
		}
		else {
			
			// write center positions
			fprintf(out,"<marker_set name=\"%s\">\n",chimout);
			int ii,jj;
			for (ii=0; ii<nseg1s; ii++) {
				i=blist[ii];
				fprintf(out,"<marker id=\"%d\" x=\"%1.3f\" y=\"%1.3f\" z=\"%1.3f\" radius=\"0.2\"/>\n",
					i,apix*(seg[i].x-data->xSize()/2),apix*(seg[i].y-data->ySize()/2),apix*(seg[i].z-data->zSize()/2));
			}
	
			// write connection info
			for (ii=0; ii<nseg1s; ii++) {
				float m1=0,m2=0;
				int mj1,mj2;
				i=blist[ii];
				for (jj=0; jj<nseg1s; jj++) {
					j=blist[jj];
					if ((segdist[i+j*nseg])>m1) {
						mj2=mj1;
						m2=m1;
						mj1=j;
						m1=segdist[i+j*nseg];
					}
					else if ((segdist[i+j*nseg])>m2) {
						mj2=j;
						m2=segdist[i+j*nseg];
					}
				}
				if (m1>0) fprintf(out,"<link id1=\"%d\" id2=\"%d\" r=\".3\" g=\".3\" b=\".7\" radius=\"%1.3f\"/>\n",
					i,mj1,segdist[i+mj1*nseg]*1.5*apix/maxc);
				if (m2>0) fprintf(out,"<link id1=\"%d\" id2=\"%d\" r=\".3\" g=\".3\" b=\".7\" radius=\"%1.3f\"/>\n",
					i,mj2,segdist[i+mj2*nseg]*1.5*apix/maxc);
				//was m2>maxc/4
	//			segdist[mj1+i*nseg]=0;
				segdist[mj1+i*nseg]*=-1.0;
				segdist[mj2+i*nseg]*=-1.0;
			}
		}
		fprintf(out,"</marker_set>\n");
		fclose(out);
		
		// generate a mask for the single segment we have determined
		EMData *newmask=segmap->copy(0,0);
		float *d=newmask->getData();
		for (i=0; i<nx*ny*nz; i++) {
			for (j=0; j<nseg1s; j++) {
				if (d[i]==blist[j]) break;
			}
			if (j<nseg1s) d[i]=1.0;
			else d[i]=0.0;
		}
		newmask->doneData();
		newmask->writeMRC("symmask.mrc");
		free(segdist);
		free(blist);
	}
}

// Write a PDB file with the vectors if requested
if (pdb) {
	/* PDB file format
	 COLUMNS        DATA TYPE       FIELD         DEFINITION
	 ---------------------------------------------------------------------------------
	  1 -  6        Record name     "ATOM  "
	  7 - 11        Integer         serial        Atom serial number.
	 13 - 16        Atom            name          Atom name.
	 17             Character       altLoc        Alternate location indicator.
	 18 - 20        Residue name    resName       Residue name.
	 22             Character       chainID       Chain identifier.
	 23 - 26        Integer         resSeq        Residue sequence number.
	 27             AChar           iCode         Code for insertion of residues.
	 31 - 38        Real(8.3)       x             Orthogonal coordinates for X in Angstroms.
	 39 - 46        Real(8.3)       y             Orthogonal coordinates for Y in Angstroms.
	 47 - 54        Real(8.3)       z             Orthogonal coordinates for Z in Angstroms.
	 55 - 60        Real(6.2)       occupancy     Occupancy.
	 61 - 66        Real(6.2)       tempFactor    Temperature factor.
	 73 - 76        LString(4)      segID         Segment identifier, left-justified.
	 77 - 78        LString(2)      element       Element symbol, right-justified.
	 79 - 80        LString(2)      charge        Charge on the atom.
	 */
	FILE *out = fopen(pdb,"wb");
	if (out) {
		for (i=0; i<nseg; i++) {
			fprintf(out,"ATOM  %5d  CA  ALA  %4d    %8.3f%8.3f%8.3f  1.00%6.2f      S_00  0 \n",
					i+1,i+1,seg[i].x*apix,seg[i].y*apix,seg[i].z*apix,sqrt((seg[i].sigx+seg[i].sigy+seg[i].sigz)/(3.*seg[i].s)));
		}
		fclose(out);
	}
}

// write out a map constructed from gaussians at the center of each segment
// note that this destroys data
if (gaussmod) {
	//calculate std dev of each region
	dd=data->getDataRO();
	ds=segmap->getData();
	memset(seg,0,nseg*sizeof(Segment));
	for (z=0,k=0; z<nz; z++) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx; x++,k++) {
				if (dd[k]>thr && ds[k]>=0) {
					seg[(int)ds[k]].x+=x*dd[k];
					seg[(int)ds[k]].y+=y*dd[k];
					seg[(int)ds[k]].z+=z*dd[k];
					seg[(int)ds[k]].s+=dd[k];	// normalization
					seg[(int)ds[k]].v+=1.0;		// count of voxels in segment
					seg[(int)ds[k]].sigx+=x*x*dd[k];
					seg[(int)ds[k]].sigy+=y*y*dd[k];
					seg[(int)ds[k]].sigz+=z*z*dd[k];
				}
			}
		}
	}
	
	for (i=0; i<nseg; i++) {
		seg[i].x/=seg[i].s;
		seg[i].y/=seg[i].s;
		seg[i].z/=seg[i].s;
		seg[i].sigx=seg[i].sigx/seg[i].s-seg[i].x*seg[i].x;
		seg[i].sigy=seg[i].sigy/seg[i].s-seg[i].y*seg[i].y;
		seg[i].sigz=seg[i].sigz/seg[i].s-seg[i].z*seg[i].z;
		seg[i].s=(seg[i].sigx+seg[i].sigy+seg[i].sigz)/1.0;
	}
	data->doneData();
	segmap->doneData();

	if (rize) {
		for (i=0; i<nseg; i++) {
			float a1=frand(0,PI), a2=frand(0,2.0*PI);
			float mx[9] = {  cos(a2) , cos(a1)*sin(a2) , sin(a1)*sin(a2) ,
							-sin(a2) , cos(a1)*cos(a2) , sin(a1)*cos(a2) ,
							0 , -sin(a1) , cos(a1) };
			float tx,ty,tz;

			tx=(seg[i].x-nx/2)*mx[0]+(seg[i].y-ny/2)*mx[1]+(seg[i].z-nz/2)*mx[2]+nx/2;
			ty=(seg[i].x-nx/2)*mx[3]+(seg[i].y-ny/2)*mx[4]+(seg[i].z-nz/2)*mx[5]+ny/2;
			tz=(seg[i].x-nx/2)*mx[6]+(seg[i].y-ny/2)*mx[7]+(seg[i].z-nz/2)*mx[8]+nz/2;

			seg[i].x=tx;
			seg[i].y=ty;
			seg[i].z=tz;
		}
	}
	
	data->zero();
	dd=data->getDataRO();

	for (z=0,k=0; z<nz; z++) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx; x++,k++) {
				for (i=0; i<nseg; i++) {
					dd[k]+=exp(-hypot3(x-seg[i].x,y-seg[i].y,z-seg[i].z)/(2.0*seg[i].s));
				}
			}
		}
	}
	data->doneData();
	data->writeMRC(gaussmod);
}


LOG(Ref,NULL,LOG_END,NULL);

return 0;
}
