// snrboxer.C  8/18/03 Mike Wakin -- wakin@rice.edu

//N snrboxer
//F This program is used for offline boxing of micrographs (MRC format ONLY!)
//T LM
//
//1
//P [input=<filespec>]	Image to be boxed	(MRC format only!!!)
//P [dbbox=<filespec>]	Input box database
//P [imgbox=<filespec>]	Read box database from particle image set headers
//P [refimg=<filespec>]	Reference particle images for autoboxing
//P [auto[=<thr1>,<thr2>,<thr3>]]	Autobox with defined thresholds
//P [xform=<dx>,<dy>,<da>]	Rotate and translate the box database before output (see alignhuge)
//P [scale=<mult>]	Scale the box locations, useful if the micrograph has been resized
//P [newsize=<size>]	New box size for all boxes (center preserved)
//P [output=<filespec>]	Output file for boxed images
//P [ctfout=<filespec>]	File to put CTF curves into (if input is focal pair merged)
//P [dbout=<filespec>]	Output box database file
//P [invert]	Invert the particle image before output
//P [adjust=<filespec>]	If images were from a focal pair, this will try to center the particles properly
//P [insideonly]	This will not box particles that extend outside the micrograph
//U snrboxer input=jj0441.mrc output=0441.hed dbout=0441.box auto=0.005,1,6 refimg=refs.hed
//D This program does offline boxing of particles using using an
//E automatic boxing routine. (Please note that it will work only on
//E MRC images.) This program is meant as an alternative to the
//E batchboxer program, with slightly different thresholds: (1) the
//E first threshold uses a minimum estimated SNR to find all candidate
//E particles, (2) the second eliminates boxes with high residual noise
//E energy, and (3) the third eliminates boxes with high
//E spectral variation in the residual noise. The second and third have
//E similar purposes, but one may be more effective in particular
//E cases. It is strongly recommended that the user specify an accurate
//E set of references (in particular, accounting for the CTF if possible).
 

#include <cstdio>
#include <cmath>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include <iostream>
#include <cstdlib>

#define CLIP	768

#define DB(x) printf("%s\n",x);

using namespace std;

// class for dealing with box databases, the x,y coordinates are 
// stored internally as the center of the box, but returned and
// passed in as the lower left corner of the image
class BOXLIST {
	int *x,*y;
	int n,na;
	int sz;
	
public:
	BOXLIST();
	~BOXLIST();
	void empty();
	void append(int x,int y);
	int atx(int i);
	int aty(int i);
	void setxy(int i,int X,int Y);
	void setsize(int s);
	int size();
	int count();
	void xform(int dx,int dy,float da,int xsize,int ysize);
	void scale(float sca);
	void write(const char *file);
	void read(const char *file);
	void readimg(const char *file);
};

BOXLIST::BOXLIST() { 
	x=(int *)malloc(8*sizeof(int)); 
	y=(int *)malloc(8*sizeof(int));
	na=8;
	n=0;
}


BOXLIST::~BOXLIST() {
	free(x);
	free(y);
}

void BOXLIST::setxy(int i,int X, int Y) {
	if (i>=n) return;
	x[i]=X;
	y[i]=Y;
}

int BOXLIST::size() { return sz; }

void BOXLIST::setsize(int s) { sz=s; }

void BOXLIST::empty() { n=0; }

void BOXLIST::append(int X, int Y) { 
	x[n]=X+sz/2; 
	y[n]=Y+sz/2;
	n++; 
	if (n>=na) { 
		na*=2;
		x=(int *)realloc(x,na*sizeof(int));
		y=(int *)realloc(y,na*sizeof(int));
	}
}

int BOXLIST::atx(int i) { if (i<n&&i>=0) return x[i]-sz/2; return 0; }
int BOXLIST::aty(int i) { if (i<n&&i>=0) return y[i]-sz/2; return 0; }

int BOXLIST::count() { return n; }

void BOXLIST::write(const char *file) {
FILE *out = fopen(file,"wb");
if (!out) return;
for (int i=0; i<n ; i++) fprintf(out,"%d\t%d\t%d\t%d\n",atx(i),aty(i),sz,sz);
fclose(out);
}

void BOXLIST::read(const char *file) {
FILE *in = fopen(file,"rb");
if (!in) return;

char s[128];
int xx,yy;
while (fgets(s,127,in)) {
	if (sscanf(s," %d %d %d",&xx,&yy,&sz)!=3) continue;
	append(xx,yy);
}
fclose(in);

}

void BOXLIST::readimg(const char *file) {
int i,j=0,x0,y0,x1,y1;
char t[80];
EMData r;
int n=fileCount(file);

for (i=0; i<n; i++) {
	r.readImage(file,i,1);
	if (sscanf(r.Name()," %d,%d",&x0,&y0)!=2) {
		if (sscanf(r.Name(),"!$ %d,%s %d,%d",&j,t,&x0,&y0)!=4)  continue;
	}
	x1=r.xSize();
	y1=r.ySize();

	append(x0,y0);
}

sz=x1; 
}

void BOXLIST::xform(int dx,int dy,float da,int xs,int ys) {
float mx0=cos(da);
float mx1=sin(da);

for (int i=0; i<n; i++) {
	float tx= (x[i]-xs/2)*mx0+(y[i]-ys/2)*mx1-dx+(xs/2);
	float ty=-(x[i]-xs/2)*mx1+(y[i]-ys/2)*mx0-dy+(ys/2);
	x[i]=tx;
	y[i]=ty;
}

}

void BOXLIST::scale(float sca) {
for (int i=0; i<n; i++) {
	x[i]*=sca;
	y[i]*=sca;
}
sz*=sca;

}

BOXLIST boxes;
unsigned char *bestmap=NULL;
float *resvars;
float *kurts;
float *specvars;

void autoBox(char *micro,char *ref,float *thresh);
void boxesFrom(EMData *map,EMData *data,List &abptcl,int x0,int y0,int sx,int sy,float *thresh,int boxsize);
EMData *ptclCorr(EMData *data, List &refs,int xo,int y0,int boxsize);

int main(int argc,char *argv[]) {
char *input=NULL,*boxed=NULL,*dbout=NULL,*dbin=NULL,*imgin=NULL,*ctfout=NULL,*ref=NULL,*adjust=NULL;
float autothr[4] = { 0,0,0,1 };
float xfdx=0,xfdy=0,xfda=0,scale=1.0;
char refs[512];
int i,newsize=0,ionly=0;
int invert=0,nx,ny;

appinit(argc,argv);

if (argc<2) appusage(argv[0]);

for (i=1; i<argc; i++) {	
	if (strncmp(argv[i],"input=",6)==0) input=argv[i]+6;
	else if (strncmp(argv[i],"output=",7)==0) boxed=argv[i]+7; 
	else if (strncmp(argv[i],"ctfout=",7)==0) ctfout=argv[i]+7; 
	else if (strncmp(argv[i],"scale=",6)==0) scale=atof(argv[i]+6); 
	else if (strncmp(argv[i],"newsize=",8)==0) newsize=atoi(argv[i]+8); 
	else if (strncmp(argv[i],"dbout=",6)==0) dbout=argv[i]+6;
	else if (strncmp(argv[i],"dbbox=",6)==0) dbin=argv[i]+6;
	else if (strncmp(argv[i],"imgbox=",7)==0) imgin=argv[i]+7;
	else if (strncmp(argv[i],"auto=",5)==0) sscanf(argv[i]+5,"%f,%f,%f,%f",autothr,autothr+1,autothr+2,autothr+3);
	else if (strncmp(argv[i],"xform=",6)==0) sscanf(argv[i]+6,"%f,%f,%f",&xfdx,&xfdy,&xfda);
	else if (strncmp(argv[i],"invert",6)==0) invert=1;
	else if (strncmp(argv[i],"refimg=",7)==0) ref=argv[i]+7; 
	else if (strncmp(argv[i],"adjust=",7)==0) adjust=argv[i]+7;
	else if (strncmp(argv[i],"insideonly",10)==0) ionly=1;
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

 
int Ref=LOGbegin(argc,argv);

EMData *dt=new EMData;
dt->readImage(input,0,1);
nx=dt->xSize();
ny=dt->ySize();
delete dt;

boxes.empty();
if (dbin) { printf("Read boxes from %s\n",dbin); boxes.read(dbin); }
if (imgin) { printf("Extract location info from %s\n",imgin); boxes.read(imgin); }
if (autothr[0]||autothr[1]||autothr[2]) {
    //autobox micrographs
    //input: entire micrographs
    //ref:  reference image sets
    //autothr: thresholds
    autoBox(input,ref,autothr);
}

if (adjust && newsize) error(ERR_EXIT,"Must specify imgin with adjust, and newsize not allowed");

if (boxes.count()==0) error(ERR_EXIT,"No boxes to process");
printf("%d Boxes to process\n",boxes.count());

if (newsize>0) boxes.setsize(newsize);

if (xfdx||xfdy||xfda) {
	if (!input) error(ERR_EXIT,"Cannot perform transformation without an input image.");
	EMData tmp;
	tmp.readImage(input,0,1);
	boxes.xform(xfdx,xfdy,xfda,tmp.xSize(),tmp.ySize());
}
  if (scale!=1.0) boxes.scale(scale);


if (boxed) {
  if (!input) error(ERR_EXIT,"Cannot output boxed particles without an input image");
  printf("Boxing %s into %s\n",input,boxed);
  EMData tmp;
  tmp.readMRC(input,1);
  char s[120];
  int ni;
  strcpy(s,tmp.Name());
  for (i=0; i<boxes.count(); i++) {
    if (adjust) {
      EMData *tmp2;
      tmp.readImage(adjust,i);
      tmp.edgeNormalize();
      tmp2=tmp.clip(-boxes.size()/2,-boxes.size()/2,boxes.size()*2,boxes.size()*2);
      tmp2->filter(2,boxes.size()/10,1);
      tmp.readMRCArea(input,boxes.atx(i)-boxes.size()/2,boxes.aty(i)-boxes.size()/2,boxes.size()*2,boxes.size()*2);
      tmp.realFilter(63);
      tmp.edgeNormalize();
      tmp.filter(2,boxes.size()/10,1);
      tmp.transAlign(tmp2,0,0,boxes.size()/4);
      //			tmp.writeIMAGIC("tst.hed",-1);
      //			tmp2->writeIMAGIC("tst.hed",-1);
      boxes.setxy(i,boxes.atx(i)+tmp.Dx(),boxes.aty(i)+tmp.Dy());
      printf("%d)  %d,%d\n",i,(int)tmp.Dx(),(int)tmp.Dy());
      delete tmp2;
    }
    if (ionly && (boxes.atx(i)<0 || boxes.aty(i)<0 || 
		  boxes.atx(i)+boxes.size()>nx || boxes.aty(i)+boxes.size()>ny)) continue;
    tmp.readMRCArea(input,boxes.atx(i),boxes.aty(i),boxes.size(),boxes.size());
    if(invert) tmp.invert();
    char t[120];
    if (s[0]=='!' && s[1]=='$') {
      if (ctfout) {
	char nm[120];
	float apix;
	int u;
	sscanf(s,"!$ %d,%f,%s",&u,&apix,nm);
	EMData tmp2;
	if (i==0) {
	  tmp2.readImage(nm,u);
	  tmp2.writeIMAGIC(ctfout,-1);
	  ni=fileCount(ctfout)-1;
	}
	sprintf(t,"!$ %d,%1.5g,%s %d,%d",ni,apix,ctfout,boxes.atx(i),boxes.aty(i));
      }
      else sprintf(t,"%s %d,%d",s,boxes.atx(i),boxes.aty(i));
    }
    else sprintf(t,"%d,%d",boxes.atx(i),boxes.aty(i));
    tmp.setName(t);
    tmp.writeIMAGIC(boxed,-1);
  }
}

if (dbout) boxes.write(dbout);

LOG(Ref,NULL,LOG_END,NULL);

return 0;
}

void autoBox(char *micro,char *ref,float *thresh) {
EMData *d1,*d2,*data,*rf;
int nx,ny,i,j;
List *refs;
List abptcl;

// autoboxing option of batchboxer program, very similar to Boxer GUI version

cout<<"thresh[0]="<<thresh[0]<<" thresh[1]="<<thresh[1]<<" thresh[2]="<<thresh[2]<<" thresh[3]="<<thresh[3]<<endl;

//read the micrograph
data=new EMData;
data->readImage(micro,0);
data->normalize();
//data->readMRC(micro,1); 
 
//read ref projection(s) 
refs=readImages(ref,-1,-1); 
d1=(EMData *)refs->objectAt(0);
boxes.setsize(d1->ySize()); 
int boxsize=d1->xSize();

if (refs->count()<1) return;
if (refs->count()>250) error(ERR_EXIT,"FATAL ERROR: Max # references 250.");

abptcl.empty();
for (i=0; i<refs->count(); i++) {
  d1=(EMData *)refs->objectAt(i);
  d1->edgeNormalize();
  d2=d1->copy(0,0);
  d2->meanShrink(2); // originally medianShrink(2);
  d2->unum3=0;
  //	d2->normalize();
  abptcl.addObject(d2);
  delete d1;
}
delete refs;

// moving window
nx=(data->xSize()-boxsize*2)/(CLIP*2-boxsize)+1;
ny=(data->ySize()-boxsize*2)/(CLIP*2-boxsize)+1;

// Starting the main loop for cross correlation and particle extraction
// Both micrograph and refrences will be shrank by a factor of two in ptclCorr
// so take that into account when finding the particle coordinates in boxesFrom

for (i=0; i<ny; i++) {
  for (j=0; j<nx; j++) {
    
    d1=ptclCorr(data,abptcl,j*(CLIP*2-boxsize),i*(CLIP*2-boxsize),boxsize);
    boxesFrom(d1,data,abptcl,j*(CLIP*2-boxsize),i*(CLIP*2-boxsize),0,0,thresh,boxsize);
    
    printf("%d,%d (%d)\n",i,j,boxes.count());
    
    delete d1;
  }
}

}

void boxesFrom(EMData *map,EMData *data,List &abptcl,int x0,int y0,int sx,int sy,float *thresh,int boxsize) {
int i,j,k,l,c,n=0,mn;
float *d,*d2,da1,da2,bd;
EMData *dt,*dtc;

// This function will extract particle locations in the map returned from ptclCorr
// The map contained the peaks value of corr coeff
// keep in mind that the micrograph and refrences have been shrank by a factor of two in ptclCorr

if (!map) return;

//dtc= new EMData;
//dtc->readMRCArea(data->Path(),x0,y0,CLIP*2,CLIP*2);
dtc=data->clip(x0,y0,CLIP*2,CLIP*2);

if (sx<boxsize/4) sx=boxsize/4;
if (sy<boxsize/4) sy=boxsize/4;

mn=4*CLIP*CLIP*4/(boxsize*boxsize);

d=map->getData();
for (j=sy; j<CLIP-boxsize/4; j++) {
  for (i=sx; i<CLIP-boxsize/4; i++) {
    if (d[i+j*CLIP]>thresh[0]) {
      
      if (n>mn ||i*2+x0>data->xSize()-boxsize || j*2+y0>data->ySize()-boxsize) continue;
      dt=dtc->clip(i*2-boxsize/2,j*2-boxsize/2,boxsize,boxsize);
      
      // check for particles with large 'drawn' areas and eliminate them
      d2=dt->getDataRO();	
      c=0;
      for (l=0; l<boxsize*boxsize; l++) {
	if (d2[l]==0) c++;
      }
      dt->doneData();
      if (c>boxsize*2) {
	delete dt;
	continue;
      }			
      
      if (sqrt(resvars[i+j*CLIP])>thresh[1]) {
        //printf("Eliminating (VAR = %5.3f)\n",resvars[i+j*CLIP]);
        delete dt;
        continue;
      }
      if (specvars[i+j*CLIP]>thresh[2]) {
	//printf("Eliminating (WHITE = %5.3f)\n",specvars[i+j*CLIP]);
        delete dt;
	continue;
      }
      if (fabs(kurts[i+j*CLIP])>thresh[3]) {
	//printf("Eliminating (KURT = %5.3f)\n",kurts[i+j*CLIP]);
	delete dt;
	continue;
      }

      // ok, we've got a good particle
      boxes.append(i*2+x0-boxsize/2,j*2+y0-boxsize/2);
      delete dt;
      n++;
    }
  } 
}
delete dtc; 
map->doneData(); 
}


EMData *ptclCorr(EMData *data,List &abptcl,int xo,int yo,int boxsize) {
EMData *dc,*dd,*dn,*ccf,*rf,*orig,*thisbox,*tempbox,*res,*bestres,*brfft;
int i,j,ii,jj,k,x1,y1,xx,yy,xmax,ymax,cy,refindex,bestref;
float *dp,*ddp,t,mean0,var0,mean1,amp1,var1,bestamp,minvar1,tot,mean;
float *tempdata,*thisdata;
float numpixels = SQR(boxsize/2);
 
int SNRDEBUG = 0;

// This function will identify candidate particles, estimate the SNR on
// each candidate box, and compute some quantities to be compared against
// the thresholds.

// From this point forward, snrboxer.C and ptclCorrSNR in Control.cpp
// should be identical, except for areas marked by BATCH.

if (!bestmap) bestmap=(unsigned char *)malloc(CLIP*CLIP);
else bestmap=(unsigned char *)realloc(bestmap,CLIP*CLIP);

if (!resvars) resvars=(float *)malloc(CLIP*CLIP*sizeof(float));
else resvars=(float *)realloc(resvars,CLIP*CLIP*sizeof(float));

if (!kurts) kurts=(float *)malloc(CLIP*CLIP*sizeof(float));
else kurts=(float *)realloc(kurts,CLIP*CLIP*sizeof(float));

if (!specvars) specvars=(float *)malloc(CLIP*CLIP*sizeof(float));
else specvars=(float *)realloc(specvars,CLIP*CLIP*sizeof(float));

// a little padding to get rid of edge effects
dd=data->clip(xo,yo,CLIP*2,CLIP*2);

dd->meanShrink(2);

dc=dd->copy(0,0);
dn=dd->copy(0,0);
orig=dd->copy(0,0);
dd->zero();

// First attempt to account for background drift
dn->filter(-5,CLIP/50.0,1);
dc->subtract(dn);

for (i=0; i<abptcl.count(); i++) {
  // BATCH
    rf = ((EMData *)abptcl.objectAt(i))->copy(0,0);
    rf = rf->clip(-(CLIP-boxsize/2)/2,-(CLIP-boxsize/2)/2,CLIP,CLIP);
  // NON-BATCH
    //rf = abptcl.at(i)->copy(0,0);
    //rf = rf->clip(-(CLIP-boxsize/2)/2,-(CLIP-boxsize/2)/2,CLIP,CLIP);

  rf->multConst(1.0/rf->Sigma());

  rf->doneData();
  rf->update();

  ccf=dc->calcCCF(rf,1);
  
  dp=ccf->getDataRO();
  ddp=dd->getData();
  for (j=0; j<CLIP*CLIP; j++) {
    if (dp[j]>ddp[j]) { ddp[j]=dp[j]; bestmap[j]=i; }
  }
  ccf->doneData();
  dd->doneData();
  rf ->doneData();
  
  delete rf;
  delete ccf;
}

// low pass filter so peaks are separated from each other
dd->filter(0,CLIP/6.0,5);
dd->doneData();
dd->update();
t=dd->Max();

dp=dd->getData();
 
//locate peaks
for(i=boxsize/4; i<CLIP-boxsize/4-1; i++) {
  for (j=boxsize/4; j<CLIP-boxsize/4-1; j++) {
    k=i+j*CLIP;
    if (dp[k]>dp[k-1] && dp[k]>dp[k+1] && dp[k]>dp[k+CLIP] && dp[k]>dp[k-CLIP] && dp[k]>0 &&
	dp[k]>dp[k-1-CLIP] && dp[k]>dp[k+1-CLIP] && dp[k]>dp[k-1+CLIP] && dp[k]>dp[k+1+CLIP]) {
      dp[k]+=t;
    }
  }
}

// eliminate boxes which are too close to each other
for(i=boxsize/4; i<CLIP-boxsize/4-1; i++) {
  for (j=boxsize/4; j<CLIP-boxsize/4-1; j++) {
    k=i+j*CLIP;
    if (dp[k]>t) {
      x1=i+boxsize/4-1;
      y1=j+boxsize/4-1;
      for (ii=i-boxsize/4; ii<x1; ii++) {
	for (jj=j-boxsize/4; jj<y1; jj++) {
	  if (dp[ii+jj*CLIP]>dp[k]) { dp[k]=0; break; }
	}
      }    
    }
  }
}

// eliminate everything but the peaks
for (i=1; i<CLIP-1; i++) {
  for (j=1; j<CLIP-1; j++) {
    k=i+j*CLIP;
    if (dp[k]<=t) dp[k]=0; 
    else {
      dp[k]-=t;
      dp[k]*=100.0;
    }
  }
}

dd->doneData();
dp=dd->getData();
 
// Estimate SNR for candidate particles
for(i=boxsize/4; i<CLIP-boxsize/4-1; i++) {
  for (j=boxsize/4; j<CLIP-boxsize/4-1; j++) {
    k=i+j*CLIP;
    if (dp[k]>0) { 
      // Get this box
      thisbox = orig->clip(i-boxsize/4,j-boxsize/4,boxsize/2,boxsize/2);
      
      if ((thisbox->Max() == 0) && (thisbox->Min() == 0)) {
	// No data here -- do nothing.
	dp[k] = -1;
	bestmap[k] = 0;
	resvars[k] = 0;
	kurts[k] = 0;
	specvars[k] = 0;
      } else {
	if (SNRDEBUG) printf("Got thisbox i=%d, j=%d... ",i,j);

	// Suppose there is only noise (Hypothesis 0). Estimate mean
	// and variance.
	mean0 = thisbox->Mean();
	var0 = SQR(thisbox->Sigma());
	if (SNRDEBUG) printf("mean0 = %4.3f, var0 = %4.3f, ",mean0,var0);
	
	// Suppose there is particle + noise (Hypothesis 1).  For each
	// reference, find the corresponding amplitude, mean and
	// variance. Keep the reference that minimizes residual noise
	// energy.
	minvar1 = 1e20;
	for (refindex=0; refindex<abptcl.count(); refindex++) {
	  // BATCH 
	    rf = ((EMData *)abptcl.objectAt(refindex))->copy(0,0);
	  // NON-BATCH
	    // rf = abptcl.at(refindex)->copy(0,0);	
	  
	  // Estimate mean and amplitude; compute variance of residual   
          tempbox = rf->copy(0,0);
          tempdata=tempbox->getDataRO();
          thisdata=thisbox->getDataRO();
	  // Careful with last argument to lsqfit (when 1, excludes
	  // zeros in fitting).
          lsqfit(numpixels,tempdata,thisdata,&amp1,&mean1,1);
          tempbox->doneData();
          thisbox->doneData();
          delete tempbox;
          tempbox = rf->copy(0,0);
          tempbox->multConst(-amp1);
          tempbox->addConst(-mean1);
          tempbox->add(thisbox);
          res = tempbox->copy(0,0);
          var1 = SQR(tempbox->Sigma());

	  if (var1 < minvar1) {
	    minvar1 = var1;
	    bestref = refindex;
	    bestres = res->copy(0,0);
	    bestamp = amp1;
	  }

	  delete tempbox;
	  rf->doneData();
	  delete rf;
	  delete res;
	}

	if (bestamp > 0) {
	  // Let's only consider particles that seem to be positive

	  // Now, estimate SNR. This quantity should be invariant to
	  // scales and mean shifts of the observations.  (Note: the
	  // proper criteria to use in distinguishing white noise from
	  // particle would be dp[k] = var0/minvar1, but this is very
	  // similar.) We divide by 4.0 to account for the downsampling.
	  dp[k] = SQR(bestamp)/minvar1/4.0; 
	  resvars[k] = minvar1;
	  kurts[k] = bestres->Kurtosis();
	  bestmap[k] = bestref;
	    
	  // Want to analyze the residual image to eliminate junk.
	  // We normalize by var0 so that the residual image is 
	  // invariant to scales and mean shifts of the observations.
	  // This is not the same as normalizing by the variance of the 
	  // residual -- some will have more energy than others.
	  bestres->multConst(1.0/sqrt(var0)); 

	  // We would like the residual to be noise only (no junk).
	  // We've already accounted for noise energy with the SNR
	  // criteria; we need to check whether the block contains
	  // junk. For this criteria, we look at the FFT of the
	  // residual at low frequencies. There should only be small
	  // variation of energy at these frequencies. We compute the
	  // variation and call it the spectral variation of the
	  // residual. The lower the value, the more the residual
	  // resembles white noise. (Typically, pure white noise will have
	  // a value of 2 or lower.)
	  brfft = bestres->doFFT();
	  bestres->gimmeFFT();
	  brfft->ri2ap();
	  ymax = 4;
	  xmax = ymax*2;
	  cy = boxsize/4;
	  tot = 0;
	  for (xx=0;xx<xmax;xx+=2) {
	    for (yy=(cy-ymax);yy<=(cy+ymax);yy++) {
	      tot += brfft->valueAt(xx,yy);
	    }
	  }
	  mean = tot/(2.0*ymax+1.0)/(xmax/2.0);
	  tot = 0; 
	  for (xx=0;xx<xmax;xx+=2) {
	    for (yy=(cy-ymax);yy<=(cy+ymax);yy++) {
	      tot += SQR(brfft->valueAt(xx,yy)-mean);
	    }
	  }
	  specvars[k] = 10000*tot/(2.0*ymax+1.0)/(xmax/2.0);

	  if (SNRDEBUG) printf("minvar1 = %4.3f, bestamp = %4.3f, refindex = %d, SNR = %4.3f, specvars = %4.3f\n",minvar1,bestamp,bestmap[k],dp[k],specvars[k]);

	  delete brfft;

	} else {
	  // bestamp was negative -- not a likely match
	  dp[k] = -1;
	  bestmap[k] = 0;
	  resvars[k] = 0;
	  kurts[k] = 0;
	  specvars[k] = 0;

	  if (SNRDEBUG) printf("NEG AMP.\n");
	}

	delete thisbox;
	delete bestres;

      }    
    }
  }
}

dd->doneData();

delete dc; 
delete dn;
delete orig;
return dd; 
}

