// EMData.C - Steve Ludtke  6/97
// This is the heart of the EMAN project. It does just about all of the
// hard stuff. It reads and writes EM images in various formats, does FFTs,
// filters, etc...
//
// $Id: EMDataB.C,v 1.245 2008/07/15 18:34:55 stevel Exp $
//
#ifndef _WIN32
 #include <unistd.h>
#else
 #include "UnixToWin.h"
#endif

#define OPTIMIZEMEM

#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include "List.h"
#include "mrc.h"
#include "imagic.h"
#include "emim.h"
#include "util.h"
#include "EMData.h"
#include "VolTable.h"

// defined from 0 to 5 (pixels) in .001 steps
extern float HYPERGEOM[];

// linear interpolation of HYPERGEOM
float hyperg(float v) {
float r;
int a;

//if (v<0||v>=4.999) { printf("Waring: hypergeom %f outside range\n",v); return(0); }
if (v<0 || v>4.998) return 0;

r=v/.001;
a=(int)floor(r);
r-=a;

//printf("%d %f\n",a,HYPERGEOM[a]*(1.0-r)+HYPERGEOM[a+1]*r);

return -(HYPERGEOM[a]*(1.0-r)+HYPERGEOM[a+1]*r);

}

// This is used to define the width of the gaussian for 5x5x5 interpolation
// in 3d slices
#define I5G	(10.4/(PI*PI))	// used for 5x5x5 interpolation
#define I4G	(8.8/(PI*PI))	// used for 4 interpolation
#define I3G	(6.4/(PI*PI))	// used for 3 and 5x5x5 interpolation
#define I2G	(4.0/(PI*PI))	// 2 interpolation


//#define DEBUG

// subtract 2 angles (radians)
inline float asub(float x,float y) {
float r;
r=fmod(fabs(x-y),(float)(2*PI));
if (r>PI) r=2.0*PI-r;
return r;
}

// subtract 2 angles (radians)
inline float asub2(float x,float y) {
float r;

r=fmod(fabs(x-y),(float)PI);
if (r>PI/2.0) r=PI-r;
return r;
}

// This does fast integer-only translation
void EMData::fastTranslate(int inplace) {
float *d;
float *dd;
int x0,x1,x2,y0,y1,y2;

if (parent&&(!inplace)) {
	d=parent->getDataRO();
	dd=getData();
}
else {
	d=dd=getData();
}

if (nz==1) {
	if (dx<0) { x0=0; x1=xSize(); x2=1; }
	else { x0=xSize()-1; x1= -1; x2=-1; }
	if (dy<0) { y0=0; y1=ySize(); y2=1; }
	else { y0=ySize()-1; y1=-1; y2=-1; }
	
	for (int x=x0; x!=x1; x+=x2) {
		for (int y=y0; y!=y1; y+=y2) {
			int xp=x-dx;
			int yp=y-dy;
			if (xp<0||yp<0||xp>=xSize()||yp>=ySize()) { dd[x+y*xSize()]=0; continue; }
			dd[x+y*xSize()]=d[xp+yp*xSize()];
		}
	}
	doneData();
	if (d==dd) dx=dy=0;
	else parent->doneData();
}
else {
	error(ERR_ERROR,"Sorry, 3d translateFast not supported yet");
}

}

// uses dx,dy,euler.alt() to rotate 'parent' into this image. If parent is
// NULL, just rotates/translates the current image
// scaling factor is optional, and not persistent
// r sets a limit on the radius
void EMData::rotateAndTranslate(float scale,float dxc,float dyc,float dzc,int r) {
float *ddata,*sdata;
float yy,x2,y2,z2,x2c,y2c;
int x, y, z;
float t,u,v,p0,p1,p2,p3,em,fl;
//static int *limit = (int *)malloc(4),nsz,nrad;
int k0,k1,k2,k3;
int i,j,k,ii,jj;
int nx2=nx, ny2=ny;	// deal with the special case of rotating different size parent to this. only 2D is supported now

// This delays in case of a radius mismatch in parallel processing
/*while (r>0 && nsz) usleep(1000);

if (r && r!=abs(nlimit)) {
	nlimit=r;
	limit=(int *)realloc(limit,r*2sizeof(int));
}*/

r=r*r;

if (scale!=0) scale=1.0/scale; else scale=1.0;

if (parent) {
	sdata=parent->getDataRO();
	ddata=getData();
	nx2=parent->xSize();
	ny2=parent->ySize();
}
else {
	sdata=getData();
	//ddata=(float *)Smalloc(nx*ny*nz*sizeof(float),1);
	ddata=(float *)malloc(nx*ny*nz*sizeof(float));
}

if (nz==1) {
//em=edgeMean();
em=0;
float mx0=scale*cos(euler.alt());
float mx1=scale*sin(euler.alt());
x2c=nx/2.0-dxc-dx;
y2c=ny/2.0-dyc-dy;
for (j=0,y=-ny/2.0+dxc; j<ny; j++,y+=1.0) {
	for (i=0,x=-nx/2.0+dyc; i<nx; i++,x+=1.0) {

		if (r && (i-nx/2)*(j-nx/2)>r) {
			ddata[i+j*nx]=0;
			continue;
		}

		// rotate coordinates
		x2=(mx0*x+mx1*y)+x2c;
		y2=(-mx1*x+mx0*y)+y2c;

		if (x2<0||x2>nx2-1.0||y2<0||y2>ny2-1.0) {
//			ddata[i+j*nx]=em;
//			ddata[i+j*nx]=grand(Mean(),Sigma());
//			ddata[i+j*nx]=Mean();
			ddata[i+j*nx]=0;
			continue;
		}

		ii=floor2(x2);
		jj=floor2(y2);
		k0=ii+jj*nx;
		k1=k0+1;
		k2=k0+nx+1;
		k3=k0+nx;

		if (ii==nx2-1) { k1--; k2--; }
		if (jj==ny2-1) { k2-=nx2; k3-=nx2; }

		// no interpolation
		//ddata[i+j*nx]=sdata[ii+jj*nx];

		// bilinear interpolation of raw data
		t=(x2-(float)ii);
		u=(y2-(float)jj);
		float tt=1.0-t;
		float uu=1.0-u;

		p0=sdata[k0]*tt*uu;
		p1=sdata[k1]*t*uu;
		p3=sdata[k3]*tt*u;
		p2=sdata[k2]*t*u;
		ddata[i+j*nx]=p0+p1+p2+p3;
	}
}
}
else if(get_map_type() == ICOS5F_HALF) { // to deal with the special icosahedral 5fold half map which has 2fold axis on y
	float mx[9];
	float x3,y3,z3, x4,y4,z4;
	int l=0,xy;

	mx[0]=scale*(cos(euler.phi())*cos(euler.az())-cos(euler.alt())*sin(euler.az())*sin(euler.phi()));
	mx[1]=-scale*(sin(euler.phi())*cos(euler.az())+cos(euler.alt())*sin(euler.az())*cos(euler.phi()));
	mx[2]=scale*sin(euler.alt())*sin(euler.az());
	mx[3]=scale*(cos(euler.phi())*sin(euler.az())+cos(euler.alt())*cos(euler.az())*sin(euler.phi()));
	mx[4]=scale*(-sin(euler.phi())*sin(euler.az())+cos(euler.alt())*cos(euler.az())*cos(euler.phi()));
	mx[5]=-scale*sin(euler.alt())*cos(euler.az());
	mx[6]=scale*sin(euler.alt())*sin(euler.phi());
	mx[7]=scale*sin(euler.alt())*cos(euler.phi());
	mx[8]=scale*cos(euler.alt());

	for (int p = 0; p < 9; p++) {
		printf("%4.2f ", mx[p]);
	}
	printf("\n");

	
	xy=nx*ny;
printf("in rotation\n");
	for (int k=0; k<nz; k++) {
		for (int j=-ny/2; j<ny-ny/2; j++) {
			for (int i=-nx/2; i<nx-nx/2; i++, l++) {
				x2=mx[0]*i+mx[1]*j+mx[2]*k+nx/2;
				y2=mx[3]*i+mx[4]*j+mx[5]*k+ny/2;
				z2=mx[6]*i+mx[7]*j+mx[8]*k+0/2;

				if (x2<0||y2<0||z2<=-(nz-1)|| x2>=nx||y2>=ny||z2>=nz-1) {
					continue;
				}

				if(z2<0){
					x2=nx/2*2-x2;
					z2=-z2;
				}

				x=floor2(x2);
				y=floor2(y2);
				z=floor2(z2);

				t=x2-x;
				u=y2-y;
				v=z2-z;

				// real part
				ii=(int)(x+y*nx+z*xy);

				ddata[l] +=	trilin(sdata[ii],sdata[ii+1],sdata[ii+nx],sdata[ii+nx+1],
							sdata[ii+nx*ny],sdata[ii+xy+1],sdata[ii+xy+nx],sdata[ii+xy+nx+1],
							t,u,v);
			}
		}
	}
}
else {
	float mx[9];
	float x3,y3,z3, x4,y4,z4;
	int l,xy,mr;

	mx[0]=scale*(cos(euler.phi())*cos(euler.az())-cos(euler.alt())*sin(euler.az())*sin(euler.phi()));
	mx[1]=-scale*(sin(euler.phi())*cos(euler.az())+cos(euler.alt())*sin(euler.az())*cos(euler.phi()));
	mx[2]=scale*sin(euler.alt())*sin(euler.az());
	mx[3]=scale*(cos(euler.phi())*sin(euler.az())+cos(euler.alt())*cos(euler.az())*sin(euler.phi()));
	mx[4]=scale*(-sin(euler.phi())*sin(euler.az())+cos(euler.alt())*cos(euler.az())*cos(euler.phi()));
	mx[5]=-scale*sin(euler.alt())*cos(euler.az());
	mx[6]=scale*sin(euler.alt())*sin(euler.phi());
	mx[7]=scale*sin(euler.alt())*cos(euler.phi());
	mx[8]=scale*cos(euler.alt());

	x4=(mx[0]*(-nx/2.0+dxc)+mx[1]*(-ny/2.0+dyc)+mx[2]*(-nz/2.0+dzc))+nx/2.0-dxc-dx;
	y4=(mx[3]*(-nx/2.0+dxc)+mx[4]*(-ny/2.0+dyc)+mx[5]*(-nz/2.0+dzc))+ny/2.0-dyc-dy;
	z4=(mx[6]*(-nx/2.0+dxc)+mx[7]*(-ny/2.0+dyc)+mx[8]*(-nz/2.0+dzc))+nz/2.0-dzc-dz;

	xy=nx*ny;
	if (nx>ny && nx>nz) mr=(nx-2)*(nx-2)/4;
	else if (ny>nz) mr=(ny-2)*(ny-2)/4;
	else mr=(nz-2)*(nz-2)/4;

	for (k=-nz/2,l=0; k<nz/2; k++,x4+=mx[2],y4+=mx[5],z4+=mx[8]) {
		x3=x4;
		y3=y4;
		z3=z4;
		for (j=-ny/2; j<ny/2; j++,x3+=mx[1],y3+=mx[4],z3+=mx[7]) {
			x2=x3;
			y2=y3;
			z2=z3;
			for (i=-nx/2; i<nx/2; i++,l++,x2+=mx[0],y2+=mx[3],z2+=mx[6]) {

				if (x2<0||y2<0||z2<0|| x2>=nx-1||y2>=ny-1||z2>=nz-1) {
//				if (i*i+j*j+k*k>=mr) {
					ddata[l]=0;
					continue;
				}

				x=floor2(x2);
				y=floor2(y2);
				z=floor2(z2);

				t=x2-x;
				u=y2-y;
				v=z2-z;

				// real part
				ii=(int)(x+y*nx+z*xy);

				ddata[l]=trilin(sdata[ii],sdata[ii+1],sdata[ii+nx],sdata[ii+nx+1],
					sdata[ii+nx*ny],sdata[ii+xy+1],sdata[ii+xy+nx],sdata[ii+xy+nx+1],
					t,u,v);
			}
		}
	}

/*	for (k=0,z=-nz/2.0+dzc; k<nz; k++,z+=1.0) {
		for (j=0,y=-ny/2.0+dyc; j<ny; j++,y+=1.0) {
			for (i=0,x=-nx/2.0+dxc; i<nx; i++,x+=1.0) {
				x2=(mx[0]*x+mx[1]*y+mx[2]*z)+nx/2.0-dxc-dx;
				y2=(mx[3]*x+mx[4]*y+mx[5]*z)+ny/2.0-dyc-dy;
				z2=(mx[6]*x+mx[7]*y+mx[8]*z)+nz/2.0-dzc-dz;

				if (x2<0||y2<0||z2<0|| x2>=nx-1||y2>=ny-1||z2>=nz-1) {
					ddata[i+j*nx+k*nx*ny]=0;
					continue;
				}

				t=x2-floor(x2);
				u=y2-floor(y2);
				v=z2-floor(z2);

				// real part
				ii=(int)(floor(x2)+floor(y2)*nx+floor(z2)*nx*ny);

				ddata[i+j*nx+k*nx*ny]=trilin(sdata[ii],sdata[ii+1],sdata[ii+nx],sdata[ii+nx+1],
					sdata[ii+nx*ny],sdata[ii+nx*ny+1],sdata[ii+nx*ny+nx],sdata[ii+nx*ny+nx+1],
					t,u,v);
			}
		}
	}*/

}
setXYZOrigin(	getXorigin()+nx/2*pixel*(1-scale)-dx*scale*pixel,
				getYorigin()+ny/2*pixel*(1-scale)-dy*scale*pixel,
				getZorigin()+nz/2*pixel*(1-scale)-dz*scale*pixel
			);

setPixel(Pixel()*scale);
if (hasCTF()) ctf[10]*=scale;
if (parent) parent->doneData();
else {
	free(rdata);
	rdata=ddata;
	dx=dy=dz=0;
	setRAlign(0,0,0);
//	flags&=~EMDATA_FLIP;
}
doneData();

update();
return;
}

// faster version of rotateAndTranslate above, but lower quality
void EMData::rotateAndTranslateFast(float scale) {
float *ddata,*sdata;
float x,y,z,yy,x2,y2,z2,mx[9];
float dxx,dxy,dyy,dyx;
float t,u,v,p[4],em,fl;
int k0,k1,k2,k3;
int i,j,k,ii,jj;


if (scale!=0) scale=1.0/scale; else scale=1.0;

if (parent) {
	sdata=parent->getDataRO();
	ddata=getData();
}
else {
	sdata=getData();
	//ddata=(float *)Smalloc(nx*ny*nz*sizeof(float),1);
	ddata=(float *)malloc(nx*ny*nz*sizeof(float));
}

if (nz==1) {
mx[0]=scale*cos(euler.alt());
mx[1]=scale*sin(euler.alt());
dxx=mx[0];
dyx=-mx[1];
dxy=mx[1];
dyy=mx[0];
for (j=k=0, x=-mx[1]*ny/2.0+nx/2.0-dx,y=-mx[0]*ny/2.0+ny/2.0-dy;   
	j<ny; j++,x+=dxy,y+=dyy) {
	for (i=0,x2=x-mx[0]*nx/2.0,y2=y+mx[1]*nx/2.0; i<nx; i++,k++,x2+=dxx,y2+=dyx) {

		if (x2<0||x2>nx-2.0||y2<0||y2>ny-2.0) {
//			ddata[k]=Mean();
//			ddata[k]=grand(Mean(),Sigma());
			ddata[k]=0;
			continue;
		}

		ii=(int)floor(x2+.5);
		jj=(int)floor(y2+.5);
		
		// no interpolation
		ddata[k]=sdata[ii+jj*nx];
	}
}
}
// not fast yet
else {
	float x3,y3,z3, x4,y4,z4;
	int l,xy,mr;
	
	mx[0]=scale*(cos(euler.phi())*cos(euler.az())-cos(euler.alt())*sin(euler.az())*sin(euler.phi()));
	mx[1]=-scale*(sin(euler.phi())*cos(euler.az())+cos(euler.alt())*sin(euler.az())*cos(euler.phi()));
	mx[2]=scale*sin(euler.alt())*sin(euler.az());
	mx[3]=scale*(cos(euler.phi())*sin(euler.az())+cos(euler.alt())*cos(euler.az())*sin(euler.phi()));
	mx[4]=scale*(-sin(euler.phi())*sin(euler.az())+cos(euler.alt())*cos(euler.az())*cos(euler.phi()));
	mx[5]=-scale*sin(euler.alt())*cos(euler.az());
	mx[6]=scale*sin(euler.alt())*sin(euler.phi());
	mx[7]=scale*sin(euler.alt())*cos(euler.phi());
	mx[8]=scale*cos(euler.alt());

	x4=(mx[0]*(-nx/2.0)+mx[1]*(-ny/2.0)+mx[2]*(-nz/2.0))+nx/2.0-dx;
	y4=(mx[3]*(-nx/2.0)+mx[4]*(-ny/2.0)+mx[5]*(-nz/2.0))+ny/2.0-dy;
	z4=(mx[6]*(-nx/2.0)+mx[7]*(-ny/2.0)+mx[8]*(-nz/2.0))+nz/2.0-dz;

	xy=nx*ny;
	if (nx<ny && nx<nz) mr=(nx-2)*(nx-2)/4;
	else if (ny<nz) mr=(ny-2)*(ny-2)/4;
	else mr=(nz-2)*(nz-2)/4;
	
	for (k=-nz/2,l=0; k<nz/2; k++,x4+=mx[2],y4+=mx[5],z4+=mx[8]) {	
		x3=x4;
		y3=y4;
		z3=z4;
		for (j=-ny/2; j<ny/2; j++,x3+=mx[1],y3+=mx[4],z3+=mx[7]) {
			x2=x3;
			y2=y3;
			z2=z3;
			for (i=-nx/2; i<nx/2; i++,l++,x2+=mx[0],y2+=mx[3],z2+=mx[6]) {
			
//				if (x2<0||y2<0||z2<0|| x2>=nx-1||y2>=ny-1||z2>=nz-1) {
				if (i*i+j*j+k*k>=mr) {
					ddata[l]=0;
					continue;
				}

				x=floor(x2+.5);
				y=floor(y2+.5);
				z=floor(z2+.5);

				// real part
				ii=(int)(x+y*nx+z*xy);

				ddata[l]=sdata[ii];
			}
		}
	}
}

if (parent) parent->doneData();
else {
	free(rdata);
	rdata=ddata;
	dx=dy=dz=0;
	setRAlign(0,0,0);
//	flags&=~EMDATA_FLIP;
}
doneData();
setPixel(Pixel()*scale);
if (hasCTF()) ctf[10]*=scale;
setXYZOrigin(getXorigin()*scale,getYorigin()*scale,getZorigin()*scale);

update();
return;
}

// Complex only, 
void EMData::applyRadFn(int n,float x0,float dx,float *y,int interp) {
int i, j, m, k=0, l;
float r,f;

if (!isComplex()) return;
//writeMRC("a.img");
ap2ri();
getData();

// 2d
if (nz==1) {
	for (j=0; j<ny; j++) {
	    for (i=0; i<nx; i+=2, k+=2) {
//	r=hypot((float)i/2.0,(float)j-ny/2.0-1.0);
		r=hypot((float)i/2.0,(float)j-ny/2.0);
		r=(r-x0)/dx;
		if (interp) l=(int)floor(r); else l=(int)floor(r+1);
		r-=l;
		if (l>=n-2) f=y[n-1];
		else {
			if (interp) f=(y[l]*(1.0-r)+y[l+1]*r);
			else f=y[l];
		}
		rdata[k]*=f;
		rdata[k+1]*=f;
		}
	}
}
else {
	for (m=k=0; m<nz; m++) {
		for (j=0; j<ny; j++) {
		    for (i=0; i<nx; i+=2, k+=2) {
//	r=hypot((float)i/2.0,(float)j-ny/2.0-1.0);
			r=sqrt((float)(i*i/4.0)+(float)((j-ny/2.0)*(j-ny/2.0))+(float)((m-nz/2.0)*(m-nz/2.0)));
			r=(r-x0)/dx;
			if (interp) l=(int)floor(r); else l=(int)floor(r+1);
			r-=l;
			if (l>=n-2) f=y[n-1];
			else {
				if (interp) f=(y[l]*(1.0-r)+y[l+1]*r);
				else f=y[l];
			}
			rdata[k]*=f;
			rdata[k+1]*=f;
			}
		}
	}

}

doneData();
update();
//writeMRC("b.img");
}

// the passed noise is a mean noise INTENSITY. The noise distribution is a gaussian. This
// fact is not based on any real data
void EMData::addRadNoise(int n,float x0,float dx,float *y,int interp) {
int h, i, j, m, k=0, l;
float r,f,a;

if (!isComplex()) return;
//writeMRC("a.img");
ap2ri();
getData();

if (nz==1) {
	for (j=k=0; j<ny; j++) {
	    for (i=0; i<nx; i+=2, k+=2) {
//	r=hypot((float)i/2.0,(float)j-ny/2.0-1.0);
			r=hypot((float)i/2.0,(float)j-ny/2.0);
			r=(r-x0)/dx;
			if (interp) l=(int)floor(r); else l=(int)floor(r+.5);
			r-=l;
			if (l>=n-2) f=y[n-1];
			else if (l<0) l=0;
			else {
				if (interp) f=(y[l]*(1.0-r)+y[l+1]*r);
				else f=y[l];
			}
			f=grand(sqrt(f),sqrt(f)/3.0);
			a=frand(0,2.0*PI);
			if (i==0) f*=sqrt(2.0);
			rdata[k]+=f*cos(a);
			rdata[k+1]+=f*sin(a);
		}
	}
}
else { 
	for (h=k=0; h<nz; h++) {
		for (j=0; j<ny; j++) {
		    for (i=0; i<nx; i+=2, k+=2) {
				r=sqrt(hypot3((float)i/2.0,(float)j-ny/2.0,(float)h-nz/2.0));
				r=(r-x0)/dx;
				if (interp) l=(int)floor(r); else l=(int)floor(r+.5);
				r-=l;
				if (l>=n-2) f=y[n-1];
				else if (l<0) l=0;
				else {
					if (interp) f=(y[l]*(1.0-r)+y[l+1]*r);
					else f=y[l];
				}
				f=grand(sqrt(f),sqrt(f)/3.0);
				a=frand(0,2.0*PI);
				if (i==0) f*=sqrt(2.0);
				rdata[k]+=f*cos(a);
				rdata[k+1]+=f*sin(a);
			}
		}
	}
}

doneData();

}

// subtract 'obj' from 'this'
void EMData::subtract(EMData *obj) {
int i,j,c;
float *dest,*src;

if (obj->isComplex()) obj->ap2ri();
if (obj->xSize()!=nx || obj->ySize()!=ny || obj->zSize()!=nz) return;

dest=getData();
src=obj->getDataRO();

for (j=0; j<nx*ny*nz; j++) dest[j]-=src[j];

obj->doneData();
doneData();

update();
}

// vFlip() of 2D image is equivalent to projection from 3D using (alt+PI,az,-phi) instead of (alt,az,phi)

// Does a vertical flip of a data set. This flips around the vertical 'phase origin'.
// ie - the bottom row/plane is not mirrored
void EMData::vFlip() {
int i,j,k;
float t;

getData();

/*
if (nz==1) {
	for (i=0; i<(ny+1)>>1; i++) {
		for (j=0; j<nx; j++) {
			t=rdata[i*nx+j];
			rdata[i*nx+j]=rdata[(ny-i-1)*nx+j];
			rdata[(ny-i-1)*nx+j]=t;
		}
	}
}
else {
	for (i=0; i<(nz+1)>>1; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nx; k++) {
				t=rdata[i*nx*ny+j*nx+k];
				rdata[i*nx*ny+j*nx+k]=rdata[(nz-i-1)*nx*ny+j*nx+k];
				rdata[(nz-i-1)*nx*ny+j*nx+k]=t;
			}
		}
	}
}
*/

if (nz==1) {
	for (i=1; i<ny/2; i++) {
		for (j=0; j<nx; j++) {
			t=rdata[i*nx+j];
			rdata[i*nx+j]=rdata[(ny-i)*nx+j];
			rdata[(ny-i)*nx+j]=t;
		}
	}
	// bottom is a copy of the row above the bottom (hopefully avoids fourier artifacts)
	for (j=0; j<nx; j++) {
		rdata[j]=rdata[nx+j];
	}
}
else {
	for (i=1; i<nz/2; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nx; k++) {
				t=rdata[i*nx*ny+j*nx+k];
				rdata[i*nx*ny+j*nx+k]=rdata[(nz-i)*nx*ny+j*nx+k];
				rdata[(nz-i)*nx*ny+j*nx+k]=t;
			}
		}
	}
	for (j=0; j<ny; j++) {
		for (k=0; k<nx; k++) {
			rdata[j*nx+k]=rdata[nx*ny+j*nx+k];
		}
	}
}

doneData();
}

// hFlip() of 2D image is equivalent to projection from 3D using (alt+PI,az,PI-phi) instead of (alt,az,phi)
void EMData::hFlip() {
float t,*d;
int x,y,z,i,j;

d=getData();
for (z=0; z<zSize(); z++) {
	for (y=0; y<ySize(); y++) {
		i=y*xSize()+z*xSize()*ySize();
		j=i+xSize();
		for (x=1; x<xSize()/2; x++) {
			t=d[i+x];
			d[i+x]=d[j-x];
			d[j-x]=t;
		}
	}
}
doneData();
			
}

// in place DWT, basis/level defined elsewhere
void EMData::doDWT(int basis, int level) {	

}

// inverse wavelet x-form
void EMData::doIWT(int basis, int level) {

}

// in place nonlinear threshold wavelet filter
// thresh is with respect to the 'ideal' threshold
void EMData::DWTFilt(int basis, int level, float thresh) {

}

// applies a circular mask to the data
// type=0 is a step cutoff to the mean value
// type=1 fills in with flatband random noise
// type=2 is a tanh cutoff to 0
// type=3 is a gaussian falloff, r is the 1/e width
// type=4 is a step cutoff to 0
// type=5 is an INNER step cutoff to 0 (inside r -> 0)
// type=6 is a cutoff to the circular edgemean value
// type=7 generates a 2 pixel ring of '1', replacing the current contents
// type=8 generates a 2 pixel ring normalized so the sum of the values in the ring is 1, replacing the current contents
// type=60 is type=6 with each pixel value subtracted by the mean at mask rad
// type=107 is an exp(x^4) to fix hypergeometric 3d interpolation
// type=105/6 is an inverse gaussian to fix gaussian 3d interpolation (modes 5-6)
// type=104 (mode 4)
// type=103 (mode 3)
// type=102 is an inverse gaussian to fix gaussian 3d interpolation (mode 2)
// type=101 is actually a null function
// type=20 is a forward gaussian to 'unfix' gaussian 3d interpolation (modes 3-5)
// type=-1 takes the fabs() of the data
// real only at the moment
void EMData::applyMask(int r,int type,float dx0, float dy0, float dz0) { // dx0,dy0,dz0 modify the mask center
int x,y,z,i;												  // relative to default center nx/2,ny/2,nz/2	
float k,rr,m;
float x1,y1,z1;
if (flags&EMDATA_COMPLEX) return;
if (type==5 &&r<=0) return;
if (r<=0) r=nx/2;

getData();
if (nz==1) {
	float xc=nx/2.0+dx0,	yc=ny/2.0+dy0;
	switch (type) {
	case 0:
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				x1=fabs(x-xc);
				y1=fabs(y-yc);
				if (x1>r || y1>r || hypot((float)x1,(float)y1)>r ) rdata[i]=Mean();
			}
		}
		break;
	case 1:
		{
		float m=Mean(),s=Sigma();
		if (!s || !goodf(&s)) s=1.0;
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				x1=fabs(x-xc);
				y1=fabs(y-yc);
				if (x1>r || y1>r || hypot((float)x1,(float)y1)>r ) rdata[i]=grand(m,s);
			}
		}
		break;
		}
	case 2:
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rr=hypot(x-xc,y-yc);
				rdata[i]*=(tanh(((float)r-rr)/2.0)/2.0+.5);
			}
		}
		break;
	case 3:
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rr=hypot(x-xc,y-yc);
				rdata[i]*=exp(-rr*rr/(r*r));
			}
		}
		break;
	case 4:
		r=r*r;
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				if (hypot2(x-xc,y-yc)>r ) rdata[i]=0;
			}
		}
		break;
	case 5:
		r=r*r;
		if (r<1) break;
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				if (hypot2(x-xc,y-yc)<r ) rdata[i]=0;
			}
		}
		break;
	case 6:
	case 60:
		rr=k=0;
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				x1=fabs(x-xc);
				y1=fabs(y-yc);
				if (hypot((float)x1,(float)y1)<=r+1 &&
					hypot((float)x1,(float)y1)>=r-1) { rr+=rdata[i]; k+=1.0; }
			}
		}
		rr/=k;
		
		if (type==6) {
			for (y=i=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					x1=fabs(x-xc);
					y1=fabs(y-yc);
					if (x1>r || y1>r || hypot((float)x1,(float)y1)>r ) rdata[i]=rr;
				}
			}
		}
		else {
			for (y=i=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rdata[i]-=rr;
					x1=fabs(x-xc);
					y1=fabs(y-yc);
					if (x1>r || y1>r || hypot((float)x1,(float)y1)>r ) rdata[i]=0;
				}
			}
		}
		break;		
	case 7:
	case 8:
		x1=(r-1)*(r-1);
		y1=(r+1)*(r+1);
		if (r<1) break;
		k=0;
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rr=hypot2(x-xc,y-yc);
				if (rr>=x1 && rr<=y1) { rdata[i]=1; k++; }
			}
		}
		if (type==8 && k>0) {
			k=1.0/k;
			for (i=0; i<nx*ny*nz; i++) rdata[i]*=k;
		}
		break;
	case -1:
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rdata[i]=fabs(rdata[i]);
			}
		}
		break;
	case 10:
	case 11:
	case 12:
		error("This program uses an old interpolation method!",ERR_EXIT);
	case 20:
		k=I5G*PI*PI/(float)(ny*ny);		// make sure we use I5G used in 3d slicing
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rr=hypot(x-xc,y-yc);
				rdata[i]*=exp(-rr*rr*k);
			}
		}
		break;
	case 107:
		m=ny*ny/4.0;
//		k=3.24*16.0/(float)(ny*ny*ny*ny);		// corrects for precalculated exp(-x^4)
		k=1.80*16.0/(float)(ny*ny*ny*ny);		// corrects for precalculated exp(-x^4)
										// kernel is hypergeometric
/*		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rr=SQR(x-nx/2.0)+SQR(y-ny/2.0);
				if (rr>m) rdata[i]=0;
				else rdata[i]/=exp(-rr*rr*k);
			}
		}*/
		break;
	case 106:
	case 105:
		k=I5G*PI*PI/(float)(ny*ny);		// make sure we use I5G used in 3d slicing
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rr=SQR(x-xc)+SQR(y-yc);
				rdata[i]/=exp(-rr*k);
			}
		}
		break;
	case 104:
		k=I4G*PI*PI/(float)(ny*ny);		// make sure we use I5G used in 3d slicing
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rr=SQR(x-xc)+SQR(y-yc);
				rdata[i]/=exp(-rr*k);
			}
		}
		break;
	case 103:
		k=I3G*PI*PI/(float)(ny*ny);		// make sure we use I5G used in 3d slicing
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rr=SQR(x-xc)+SQR(y-yc);
				rdata[i]/=exp(-rr*k);
			}
		}
		break;
	case 102:
		k=I2G*PI*PI/(float)(ny*ny);		// make sure we use I2G used in 3d slicing
		for (y=i=0; y<ny; y++) {
			for (x=0; x<nx; x++,i++) {
				rr=hypot(x-xc,y-yc);
				rdata[i]/=exp(-rr*rr*k);
			}
		}
		break;
	case 101:
		break;
	}
}
else {
	float xc=nx/2.0+dx0,	yc=ny/2.0+dy0,	zc=nz/2.0+dz0;
	switch (type) {
	case 0:
		r=r*r;
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					if (hypot3(x-xc,y-yc,z-zc)<r) continue;
					rdata[i]=Mean();
				}
			}
		}
		break;
	case 1:
		r=r*r;
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					if (hypot3(x-xc,y-yc,z-zc)<r) continue;
					rdata[i]=grand(Mean(),Sigma());
				}
			}
		}
		break;
	case 2:
		k=16.0/(float)(ny*ny);
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rr=sqrt(hypot3(x-xc,y-yc,z-zc));
					rdata[i]*=(tanh(((float)r-rr)/2.0)/2.0+.5);
				}
			}
		}
		break;
	case 3:
		r=r*r;
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rr=hypot3(x-xc,y-yc,z-zc);
					rdata[i]*=exp(-rr/r);
				}
			}
		}
		break;
	case 4:
		r=r*r;
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					if (hypot3(x-xc,y-yc,z-zc)<r) continue;
					rdata[i]=0;
				}
			}
		}
		break;
	case 5:
		if (r<1) break;
		r=r*r;
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					if (hypot3(x-xc,y-yc,z-zc)>r) continue;
					rdata[i]=0;
				}
			}
		}
		break;
	case 6:
	case 60:
		rr=k=0;
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					x1=sqrt(hypot3(x-xc,y-yc,z-zc));
					if (x1<=r+1 && x1>=r-1) { rr+=rdata[i]; k+=1.0; }
				}
			}
		}
		rr/=k;

		if (type==6) {
			for (z=i=0; z<nz; z++) {
				for (y=0; y<ny; y++) {
					for (x=0; x<nx; x++,i++) {
						if (hypot3(x-xc,y-yc,z-zc)<r*r) continue;
						rdata[i]=rr;
					}
				}
			}
		}
		else {
			for (z=i=0; z<nz; z++) {
				for (y=0; y<ny; y++) {
					for (x=0; x<nx; x++,i++) {
						rdata[i]-=rr;
						if (hypot3(x-xc,y-yc,z-zc)<r*r) continue;
						rdata[i]=0;
					}
				}
			}
		}
		break;		
	case 7:
	case 8:
		x1=(r-1)*(r-1);
		y1=(r+1)*(r+1);
		k=0;
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rr=hypot3(x-xc,y-yc,z-zc);
					if (rr>=x1 && rr<=y1) { rdata[i]=1.0; k+=1.0; }
				}
			}
		}
		if (type==8 && k>0) {
			k=1.0/k;
			for (i=0; i<nx*ny*nz; i++) rdata[i]*=k;
		}
		break;
	case 10:
	case 11:
	case 12:
		error("This program uses an old interpolation method!\n",ERR_EXIT);
	case 20:
		k=I5G*PI*PI/(float)(ny*ny);		// make sure we use I5G used in 3d slicing
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rr=sqrt(hypot3(x-xc,y-yc,z-zc));
					rdata[i]*=exp(-rr*rr*k);
				}
			}
		}
		break;
	case 107:
		m=ny*ny/4.0;
		k=3.24*16.0/(float)(ny*ny*ny*ny);		// corrects for precalculated exp(-x^4)
//		k=1.8*16.0/(float)(ny*ny*ny*ny);		// corrects for precalculated exp(-x^4)
										// kernel is hypergeometric
/*		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rr=hypot3(x-nx/2.0,y-ny/2.0,z-nz/2.0);
					if (rr>m) rdata[i]=0;
					else rdata[i]/=exp(-rr*rr*k);
				}
			}
		}*/
		break;
	case 106:
	case 105:
		k=I5G*PI*PI/(float)(ny*ny);		// make sure we use I5G used in 3d slicing
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rr=sqrt(hypot3(x-xc,y-yc,z-zc));
					rdata[i]/=exp(-rr*rr*k);
				}
			}
		}
		break;
	case 104:
		k=I4G*PI*PI/(float)(ny*ny);		// make sure we use I5G used in 3d slicing
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rr=sqrt(hypot3(x-xc,y-yc,z-zc));
					rdata[i]/=exp(-rr*rr*k);
				}
			}
		}
		break;
	case 103:
		k=I3G*PI*PI/(float)(ny*ny);		// make sure we use I5G used in 3d slicing
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rr=sqrt(hypot3(x-xc,y-yc,z-zc));
					rdata[i]/=exp(-rr*rr*k);
				}
			}
		}
		break;
	case 102:
		k=I2G*PI*PI/(float)(ny*ny);		// make sure we use I2G used in 3d slicing
		for (z=i=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,i++) {
					rr=sqrt(hypot3(x-xc,y-yc,z-zc));
					rdata[i]/=exp(-rr*rr*k);
				}
			}
		}
		break;
	case 101:
		break;
	}
}

doneData();

update();
}

// calculates radial distribution, works for real
// and complex images. Note that this returns the mean INTENSITY
// for complex images !!!
// For 3D data, only complex form currently supported
// x must be defined on input and y must be allocated.
void EMData::calcRadDist(int n,float x0,float dx,float *d) {
int x,y,z;
int i,c;
float r,h;
float *yc;

getDataRO();

yc=(float *)malloc(n*sizeof(float));
for (i=0; i<n; i++) {
	d[i]=0;
	yc[i]=0;			// for empty bins
}
if (nz==1) {
	if (flags&EMDATA_COMPLEX) {
		if (flags&EMDATA_RI) {
			for (y=c=0; y<ny; y++) {
				for (x=0; x<nx; x+=2,c+=2) {
					r=(hypot(x/2.0,y-ny/2.0)-x0)/dx;

					i=(int)floor(r);						// lower bin to put data in
					r-=i;							// float remainder, how much to put in each bin
					if (i<0 || i>=n) continue;		// beyond the range we're calculating
					if (i==0) { d[0]+=SQR(rdata[c])*(1.0-r); yc[0]+=(1.0-r); }
					else if (i==n-1) { d[n-1]+=SQR(rdata[c])*(r); yc[n-1]+=(r); }
					else {
						h=SQR(rdata[c])+SQR(rdata[c+1]);
						d[i]+=h*(1.0-r);
						yc[i]+=(1.0-r);
						d[i+1]+=h*r;
						yc[i+1]+=(r);
					}
				}
			}
		}
		else {
			for (y=c=0; y<ny; y++) {
				for (x=0; x<nx; x+=2,c+=2) {
					r=(hypot(x/2.0,y-ny/2.0)-x0)/dx;

					i=(int)floor(r);						// lower bin to put data in
					r-=i;							// float remainder, how much to put in each bin
					if (i<0 || i>=n) continue;		// beyond the range we're calculating
					if (i==0) { d[0]+=SQR(rdata[c])*(1.0-r); yc[0]+=(1.0-r); }
					else if (i==n-1) { d[n-1]+=SQR(rdata[c])*(r); yc[n-1]+=(r); }
					else {
						h=rdata[c]*rdata[c];
						d[i]+=h*(1.0-r);
						yc[i]+=(1.0-r);
						d[i+1]+=h*r;
						yc[i+1]+=(r);
					}
				}
			}
		}
	}
	else {
		for (y=c=0; y<ny; y++) {
			for (x=0; x<nx; x++,c++) {
				r=(hypot(x-(nx)/2.0,y-(ny)/2.0)-x0)/dx;

				i=(int)floor(r);						// lower bin to put data in
				r-=i;							// float remainder, how much to put in each bin
				if (i<0 || i>=n) continue;		// beyond the range we're calculating
				if (i==0) { d[0]+=SQR(rdata[c])*(1.0-r); yc[0]+=(1.0-r); }
				else if (i==n-1) { d[n-1]+=SQR(rdata[c])*(r); yc[n-1]+=(r); }
				else {
					d[i]+=rdata[c]*(1.0-r); 
					yc[i]+=(1.0-r); 
					d[i+1]+=rdata[c]*(r); 
					yc[i+1]+=(r); 
				}
			}
		}
	}
}
else {
	if (flags&EMDATA_COMPLEX) {
//		if (!flags&EMDATA_RI) ap2ri();
		for (z=c=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x+=2,c+=2) {
					r=(sqrt(SQR(x/2.0)+SQR(y-ny/2.0)+SQR(z-nz/2.0))-x0)/dx;
				
					i=(int)floor(r);						// lower bin to put data in
					r-=i;							// float remainder, how much to put in each bin
					if (i<0 || i>=n) continue;		// beyond the range we're calculating
					if (i==0) { d[0]+=SQR(rdata[c])*(1.0-r); yc[0]+=(1.0-r); }
					else if (i==n-1) { d[n-1]+=SQR(rdata[c])*(r); yc[n-1]+=(r); }
					else {
						h=SQR(rdata[c])+SQR(rdata[c+1]);
						d[i]+=h*(1.0-r); 
						yc[i]+=(1.0-r); 
						d[i+1]+=h*r; 
						yc[i+1]+=(r); 
					}
				}
			}
		}
	}
	else {	// 3D real space map
		for (z=c=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,c++) {
					r=(sqrt(SQR(x-nx/2.0)+SQR(y-ny/2.0)+SQR(z-nz/2.0))-x0)/dx;

					i=(int)floor(r);						// lower bin to put data in
					r-=i;							// float remainder, how much to put in each bin
					if (i<0 || i>=n) continue;		// beyond the range we're calculating
					if (i==0) { d[0]+=SQR(rdata[c])*(1.0-r); yc[0]+=(1.0-r); }
					else if (i==n-1) { d[n-1]+=SQR(rdata[c])*(r); yc[n-1]+=(r); }
					else {
						d[i]+=rdata[c]*(1.0-r); 
						yc[i]+=(1.0-r); 
						d[i+1]+=rdata[c]*(r); 
						yc[i+1]+=(r); 
					}
				}
			}
		}
	}
}

for (i=0; i<n; i++) {
	if (yc[i]!=0) d[i]/=yc[i];
}

// linear interpolation for 'holes'
// does NOT deal with 2 sequential holes
for (i=1; i<n-1; i++) {
	if (yc[i]<0.1) d[i]=(d[i-1]+d[i+1])/2.0;
}

doneData();
free(yc);
}

// Same as above, but only for a range of angles. acen the direction,arange is angular range around the direction in radians
void EMData::calcRadDist(int n,float x0,float dx,float *d,float acen,float arange, bool smoothing)  {
int x,y;
int i,c;
float r,h,a;
float *yc,*yc2,*d2;

getDataRO();

yc=(float *)malloc(n*sizeof(float));
yc2=(float *)malloc(n*sizeof(float));
d2=(float *)malloc(n*sizeof(float));

for (i=0; i<n; i++) { 
	d[i]=0; 
	yc[i]=0;
}

if (nz==1) {
	if (flags&EMDATA_COMPLEX) {
		if (flags&EMDATA_RI) {
			for (y=c=0; y<ny; y++) {
				for (x=0; x<nx; x+=2,c+=2) {
					if (y==ny/2.0 && x==0) a=0;
					else a=atan2(y-ny/2.0,x/2.0);
					if (fabs(asub2(a,acen))>arange) continue;
					r=(hypot(x/2.0,y-ny/2.0)-x0)/dx;
					
					i=(int)floor(r);				// lower bin to put data in
					r-=i;							// float remainder, how much to put in each bin
					if (i<0 || i>=n) continue;		// beyond the range we're calculating
					if (i==0) { d[0]+=SQR(rdata[c])*(1.0-r); yc[0]+=(1.0-r); }
					else if (i==n-1) { d[n-1]+=SQR(rdata[c])*(r); yc[n-1]+=(r); }
					else {
						h=SQR(rdata[c])+SQR(rdata[c+1]);
						d[i]+=h*(1.0-r); 
						yc[i]+=(1.0-r); 
						d[i+1]+=h*r; 
						yc[i+1]+=(r); 
					}
				}
			}
		}
		else {
			for (y=c=0; y<ny; y++) {
				for (x=0; x<nx; x+=2,c+=2) {
					if (y==ny/2.0 && x==0) a=0;
					else a=atan2(y-ny/2.0,x/2.0);
					if (fabs(asub2(a,acen))>arange) continue;
					r=(hypot(x/2.0,y-ny/2.0)-x0)/dx;
					
					i=(int)floor(r);						// lower bin to put data in
					r-=i;							// float remainder, how much to put in each bin
					if (i<0 || i>=n) continue;		// beyond the range we're calculating
					if (i==0) { d[0]+=SQR(rdata[c])*(1.0-r); yc[0]+=(1.0-r); }
					else if (i==n-1) { d[n-1]+=SQR(rdata[c])*(r); yc[n-1]+=(r); }
					else {
						h=rdata[c]*rdata[c];
						d[i]+=h*(1.0-r); 
						yc[i]+=(1.0-r); 
						d[i+1]+=h*r; 
						yc[i+1]+=(r); 
					}
				}
			}
		}
	}
	else {
		for (y=c=0; y<ny; y++) {
			for (x=0; x<nx; x++,c++) {
				if (y==ny/2.0 && x==nx/2.0) a=0;
				else a=atan2(y-ny/2.0,x-nx/2.0);
				if (fabs(asub2(a,acen))>arange) continue;
				r=(hypot(x-(nx-1)/2.0,y-(ny-1)/2.0)-x0)/dx;
				
				i=(int)floor(r);						// lower bin to put data in
				r-=i;							// float remainder, how much to put in each bin
				if (i<0 || i>=n) continue;		// beyond the range we're calculating
				if (i==0) { d[0]+=rdata[c]*(1.0-r); yc[0]+=(1.0-r); }
				else if (i==n-1) { d[n-1]+=rdata[c]*(r); yc[n-1]+=(r); }
				else { 
					d[i]+=rdata[c]*(1.0-r); 
					yc[i]+=(1.0-r); 
					d[i+1]+=rdata[c]*(r); 
					yc[i+1]+=(r); 
				}
			}
		}
	}
}
else {

}

if (smoothing) {
	for (i=0; i<n; i++) { d2[i]=d[i]; yc2[i]=yc[i]; }
	for (i=1; i<n; i++) { d2[i-1]+=d[i]*.57; yc2[i-1]+=yc[i]*.57; }
	for (i=2; i<n; i++) { d2[i-2]+=d[i]*.105; yc2[i-2]+=yc[i]*.105; }
	for (i=0; i<n-1; i++) { d2[i+1]+=d[i]*.57; yc2[i+1]+=yc[i]*.57; }
	for (i=0; i<n-2; i++) { d2[i+2]+=d[i]*.105; yc2[i+2]+=yc[i]*.105; }
	
	for (i=0; i<n; i++) { if (yc2[i]!=0) d[i]=d2[i]/yc2[i]; else d[i]=0; }
}
else {
	for (i=0; i<n; i++) { if (yc[i]!=0) d[i]=d[i]/yc[i]; else d[i]=0; }
	for (i=1; i<n-1; i++) {
		if (yc[i]<0.1) d[i]=(d[i-1]+d[i+1])/2.0;
	}
}

/*
for (i=0; i<n; i++) { if (yc[i]!=0) d[i]/=yc[i]; } 

for (i=1; i<n-1; i++) {
	if (yc[i]==0) d[i]=(d[i-1]+d[i+1])/2.0;
}
*/

doneData();
free(yc);
free(yc2);
free(d2);
}

// works for real or complex images, 2d only for now
void EMData::calcAzDist(int n,float a0,float da,float *d,float rmin,float rmax) {
int x,y;
int i,c;
float r,h,a,w;
float *yc;

getData();



yc=(float *)malloc(n*sizeof(float));
for (i=0; i<n; i++) { 
	d[i]=0; 
	yc[i]=.00001;			// for empty bins
}

if (nz==1) {
	if (flags&EMDATA_COMPLEX) {
		if (flags&EMDATA_RI) {
			for (y=c=0; y<ny; y++) {
				for (x=0; x<nx; x+=2,c+=2) {
					r=hypot(x/2.0,y-ny/2.0);
					if (r<rmin||r>rmax) continue;
					if (y==ny/2.0 && x==0) a=0;
					else a=(atan2(y-ny/2.0,x/2.0)-a0)/da;
					
					i=(int)floor(a);						// lower bin to put data in
					a-=i;							// float remainder, how much to put in each bin
					if (i<0 || i>=n) continue;		// beyond the range we're calculating
					if (i==0) { d[0]+=rdata[c]*(1.0-a); yc[0]+=(1.0-a); }
					else if (i==n-1) { d[n-1]+=rdata[c]*(a); yc[n-1]+=(a); }
					else {
						h=hypot(rdata[c],rdata[c+1]);
						d[i]+=h*(1.0-a); 
						yc[i]+=(1.0-a); 
						d[i+1]+=h*a; 
						yc[i+1]+=(a); 
					}
				}
			}
		}
		else {
			for (y=c=0; y<ny; y++) {
				for (x=0; x<nx; x+=2,c+=2) {
					r=hypot(y-ny/2.0,x/2.0);
					if (y==ny/2.0 && x==0) a=0;
					else a=(atan2(y-ny/2.0,x/2.0)-a0)/da;
					
					i=(int)floor(a);				// lower bin to put data in
					a-=i;							// float remainder, how much to put in each bin
					if (i<0 || i>=n) continue;		// beyond the range we're calculating
					if (i==0) { d[0]+=rdata[c]*(1.0-a); yc[0]+=(1.0-a); }
					else if (i==n-1) { d[n-1]+=rdata[c]*(a); yc[n-1]+=(a); }
					else {
						h=rdata[c];
						d[i]+=h*(1.0-a); 
						yc[i]+=(1.0-a); 
						d[i+1]+=h*a;
						yc[i+1]+=(a); 
					}
				}
			}
		}
	}
	else {
		for (y=c=0; y<ny; y++) {
			for (x=0; x<nx; x++,c++) {
				//r=hypot(y-(ny-1)/2.0,x-(nx-1)/2.0);
				//if (r<rmin||r>rmax) continue;
				//if (y-(ny-1)/2.0==0 && x-(nx-1)/2.0==0) a=0;
				//else a=atan2(y-(ny-1)/2.0,x-(nx-1)/2.0);
				r=hypot(y-ny/2,x-nx/2);
				if (r<rmin||r>rmax) continue;
				if (y-ny/2==0 && x-nx/2==0) a=0;
				else a=atan2(float(y-ny/2),float(x-nx/2));
				if (a<0) a+=PI*2.0;
				a=(a-a0)/da;
				
				i=(int)floor(a);				// lower bin to put data in
				a-=i;							// float remainder, how much to put in each bin
				if (i<0 || i>=n) continue;		// beyond the range we're calculating
				if (i==0) { d[0]+=rdata[c]*(1.0-a); yc[0]+=(1.0-a); }
				else if (i==n-1) { d[n-1]+=rdata[c]*(a); yc[n-1]+=(a); }
				else { 
					d[i]+=rdata[c]*(1.0-a); 
					yc[i]+=(1.0-a); 
					d[i+1]+=rdata[c]*(a); 
					yc[i+1]+=(a); 
				}
			}
		}
	}
}
else {
	error("Cannot calcazdist 3d images",ERR_CRIT);
}

for (i=0; i<n; i++) d[i]/=yc[i];

doneData();
free(yc);
}

// makes image circularly symmetric (real only)
void EMData::radialAverage() {
float *dist;
int i,x,y,c;
float r;

if (flags&EMDATA_COMPLEX) return;

dist=(float *)malloc(nx/2*sizeof(float));
calcRadDist(nx/2,0,1.0,dist);	// calculate the radial distribution

getData();

for (y=c=0; y<ny; y++) {
	for (x=0; x<nx; x++,c++) {
		r=hypot(x-nx/2.0,y-ny/2.0);

		i=(int)floor(r);						// lower bin to put data in
		r-=i;
		if (i>=0&&i<nx/2-1) rdata[c]=dist[i]*(1.0-r)+dist[i+1]*r;
		else if (i<0) rdata[c]=dist[0];
		else rdata[c]=0;			// outside calculated area set to zero
//		else rdata[c]=dist[nx/2-1];	// outside calculated area set to last value
	}
}

doneData();
update();

free(dist);
}

// subtracts the rotationally averaged image to leave only the asymmetric part
void EMData::radialSubtract() {
float *dist;
int i,x,y,c;
float r;

if (flags&EMDATA_COMPLEX) return;

dist=(float *)malloc(nx/2*sizeof(float));
calcRadDist(nx/2,0,1,dist);	// calculate the radial distribution

getData();

for (y=c=0; y<ny; y++) {
	for (x=0; x<nx; x++,c++) {
		r=hypot(x-nx/2,y-ny/2);

		i=(int)floor(r);						// lower bin to put data in
		r-=i;
		if (i>=0&&i<nx/2-1) rdata[c]-=dist[i]*(1.0-r)+dist[i+1]*r;
		else if (i<0) rdata[c]=0;
		else rdata[c]=0;
	}
}

doneData();
update();

free(dist);
}


// compares 2 data sets. 'this' should be low noise, 'noisy' may be noisy
//  Scaling of 'this' is determined to make the density histogram of
// the difference between the data sets as symmetric as possible
// scale will optionally return the scale factor which would be multiplied by
// 'this' to achieve this normalization
// shift will return the corresponding shift. If modifying 'this', scale should be
// applied first, then b should be added
//
// if invert is 1, then 'this' should be noisy and 'noisy' should not be
//
// smaller is a better match
float EMData::lcmp(EMData *noisy,int keepzero,int invert,float *scale,float *shift) {
float ret=0,*dx,*dy;
int i,n=nx*ny*nz;
float m=1.0,b=0.0;

// First we must determine the proper normalization for 'this' for optimal subtraction
if (!invert) {
	dx=getDataRO();
	dy=noisy->getDataRO();

	lsqfit(n,dx,dy,&m,&b,1);

	doneData();
	noisy->doneData();
}


// If m<0 this implies quite strongly that 'this' is noisy
// Rather than simply fail, we try inverting the least-squares
//if (invert || m<0) { 
//if (m<0) error(ERR_WARN,"Odd normalization in lcmp");
if (invert) { 
	dx=getDataRO();
	dy=noisy->getDataRO();

	lsqfit(n,dy,dx,&m,&b,1);

	doneData();
	noisy->doneData();

	if (m==0) m=1e-20;
	b=-b/m;
	m=1.0/m;

	if (m<0) { b=0; m=1.0; }	// Ok. Lost cause, give up
}

float *d1,*d2,norm=0;
ret=0;
d1=getDataRO();
d2=noisy->getDataRO();
if (keepzero) {
	for (int i=0; i<nx*ny*nz; i++) if (d1[i]&&d2[i]) { ret+=SQR((d1[i]*m)+b-d2[i]); norm+=1.0; }
}
else {
	for (int i=0; i<nx*ny*nz; i++) if (d1[i]&&d2[i]) ret+=SQR((d1[i]*m)+b-d2[i]);
}
doneData();
noisy->doneData();

if (scale) *scale=m;
if (shift) *shift=b;
if (keepzero) return sqrt(ret/norm);
return sqrt(ret/(nx*ny*nz));
}

// Amplitude weighted mean phase difference (radians) with optional
// SNR weight. SNR should be an array as returned by ctfcurve()
// 'data' should be the less noisy image, since it's amplitudes will be used
// to weight the phase residual
// 2d only
float EMData::pcmp(EMData *data,float *snr) {
static float *dfsnr = NULL; //(float *)malloc(200*sizeof(float));
static int nsnr=0;
if (nz>1 || data->zSize()>1 || nx!=data->xSize() || ny!=data->ySize()) return -1.;

if (!snr) {
	snr=dfsnr;
	int NP=ceil(CTFOS*sqrt(2.0)*ny/2)+2;
	if (nsnr!=NP) {
		nsnr=NP;
		dfsnr=(float *)realloc(dfsnr,NP*sizeof(float));
		snr=dfsnr;

		float w=SQR(xSize()/4.0);
		for (int i=0; i<NP; i++) {
			float x2=SQR(i/(float)CTFOS);
			snr[i]=(1.0-exp(-x2/4.0))*exp(-x2/w);
		}
		save_data(0,1.0/CTFOS, snr, NP, "filt.txt");
	}
}

EMData *F1 = doFFT();
//F1->ap2ri();
F1->ri2ap();
EMData *F2 = data->doFFT();
//F2->ap2ri();
F2->ri2ap();
int i;
float x,y;

float *f1=F1->getDataRO();
float *f2=F2->getDataRO();

double sum=0,norm=0.000001;
for (y=-ny/2.0,i=0; y<ny/2.0; y+=1.0) {
	for (x=0; x<nx+2.0; x+=2.0,i+=2) {
		float a2=f2[i];
		int r=round(hypot(x/2,y)*CTFOS);
		float a3=snr[r]*a2;

		float p1=f1[i+1];
		float p2=f2[i+1];
		sum+=asub(p1,p2)*a3;
		norm+=a3;
/*		float a1=hypot(f1[i],f1[i+1]);
		float a2=hypot(f2[i],f2[i+1]);
		int r=round(hypot(x/2,y)*CTFOS);
		float a3=snr[r]*a2;

		float a=acos((f1[i]*f2[i]+f1[i+1]*f2[i+1])/(a1*a2));
		norm+=a3;*/
	}
}

F1->doneData();
F2->doneData();

return (sum/norm);
}

// Dot product of rotated/translated image in-place
// inputs must be same size
// 'this' is rotated/translated
double EMData::dotRT(EMData *with,float dx,float dy,float da) {
double ret=0;
float *d1,*d2;
int i,j;
float y;

if (nx!=with->xSize() || ny!=with->ySize() || nz!=1 || with->zSize()!=1) return 0;

if (parent) d1=parent->getDataRO();
else d1=getDataRO();
d2=with->getDataRO();

float mx0=cos(da);
float mx1=sin(da);
float x2c=nx/2.0-dx;
float y2c=ny/2.0-dy;
for (j=0,y=-ny/2.0; j<ny; j++,y+=1.0) {
	float x2=(mx0*(-nx/2.0-1.0)+mx1*y)+x2c;		// ideally we'd move a bit more of this outside this loop
	float y2=(-mx1*(-nx/2.0-1.0)+mx0*y)+y2c;	// but it might cause too much rounding error
	int ii=floor2(x2);
	int jj=floor2(y2);
	float t=x2-(float)ii;
	float u=y2-(float)jj;
	for (i=0; i<nx; i++) {
		// rotate coordinates, this approach could cause a bit of rounding error
		t+=mx0;		// update x remainder
		u-=mx1;		// "	y remainder
		if (t>=1.0) { ii++; t-=1.0; }	// when remainders are more than 1, update the integers
		if (u>=1.0) { jj++; u-=1.0; }
		if (t<0) { ii--; t+=1.0; }
		if (u<0) { jj--; u+=1.0; }

		if (ii<0||ii>nx-2||jj<0||jj>ny-2) continue;

		int k0=ii+jj*nx;
		int k1=k0+1;
		int k2=k0+nx+1;
		int k3=k0+nx;

		// bilinear interpolation of raw data
		float tt=1.0-t;
		float uu=1.0-u;

		ret+=(d1[k0]*tt*uu+d1[k1]*t*uu+d1[k2]*t*u+d1[k3]*tt*u)*d2[i+j*nx];

	}
}

if (parent) parent->doneData();
else doneData();
with->doneData();

return ret;
}


// dot product of 2 images (same size)
// complex does not check r/i vs a/p
// ummm.... perhaps the r? values should be double...
float EMData::dot(EMData *data,int evenonly) {
float r0=0,r1=0,r2=0,r3=0;
float *d1,*d2,*de;
unsigned int i,j,n,p;

if (nx!=data->xSize() || ny!=data->ySize() || nz!=data->zSize()) return 0;

d1=getDataRO();
d2=data->getDataRO();
n=nx*ny*nz;

if (evenonly) {
	for (i=0; i<n; i+=2,d1+=2,d2+=2) r0+=*d1 * *d2;
} else {
#ifdef OPTIMIZEMEM
	if (n%4==0) {
		for (j=0; j<n; j+=1024) {
			if (j<n-1024) {
				BLOCK_PREFETCH_4K(&d1[j]);
				BLOCK_PREFETCH_4K(&d2[j]);
			}
			int jn=j+1024;
			if (jn>n) jn=n;
			for (i=j; i<jn; i+=4) {
				r0+=d1[i]*d2[i];
				r1+=d1[i+1]*d2[i+1];
				r2+=d1[i+2]*d2[i+2];
				r3+=d1[i+3]*d2[i+3];
			}
		}
	}
	else
#endif
	{
		de=d1+nx*ny*nz;
		while (d1!=de) r0+=*(d1++) * *(d2++);

//		for (i=0; i<n; i++) {
//			r0+=d1[i]*d2[i];
//		}
	}
}

doneData();
data->doneData();
return r0+r1+r2+r3;
}

// similarity between 2 images using FSC. Larger is better.
float EMData::fscmp(EMData *with,float *snr) {
static float *dfsnr = NULL; //(float *)malloc(200*sizeof(float));
static int nsnr=0;

float *FSC = fsc(with);
int i;

if (!snr) {
	snr=dfsnr;
	int NP=ceil(CTFOS*sqrt(2.0)*ny/2)+2;
	if (nsnr!=NP) {
		nsnr=NP;
		dfsnr=(float *)realloc(dfsnr,NP*sizeof(float));
		snr=dfsnr;

//		float w=SQR(xSize()/4.0);
		float w=SQR(xSize()/8.0);
		for (int i=0; i<NP; i++) {
			float x2=SQR(i/(float)CTFOS);
			snr[i]=(1.0-exp(-x2/4.0))*exp(-x2/w);
		}
//		save_data(0,1.0/CTFOS, snr, NP, "filt.txt");
	}
}
//save_data(0,1.0, FSC, ny/2, "fsc.txt");

float sum=0,norm=0;
for (i=0; i<ny/2; i++) {
	sum+=FSC[i]*i*snr[i*CTFOS+CTFOS/2];
	norm+=i*snr[i*CTFOS+CTFOS/2];
//	sum+=FSC[i]*snr[i*CTFOS+CTFOS/2];
//	norm+=snr[i*CTFOS+CTFOS/2];
}

//for (i=0; i<ny/2; i++) FSC[i]=FSC[i]*i*snr[i*CTFOS+CTFOS/2];
//save_data(0,1.0, FSC, ny/2, "fsc2.txt");

free(FSC);
return sum/norm;
}

float *EMData::fsc(EMData *with) {
if (!with || !this) { error(ERR_ERROR,"Null FSC"); return NULL; }
if (with->xSize() != xSize() || with->ySize() != ySize() || with->zSize()!=zSize()) {
	error(ERR_ERROR,"Size mismatch in FSC");
	return NULL;
}

float *ret = (float *)calloc(1,ySize()/2*sizeof(float));
float *n1 = (float *)calloc(1,ySize()/2*sizeof(float));
float *n2 = (float *)calloc(1,ySize()/2*sizeof(float));

EMData *f1;
if (isComplex()) f1=this; else f1=doFFT();
f1->ap2ri();

EMData *f2;
if (with->isComplex()) f2=with; else f2=with->doFFT();
f2->ap2ri();

float *d1 = f1->getDataRO();
float *d2 = f2->getDataRO();

if (nz==1) {
	int i,j;

	for (j=0; j<ySize(); j++) {
		for (i=0; i<f1->xSize(); i+=2) {
			float r=hypot((float)(i/2),(float)j-ySize()/2);
			int ri=round(r);
			if (ri<1||ri>=ySize()/2) continue;
			int ii=i+j*f1->xSize();
			ret[ri]+=d1[ii]*d2[ii]+d1[ii+1]*d2[ii+1];
			n1[ri]+=d1[ii]*d1[ii]+d1[ii+1]*d1[ii+1];
			n2[ri]+=d2[ii]*d2[ii]+d2[ii+1]*d2[ii+1];
		}
	}
}
else {
	int i,j,k;
	
	for (k=0; k<zSize(); k++) {
		for (j=0; j<ySize(); j++) {
			for (i=0; i<f1->xSize(); i+=2) {
				float r=sqrt(hypot3((float)(i/2),(float)j-ySize()/2,(float)k-zSize()/2));
				int ri=round(r);
				if (ri<1||ri>=ySize()/2) continue;
				int ii=i+j*f1->xSize()+k*f1->xSize()*ySize();
				ret[ri]+=d1[ii]*d2[ii]+d1[ii+1]*d2[ii+1];
				n1[ri]+=d1[ii]*d1[ii]+d1[ii+1]*d1[ii+1];
				n2[ri]+=d2[ii]*d2[ii]+d2[ii+1]*d2[ii+1];
			}
		}
	}
}

int i;
for (i=0; i<ySize()/2; i++) ret[i]=ret[i]/(sqrt(n1[i]*n2[i]));

f1->doneData();
f2->doneData();
ret[0]=1.0;
free(n1);
free(n2);
return ret;
}

// This does a normalized dot product of a little image with a big image using
// real-space methods. The result is the same size as 'this', but
// but a border 1/2 the size of 'with' will be zero
// This routine is only efficient when 'with' is fairly small
//
// 2D only
EMData *EMData::littleBigDot(EMData *with,int sigma) {
EMData *ret=copyHead();
ret->zero();
int i,j,ii,jj,k;
int nx2=with->xSize(),ny2=with->ySize();

float em = with->edgeMean();

float *d1=getDataRO();
float *d2=with->getDataRO();
float *d3=ret->getData();

float sum2=(SQR(with->Sigma())+SQR(with->Mean()));

// The SIGMA option is broken...
if (sigma) {
	// Outer loop over results pixels
	for (j=ny2/2; j<ny-ny2/2; j++) {
		for (i=nx2/2; i<nx-nx2/2; i++) {
			float sum=0;
			float sum1=0;
			float summ=0;
		
			// inner loop over 'with' pixels
			for (jj=j-ny2/2,k=0; jj<j+ny2/2; jj++) {
				for(ii=i-nx2/2; ii<i+nx2/2; ii++,k++) { 
					sum1+=SQR(d1[ii+jj*nx]);
					summ+=d1[ii+jj*nx];
					sum+=d1[ii+jj*nx]*d2[k];
				}
			}
			d3[i+j*nx]=sum2+(sum1/2.0-sum)/(nx2*ny2)-SQR(with->Mean()-summ/(nx2*ny2));
		}
	}
}
else {
	// Outer loop over results pixels
	for (j=ny2/2; j<ny-ny2/2; j++) {
		for (i=nx2/2; i<nx-nx2/2; i++) {
			float eml=0;
			float dot=0;
			float dot2=0;

			// this calculates the edgemean of the local box in 'this'
			for (ii=i-nx2/2; ii<i+nx2/2; ii++) eml+=d1[ii+(j-ny2/2)*nx]+d1[ii+(j+ny2/2-1)*nx];
			for (jj=j-ny2/2; jj<j+ny2/2; jj++) eml+=d1[i-nx2/2+jj*nx]+d1[i+nx2/2-1+jj*nx];
			eml/=(nx2+ny2)*2.0;
		
			// inner loop over 'with' pixels
			// Note that this isn't the most efficient way to implement this math
			// but it's fairly transparent and functional
			for (jj=j-ny2/2,k=0; jj<j+ny2/2; jj++) {
				for(ii=i-nx2/2; ii<i+nx2/2; ii++,k++) { 
					dot+=(d1[ii+jj*nx]-eml)*(d2[k]-em);
					dot2+=SQR(d1[ii+jj*nx]-eml);
				}
			}
			dot2=sqrt(dot2);
//			dot2=pow(dot2,(float).25);
			if (dot2==0) d3[i+j*nx]=0;
			else d3[i+j*nx]=dot/(nx2*ny2*dot2*with->Sigma());
//			d3[i+j*nx]=dot/(nx2*ny2);
		}
	}


}

doneData();
with->doneData();
ret->doneData();

return ret;
}

// variety of real space filters
// type 0 is a threshold filter, x<val1 -> 0
// type 1 is just like type 0 with a final normalize
// type 2 is similar, but sets x<val1 -> 0 and x>val -> 1
// type 3 x>val1 -> x, 0<x<val1->0, x<0 -> val2, then normalize
// type 4 x -> fabs(x)
// type 5 |x|-=val, x->0 if |x|<0
// type 6 x!=0 x->1
// type 7 threshold as a fraction of maxval, x<val1*maxval -> 0, normalize
// type 8 Another threshold, x>val1 ->0
// type 9 x = (x+val1)*val2
// type 10 x = sqrt(x)  or -sqrt(-x)
// type 11 x -> exp(x/val1-val2)
//			used in max liklihood, x/val1-val2>40 -> 40
// type 12 x -> 2*(val1-x)		used in max liklihood
// type 13 x>val1 -> val1, x<val2 x-> val2   (histogram tail chopper)
// type 14 val1<=x<=val2 -> 1.0 else 0.0
// type 15 x -> log(x)		behavior complex for complex data
// type 16 x -> e^(x)
// type 17 a 'tail' filter. Sets values >mean+val1*sigma or <mean-val2*sigma to mean
// type 18 x -> x^val1	(val1 defaults to 2.0 if not provided)
// type 19 x->sqrt(x)
// type 20 is a median filter with a +-val1 size, val1=1 -> 3x3 median
// type 21 same box as median filter, but fills in the std dev in the box
// type 22 same box as median filter, but takes max instead of median
// type 30 a 'true' normalization, mean->0, vector len = 1
// type 50 Gradient remover, does a rough plane fit to find linear gradients
// type 51 Vertical stripe filter, tries to fix images scanned on the zeiss for poor ccd normalization
// type 52 Block filter, val1 is dx/dy for calculating an average, val2 is dx/dy for fill/step
// type 53 Block filter, val1 is dx/dy, val2 is lp freq cutoff in pixels
// type 60 fill zeroes at edges with nearest horizontal/vertical value
// type 61 fill constant areas with zero
// type 62 try to eliminate beamstop in electron diffraction patterns, val1=sig multiplier, val2,3 are x,y of center, if val1<0 also does radial subtract
// type 63 fill zeroes at edges with nearest horizontal/vertical value damped towards Mean2
// type 80 Laplacian filter (x -> d^2/dx^2+d^2/dy^2+d^2/dz^2)
// type 90 peak filter -> pixel = pixel - max of values surrounding pixel
// type 91 peak filter -> if surrounding value greater than pixel value -> 0
// type 92 peak filter -> if more than val1 surrounding peaks (26 max) greater than value, value->0
// type 100 x -> radius^2		used in max liklihood
// type 101 x -> ver avg at x, this is ~ what is subtracted in 51
// type 102 blank-out filter for top and bottom blocks; val1: nrows on top; val2: nrows on bottom;
// type 103 x -> radius
// type 104 transmittance to optical density conversion for images
//          acquired by Nikon scanner
// type 105: zero edges of image, val1 is rows on top and bottom(y), val2 is rows on left and right(x), for 3d, val3 is along z axis
// type 106: x -> cylindrical radius from z axis (for cylindrical colormaps)
// type 110: Iterative expansion of a binary mask, val1 is number of pixels to expand, if val2!=0 will make a soft Gaussian edge starting after val2 pixels
// type 111: Iterative expansion of a mulitlevel integer mask
// type 112: Iterative contraction of data, if any nearest neighbor is 0, value -> 0
// type 113 :substract the local mean to remove gradient 
// type 114 :int(x) != x -> x=0, for building complicated masks 
// type 120: interlacing translation repair filter
// type 130: remove gradient by least square fitting a plane to the pixels in the radius range of val1 nad val2
// type 140: for A/P complex image, filter amplitude by threshhold, < threshold -> 0, >= threshold keep unchanged
// type 141: for A/P complex image, filter amplitude by threshhold, < threshold -> 0, >= threshold -> 1
void EMData::realFilter(int type, float val1, float val2,float val3) {
float *d,*d2,v;
int i,j,k,l,n;

switch (type) {
case 0:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) if (*d<val1) *d=0;
	doneData();
	break;
case 1:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) if (*d<val1) *d=0;
	doneData();
	normalize();
	break;
case 2:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) if (*d<val1) *d=0; else *d=1.0;
	doneData();
//	normalize();
	break;
case 3:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		if (*d<val1) {
			if (*d>0) *d=0;
			else *d=val2;
		}
	}
	doneData();
	normalize();
	break;
case 4:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) *d=fabs(*d);
	doneData();
	break;
case 5:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		if (*d<0) { *d+=val1; if (*d>0) *d=0; }
		else { *d-=val1; if (*d<0) *d=0; }
	}
	doneData();
	break;
case 6:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) if (*d!=0) *d=1.0;
	doneData();
	break;
case 7:
	d=getData();
	val1*=Max();
	for (i=0; i<nx*ny*nz; i++,d++) {
		if (*d<val1) *d=0;
	}
	doneData();
	normalize();
	break;
case 8:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) if (*d>val1) *d=0;
	doneData();
	break;
case 9:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		*d=(*d+val1)*val2;
	}
	doneData();
	break;
case 10:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		*d=copysign(sqrt(*d<0?(-*d):(*d)),*d);
	}
	doneData();
	normalize();
	break;
case 11:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		v=*d/val1-val2;
		if (v>40.0) { /*printf("overval %f\t%f\t%f\n",*d,val1,val2*val1);*/ *d=exp(40.0); }
		else *d=exp(v);
	}
	doneData();
	break;
case 12:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		*d=2.0*(val1- *d);
	}
	doneData();
	break;
case 13:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		if (*d>val1) *d=val1;
		if (*d<val2) *d=val2;
	}
	doneData();
	break;
case 14:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		if (*d>=val1 && *d<=val2) *d=1.0;
		else *d=0.0;
	}
	doneData();
	break;
case 15:
	if (isComplex()){
		ri2ap();
		d=getData();
/*		float m=1.0e20;
		for (i=0; i<nx*ny*nz; i+=2,d+=2) {
			if (*d>0 && *d<m) m=*d;
		}*/
//		printf("min %f\n",m);

		for (i=0; i<nx*ny*nz; i+=2,d+=2) {
			if (*d<=1.0e-6) *d=.1;
			else *d=log10(*d)+6.1;
		}
		doneData();
		ap2ri();
	}
	else {
		d=getData();
		for (i=0; i<nx*ny*nz; i++,d++) {
			*d=log10(*d);
		}
		doneData();
	}
	break;
case 16:
	if (isComplex()){
		ri2ap();
		d=getData();
		for (i=0; i<nx*ny*nz; i+=2,d+=2) {
			if (*d>26.1) *d=20.;
			*d=pow(10.0,*d-6.1);
		}
		doneData();
		ap2ri();
	}
	else {
		d=getData();
		for (i=0; i<nx*ny*nz; i++,d++) {
			*d=exp(*d);
		}
		doneData();
	}
	break;
case 17:
	{
	float th1=Mean()-val2*Sigma();
	float th2=Mean()+val1*Sigma();
	float v=Mean();
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		if ((*d<th1) | (*d>th2)) *d=v;
	}
	doneData();
	break;
	}
case 18:
	{
	d=getData();
	if (val1==0 || val1==2.0) {
		for (i=0; i<nx*ny*nz; i++) {
			d[i]*=d[i];
		}
	}
	else {
		for (i=0; i<nx*ny*nz; i++) {
			d[i]=pow(d[i],val1);
		}
	}

	doneData();
	break;
	}
case 19:
	d=getData();
	for (i=0; i<nx*ny*nz; i++,d++) {
		*d=sqrt(*d);
	}
	doneData();
	break;
case 20:
	{
	int i2,j2,k2;
	EMData *dat=copy(0,0);
	n=round(val1);
	floatStat stat((n*2+1)*(n*2+1)*(n*2+1));	// maximum size
	normalize();
	d=getData();
	d2=dat->getDataRO();
	if (nz>1) {
		for (k=n; k<nz-n; k++) {
			if (updProgress && updProgress(k-n,nz-2*n,"Filtering")) break;
			for (j=n; j<ny-n; j++) {
				for (i=n; i<nx-n; i++) {
					stat.clear();

					for (i2=i-n; i2<=i+n; i2++) {
						for (j2=j-n; j2<=j+n; j2++) {
							for (k2=k-n; k2<=k+n; k2++) {
								stat.qadd(d2[i2+j2*nx+k2*nx*ny]);
							}
						}
					}

					d[i+j*nx+k*nx*ny]=stat.median();
				}
			}
		}
	}
	else {
		for (j=n; j<ny-n; j++) {
			if (updProgress && updProgress(j-n,ny-2*n,"Filtering")) break;
			for (i=n; i<nx-n; i++) {
				stat.clear();

				for (i2=i-n; i2<=i+n; i2++) {
					for (j2=j-n; j2<=j+n; j2++) {
							stat.qadd(d2[i2+j2*nx]);
						}
					}

				d[i+j*nx]=stat.median();
			}
		}
	}
	if (updProgress) updProgress(0,0,NULL);

	dat->doneData();
	doneData();
	delete dat;
	break;
	}
case 21:
	{
	int i2,j2,k2;
	EMData *dat=copy(0,0);
	n=round(val1);
	floatStat stat((n*2+1)*(n*2+1)*(n*2+1));	// maximum size
	normalize();
	d=getData();
	d2=dat->getDataRO();
	if (nz>1) {
		printf("Realfilter type 21 not supported in 3D models\n");
		exit(1);
	}
	else {
		for (j=n; j<ny-n; j++) {
			if (updProgress && updProgress(j-n,ny-2*n,"Filtering")) break;
			for (i=n; i<nx-n; i++) {
				stat.clear();

				for (i2=i-n; i2<=i+n; i2++) {
					for (j2=j-n; j2<=j+n; j2++) {
							stat.qadd(d2[i2+j2*nx]);
						}
					}

				d[i+j*nx]=stat.sigma();
			}
		}
	}
	if (updProgress) updProgress(0,0,NULL);

	dat->doneData();
	doneData();
	delete dat;
	break;
	}
case 22:
	{
	int i2,j2,k2;
	EMData *dat=copy(0,0);
	n=round(val1);
	floatStat stat((n*2+1)*(n*2+1)*(n*2+1));	// maximum size
	normalize();
	d=getData();
	d2=dat->getDataRO();
	if (nz>1) {
		printf("Realfilter type 21 not supported in 3D models\n");
		exit(1);
	}
	else {
		for (j=n; j<ny-n; j++) {
			if (updProgress && updProgress(j-n,ny-2*n,"Filtering")) break;
			for (i=n; i<nx-n; i++) {
				stat.clear();

				for (i2=i-n; i2<=i+n; i2++) {
					for (j2=j-n; j2<=j+n; j2++) {
							stat.qadd(d2[i2+j2*nx]);
						}
					}

				d[i+j*nx]=stat.Max();
			}
		}
	}
	if (updProgress) updProgress(0,0,NULL);

	dat->doneData();
	doneData();
	delete dat;
	break;
	}
case 30:
	d=getData();
	val1=-Mean();
	val2=1.0/(Sigma()*xSize()*ySize());
	for (i=0; i<nx*ny*nz; i++,d++) {
		*d=(*d+val1)*val2;
	}
	doneData();
	break;

// see mode 130 for a least square 2-D plane fit - Wen Jiang
case 50:		// stupid, this should be done in 2d not 2x 1d, but I'm in a hurry
	{
	d=getData();
	float *dy,sly=0,slx;
	float m,b;
	dy=(float *)malloc(ny*sizeof(float));
	for (i=0; i<ny; i++) {
		lsqfit(nx,NULL,d+i*nx,&m,&b,0);
		dy[i]=b;
		sly+=m;
	}
	sly/=ny;
	lsqfit(ny,NULL,dy,&slx,&b,0);
	
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i++) {
			d[i+j*nx]-=i*slx+j*sly+b;
		}
	}
	doneData();
	free(dy);
	break;
	}
case 51:
	{
	float sig=Sigma();
	float *dat=getData();
	double av;
	int x,y;
	for (x=0; x<nx; x++) {
		av=0;
		for (y=ny/4; y<3*ny/4; y++) {	// average over only the middle 1/2
			av+=dat[x+y*nx];
		}
		av/=(ny/2);
	
		for (y=0; y<ny; y++) {
			dat[x+y*nx]=(dat[x+y*nx]-av)/sig;
		}
	}
	doneData();
	}
	break;
case 52:
	{
	float *dat=getData();
	float av;
	int x,y,x1,y1;
	int v1=(int)val1,v2=(int)val2,v0=v1>v2?v1:v2;
	if (v2<=0) v2=v1;
	for (y=v0; y<=ny-v0-1; y+=v2) {
		for (x=v0; x<=nx-v0-1; x+=v2) {
			av=0;
			for (y1=y-v1; y1<=y+v1; y1++) {
				for (x1=x-v1; x1<=x+v1; x1++) {
					av+=dat[x1+y1*nx];
				}
			}
			av/=(v1*2+1)*(v1*2+1);
			for (y1=y-v2; y1<=y+v2; y1++) {
				for (x1=x-v2; x1<=x+v2; x1++) {
					dat[x1+y1*nx]=av;
				}
			}
		}
	}
	doneData();
	}
	break;
case 53:
	{
	float av;
	int x,y,x1,y1;
	int v1=(int)val1,v2=(int)val2;
	if (v2>v1/2) { error(ERR_WARN,"invalid val2 in realfilter 53"); v2=v1/2-1; }
	if (v2<=0) v2=v1;
	for (y=0; y<=ny-v1; y+=v1) {
		for (x=0; x<=nx-v1; x+=v1) {
			EMData *cl = clip(x,y,v1,v1); 
			EMData *ff = cl->doFFT();
			float *dt=ff->getDataRO();
			float av=0,no=0;
			for (y1=-v2; y1<v2; y1++) {
				for (x1=0; x1<v2; x1++) {
					if (x1==0 && y1==0) continue;
					if (hypot((float)x1,(float)y1)<val2) { av+=hypot2(dt[x1*2+(y1+v1/2)*(v1+2)],dt[x1*2+(y1+v1/2)*(v1+2)+1]); no+=1.0; }
				}
			}
			ff->doneData();
			delete cl;
			av/=no;
			for (y1=y; y1<y+v1; y1++) {
				for (x1=x; x1<x+v1; x1++) {
					setValueAt(x1,y1,0,av);
				}
			}
		}
	}

	for (y1=y; y1<ny; y1++) {
		for (x1=0; x1<nx; x1++) {
			setValueAt(x1,y1,0,0);
		}
	}
	for (y1=0; y1<ny; y1++) {
		for (x1=x; x1<nx; x1++) {
			setValueAt(x1,y1,0,0);
		}
	}

	getData();
	doneData();
	}
	break;
case 60:
	d=getData();
	for (j=0; j<ny; j++) {
		// left fill
		for (i=0; i<nx-1; i++) if (d[i+j*nx]!=0) break;
		v=d[i+j*nx];
		while (i>=0) { d[i+j*nx]=v; i--; }
		
		// right fill
		for (i=nx-1; i>0; i--) if (d[i+j*nx]!=0) break;
		v=d[i+j*nx];
		while (i<nx) { d[i+j*nx]=v; i++; }
	}

	for (i=0; i<nx; i++) {
		// bottom fill
		for (j=0; j<ny; j++) if (d[i+j*nx]!=0) break;
		v=d[i+j*nx];
		while (j>=0) { d[i+j*nx]=v; j--; }

		// top fill
		for (j=ny-1; j>0; j--) if (d[i+j*nx]!=0) break;
		v=d[i+j*nx];
		while (j<ny) { d[i+j*nx]=v; j++; }
	}
	doneData();
	break;
case 61:
	d=getData();
	for (j=1; j<ny-1; j++) {
		for (i=1; i<nx-1; i++) {
			int k=i+j*nx;
			float v1=d[k];
			if (v1==0) continue;
			if (v1==d[k-1] || v1==d[k+1] || v1==d[k-nx] || v1==d[k+nx] ||
				0==d[k-1] || 0==d[k+1] || 0==d[k-nx] || 0==d[k+nx]) d[k]=0;
		}
	}
	doneData();
	break;
case 62:
	{
	float mi=Min();
//	float thr=Min()+Sigma()/10.0;
	float thr=fabs(val1);
	d=getData();
	int cenx=(int)val2;
	int ceny=(int)val3;
	if (cenx<=0) cenx=xSize()/2;
	if (ceny<=0) ceny=ySize()/2;
	
	int mxr=(int)floor(sqrt(2.0)*xSize()/2);
	float *val = (float *)malloc(mxr*sizeof(float));
	float *vals = (float *)malloc(mxr*sizeof(float));
	double sum,cnt,sig;
	//calculate mean values around rings ignoring values below threshold
	for (i=0; i<mxr; i++) {
		sum=cnt=sig=1.0e-30;
		for (j=0; j<6*i+2; j++) {
			float ang=j*2.*PI/(6.*i+2.);
			int x0=(int)floor(cos(ang)*i+cenx);
			int y0=(int)floor(sin(ang)*i+ceny);
			if (x0<0||y0<0||x0>=xSize()||y0>=ySize()) continue;
			sum+=d[x0+y0*xSize()];
			sig+=SQR(d[x0+y0*xSize()]);
			cnt+=1.0;
		}
		val[i]=sum/cnt;
		vals[i]=sqrt(sig/cnt-SQR(val[i]));
//		printf("%d. %g\t%g\t%g\t%g\t%g\n",i,sig/cnt,sum*sum/(cnt*cnt),cnt,val[i],vals[i]);
	}
	save_data(0,1, val, mxr, "mean.txt");
	save_data(0,1, vals, mxr, "sig.txt");

	//now we move away from the mean by excluding values far from the mean
	// this will reduce the influence of the diffraction spots if present
	for (k=0; k<5; k++) {
	for (i=0; i<mxr; i++) {
		sum=cnt=sig=1.0e-30;
		for (j=0; j<6*i+2.; j++) {
			float ang=j*2.*PI/(6.*i+2.);
			int x0=(int)floor(cos(ang)*i+cenx);
			int y0=(int)floor(sin(ang)*i+ceny);
			if (x0<0||y0<0||x0>=xSize()||y0>=ySize()||
				d[x0+y0*xSize()]<val[i]-vals[i]*thr || d[x0+y0*xSize()]>val[i]+vals[i]) continue;
			sum+=d[x0+y0*xSize()];
			sig+=SQR(d[x0+y0*xSize()]);
			cnt+=1.0;
		}
		val[i]=sum/cnt;
		vals[i]=sqrt(sig/cnt-val[i]*val[i]);
	}
	}
//	save_data(0,1, val, mxr, "mean.txt");
//	save_data(0,1, vals, mxr, "sig.txt");
	
	
	for (i=0; i<xSize(); i++) {
		for (j=0; j<ySize(); j++) {
			int r=round(hypot((float)i-cenx,(float)j-ceny));
			if (val1<0) {
				if (d[i+j*xSize()]<val[r]-vals[r]*thr) d[i+j*xSize()]=0;
				else d[i+j*xSize()]-=val[r];
				continue;
			}
			if (d[i+j*xSize()]>val[r]-vals[r]*thr) continue;
			d[i+j*xSize()]=val[r];
		}
	}
			
	free(val);
	free(vals);
	doneData();
	break;
	}
case 63:
	{
	float m2=Mean2();
	d=getData();
	for (j=0; j<ny; j++) {
		// left fill
		for (i=0; i<nx-1; i++) if (d[i+j*nx]!=0) break;
		if (i==nx-1) i=-1;
		v=d[i+j*nx]-m2;
		while (i>=0) { v*=.9; d[i+j*nx]=v+m2; i--; }
		
		// right fill
		for (i=nx-1; i>0; i--) if (d[i+j*nx]!=0) break;
		if (i==0) i=nx;
		v=d[i+j*nx]-m2;
		while (i<nx) { v*=.9; d[i+j*nx]=v+m2; i++; }
	}

	for (i=0; i<nx; i++) {
		// bottom fill
		for (j=0; j<ny; j++) if (d[i+j*nx]!=0) break;
		v=d[i+j*nx]-m2;
		while (j>=0) { v*=.9; d[i+j*nx]=v+m2; j--; }

		// top fill
		for (j=ny-1; j>0; j--) if (d[i+j*nx]!=0) break;
		v=d[i+j*nx]-m2;
		while (j<ny) { v*=.9; d[i+j*nx]=v+m2; j++; }
	}
	doneData();
	}
	break;
case 80:		// Laplacian, not done very intelligently yet
  
  {
   
    d=getData();

    if (nz==1){// 2d images
	d2 = (float *)malloc(sizeof(float)*nx*ny);
	memcpy(d2,d,sizeof(float)*nx*ny);
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i++) {
			if (i==0||j==0||i==nx-1||j==ny-1) { d[i+j*nx]=0; continue; }
			d[i+j*nx]-=.25*(d2[i+j*nx-1]+d2[i+j*nx+1]+d2[i+j*nx-nx]+d2[i+j*nx+nx]);

			
		}
	}
	}
    else if (nz>1){//3d image: need to optimize for memory-crrently it save copy of the whole map. htet khant
	d2 = (float *)malloc(sizeof(float)*nx*ny*nz);
	memcpy(d2,d,sizeof(float)*nx*ny*nz);
	for (k=0; k <nz; k++) {
	for (j=0; j <ny; j++) {
		for (i=0; i<nx; i++) {
		  if (i==0||j==0||k==0||i==nx-1||j==ny-1||k==nz-1) { d[i+j*nx+k*nx*nx]=0; continue; }
		  d[i+j*nx+k*nx*ny]-=(1.0/6.0)*(d2[i+j*nx+k*nx*ny-1]+d2[i+j*nx+k*nx*ny+1]+d2[i+j*nx-nx+k*nx*ny]+d2[i+j*nx+nx+k*nx*ny]+d2[i+j*nx+k*nx*ny+nx*ny]+d2[i+j*nx+k*nx*ny-nx*ny]);

		}
	}
	}
	}
	doneData();
	free(d2);
}
	break;
case 90:		// this one looks at the surrounding pixels and subtracts
				// the largest value from the current pixel
				// This is a sort of positive peak-finding algorithm
  //if (nz!=1) { printf("Realfilter 90 works only on 2d images\n"); exit(1); }
	d=getData();
	d2 = (float *)malloc(sizeof(float)*nx*ny);
	memcpy(d2,d,sizeof(float)*nx*ny);
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i++) {
			if (i<=1||j<=1||i>=nx-2||j>=ny-2) { d[i+j*nx]=0; continue; }
			float v=d2[i+j*nx-1];
			int ii,jj;
			for (jj=j-2; jj<=j+2; jj++) {
				for (ii=i-2; ii<=i+2; ii++) {
					if (ii==i && jj==j) continue;
					if (d2[ii+jj*nx]>v) v=d2[ii+jj*nx];
				}
			}
			d[i+j*nx]-=v;
		}
	}
	doneData();
	free(d2);
	break;
case 91:		// this one looks at the surrounding pixels
				// and zeroes pixels with a higher value nearby
				// finds only peaks with positive values
				// note in 2d, the search range is +-2 pixels
				// and in 3d only +-1 pixel
	if (nz==1) {
		d=getData();
		d2 = (float *)malloc(sizeof(float)*nx*ny);
		memcpy(d2,d,sizeof(float)*nx*ny);
		for (j=0; j<ny; j++) {
			for (i=0; i<nx; i++) {
				if (i<=1||j<=1||i>=nx-2||j>=ny-2) { d[i+j*nx]=0; continue; }
				float v=d2[i+j*nx-1];
				int ii,jj;
				for (jj=j-2; jj<=j+2; jj++) {
					for (ii=i-2; ii<=i+2; ii++) {
						if (ii==i && jj==j) continue;
						if (d2[ii+jj*nx]>v) v=d2[ii+jj*nx];
					}
				}
				if (v>d[i+j*nx]) d[i+j*nx]=0;
			}
		}
		doneData();
		free(d2);
		break;
	}
	// 3D
	d=getData();
	d2 = (float *)malloc(sizeof(float)*nx*ny*nz);
	memcpy(d2,d,sizeof(float)*nx*ny*nz);
	for (k=0; k<nz; k++) {
		for (j=0; j<ny; j++) {
			for (i=0; i<nx; i++) {
				if (i<1||j<1||k<1||i>nx-2||j>ny-2||k>nz-2) { setValueAt(i,j,k,0); continue; }
				float v=d2[i+j*nx+k*nx*ny-1];
				int ii,jj,kk;
				for (kk=k-1; kk<=k+1; kk++) {
					for (jj=j-1; jj<=j+1; jj++) {
						for (ii=i-1; ii<=i+1; ii++) {
							if (ii==i && jj==j && kk==k) continue;
							if (d2[ii+jj*nx+kk*nx*ny]>v) v=d2[ii+jj*nx+kk*nx*ny];
						}
					}
				}
				if (v>d[i+j*nx+k*nx*ny]) setValueAt(i,j,k,0);
			}
		}
	}
	doneData();
	free(d2);
	break;
case 92:
	if (nz==1) {
		d=getData();
		d2 = (float *)malloc(sizeof(float)*nx*ny);
		memcpy(d2,d,sizeof(float)*nx*ny);
		for (j=0; j<ny; j++) {
			for (i=0; i<nx; i++) {
				if (i<1||j<1||i>nx-2||j>ny-2) { d[i+j*nx]=0; continue; }
				float v=d2[i+j*nx];
				int ii,jj;
				int ng=0;
				for (jj=j-1; jj<=j+1; jj++) {
					for (ii=i-1; ii<=i+1; ii++) {
						if (ii==i && jj==j) continue;
						if (d2[ii+jj*nx]>v) ng++;
					}
				}
				if (ng>=val1) d[i+j*nx]=0;
			}
		}
		doneData();
		free(d2);
		break;
	}
	// 3D
	d=getData();
	d2 = (float *)malloc(sizeof(float)*nx*ny*nz);
	memcpy(d2,d,sizeof(float)*nx*ny*nz);
	for (k=0; k<nz; k++) {
		for (j=0; j<ny; j++) {
			for (i=0; i<nx; i++) {
				if (i<1||j<1||k<1||i>nx-2||j>ny-2||k>nz-2) { setValueAt(i,j,k,0); continue; }
				float v=d2[i+j*nx+k*nx*ny];
				int ii,jj,kk;
				int ng=0;
				for (kk=k-1; kk<=k+1; kk++) {
					for (jj=j-1; jj<=j+1; jj++) {
						for (ii=i-1; ii<=i+1; ii++) {
							if (ii==i && jj==j && kk==k) continue;
							if (d2[ii+jj*nx+kk*nx*ny]>v) ng++;
						}
					}
				}
				if (ng>=val1) setValueAt(i,j,k,0);
			}
		}
	}
	doneData();
	free(d2);
	break;
case 100:
	if (nz!=1) { printf("Realfilter 100 works only on 2d images\n"); exit(1); }
	d=getData();
	for (i=0; i<ny; i++) {
		for (j=0; j<nx; j++,d++) {
			*d=(i-ny/2)*(i-ny/2)+(j-nx/2)*(j-nx/2);
		}
	}
	doneData();
	break;
case 101:
	{
	float *dat=getData();
	double av;
	int x,y;
	for (x=0; x<nx; x++) {
		av=0;
		for (y=0; y<ny; y++) {
			av+=dat[x+y*nx];
		}
		av/=(ny);

		for (y=0; y<ny; y++) {
			dat[x+y*nx]=av;
		}
	}
	doneData();	if (nz!=1) { printf("Realfilter 100 works only on 2d images\n"); exit(1); }

	}
	break;
case 102:
     {
	 float* dat = getData();
	 int bnrows = (int) val1;
	 int tnrows = (int) val2;

	 if (bnrows > 0) {		memset(d+i*nx,0,(int)val2*4);

	     double bsum = 0;;
	     int i = 0;
	     int nitems = bnrows * nx;

	     for (i = 0; i < nitems; i++) {
		 bsum += dat[i];
	     }

	     double bavg = bsum / nitems;

	     for ( i = 0; i < nitems; i++) {
		 dat[i] = 0;
	     }
	 }

	 if (tnrows > 0) {
	     double tsum = 0;
	     int j = 0;
	     int nitems = (tnrows * nx);
	     int total_items = nx * ny;

	     for (j = total_items - nitems; j < total_items; j++) {
		 tsum += dat[j];
	     }

	     double tavg = tsum / nitems;

	     for (j = total_items - nitems; j < total_items; j++) {
		 dat[j] = 0;
	     }
	 }
	 doneData();
     }
     break;
case 103:
	if (nz!=1) {
		d=getData();
		for (k=0; k<nz; k++) {
			for (j=0; j<ny; j++) {
				for (i=0; i<nx; i++,d++) {
					*d=sqrt(hypot3(i-nx/2,j-ny/2,k-nz/2));
				}
			}
		}
		doneData();
		break;
	}
	d=getData();
	for (i=0; i<ny; i++) {
		for (j=0; j<nx; j++,d++) {
			*d=(i-ny/2)*(i-ny/2)+(j-nx/2)*(j-nx/2);
		}
	}
	doneData();
	break;
#ifdef TIF
 case 104://transmittance to optical density conversion for Nikon
        d=getData();
	//cout<<"(***)samplebits  "<<samplebits<<endl;
	if(samplebits==8){
	   for (i=0; i<nx*ny*nz; i++){
	      if(d[i]>0) d[i]=log10(d[i]/255);//for 8-bits format
	      else d[i]=0;
	   }
	}
	else if(samplebits==16){
	   for (i=0; i<nx*ny*nz; i++){
	       if(d[i]>0) d[i]=log10(d[i]/65535);//for 16-bits format
	       else d[i]=0;
	   }
	}
	else{
	   printf("can't convert trans -> Opt Den");
	   exit(1);
	}
	doneData();
        break;
#endif
case 105:
	if (nz==1) { 
		d=getData();
		memset(d,0,nx*(int)val1*4);
		memset(d+(ny-(int)val1)*nx,0,nx*(int)val1*4);
		for (i=(int)val1; i<ny-(int)val1; i++) {
			memset(d+i*nx,0,(int)val2*4);
			memset(d+i*nx+nx-(int)val2,0,(int)val2*4);
		}
		doneData();
	}
	else {	// 3D
		d=getData();
		// doesn't need to be fast, do it the easy/very slow way
		if (val1<0) val1=ny/2-val1;
		if (val2<0) val2=nx/2-val2;
		if (val3<0) val3=nz/2-val3;
		for (k=0; k<nz; k++) {
			for (j=0; j<ny; j++) {
				for (i=0; i<nx; i++) {
					if (i<val2 || i>nx-val2 || j<val1 || j>ny-val1 || k<val3 || k>nz-val3) d[i+j*nx+k*nx*ny]=0;
				}
			}
		}
		doneData();
	}
	break;
case 106:
	if (nz!=1) {
		d=getData();
		for (k=0; k<nz; k++) {
			for (j=0; j<ny; j++) {
				for (i=0; i<nx; i++,d++) {
					*d=hypot(i-nx/2,j-ny/2);
				}
			}
		}
		doneData();
		break;
	}
	d=getData();
	for (i=0; i<ny; i++) {
		for (j=0; j<nx; j++,d++) {
			*d=(i-ny/2)*(i-ny/2)+(j-nx/2)*(j-nx/2);
		}
	}
	doneData();
	break;
case 107:
	/* COSINE EDGE MASKING		--Rui Zhang
	 * val1 	rad
	 * val2		falloff				
	 * val3		B			background
	 * */
	{	
		d = getData();
		int dx = 0; //the distance from the pixel to nx/2
		float weight = 0.0f;
		if(nz==1){
			float total = 0;
			if(val3 == 1){
				for(j=0; j<ny; ++j){
					for(i=0; i<nx; ++i, d++){
						total += *d;
					}
				}
				val3 = total/(nx*ny);
			}
			for(j=0; j<ny; ++j){
				for(i=0; i<nx; ++i, ++d){
					dx = abs(i-(int)(nx/2));
					if(dx > val1) {
#ifdef _WIN32
						weight = (1.0f + cos(PI * _cpp_min(1.0f, (float)(dx-val1)/(float)val2))) * 0.5f;
#else					
						weight = (1.0f + cos(PI * std::min(1.0f, (float)(dx-val1)/(float)val2))) * 0.5f;
#endif	//_WIN32
						*d = val3 + (*d - val3) * weight;	
					}
				}
			}
		}
		else {
			printf("2D image only");
		}
		doneData();
	}
	break;
case 110:
	if (nz!=1) {
		d=getData();
		for (i=0; i<nx*ny*nz; i++) {
			if (d[i]==0) d[i]=100000.0;
			else d[i]=1.0;
		}
		for (l=2; l<=(int)val1; l++) {
			for (k=1; k<nz-1; k++) {
				for (j=1; j<ny-1; j++) {
					for (i=1; i<nx-1; i++) {
						int t=i+j*nx+k*(nx*ny);
						if (d[t]==100000.0 && (d[t-1]<=l || d[t+1]<=l || d[t+nx]<=l || d[t-nx]<=l || d[t+nx*ny]<=l || d[t-nx*ny]<=l)) d[t]=l+1;
					}
				}
			}
		}
		for (i=0; i<nx*ny*nz; i++) {
			if (d[i]==100000.0) d[i]=0;
			else if (d[i]>val2) d[i]=exp(-3.0*SQR(d[i]-val2)/SQR(val1-val2));
			else d[i]=1.0;
		}
		doneData();
		break;
	}
	d=getData();
	for (i=0; i<nx*ny; i++) if (d[i]==0) d[i]=100000.0;
	for (l=2; l<=(int)val1; l++) {
		for (j=1; j<ny-1; j++) {
			for (i=1; i<nx-1; i++) {
				int t=i+j*nx;
				if (d[t]==100000.0 &&(d[t-1]<=l || d[t+1]<=l || d[t+nx]<=l || d[t-nx]<=l)) d[t]=l+1;
			}
		}
	}
	for (i=0; i<nx*ny; i++) {
		if (d[i]==100000.0) d[i]=0;
		else if (d[i]>val2) d[i]=exp(-3.0*SQR(d[i]-val2)/SQR(val1-val2));
		else d[i]=1.0;
	}
	doneData();
	break;
case 111:
	if (nz==0) { printf("Realfilter 111 only works on 3d volumes\n"); return; }
// note that this algorithm is limited to #levels < ~10000
	d=getData();
	for (v=0.001; v<val1/1000.0+.001; v+=.001) {
		for (k=1; k<nz-1; k++) {
			for (j=1; j<ny-1; j++) {
				for (i=1; i<nx-1; i++) {
					int t=i+j*nx+k*(nx*ny);
					if (d[t]) continue;
					float lv[6];
					lv[0]=d[t-1];
					lv[1]=d[t+1];
					lv[2]=d[t-nx];
					lv[3]=d[t+nx];
					lv[4]=d[t-nx*ny];
					lv[5]=d[t+nx*ny];
					for (l=0; l<6; l++) {
						if (lv[l] && lv[l]-floor(lv[l])<v-.0005) break;
					}
					if (l<6) d[t]=floor(lv[l])+v;
				}
			}
		}
	}
	for (i=0; i<nx*ny*nz; i++) d[i]=floor(d[i]);
	doneData();
	break;
case 112:
	if (nz!=1) {
		d=getData();
		for (l=1; l<=(int)val1; l++) {
			for (k=1; k<nz-1; k++) {
				for (j=1; j<ny-1; j++) {
					for (i=1; i<nx-1; i++) {
						int t=i+j*nx+k*(nx*ny);
						if (!d[t]) continue;
						if (d[t-1]==0 || d[t+1]==0 || d[t+nx]==0 || d[t-nx]==0 || d[t+nx*ny]==0 || d[t-nx*ny]==0)
							d[t]=1.011e6;
					}
				}
			}
			for (i=0; i<nx*ny*nz; i++) if (d[i]==1.011e6) d[i]=0;
		}
		doneData();
		break;
	}
	d=getData();
	for (l=1; l<=(int)val1; l++) {
		for (j=1; j<ny-1; j++) {
			for (i=1; i<nx-1; i++) {
				int t=i+j*nx;
				if (!d[t]) continue;
				if (d[t-1]==0 || d[t+1]==0 || d[t+nx]==0 || d[t-nx]==0)
					d[t]=1.011e6;
			}
		}
		for (i=0; i<nx*ny*nz; i++) if (d[i]==1.011e6) d[i]=0;
	}
	doneData();
	break;
case 113:
{	// val1 is the box size to compute the local mean
	int boxsize = val1;
	EMData* tmp_mean = copy();
	tmp_mean->meanShrink(boxsize);
	int mnx = tmp_mean->xSize(), mny = tmp_mean->ySize();
	float *fp=getData();
	float *tfp=tmp_mean->getDataRO();
	for(int j=0; j<ny; j++) {
		for(int i=0; i<nx; i++){
			float y = float(j-boxsize/2)/float(boxsize);
			float x = float(i-boxsize/2)/float(boxsize);
			int jj = int(y);
			int ii = int(x);
			if (jj<0) jj=0;
			else if (jj>=mny-1) jj=mny-2;
			if (ii<0) ii=0;
			else if (ii>=mnx-1) ii=mnx-2;
			float t = bilin(tfp[jj*mnx+ii], tfp[jj*mnx+ii+1], tfp[jj*mnx+ii+mnx+1], tfp[jj*mnx+ii+mnx], x-ii, y-jj);
			fp[j*nx+i] = (fp[j*nx+i]-t);
		}
	}	
	tmp_mean->doneData();
	doneData();
	delete 	tmp_mean;
	break;	
}
case 114:
{
	float *fp=getData();
	for(int i=0; i<nx*ny*nz; i++) {
		if(float(int(fp[i]))!= fp[i]) fp[i]=0;
		//fp[i]=round(fp[i]);
	}
	doneData();
	break;	
}
case 120:
// This is an interlacing frame repair filter
// if there is overall motion in a static image (camera shift), this 
// can create interlacing artifacts in the image. This will attempt
// to repair this problem
	{
	EMData tmp1,tmp2;
	tmp1.setSize(nx,ny/2,1);
	tmp2.setSize(nx,ny/2,1);
	
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i++) {
			if (j&1)  tmp2.setValueAt(i,j/2,0,valueAt(i,j));
			else tmp1.setValueAt(i,j/2,0,valueAt(i,j));
		}
	}
	tmp1.update();
	tmp2.update();

	tmp1.transAlign(&tmp2,0,1,nx/4);
	tmp1.rotateAndTranslate();

	for (j=0; j<ny; j+=2) {
		for (i=0; i<nx; i++) {
			setValueAt(i,j,0,tmp1.valueAt(i,j/2,0));
		}
	}
	}
	update();
	break;
case 130:
{
	if(nz>1) {
		fprintf(stderr,"EMData::realFilter() type 130 only works for 2D images at this moment!\n");
		exit(1);
	}
	float rmin2=val1*val1, rmax2=val2*val2;
	EMData* tmp_mask = new EMData();
	tmp_mask->setSize(nx,ny);
	tmp_mask->zero();
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i++) {
			float r2 = hypot2(i-nx/2, j-ny/2);
			if (r2>=rmin2 && r2<=rmax2)	tmp_mask->setValueAt(i,j,0,valueAt(i,j,0));
		}
	}
	EMData* gradient = tmp_mask->leastSquarePlaneFit();
	subtract(gradient);
	delete gradient;
	delete tmp_mask;
}
	break;
case 140:
{
	if (!(flags&EMDATA_COMPLEX)) {
		printf("Warning: This option only apply to complex image.\n");
		return;
	}
	
	ri2ap();
	d = getData();
	for(i=0; i<nx*ny*nz; i+=2, d+=2) {
		if(*d < val1) *d = 0;
	}
	doneData();
}
	break;
case 141:
{
	if (!(flags&EMDATA_COMPLEX)) {
		printf("Warning: This option only apply to complex image.\n");
		return;
	}
	
	ri2ap();
	d = getData();
	for(i=0; i<nx*ny*nz; i+=2, d+=2) {
		if(*d < val1) {
			*d = 0;
		}
		else {
			*d = 1;
		}
	}
	doneData();
}
	break;
default:
	printf("UNKNOWN REAL FILTER!\n");
	break;
}

return;
}

// Fourier filter so radial power spectrum of 'this' matches 'to'
EMData* EMData::matchFilter(EMData *to) {
float *sfcurve1 = (float *)malloc(nx*4*sizeof(float));
float *sfcurve2 = (float *)malloc(nx*4*sizeof(float));

EMData *tof=to->doFFT();
tof->calcRadDist(nx,0,.5,sfcurve1);

EMData *f=doFFT();
gimmeFFT();
f->calcRadDist(nx,0,.5,sfcurve2);

for (int j=0; j<nx; j++) {
	if (sfcurve1[j]>0 && sfcurve2[j]>0) sfcurve2[j]=sqrt(sfcurve1[j]/sfcurve2[j]);
	else sfcurve2[j]=0;
}

f->applyRadFn(nx,0,.5,sfcurve2);
EMData *ret=f->doIFT();
f->gimmeFFT();
delete f;

free(sfcurve1);
free(sfcurve2);

return ret;
}

// fourier filters an image, real or imaginary
// note, this routine reallocates the data pointer if a real image is passed
// type 0 -> sharp cutoff
// type 1 -> gaussian lowpass (tanh highpass)
// type 2 -> hyperbolic tangent
// type 3 -> butterworth highpass
// type 4 -> undo tanh highpass, lowpass ignored
// type 5 -> gaussian tanh mix
// type 6 -> linear ramp
// type 7 -> inverse gaussian highpass
// type 8 -> Gaussian highpass & lowpass
// type 10 -> *s linear filter for back-projection
// type 99 -> strange low freq, attempt to roughly fix bad struc fac
void EMData::filter(float highpass, float lowpass,int type) {
float *y,*t;
float x;
int i;
EMData *tmp,*tmp2;

y=(float *)malloc(sizeof(float)*ny*6);  // 8 samples per pixel to handle non integral r

switch (type) {
case 0:
	for (i=0,x=0; i<ny*6; i++,x+=.125) {
		if (x>=highpass && x<=lowpass) y[i]=1.0; else y[i]=0;
	}
	break;
case 1:
	if (lowpass>0) {
		for (i=0,x=0; i<ny*6; i++,x+=.125) {
			y[i]=(tanh(x-highpass)/2.0+.5)*exp(-x*x/(lowpass*lowpass));
//		y[i]=1./(1.  + (highpass/1.5/ (x+.001)) * (highpass/1.5/ (x+.001)) );
//			y[i] *= exp(-x*x/(lowpass*lowpass));
		}
	}
	// negative lowpass denotes sharpening rather than blurring
	else {
		for (i=0,x=0; i<ny*6; i++,x+=.125) {
			y[i]=(tanh(x-highpass)/2.0+.5)*exp(x*x/(lowpass*lowpass));
		}
	}	
	break;	
case 2:
	for (i=0,x=0; i<ny*6; i++,x+=.125) {
		y[i]=(tanh(x-highpass)/2.0+.5)*(tanh(lowpass-x)/2.0+.5);
	}
	break;	
case 3:// Butterworth High Pass
	for (i=0,x=0; i<ny*6; i++,x+=.125) {
		y[i]=1./(1.  + (highpass/1.5/ (x+.001)) * (highpass/1.5/ (x+.001)) );
	}
	break;	
case 4:
	for (i=0,x=0; i<ny*6; i++,x+=.125) {
		y[i]=1.0/(tanh(x-highpass)/2.0+.5);
		y[0]=0;
	}
	break;	
case 5:
	for (i=0,x=0; i<ny*6; i++,x+=.125) {
		y[i]=exp(-x*x/(lowpass*lowpass*4))*(tanh(lowpass-x)/2.0+.5);
	}
	break;
case 6:  // Linear Ramp
	for (i=0,x=0; i<ny*6; i++,x+=.125) {
		y[i]= lowpass + ((highpass-lowpass)*i)/(ny*6.-1);
	}
	break;
case 7:
	if (lowpass>0) {
		for (i=0,x=0; i<ny*6; i++,x+=.125) {
			y[i]=(1.0-exp(-x*x/(highpass*highpass)))*exp(-x*x/(lowpass*lowpass));
		}
	}
	break;
case 8:	// gaussian hig&low
	if (highpass<=0) highpass=1.0e-10;
	if (lowpass<=0) lowpass=1.0e10;
	for (i=0,x=0; i<ny*6; i++,x+=.125) {
		y[i]=(1.0-exp(-x*x/(highpass*highpass)))*exp(-x*x/(lowpass*lowpass));
	}
	break;
case 10:  // basic inverse filter for back-projection
	for (i=0,x=0; i<ny*6; i++,x+=.125) {
		y[i]= x+.5;
	}
	break;
case 99:
	for (i=0,x=0; i<ny*6; i++,x+=.125) {
		y[i]=(tanh(x-highpass)*.45+.55);
		y[0]=0;
	}
	break;
}

if (isComplex()) {
	applyRadFn(ny*6,0,.125,y);
	doneData();
	free(y);
	return;
}

//save_data(0,.5, y, nx, "xxx");

//toCorner();
tmp=doFFT();	// transform
gimmeFFT();		// make sure we have control of transform
tmp->applyRadFn(ny*6,0,.125,y);		// filter
tmp2=tmp->doIFT();				// invert transform
//tmp2->toCorner();
tmp->gimmeFFT();				// take control of the inverse

// now swap the data pointers for the result and the original, this
// is dangerous, but a lot faster than doing a copy  :^)
getData();
tmp2->getData();
t=rdata;
rdata=tmp2->rdata;
tmp2->rdata=t;
tmp2->doneData();

delete tmp;
delete tmp2;

doneData();
update();
free(y);
}

// Uses functional parameters to calculate various CTF curves
// type 0 -> ctf amp vs s
// type 1 -> noise inten vs s
// type 2 -> fabs(ctf amp) vs s
// type 3 -> SNR (relative, no struc fac)
// type 4 -> SNR (absolute, requires SF data), pass S vs log10(SF intensity)
// type 5 -> Wiener filter using SNR, must pass S as in 4
// type 6 -> Wiener filter with 1/ctf correction, must pass S as in 4
// type 7 -> Wiener filter with 1/ctf correction, must pass S as in 4, includes phase flip
// type 8 -> Total curve, ctf inten * SF + noise inten 
// type 9 -> SNR (absolute, requires SF data), pass S vs SF intensity
// type 10 -> pure ctf only [-1, 1]
// type 11 -> pure ctf only with absolute value[0, 1]
// type 12 -> Wiener filter using SNR, pass S vs SF intensity
//#define CTFOS	5			// curve oversampling, generates ny*CTFOS/2 points (see EMData.h)
float *EMData::ctfCurve(int type, XYData *sf) {
float *r,ds;
float *p=ctf;
float lambda,gamma,s;
int i;

int NP=ceil(CTFOS*sqrt(2.0)*ny/2)+2;

if (hasCTFF()) {
	if (type==0 || type==1 || type==5 || type==6 || type==7 || type==8) {
		printf("%d:",type);
		error(ERR_EXIT," Cannot generate this CTF info from a file.");
	}

	r=(float *)calloc(NP*sizeof(float),1);

	EMData t;
	char s[80];
	float pix;
	int cn;
	if (3!=sscanf(name,"!$ %d,%f,%s",&cn,&pix,s)) error(ERR_EXIT,"Reported HASCTFF, but filename not present in name.");
	if (t.readImage(s,cn)) { printf("%s,%d: ",s,cn); error(ERR_EXIT,"Cannot read HASCTFF file"); }
	float sca=t.xSize()/(CTFOS*ySize()/2.0);
	float *sr=t.getDataRO();
//printf("%s,%d\n",s,cn);
	switch (type) {
	case 2:
		for (i=0; i<CTFOS*ySize()/2; i++) {
			int fl=floor(i*sca);
			r[i]=sr[fl+1]*(i*sca-fl)+sr[fl]*(1.+fl-i*sca);
		}

		break;
	case 3:
		sr+=t.xSize();
		for (i=0; i<CTFOS*ySize()/2; i++) {
			int fl=floor(i*sca);
			r[i]=sr[fl+1]*(i*sca-fl)+sr[fl]*(1.+fl-i*sca);
		}
	
		break;
	case 4:
		sr+=2*t.xSize();
		for (i=0; i<CTFOS*ySize()/2; i++) {
			int fl=floor(i*sca);
			r[i]=sr[fl+1]*(i*sca-fl)+sr[fl]*(1.+fl-i*sca);
		}
	
		break;
	}
	t.doneData();
	return r;
}

if (!flags&EMDATA_HASCTF) { error(ERR_WARN,"ctfCurve with no CTF parameters\n"); return NULL; }

r=(float *)malloc(NP*sizeof(float));

ds=1.0/(p[10]*ny*CTFOS);

lambda=12.2639/sqrt(p[8]*1000.0+.97845*p[8]*p[8]);

switch (type) {
case 0:
	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(sqrt(-p[1])*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
		}
	}
	else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
		}
	}
	break;
case 1:
	for (i=0,s=0; i<NP; i++,s+=ds) {
		r[i]=(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
	}
	break;
case 2:
	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=fabs(p[2]*exp(-(sqrt(-p[1])*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma)));
		}
	} else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=fabs(p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma)));
		}
	}
	break;
case 3:
	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(sqrt(-p[1])*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
		}
	} else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
		}
	}
	break;
case 4:
	if (!sf) { error(ERR_WARN,"CTF computation error, no SF found\n"); return NULL; }

	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(sqrt(-p[1])*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=pow((float)10.0,sf->yatx(s));
		}
	} else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=pow((float)10.0,sf->yatx(s));
		}
	}
	break;
case 5:
	if (!sf) { error(ERR_WARN,"CTF computation error, no SF found\n"); return NULL; }

	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(sqrt(-p[1])*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=pow((float)10.0,sf->yatx(s));
			r[i]=1./(1.+1./r[i]);
		}
	} else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=pow((float)10.0,sf->yatx(s));
			r[i]=1./(1.+1./r[i]);
		}
	}
	break;
case 6:
	if (!sf) { error(ERR_WARN,"CTF computation error, no SF found\n"); return NULL; }

	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(sqrt(-p[1])*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=pow((float)10.0,sf->yatx(s));
			float v=fabs(p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma)));
			if (r[i]==0 || v==0) r[i]=0;
			else r[i]=(1.0/(1.0+1.0/r[i]))/v;
		}
	} else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=pow((float)10.0,sf->yatx(s));
			float v=fabs(p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma)));
			if (r[i]==0 || v==0) r[i]=0;
			else r[i]=(1.0/(1.0+1.0/r[i]))/v;
		}
	}
	break;
case 7:
	if (!sf) { error(ERR_WARN,"CTF computation error, no SF found\n"); return NULL; }

	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(sqrt(-p[1])*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=pow((float)10.0,sf->yatx(s));
			float v=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			if (r[i]==0 || v==0) r[i]=0;
			else r[i]=(1.0/(1.0+1.0/r[i]))/v;
		}
	} else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=pow((float)10.0,sf->yatx(s));
			float v=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			if (r[i]==0 || v==0) r[i]=0;
			else r[i]=(1.0/(1.0+1.0/r[i]))/v;
		}
	}
	break;
case 8:
	if (!sf) { error(ERR_WARN,"CTF computation error, no SF found\n"); return NULL; }

	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			if (sf->validx(s)) {
				r[i]=p[2]*exp(-(sqrt(-p[1])*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
				r[i]=r[i]*r[i]*pow((float)10.0,sf->yatx(s))+(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			}
			else r[i]=0;
		}
	} else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			if (sf->validx(s)) {
				r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
				r[i]=r[i]*r[i]*pow((float)10.0,sf->yatx(s))+(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			}
			else r[i]=0;
		}
	}
	break;
case 9:
	if (!sf) { error(ERR_WARN,"CTF computation error, no SF found\n"); return NULL; }

	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=sf->yatx(s);
		}
	} else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=sf->yatx(s);
		}
	}
	break;
case 10:
                for (i=0,s=0; i<NP; i++,s+=ds) {
                        gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
                        r[i]=sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma);
                }
        break;
case 11:
                for (i=0,s=0; i<NP; i++,s+=ds) {
                        gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
                        r[i]=fabs(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
                }
        break;
case 12: // Wiener filter, sf as intensity instead of log10(intensity)
	if (!sf) { error(ERR_WARN,"CTF computation error, no SF found\n"); return NULL; }

	if (p[1]<0) {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(sqrt(-p[1])*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=sf->yatx(s);
			r[i]=1./(1.+1./r[i]);
		}
	} else {
		for (i=0,s=0; i<NP; i++,s+=ds) {
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			r[i]=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			r[i]=r[i]*r[i]/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) r[i]*=sf->yatx(s);
			r[i]=1./(1.+1./r[i]);
		}
	}
	break;
default:
        printf("The mode %d is not supported in EMData::ctfCurve!\n",type);
        exit(0);
}

//save_data(0,ds,r,ny/2,"ctf.dat");

return r;
}

// generates 2D CTF modification maps (fourier space)
// 0 - noise curve
// 1 - fabs(ctf)
// 2 - sign(ctf)
// 4 - B factor
// 8 - Wiener filter
// 16 - SNR
// 32 - Wiener filter with 1/CTF (including sign) correction
// 64 - Wiener filter with 1/CTF (no sign) correction
void EMData::ctfMap(int type,XYData *sf) {
float s,ds;
float *p=ctf;
float lambda,gamma;
int x,y;
int typeExist = 0;
//printf("ctfMap %d\n",type);
//for (x=0; x<11; x++) printf("%f\t",p[x]);
//printf("\n");

if (!flags&EMDATA_HASCTF) { error(ERR_WARN,"ctfCurve with no CTF parameters\n"); return; }
if (!isComplex()) { error(ERR_CRIT,"ctfMap only works on complex images"); return; }
if (nx!=ny+2) { error(ERR_CRIT,"ctfMap only works on nx x nx-2 images"); return; }

ds=1.0/(p[10]*ny);

lambda=12.2639/sqrt(p[8]*1000.0+.97845*p[8]*p[8]);

one();
float *d=getData();

if (type==0) {
	for (y=0; y<ny; y++) {
		for (x=0; x<nx/2; x++) {
			s=hypot((float)x,(float)y-ny/2)*ds;
			d[x*2+y*nx]=(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			d[x*2+y*nx+1]=d[x*2+y*nx];
		}
	}
	typeExist = 1;
}
if ((type&3)==3) {
	for (y=0; y<ny; y++) {
		for (x=0; x<nx/2; x++) {
			s=hypot((float)x,(float)y-ny/2)*ds;
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			float v=p[2]*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			d[x*2+y*nx]*=v;
			d[x*2+y*nx+1]=d[x*2+y*nx];
		}
	}
        typeExist = 1;
}
else {
	if (type&1) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx/2; x++) {
				s=hypot((float)x,(float)y-ny/2)*ds;
				gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
				float v=fabs(p[2]*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma)));
				d[x*2+y*nx]*=v;
				d[x*2+y*nx+1]=d[x*2+y*nx];
			}
		}
	}
	if (type&2) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx/2; x++) {
				s=hypot((float)x,(float)y-ny/2)*ds;
				gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
				float v=(p[2]*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma)))>0?1.0:-1.0;
				d[x*2+y*nx]*=v;
				d[x*2+y*nx+1]=d[x*2+y*nx];
			}
		}
	}
        typeExist = 1;
}
if (type&4) {
	for (y=0; y<ny; y++) {
		for (x=0; x<nx/2; x++) {
			s=hypot((float)x,(float)y-ny/2)*ds;
			float v=exp(-(p[1]*s*s));
			d[x*2+y*nx]*=v;
			d[x*2+y*nx+1]=d[x*2+y*nx];
		}
	}
        typeExist = 1;
}
if (type&8) {
	for (y=0; y<ny; y++) {
		for (x=0; x<nx/2; x++) {
			float f;
			s=hypot((float)x,(float)y-ny/2)*ds;
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			f=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			f=f*f/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) f*=pow((float)10.0,sf->yatx(s));
			f=1./(1.+1./f);
			d[x*2+y*nx]*=f;
			d[x*2+y*nx+1]=d[x*2+y*nx];
		}
	}
        typeExist = 1;
}
if (type&16) {	// SNR
	for (y=0; y<ny; y++) {
		for (x=0; x<nx/2; x++) {
			float f;
			s=hypot((float)x,(float)y-ny/2)*ds;
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			f=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			f=f*f/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) f*=pow((float)10.0,sf->yatx(s));
			d[x*2+y*nx]*=f;
			d[x*2+y*nx+1]=d[x*2+y*nx];
		}
	}
        typeExist = 1;
}
if (type&32) {	// Wiener filter with 1/CTF (including sign) correction
	for (y=0; y<ny; y++) {
		for (x=0; x<nx/2; x++) {
			float f, v;
			s=hypot((float)x,(float)y-ny/2)*ds;
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			f=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			v = f;
			f=f*f/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) f*=pow((float)10.0,sf->yatx(s));
			if (!f || !v) f=0.; 
			else f=1./(v*(1.+1./f));
			d[x*2+y*nx]*=f;
			d[x*2+y*nx+1]=d[x*2+y*nx];
		}
	}
        typeExist = 1;
}
if (type&64) {	// Wiener filter with 1/CTF (no sign) correction
	for (y=0; y<ny; y++) {
		for (x=0; x<nx/2; x++) {
			float f, v;
			s=hypot((float)x,(float)y-ny/2)*ds;
			gamma=-2*PI*(p[9]*2.5e6*lambda*lambda*lambda*s*s*s*s+p[0]*5000.0*lambda*s*s);
			f=p[2]*exp(-(p[1]*s*s))*(sqrt(1.0-p[3]*p[3])*sin(gamma)+p[3]*cos(gamma));
			v = fabs(f);
			f=f*f/(p[6]*exp(-(SQR(PI/2*p[7]*s))-s*p[5]-sqrt(fabs(s))*p[4]));
			if (s) f*=pow((float)10.0,sf->yatx(s));
			if (!f || !v) f=0.; 
			else f=1./(v*(1.+1./f));
			d[x*2+y*nx]*=f;
			d[x*2+y*nx+1]=d[x*2+y*nx];
		}
	}
        typeExist = 1;
}
//save_data(0,ds,r,ny/2,"ctf.dat");

doneData();

if (typeExist == 0) {
        printf("The mode %d is not supported EMData::ctfMap!\n",type);
        exit(0);
}


return;
}


// Subtract average noise from intensity. Negative results allowed
// WARNING - may be broken
void EMData::subNoise() {
float *y,r,f;
int i,j,k,l;
if (!isComplex()) { printf("subnoise on  FFT's only!\n"); return; }
if (!hasCTF()) { /*printf("No CTF to subtract\n");*/ return; }
ri2ap();

y=ctfCurve(1);
//save_data(0,1,y,ny/2,"ctf.dat");

getData();
for (j=k=0; j<ny; j++) {
    for (i=0; i<nx; i+=2, k+=2) {
		r=hypot((float)i/2.0,(float)j-ny/2.0)*CTFOS;
		l=(int)floor((r+.5));
		r-=l;
		if (l>=ny/2-2) f=y[CTFOS*ny/2-1];
		else f=y[l];
		rdata[k]=rdata[k]*rdata[k]-f;
		rdata[k+1]=0;
	}
}
doneData();
update();
free(y);
}

// averages the images in the list, result in 'this'
void EMData::makeAverage(List *in,EMData *sig,int ignore0,int nweight) {
int i,j,nx2,ny2,nz2;
EMData *obj;
float *dest,*d2,*src;
float c,f,g;

if (in->count()==0) return;

// Start by copying the first object in the list and getting the size to
// make sure they're all the correct size
obj=(EMData *)in->objectAt(0);
nx2=obj->xSize();
ny2=obj->ySize();
nz2=obj->zSize();
setSize(nx2,ny2,nz2);
if (sig) sig->setSize(nx2,ny2,nz2);
if (ignore0 && nweight) error(ERR_EXIT,"only one of ignore0 and nweight can be used in makeAverage");

zero();
dest=getData();
if (sig) d2=sig->getData(); else d2=NULL;

if (ignore0) {
	for (j=0; j<nx*ny*nz; j++) {
		for (i=0,g=0; i<in->count(); i++) {
			f=((EMData *)in->objectAt(i))->valueAt(j,0,0);
			if (!f) continue;
			g+=1.0;
			dest[j]+=f;
			if (d2) d2[j]+=f*f;
		}
		if (g) dest[j]/=g;
		if (d2) d2[j]=sqrt(d2[j]/g-SQR(dest[j]));
	}
	setNImg(in->count());
}
else {
	src=obj->getDataRO();
	if (d2) {
		memcpy(d2,src,nx*ny*nz*sizeof(float));
		for (j=0; j<nx*ny*nz; j++) d2[j]*=d2[j];
	}
	obj->doneData();
	memcpy(dest,src,nx*ny*nz*sizeof(float));
	float ntot=nweight ? obj-> NImg() : 1.0;
	for (i=1; i<in->count(); i++) {
		obj=(EMData *)in->objectAt(i);

		// skip records that are different sizes
		if (obj->xSize()!=nx || obj->ySize()!=ny || obj->zSize()!=nz) continue;
		float nim = nweight ? obj->NImg() : 1.0;
		ntot+=nim;

		// add each image
		src=obj->getDataRO();
		for (j=0; j<nx*ny*nz; j++) dest[j]+=nim*src[j];
		if (d2) {
			for (j=0; j<nx*ny*nz; j++) d2[j]+=nim*src[j]*src[j];
		}
		obj->doneData();
	}
	for (j=0; j<nx*ny*nz; j++) dest[j]/=ntot;
	if (d2) {
		for (j=0; j<nx*ny*nz; j++) d2[j]=sqrt(d2[j]/ntot-SQR(dest[j]));
	}
	setNImg((int)ntot);
}


//printf("%1.0f in ave\n",c);
// normalize
doneData();
if (d2) sig->doneData();
update();
}

// averages the images in the list, result in 'this'
void EMData::makeAverageIter(List *in) {
int i,j,k,l,nx2,ny2,nz2;
EMData *obj,*sig;
float *dest,*d2,*src;
float c,c2,sigma;

if (in->count()==0) return;

sig=new EMData;

// Start by copying the first object in the list and getting the size to
// make sure they're all the correct size
obj=(EMData *)in->objectAt(0);
nx2=obj->xSize();
ny2=obj->ySize();
nz2=obj->zSize();
setSize(nx2,ny2,nz2);
sig->setSize(nx2,ny2,nz2);

dest=getData();
d2=sig->getData();
src=obj->getDataRO();
memcpy(dest,src,nx*ny*nz*sizeof(float));
memcpy(d2,src,nx*ny*nz*sizeof(float));
for (j=0; j<nx*ny*nz; j++) d2[j]*=d2[j];

obj->doneData();

for (i=1,c=1.0; i<in->count(); i++) {
	obj=(EMData *)in->objectAt(i);
	
	// skip records that are different sizes
	if (obj->xSize()!=nx || obj->ySize()!=ny || obj->zSize()!=nz) continue;

	// add each image
	src=obj->getDataRO();
	for (j=0; j<nx*ny*nz; j++) dest[j]+=src[j];
	for (j=0; j<nx*ny*nz; j++) d2[j]+=src[j]*src[j];
		
	obj->doneData();
	c+=1.0;
}

// This one is std dev
for (j=0; j<nx*ny*nz; j++) d2[j]=sqrt(d2[j]/c-SQR(dest[j]/c))/sqrt(c);

// or std dev of mean
//for (j=0; j<nx*ny*nz; j++) d2[j]=sqrt(d2[j]/c-SQR(dest[j]/c));

printf("%1.0f in ave\n",c);
// normalize
for (j=0; j<nx*ny*nz; j++) dest[j]/=c;
doneData();
sig->doneData();
update();
sig->update();
writeIMAGIC("iter.hed",-1);

// now do the smoothing iteration
sigma=sig->Mean();
src=sig->getData();
dest=getData();
d2=(float *)malloc(nx*ny*sizeof(float));
memcpy(d2,dest,nx*ny*sizeof(float));
memcpy(src,dest,nx*ny*sizeof(float));

printf("Iter sigma=%f\n",sigma);

for (k=0; k<1000; k++) {
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			l=i+j*nx;
			c=(d2[l-1]+d2[l+1]+d2[l-nx]+d2[l+nx])/4.0-d2[l];
			c2=fabs(dest[l]-src[l])/sigma;
			dest[l]+=c*erfc(c2)/100.0;
//if (i==nx/2) printf("%f %f %f\n",c,c2,erfc(c2));
		}
	}
	memcpy(d2,dest,nx*ny*sizeof(float));
}
free(d2);
doneData();
sig->doneData();
delete sig;
update();

writeIMAGIC("iter.hed",-1);

}

// averages the images in the list, result in 'this'
// Also does CTF correction if possible
// 2D ONLY!!!!
// filtr is in Angstroms!
void EMData::makeAverageCTFC(List *in,float filtr) {
int i,j,x,y,l,nx2,ny2,nz2;
EMData *obj,*res;
float *dest,*s2,**src;
float **Ctf,**Ctfn,ctf0,f,r,ds;
char *s;

if (in->count()==0) return;

obj=((EMData *)in->objectAt(0));
if (!obj->hasCTF()) return;
ds=1.0/((obj->getCTF())[10]*ny);

// get data pointers and calculate individual CTF curves
Ctf=(float **)malloc(sizeof(float *)*in->count());
Ctfn=(float **)malloc(sizeof(float *)*in->count());
src=(float **)malloc(sizeof(float *)*in->count());
for (i=0; i<in->count(); i++) {
	obj=((EMData *)in->objectAt(i))->doFFT();
	obj->ap2ri();
	src[i]=obj->getDataRO();
	Ctf[i]=obj->ctfCurve(2);
	if (Ctf[i]==NULL) error("Unexpected CTF correction problem\n",ERR_EXIT);
	Ctfn[i]=obj->ctfCurve(3);
}


// Start by copying the first object in the list and getting the size to
// make sure they're all the correct size
obj=((EMData *)in->objectAt(0))->doFFT();
nx2=obj->xSize();
ny2=obj->ySize();
nz2=obj->zSize();
if (nz2!=1) { printf("Only 2d currently supported\n"); return; }
setSize(nx2-2,ny2,nz2);

ds=1.0/((((EMData *)in->objectAt(0))->getCTF())[10]*ny*CTFOS);
if (filtr==0) filtr=22.0;	// 22A default filter
//filtr=sqrt(2.0)/(ds*filtr*M_PI);		// filtr is now the radius in pixels
filtr=1.0/(filtr*ds);		// we use gaussian half-width in fourier space
//printf("Filter = %f %f\n",filtr,ds);

// Normalize CTF curves
for (i=0; i<CTFOS*ny/2; i++) {
	ctf0=0;
	for (j=0; j<in->count(); j++) {
		ctf0+=Ctfn[j][i];
		if (Ctf[j][i]==0) Ctf[j][i]=1.0e-12;	// just to avoid infinities
	}
//printf("%f\n",ctf0);
	if (ctf0<=0.0001) { /*printf("Warning CTF==0\n");*/ ctf0=.1; }
//	for (j=0; j<in->count(); j++) Ctf[j][i]=exp(-i*i/((float)filtr*filtr*4.0))*(tanh(filtr-i)/2.0+.5)*Ctfn[j][i]/(fabs(Ctf[j][i])*ctf0);
	for (j=0; j<in->count(); j++) Ctf[j][i]=exp(-i*i/((float)filtr*filtr))*Ctfn[j][i]/(fabs(Ctf[j][i])*ctf0);
}
//save_data(0,1.0,Ctf[0],ny/2,"ctf.dat");

res=obj->copyHead();
res->ap2ri();

dest=res->getData();

for (y=j=0; y<ny2; y++) {
	for (x=0; x<nx2/2; x++,j+=2) {
		r=CTFOS*sqrt((float)(x*x)+(float)((y-ny2/2.0)*(y-ny2/2.0)));
		l=(int)floor2(r);
		r-=l;
		dest[j]=0;
		dest[j+1]=0;
		for (i=0; i<in->count(); i++) {
			if (l>CTFOS*ny/2-2) f=0;
			else f=(Ctf[i][l]*(1.0-r)+Ctf[i][l+1]*r);

			dest[j]  +=src[i][j]  *f;
			dest[j+1]+=src[i][j+1]*f;
		}
	}
}

res->doneData();

dest=getData();
obj=res->doIFT();
s2=obj->getData();
memcpy(dest,s2,sizeof(float)*(nx2-2)*ny);
doneData();
obj->doneData();

delete res;

for (i=0; i<in->count(); i++) {
	obj=((EMData *)in->objectAt(i))->doFFT();
	obj->doneData();
	free(Ctf[i]);
	free(Ctfn[i]);
}
free(src);
free(Ctf);
free(Ctfn);
setNImg(in->count());
update();
}



// averages the images in the list, result in 'this'
// Also does CTF correction, and applies a Wiener filter with the SNR
// estimated from the data. If specified the SNR estimate will be
// written to filespec
// 2D ONLY!!!
void EMData::makeAverageCTFCWauto(List *in, char *filespec) {
int i,j,x,y,l,m,nx2,ny2,nz2;
EMData *obj,*res;
float *dest,*s2,**src;
float **Ctf,**Ctfn,ctf0,ctf1,f,r,ds;
float *snri,*snrn;
char *s;

if (in->count()==0) return;

obj=((EMData *)in->objectAt(0));
if (!obj->hasCTF()&&!obj->hasCTFF()) return;

if (obj->hasCTF()) ds=1.0/((obj->getCTF())[10]*obj->ySize());
else {
	float f;
	sscanf(Name(),"!$ %d,%f",&i,&f);
	ds=1.0/(f*obj->ySize());
}

// get data pointers and calculate individual CTF curves
Ctf=(float **)malloc(sizeof(float *)*in->count());
Ctfn=(float **)malloc(sizeof(float *)*in->count());
src=(float **)malloc(sizeof(float *)*in->count());
for (i=0; i<in->count(); i++) {
	obj=((EMData *)in->objectAt(i))->doFFT();
	obj->ap2ri();
	src[i]=obj->getDataRO();
	Ctf[i]=obj->ctfCurve();
	if (Ctf[i]==NULL) error("Unexpected CTF correction problem\n",ERR_EXIT);
	Ctfn[i]=obj->ctfCurve(3);
}


// Start by copying the first object in the list and getting the size to
// make sure they're all the correct size
obj=((EMData *)in->objectAt(0))->doFFT();
ds=1.0/((obj->getCTF())[10]*ny);

nx2=obj->xSize();
ny2=obj->ySize();
nz2=obj->zSize();
if (nz2!=1) { printf("Only 2d currently supported\n"); return; }
setSize(nx2-2,ny2,nz2);

//ds=1.0/((((EMData *)in->objectAt(0))->getCTF())[10]*ny*CTFOS);


// now we estimate the SNR directly from the data
snri=(float *)calloc(ny/2,sizeof(float));
snrn=(float *)calloc(ny/2,sizeof(float));

for (y=j=0; y<ny2; y++) {
	for (x=0; x<nx2/2; x++,j+=2) {
		r=hypot((float)(x),(float)(y-ny2/2.0));
		m=(int)floor2(r*CTFOS);
		l=(int)floor2(r);
		if (l<0||l>=ny/2) continue;		// we're not calculating outside a box-sized circle
		float ti=0;		// mean pixel intensity
		float tsr=0,tsi=0;	// mean pixel vector (coherent average)
		float tn=0;
		for (i=0; i<in->count(); i++) {
			ti+=hypot2(src[i][j],src[i][j+1]);
			tsr+=src[i][j];
			tsi+=src[i][j+1];
			tn+=Ctfn[i][m];
		}
		ti/=in->count();	// mean intensity of this pixel 'signal^2+noise^2'
		tsr/=in->count();
		tsi/=in->count();
		tsr=hypot2(tsr,tsi);	// intensity of mean value	'signal^2'
		if (ti<=tsr) ti=tsr+.00001;		// shouldn't happen

		snri[l]+=tsr/(ti-tsr);	// now add in to calculation of mean snr for this radius
		snrn[l]+=1.0;
	}
}
// normalize the SNR, multiply by count to get snr of average
// rather than mean snr of 1 image 
for (i=0; i<ny/2; i++) snri[i]*=in->count()/snrn[i];	


// now we estimate the SNR directly from the data
/*
snri=(float *)calloc(ny/2,sizeof(float));
snrn=(float *)calloc(ny/2,sizeof(float));

for (y=j=0; y<ny2; y++) {
	for (x=0; x<nx2/2; x++,j+=2) {
		r=hypot((float)(x),(float)(y-ny2/2.0));
		m=(int)floor2(r*CTFOS);
		l=(int)floor2(r);
		if (l<0||l>=ny/2) continue;		// we're not calculating outside a box-sized circle
		float tdr=1.0;		// pixel dot product
		float tdi=1.0;
		float tn=1.0;		// normalization
		for (i=0; i<in->count(); i++) {
			tdr*=src[i][j];
			tdi*=src[i][j+1];
			tn*=hypot(src[i][j],src[i][j+1]);
		}
		tdr=tdr+tdi/tn;		// normalized dot product of all values for this pixel

		snri[l]+=tdr;	// now add in to calculation of mean snr for this radius
		snrn[l]+=1.0;
	}
}
// normalize the SNR, multiply by count to get snr of average
// rather than mean snr of 1 image 
for (i=0; i<ny/2; i++) snri[i]*=in->count()/snrn[i];	
*/

if (filespec) save_data(0,1.0,snri,ny/2,filespec);

// Normalize CTF curves
for (i=0; i<CTFOS*ny/2; i++) {
	ctf1=0;
	for (j=0; j<in->count(); j++) {
		ctf1+=Ctfn[j][i];
		if (Ctf[j][i]==0) Ctf[j][i]=1.0e-12;	// just to avoid infinities
	}
	ctf0=snri[i/CTFOS];
//printf("%f\n",ctf0);
	if (ctf0<=0.0001) { /*printf("Warning CTF==0\n");*/ ctf0=.1; }
//	for (j=0; j<in->count(); j++) Ctf[j][i]=(ctf0/(ctf0+1.0))*Ctfn[j][i]/(fabs(Ctf[j][i])*ctf1);
	for (j=0; j<in->count(); j++) Ctf[j][i]=(ctf0)*Ctfn[j][i]/(fabs(Ctf[j][i])*ctf1);
}
//save_data(0,1.0,Ctf[0],ny/2,"ctf.dat");

res=obj->copyHead();
res->ap2ri();

dest=res->getData();

for (y=j=0; y<ny2; y++) {
	for (x=0; x<nx2/2; x++,j+=2) {
		r=CTFOS*sqrt((float)(x*x)+(float)((y-ny2/2.0)*(y-ny2/2.0)));
		l=(int)floor2(r);
		r-=l;
		dest[j]=0;
		dest[j+1]=0;
		for (i=0; i<in->count(); i++) {
			if (l>CTFOS*ny/2-2) f=0;
			else f=(Ctf[i][l]*(1.0-r)+Ctf[i][l+1]*r);

			dest[j]  +=src[i][j]  *f;
			dest[j+1]+=src[i][j+1]*f;
		}
	}
}

res->doneData();

dest=getData();
obj=res->doIFT();
s2=obj->getData();
memcpy(dest,s2,sizeof(float)*(nx2-2)*ny);
doneData();
obj->doneData();

delete res;

for (i=0; i<in->count(); i++) {
	obj=((EMData *)in->objectAt(i))->doFFT();
	obj->doneData();
	free(Ctf[i]);
	free(Ctfn[i]);
}
free(src);
free(Ctf);
free(Ctfn);
free(snri);
free(snrn);
setNImg(in->count());
update();
}

// averages the images in the list, result in 'this'
// Also does CTF correction if possible
// 2D ONLY!!!!
// The sf data set represents the log10() of the structure factor
// Individual particles must be properly normalized to a reference projection
// before processing in this routine, otherwise weighting will be incorrect
// SNR (if not NULL) will be filled with the total SNR of the average, it must
// point to an array of at least CTFOS*ny/2 elements
void EMData::makeAverageCTFCW(List *in,XYData *sf,float *SNR,int phasematch) {
int i,j,x,y,l,nx2,ny2,nz2;
EMData *obj,*res;
float *dest,*s2,**src;
float **Ctf,**Ctfn,ctf0,f,r,ds;
char *s;

if (in->count()==0) { error(ERR_WARN,"Warning, average of 0 objects\n"); return; }

obj=((EMData *)in->objectAt(0));
if (!obj->hasCTF() && !obj->hasCTFF()) { error(ERR_WARN,"Attempted CTFCW with no ctf parameters"); return; }

// get data pointers and calculate individual CTF curves
Ctf=(float **)malloc(sizeof(float *)*in->count());
Ctfn=(float **)malloc(sizeof(float *)*in->count());
src=(float **)malloc(sizeof(float *)*in->count());
for (i=0; i<in->count(); i++) {
	obj=((EMData *)in->objectAt(i))->doFFT();
	obj->ap2ri();
	src[i]=obj->getDataRO();
	Ctf[i]=obj->ctfCurve(2);		// ctf amp
	if (Ctf[i]==NULL) error("Unexpected CTF correction problem\n",ERR_EXIT);
	Ctfn[i]=obj->ctfCurve(4,sf);	// absolute SNR
}


// Start by copying the first object in the list and getting the size to
// make sure they're all the correct size
obj=((EMData *)in->objectAt(0))->doFFT();
nx2=obj->xSize();
ny2=obj->ySize();
nz2=obj->zSize();
if (nz2!=1) { printf("Only 2d currently supported\n"); return; }
setSize(nx2-2,ny2,nz2);

//save_data(0,1.0/CTFOS,Ctf[0],CTFOS*ny/2,"ctfo.dat");

ds=1.0/((((EMData *)in->objectAt(0))->getCTF())[10]*ny*CTFOS);

// Normalize CTF curves
for (i=0; i<CTFOS*ny/2; i++) {
	ctf0=0;		// ctf0 contains the total SNR at each spatial frequency
	for (j=0; j<in->count(); j++) {
		ctf0+=Ctfn[j][i];
		if (Ctf[j][i]==0) Ctf[j][i]=1.0e-12;	// just to avoid infinities
	}
//printf("%f\n",ctf0);
//printf("cw   %d\t%f\t%f\n",i,Ctf[0][i],Ctfn[0][i]);

	if (SNR) SNR[i]=ctf0;
	if (ctf0<=0.0001) { /*printf("Warning CTF==0\n");*/ ctf0=.1; }
	for (j=0; j<in->count(); j++) Ctf[j][i]=(ctf0/(ctf0+1.0))*Ctfn[j][i]/(Ctf[j][i]*ctf0);
}

//save_data(0,1.0/CTFOS,Ctfn[0],CTFOS*ny/2,"ctfn.dat");
//save_data(0,1.0/CTFOS,Ctf[0],CTFOS*ny/2,"ctf.dat");

// save curve(s)
/*for (i=0; i<in->count(); i++) {
	char s[80];
	if (i!=0 && Ctf[i][25]==Ctf[i-1][25]) continue;
	sprintf(s,"ctf.%d.dat",i);
	save_data(0,1.0,Ctf[i],ny/2,s);
}*/

res=obj->copyHead();
res->ap2ri();

dest=res->getData();

for (y=j=0; y<ny2; y++) {
	for (x=0; x<nx2/2; x++,j+=2) {
		r=CTFOS*sqrt((float)(x*x)+(float)((y-ny2/2.0)*(y-ny2/2.0)));
		l=(int)floor2(r);
		r-=l;
		dest[j]=0;
		dest[j+1]=0;
		for (i=0; i<in->count(); i++) {
			if (l>CTFOS*ny/2-2) f=0;
			else f=(Ctf[i][l]*(1.0-r)+Ctf[i][l+1]*r);

			dest[j]  +=src[i][j]  *f;
			dest[j+1]+=src[i][j+1]*f;
		}
		if (phasematch) {
			for (i=0; i<in->count(); i++) {
				if (l>CTFOS*ny/2-2) f=0;
				else f=(Ctf[i][l]*(1.0-r)+Ctf[i][l+1]*r);
				if (src[i][j]*dest[j]+src[i][j+1]*dest[j+1]<0) {
					dest[j]  -=2.0*src[i][j]  *f;
					dest[j+1]-=2.0*src[i][j+1]*f;
				}
			}		
		}
	}
}

res->doneData();

dest=getData();
obj=res->doIFT();
s2=obj->getData();
memcpy(dest,s2,sizeof(float)*(nx2-2)*ny);
doneData();
obj->doneData();

delete res;

for (i=0; i<in->count(); i++) {
	obj=((EMData *)in->objectAt(i))->doFFT();
	obj->doneData();
	free(Ctf[i]);
	free(Ctfn[i]);
}
free(src);
free(Ctf);
free(Ctfn);
setNImg(in->count());
update();
}

// averages the images in the list, result in 'this'
// Performs SNR weighting, but NO CTF correction
// 2D ONLY!!!!
// if provided, curves will contain an image containing
// CTF curves for the average
void EMData::makeAverageWT(List *in,EMData *curves,XYData *sf) {
int i,j,x,y,l,nx2,ny2,nz2;
EMData *obj,*res;
float *dest,*s2,**src;
float **Ctf,**Ctfn,ctf0,f,r,ds;
char *s;

if (in->count()==0) return;

obj=((EMData *)in->objectAt(0));
if (!obj->hasCTF()) return;
ds=1.0/((obj->getCTF())[10]*ny);

// get data pointers and calculate individual CTF curves
Ctf=(float **)malloc(sizeof(float *)*in->count());
Ctfn=(float **)malloc(sizeof(float *)*in->count());
src=(float **)malloc(sizeof(float *)*in->count());
for (i=0; i<in->count(); i++) {
	obj=((EMData *)in->objectAt(i))->doFFT();
	obj->ap2ri();
	src[i]=obj->getDataRO();
//	Ctf[i]=(float*)malloc(sizeof(float)*(CTFOS*ny/2+1));
	Ctf[i]=obj->ctfCurve(2);
	if (sf) Ctfn[i]=obj->ctfCurve(4,sf);	// absolute SNR
	else Ctfn[i]=obj->ctfCurve(3);	// relative SNR
}


// Start by copying the first object in the list and getting the size to
// make sure they're all the correct size
obj=((EMData *)in->objectAt(0))->doFFT();
nx2=obj->xSize();
ny2=obj->ySize();
nz2=obj->zSize();
if (nz2!=1) { printf("Only 2d currently supported\n"); return; }
setSize(nx2-2,ny2,nz2);

float *cd = NULL;
if (curves) {
	if (!sf) error(ERR_WARN,"CTF curve in file will contain relative, not absolute SNR!");
	curves->setSize(CTFOS*ny/2,3);
	curves->zero();
//	printf("db- %d,%d\n",curves->xSize(),curves->ySize());
	cd=curves->getData();
}

//save_data(0,1.0/CTFOS,Ctf[0],CTFOS*ny/2,"ctfo.dat");

ds=1.0/((((EMData *)in->objectAt(0))->getCTF())[10]*ny*CTFOS);

// Normalize CTF curves
for (i=0; i<CTFOS*ny/2; i++) {
	ctf0=0;		// ctf0 contains the total SNR at each spatial frequency
	for (j=0; j<in->count(); j++) {
		ctf0+=Ctfn[j][i];
		if (curves) {
			cd[i]+=Ctf[j][i]*Ctfn[j][i];	// This is an amplitide weighted average of CTF amplitude
			cd[i+CTFOS*ny/2]+=Ctfn[j][i];	// this should be relative SNR, but we're using absolute instead
			cd[i+2*CTFOS*ny/2]+=Ctfn[j][i];	// This will only be absolute if sf is passed
		}
	}
	if (ctf0<=0.0001) { /*printf("Warning CTF==0\n");*/ ctf0=.1; }
	for (j=0; j<in->count(); j++) Ctf[j][i]=Ctfn[j][i]/ctf0;
}
if (curves) {
	for (i=0; i<CTFOS*ny/2; i++) cd[i]/=cd[i+CTFOS*ny/2];
	curves->doneData();
}

//printf("%p\n",obj);
res=obj->copyHead();
res->ap2ri();

dest=res->getData();

for (y=j=0; y<ny2; y++) {
	for (x=0; x<nx2/2; x++,j+=2) {
		r=CTFOS*sqrt((float)(x*x)+(float)((y-ny2/2.0)*(y-ny2/2.0)));
		l=(int)floor2(r);
		r-=l;
		dest[j]=0;
		dest[j+1]=0;
		for (i=0; i<in->count(); i++) {
			if (l>CTFOS*ny/2-2) f=0;
			else f=(Ctf[i][l]*(1.0-r)+Ctf[i][l+1]*r);

			dest[j]  +=src[i][j]  *f;
			dest[j+1]+=src[i][j+1]*f;
		}
	}
}

res->doneData();

dest=getData();
obj=res->doIFT();
s2=obj->getData();
memcpy(dest,s2,sizeof(float)*(nx2-2)*ny);
doneData();
obj->doneData();

delete res;

for (i=0; i<in->count(); i++) {
	obj=((EMData *)in->objectAt(i))->doFFT();
	obj->doneData();
	free(Ctf[i]);
	free(Ctfn[i]);
}
free(src);
free(Ctf);
free(Ctfn);

update();
}

// Median of the images in the list, result in 'this'
void EMData::makeMedian(List *in) {
int i,j,k,c,nx2,ny2,nz2,ni;
EMData *obj;
float v,*dest,**src,*srt;

if (in->count()==0) return;

// Start by copying the first object in the list and getting the size to
// make sure they're all the correct size
obj=(EMData *)in->objectAt(0);
nx2=obj->xSize();
ny2=obj->ySize();
nz2=obj->zSize();
setSize(nx2,ny2,nz2);
dest=getData();

ni=in->count();
srt=(float *)malloc(sizeof(float)*ni);
src=(float **)malloc(sizeof(float *)*ni);
for(i=0; i<ni; i++) src[i]=((EMData *)in->objectAt(i))->getDataRO();

for (i=0; i<nx*ny*nz; i++) {
	for (j=0; j<ni; j++) srt[j]=src[j][i];
	
	// not much data, just shell sort
	for (j=0; j<ni; j++) {
		for (k=j+1; k<ni; k++) {
			if (srt[j]<srt[k]) { v=srt[j]; srt[j]=srt[k]; srt[k]=v; }
		}
	}
	
	if (ni<3) dest[i]=srt[ni/2]; 
	else dest[i]=(srt[ni/2]+srt[ni/2+1]+srt[ni/2-1])/3.0;
}

setNImg(in->count());

doneData();
for (i=0; i<ni; i++) ((EMData *)in->objectAt(i))->doneData();
free(srt);
free(src);
update();
}


// adds 'obj' to 'this' incoherently
void EMData::addIncoherent(EMData *obj) {
int j;
float *dest,*src;

if (obj->xSize()!=nx || obj->ySize()!=ny || obj->zSize()!=nz) return;
if (!obj->isComplex() || !isComplex()) return;
ri2ap();
obj->ri2ap();

dest=getData();
src=obj->getDataRO();

for (j=0; j<nx*ny*nz; j+=2) {
	dest[j]=hypot(src[j],dest[j]);
	dest[j+1]=0;
}

obj->doneData();
doneData();

update();
}
// adds 'obj' to 'this'
void EMData::add(EMData *obj) {
int i,j,c;
float *dest,*src;

if (obj->isComplex()) obj->ap2ri();
if (obj->xSize()!=nx || obj->ySize()!=ny || obj->zSize()!=nz) return;

dest=getData();
src=obj->getDataRO();

for (j=0; j<nx*ny*nz; j++) dest[j]+=src[j];

obj->doneData();
doneData();

update();
}

// mult 'obj' by 'this'
void EMData::mult(EMData *obj, int recip) {
int i,j,c;
float *dest,*src;

if (isComplex()) ap2ri();
if (obj->isComplex()) obj->ap2ri();
if (obj->xSize()!=nx || obj->ySize()!=ny || obj->zSize()!=nz) { printf("multiply sizes different\n"); return; }

dest=getData();
if (obj!=this) src=obj->getDataRO(); else src=dest;

if (recip) {
	for (j=0; j<nx*ny*nz; j++) {
		if (src[j]) dest[j]/=src[j];
		else dest[j]=0;
	}
}
else {
	for (j=0; j<nx*ny*nz; j++) dest[j]*=src[j];
}

if (obj!=this) obj->doneData();
doneData();

}

// mult 'obj' by 'this' then do sqrt, keep the sign of this object
void EMData::mult_sqrt(EMData *obj) {
int i,j,c;
float *dest,*src;

if (isComplex()) ap2ri();
if (obj->isComplex()) obj->ap2ri();
if (obj->xSize()!=nx || obj->ySize()!=ny || obj->zSize()!=nz) { printf("multiply sizes different\n"); return; }

dest=getData();
if (obj!=this) src=obj->getDataRO(); else src=dest;

for (j=0; j<nx*ny*nz; j++) {
	dest[j]*=src[j];  //multiple
        if(src[j]<0) dest[j] = -dest[j];
        if(dest[j] >=0) dest[j] = sqrt(dest[j]); // >=0, just sqrt
	else            dest[j] = -sqrt(-dest[j]); // <0,  keep - outside of sqrt, we just want to keep the sign of this object
}

if (obj!=this) obj->doneData();
doneData();

}


void EMData::addConst(float c) {
int i;

if (isComplex()) ap2ri();

getData();
for (i=0; i<nx*ny*nz; i++) rdata[i]+=c;
doneData();

}

void EMData::multConst(float c) {
int i;

if (isComplex()) ap2ri();

getData();
for (i=0; i<nx*ny*nz; i++) rdata[i]*=c;
doneData();
update();
}


// This moves the center of the image to the lower left corner
// ie - phase origin shift from the center of the image to the
// phase origin used by the fft routines
// 2d and 3d only
// note that the complex version is MUCH simpler and faster
void EMData::toCorner() {
int i,j,k;
float t;

getData();

if (isComplex()) {
	int of=0;
	if (((ny/2)%2)+((nz/2)%2)==1) of=1;
	for (k=0; k<nz; k++) {
		for (j=0; j<ny; j++) {
			for (i=((k+j)%2==of?2:0); i<nx; i+=4) {
				rdata[i+j*nx+k*nx*ny]*= -1.0;
				rdata[i+j*nx+k*nx*ny+1]*= -1.0;
			}
		}
	}
	doneData();
	return;
}

if (nz==1) {
	for (j=0; j<ny/2; j++) {
		for (i=0; i<nx/2; i++) {
			t=rdata[i+j*nx];
			rdata[i+j*nx]=rdata[i+nx/2+(j+ny/2)*nx];
			rdata[i+nx/2+(j+ny/2)*nx]=t;
		}
	}

	for (j=0; j<ny/2; j++) {
		for (i=0; i<nx/2; i++) {
			t=rdata[i+nx/2+j*nx];
			rdata[i+nx/2+j*nx]=rdata[i+(j+ny/2)*nx];
			rdata[i+(j+ny/2)*nx]=t;
		}
	}
}
else {
	for (k=0; k<nz/2; k++) {
		for (j=0; j<ny/2; j++) {
			for (i=0; i<nx/2; i++) {
				t=rdata[i+j*nx+k*nx*ny];
				rdata[i+j*nx+k*nx*ny]=rdata[i+nx/2+(j+ny/2)*nx+(k+nz/2)*nx*ny];
				rdata[i+nx/2+(j+ny/2)*nx+(k+nz/2)*nx*ny]=t;
			}
		}
	}

	for (k=0; k<nz/2; k++) {
		for (j=0; j<ny/2; j++) {
			for (i=0; i<nx/2; i++) {
				t=rdata[i+nx/2+j*nx+k*nx*ny];
				rdata[i+nx/2+j*nx+k*nx*ny]=rdata[i+(j+ny/2)*nx+(k+nz/2)*nx*ny];
				rdata[i+(j+ny/2)*nx+(k+nz/2)*nx*ny]=t;
			}
		}
	}

	for (k=0; k<nz/2; k++) {
		for (j=0; j<ny/2; j++) {
			for (i=0; i<nx/2; i++) {
				t=rdata[i+nx/2+(j+ny/2)*nx+k*nx*ny];
				rdata[i+nx/2+(j+ny/2)*nx+k*nx*ny]=rdata[i+j*nx+(k+nz/2)*nx*ny];
				rdata[i+j*nx+(k+nz/2)*nx*ny]=t;
			}
		}
	}

	for (k=0; k<nz/2; k++) {
		for (j=0; j<ny/2; j++) {
			for (i=0; i<nx/2; i++) {
				t=rdata[i+(j+ny/2)*nx+k*nx*ny];
				rdata[i+(j+ny/2)*nx+k*nx*ny]=rdata[i+nx/2+j*nx+(k+nz/2)*nx*ny];
				rdata[i+nx/2+j*nx+(k+nz/2)*nx*ny]=t;
			}
		}
	}

}

doneData();
}

void EMData::commonLinesR(EMData *data1,EMData *data2,int steps,int horiz) {
float *im1,*im2,*d1,*d2;
int i,j,k,l,nny;
float a,f,g;

if (data1->xSize()!=data1->ySize() || data2->xSize()!=data2->ySize() || data1->xSize()!=data2->xSize()) {
	
}
nny=data1->ySize();

data1=data1->copy();
data2=data2->copy();

im1=(float *)malloc(2*steps*nny*sizeof(float));
im2=(float *)malloc(2*steps*nny*sizeof(float));

for (i=0,a=PI/(steps*2.0); i<2.0*steps; i++,a+=PI/(float)steps) {
	// Calculate 1d projection at angle a
	data1->setRAlign(-a);
	data1->rotateAndTranslate();
	d1=data1->getDataRO();
	for (j=0; j<nny; j++) {
		f=0;
		for (k=0; k< nny; k++) {
			f+=d1[j*nny+k];
		}
		im1[i*nny+j]=f;
	}
	
	// calc mean and sigma for line
	f=g=0;
	for (j=0; j<nny; j++) {
		f+=im1[i*nny+j];
		g+=SQR(im1[i*nny+j]);
	}
	f/=nny;
	g=sqrt(g/nny-f*f);

	// normalize line
	for (j=0; j<nny; j++) im1[i*nny+j]=(im1[i*nny+j]-f)/g;
	
	data1->doneData();
}

for (i=0,a=PI/(steps*2.0); i<2.0*steps; i++,a+=PI/(float)steps) {
	data2->setRAlign(-a);
	data2->rotateAndTranslate();
	d2=data2->getDataRO();
	for (j=0; j<nny; j++) {
		f=0;
		for (k=0; k< nny; k++) {
			f+=d2[j*nny+k];
		}
		im2[i*nny+j]=f;
	}

	// calc mean and sigma for line
	f=g=0;
	for (j=0; j<nny; j++) {
		f+=im2[i*nny+j];
		g+=SQR(im2[i*nny+j]);
	}
	f/=nny;
	g=sqrt(g/nny-f*f);

	// normalize line
	for (j=0; j<nny; j++) im2[i*nny+j]=(im2[i*nny+j]-f)/g;

	data2->doneData();
}

setSize(steps*2,steps*2,1);

// Now we calclulate the final common-lines matrix
d1=getData();
// This option calculates y relative to x. That is in a self-CL map, the horizontal
// axis will have 1 rather than the diagonal
if (horiz) {
	for (i=0; i<steps*2; i++) {
		for (j=0,l=i; j<steps*2; j++,l++) {
			if (l==steps*2) l=0;
			f=0;
			for (k=0; k<nny; k++) {
				f+=im1[i*nny+k]*im2[l*nny+k];
			}
			d1[i+j*steps*2]=f;
		}
	}
}
else {
	for (i=0; i<steps*2; i++) {
		for (j=0; j<steps*2; j++) {
			f=0;
			for (k=0; k<nny; k++) {
				f+=im1[i*nny+k]*im2[j*nny+k];
			}
			d1[i+j*steps*2]=f;
		}
	}
}
doneData();

delete data1;	// note these are copies
delete data2;
free(im1);
free(im2);
}

// calculates common lines of 2 data sets (assumes center is already correct)
// mode 0 is dot product		(1 best match)
// mode 1 is weighted phase residual, default (0 best match, PI bad), amplitude weight is from d1
// mode 2 is amplitude only dot product (doesn't require a good center)
// NOTE : for this to work well, you should edgenormalize your images, then pad them by 2x
//		before calling this routine
#define CCLSTEPS	180			// 1/2 the resolution of the map
void EMData::commonLines(EMData *d1,EMData *d2,int mode,int steps,int horiz) {
float *im1,*im2;
int rmax;
int i,j,k,l,xp,yp,nx,ny,jmax;
float da=PI/steps;
float a,a1,a2,p1,p2,r,x,y,s1,s2;
float *r1,*r2;

if (!d1->isComplex()) d1=d1->doFFT();
if (!d2->isComplex()) d2=d2->doFFT();
d1->ap2ri();
d2->ap2ri();

if (d1->xSize()!=d2->xSize() || d1->ySize()!=d2->ySize()) return;
nx=d2->xSize();
ny=d2->ySize();

//rmax=ny/2-1;
rmax=ny/4-1;	// only use the inner 1/2 of fourier space
im1=(float *)calloc(steps*rmax*2,sizeof(float));
im2=(float *)calloc(steps*rmax*2,sizeof(float));

setSize(steps*2,steps*2,1);	// Final map is twice as big with complex conjugates
//setSize(rmax*2,CCLSTEPS,1);

//d1->writeMRC("xx.img");

r1=d1->getDataRO();
r2=d2->getDataRO();
getData();

// this generates radial lines
for (i=0,a=-PI/2.0+da/2.0; i<steps*2; i+=2,a+=da) {
	s1=s2=0;
	for (j=0,r=3.0; r<rmax-3.0; j+=2,r+=1.0) {
		x=r*cos(a);
		y=r*sin(a);
		if (x<0) { x=-x; y=-y; printf("CCL ERROR %d,%f !\n",i,-x); }
		k=(int)(floor(x)*2+floor(y+ny/2)*nx);

// real
		im1[i*rmax+j]=bilin(r1[k],r1[k+2],r1[k+2+nx],r1[k+nx],x-floor(x),y-floor(y));
		im2[i*rmax+j]=bilin(r2[k],r2[k+2],r2[k+2+nx],r2[k+nx],x-floor(x),y-floor(y));

// imaginary
		k++;
		im1[i*rmax+j+1]=bilin(r1[k],r1[k+2],r1[k+2+nx],r1[k+nx],x-floor(x),y-floor(y));
		im2[i*rmax+j+1]=bilin(r2[k],r2[k+2],r2[k+2+nx],r2[k+nx],x-floor(x),y-floor(y));

// complex conjugate for quad II & III
//		if (a>PI/2.0) { im1[i*rmax+j+1]*=-1.0; im2[i*rmax+j+1]*=-1.0; }

// normalization (sum of amplitudes^2)
		s1+=hypot2(im1[i*rmax+j],im1[i*rmax+j+1]);
		s2+=hypot2(im2[i*rmax+j],im2[i*rmax+j+1]);
	}
	jmax=j-1;
	
// normalize each line (sum of amplitudes^2 =1 )
	s1=sqrt(s1);
	s2=sqrt(s2);
	for (j=0,r=1; r<rmax; j+=2,r+=1.0) {
		im1[i*rmax+j]/=s1;
		im1[i*rmax+j+1]/=s1;
		im2[i*rmax+j]/=s2;
		im2[i*rmax+j+1]/=s2; 
	}
}


// map of line-line dot products
// each quadrant done individually for effeciency
// really should combine these into a single set of loops
switch (mode) {
case 0:
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+j*steps*2;
			rdata[l]=0;
//		for (k=4; k<rmax*2-4; k++) {
			for (k=0; k<jmax; k++) {
				rdata[l]+=im1[i*rmax*2+k]*im2[j*rmax*2+k];
			}
		}
	}
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
//		l=i+(CCLSTEPS*2-j-1)*CCLSTEPS*2;
			l=i+steps+j*steps*2;
			rdata[l]=0;
			for (k=0; k<jmax; k+=2) {
				rdata[l]+=im1[i*rmax*2+k]*im2[j*rmax*2+k];
				rdata[l]+=-im1[i*rmax*2+k+1]*im2[j*rmax*2+k+1];
			}
		}
	}
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
//		l=(CCLSTEPS*2-i-1)+j*CCLSTEPS*2;
			l=i+(j+steps)*steps*2;
			rdata[l]=0;
			for (k=0; k<jmax; k+=2) {
				rdata[l]+=im1[i*rmax*2+k]*im2[j*rmax*2+k];
				rdata[l]+=-im1[i*rmax*2+k+1]*im2[j*rmax*2+k+1];
			}
		}
	}
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+steps+(j+steps)*steps*2;
			rdata[l]=0;
			for (k=0; k<jmax; k+=2) {
				rdata[l]+=im1[i*rmax*2+k]*im2[j*rmax*2+k];
				rdata[l]+=im1[i*rmax*2+k+1]*im2[j*rmax*2+k+1];
			}
		}
	}
	break;
case 1:
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+j*steps*2;
			rdata[l]=a=0;
			for (k=0; k<jmax; k+=2) {
				a1=hypot(im1[i*rmax*2+k],im1[i*rmax*2+k+1]);
				a2=hypot(im2[j*rmax*2+k],im2[j*rmax*2+k+1]);
				p1=atan2(im1[i*rmax*2+k+1],im1[i*rmax*2+k]);
				p2=atan2(im2[j*rmax*2+k+1],im2[j*rmax*2+k]);
				rdata[l]+=asub(p1,p2)*a1;
				a+=a1;
			}
			rdata[l]/=a*PI/180.0;
		}
	}
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+steps+j*steps*2;
			rdata[l]=a=0;
			for (k=0; k<jmax; k+=2) {
				a1=hypot(im1[i*rmax*2+k],im1[i*rmax*2+k+1]);
				a2=hypot(im2[j*rmax*2+k],im2[j*rmax*2+k+1]);
				p1=atan2(im1[i*rmax*2+k+1],im1[i*rmax*2+k]);
				p2=atan2(im2[j*rmax*2+k+1],im2[j*rmax*2+k]);
				rdata[l]+=asub(-p1,p2)*a1;
				a+=a1;
			}
			rdata[l]/=a*PI/180.0;
		}
	}
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+(j+steps)*steps*2;
			rdata[l]=a=0;
			for (k=0; k<jmax; k+=2) {
				a1=hypot(im1[i*rmax*2+k],im1[i*rmax*2+k+1]);
				a2=hypot(im2[j*rmax*2+k],im2[j*rmax*2+k+1]);
				p1=atan2(im1[i*rmax*2+k+1],im1[i*rmax*2+k]);
				p2=atan2(im2[j*rmax*2+k+1],im2[j*rmax*2+k]);
				rdata[l]+=asub(-p1,p2)*a1;
				a+=a1;
			}
			rdata[l]/=a*PI/180.0;
		}
	}
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+steps+(j+steps)*steps*2;
			rdata[l]=a=0;
			for (k=0; k<jmax; k+=2) {
				a1=hypot(im1[i*rmax*2+k],im1[i*rmax*2+k+1]);
				a2=hypot(im2[j*rmax*2+k],im2[j*rmax*2+k+1]);
				p1=atan2(im1[i*rmax*2+k+1],im1[i*rmax*2+k]);
				p2=atan2(im2[j*rmax*2+k+1],im2[j*rmax*2+k]);
				rdata[l]+=asub(p1,p2)*a1;
				a+=a1;
			}
			rdata[l]/=a*PI/180.0;
		}
	}
	break;
case 2:
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+j*steps*2;
			rdata[l]=a=0;
			for (k=0; k<jmax; k+=2) {
				a1=hypot(im1[i*rmax*2+k],im1[i*rmax*2+k+1]);
				a2=hypot(im2[j*rmax*2+k],im2[j*rmax*2+k+1]);
				rdata[l]+=a1*a2;
			}
		}
	}
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+steps+j*steps*2;
			rdata[l]=a=0;
			for (k=0; k<jmax; k+=2) {
				a1=hypot(im1[i*rmax*2+k],im1[i*rmax*2+k+1]);
				a2=hypot(im2[j*rmax*2+k],im2[j*rmax*2+k+1]);
				rdata[l]+=a1*a2;
			}
		}
	}
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+(j+steps)*steps*2;
			rdata[l]=a=0;
			for (k=0; k<jmax; k+=2) {
				a1=hypot(im1[i*rmax*2+k],im1[i*rmax*2+k+1]);
				a2=hypot(im2[j*rmax*2+k],im2[j*rmax*2+k+1]);
				rdata[l]+=a1*a2;
			}
		}
	}
	for (i=0; i<steps; i++) {
		for (j=0; j<steps; j++) {
			l=i+steps+(j+steps)*steps*2;
			rdata[l]=a=0;
			for (k=0; k<jmax; k+=2) {
				a1=hypot(im1[i*rmax*2+k],im1[i*rmax*2+k+1]);
				a2=hypot(im2[j*rmax*2+k],im2[j*rmax*2+k+1]);
				rdata[l]+=a1*a2;
			}
		}
	}
	break;
}	
	
//memcpy(rdata,im1,rmax*CCLSTEPS*2*4);
if (horiz) {
	float *tmp=(float *)malloc(ySize()*sizeof(float));
	for (i=1; i<xSize(); i++) {
		for (j=0; j<ySize(); j++) tmp[j]=valueAt(i,j);
		for (j=0; j<ySize(); j++) setValueAt(i,j,0,tmp[(j+i)%ySize()]);
	}
	free(tmp);
}

free(im1);
free(im2);
d1->doneData();
d2->doneData();
doneData();
update();
}

// This will use fftslice or make a real-space projection of a 3d model (positive modes)
// mode=-1 is based on Pawel Penczek's optimized projection routine, subject to some small artifacts
//         due to interpolation scheme used
// mode=-2 fast inaccurate real-space projection
// mode=-3 quick and dirty real-space isosurface
// mode=-4 is real-space with trilinear interpolation
// mode=-5 is real-space for non-cubic volumes
// mode=-6 improved mode -4
// 'toCorner()' MUST be applied before fourier projection
// it must NOT be applied before real-space projection
// real space projection will cause the current data set to be rotated
EMData *EMData::project3d(float alt,float az, float phi,int mode, float thr,int ri,int dcx,int dcy,int dcz)
{
EMData *ret,*tmp,*f;
float *dat,*din,mx[9],rm,x,y,z;
int i,j,k,ii,x2,y2;
float *norm;

if (mode>0) {
	if (!isComplex()) f=doFFT(); else f=this;
	f->setup4Slice(0);
	tmp=f->fftSlice(alt,az,phi,mode);
	tmp->toCorner();
	ret=tmp->doIFT();
	tmp->gimmeFFT();
	ret->applyMask(-1,100+mode);
	ret->setRAlign(alt,az,phi);
	delete tmp;
	ret->update();
	return ret;
}
//Dr.pawel's project algorithm
 if(mode==-1){

     return project3dOpt(alt,az,phi,ri,dcx,dcy,dcz);

 }

if (mode==-6) { // this is essentially identical to mode=-4. but this mode accumulate the results directly in the 2D image
               // instead of storing in the another rotated copy then accumulating (ie: less memory requirement)
               // this mode does not modify the self data either
			   // in this mode ri is used as mask radius, EMData variable scale is used as scale, EMData variable dx and dy are used as translation for the 2-D projection
			   
	//printf("EMData::project3d(): mode = -6\n");
	float mx[9],x2,y2,z2,u,v,t;
	long l=0,xy,ii,x,y,z;
	int rmax;
	float r2 = 0;
	rmax = int(ri * scale + 0.5);	// proj = scale * map
	if(rmax<=0) {
		rmax = nx/2;
	}
	r2 = rmax * rmax;
	
	mx[0]=(cos(phi)*cos(az)-cos(alt)*sin(az)*sin(phi))/scale;
	mx[1]=-(sin(phi)*cos(az)+cos(alt)*sin(az)*cos(phi))/scale;
	mx[2]=sin(alt)*sin(az)/scale;
	mx[3]=(cos(phi)*sin(az)+cos(alt)*cos(az)*sin(phi))/scale;
	mx[4]=(-sin(phi)*sin(az)+cos(alt)*cos(az)*cos(phi))/scale;
	mx[5]=-sin(alt)*cos(az)/scale;
	mx[6]=sin(alt)*sin(phi)/scale;
	mx[7]=sin(alt)*cos(phi)/scale;
	mx[8]=cos(alt)/scale;

	xy=nx*ny;
	
	int jmin = int(-rmax+dy+0.5), jmax = int(rmax+dy+0.5);
	int imin = int(-rmax+dx+0.5),imax = int(rmax+dx+0.5);
	if(jmin<-ny/2) jmin=-ny/2;
	if(jmax>ny/2) jmax=ny/2;
	if(imin<-nx/2) imin=-nx/2;
	if(imax>nx/2) imax=nx/2;
	
	EMData* proj=new EMData;
	proj->setSize(nx,ny);
	proj->zero();
	float *sdata=getDataRO();
	float *ddata=proj->getData();
	
	if(get_map_type() == ICOS5F_HALF) {
		for (int k=-rmax; k<rmax; k++) {
			float kz = k;
			for (int j=jmin; j<jmax; j++) {
				float jy = j - dy;
				for (int i=imin; i<imax; i++) {
					float ix = i - dx;
					if (r2 < (ix*ix + jy*jy + kz*kz)) continue;
				
					x2=mx[0]*ix+mx[1]*jy+mx[2]*kz;
					y2=mx[3]*ix+mx[4]*jy+mx[5]*kz;
					z2=mx[6]*ix+mx[7]*jy+mx[8]*kz;

					if (z2<0) {
						x2*=-1.0;
						z2*=-1.0;
					}
					x2 += nx/2;
					y2 += ny/2;
					
					if (x2<0||y2<0||z2<0||x2>=(nx-1)||y2>=(ny-1)||z2>=(nz-1)) {
						continue;
					}

					x=int(floor2(x2));
					y=int(floor2(y2));
					z=int(floor2(z2));

					t=x2-x;
					u=y2-y;
					v=z2-z;

					ii=x+y*nx+z*xy;
					l = (j+ny/2) * nx + i + nx/2;

					ddata[l] +=	trilin(sdata[ii],sdata[ii+1],sdata[ii+nx],sdata[ii+nx+1],
								sdata[ii+nx*ny],sdata[ii+xy+1],sdata[ii+xy+nx],sdata[ii+xy+nx+1],
								t,u,v);
				}			
			}
		}
	}
	else {
		for (int k=-rmax; k<rmax; k++) {
			float kz = k;
			for (int j=jmin; j<jmax; j++) {
				float jy = j - dy;
				for (int i=imin; i<imax; i++) {
					float ix = i - dx;
					if (r2 < (ix*ix + jy*jy + kz*kz)) continue;
				
					x2=mx[0]*ix+mx[1]*jy+mx[2]*kz+nx/2;
					y2=mx[3]*ix+mx[4]*jy+mx[5]*kz+ny/2;
					z2=mx[6]*ix+mx[7]*jy+mx[8]*kz+nz/2;

					if (x2<0||y2<0||z2<0|| x2>=(nx-1)||y2>=(ny-1)||z2>=(nz-1)) {
						continue;
					}

					x=int(floor2(x2));
					y=int(floor2(y2));
					z=int(floor2(z2));

					t=x2-x;
					u=y2-y;
					v=z2-z;

					ii=x+y*nx+z*xy;
					l = (j+ny/2) * nx + i + nx/2;

					ddata[l] +=	trilin(sdata[ii],sdata[ii+1],sdata[ii+nx],sdata[ii+nx+1],
								sdata[ii+nx*ny],sdata[ii+xy+1],sdata[ii+xy+nx],sdata[ii+xy+nx+1],
								t,u,v);
				}			
			}
		}
	}
	doneData();
	proj->doneData();
	proj->update();
	return proj;
}

// slow accurate real space projection
if (mode==-4) {
	if (!parent) {
		parent=copy();
		parent->setParent(NULL);
//		error("Warning: parent allocated\n",ERR_WARN);
	}
	setRAlign(alt,az,phi);
	setTAlign(0,0,0);
	rotateAndTranslate();

	ret=new EMData;
	ret->setSize(nx,ny,1);
	dat=ret->getData();
	din=getDataRO();
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			dat[i+j*nx]=0;
			for (k=0; k<nz; k++) {
				dat[i+j*nx]+=din[i+j*nx+k*nx*ny];
			}
//			dat[i+j*nx]/=(nz/12.0);
		}
	}
	doneData();
	ret->doneData();
	ret->update();
	return ret;
}
// slow accurate real space projection along y, for helices, etc.
// PHI does not work in this mode, it will be zeroed
if (mode==-5) {
	if (!parent) {
		parent=copy();
		parent->setParent(NULL);
//		error("Warning: parent allocated\n",ERR_WARN);
	}
	setRAlign(alt-PI/2.0,az,0);		// phi is zero
	setTAlign(0,0,0);
	rotateAndTranslate();

	ret=new EMData;
	ret->setSize(nx,nz,1);
	dat=ret->getData();
	din=getDataRO();
	for (i=0; i<nx; i++) {
		for (j=0; j<nz; j++) {
			dat[i+j*nx]=0;
			for (k=0; k<ny; k++) {
				dat[i+j*nx]+=din[i+k*nx+j*nx*ny];
			}
//			dat[i+j*nx]/=(nz/12.0);
		}
	}
	doneData();
	ret->doneData();
	ret->update();
	return ret;
}

// quick and dirty real space projection
if (mode==-2) {
	if (!parent) {
		parent=copy();
		parent->setParent(NULL);
//		error("Warning: parent allocated\n",ERR_WARN);
	}
	setRAlign(alt,az,phi);
	setTAlign(0,0,0);
	rotateAndTranslateFast();

	ret=new EMData;
	ret->setSize(nx,ny,1);
	dat=ret->getData();
	din=getDataRO();
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			dat[i+j*nx]=0;
			for (k=0; k<nz; k++) {
				dat[i+j*nx]+=din[i+j*nx+k*nx*ny];
			}
//			dat[i+j*nx]/=(nz/12.0);
		}
	}
	doneData();
	ret->doneData();
	ret->update();
	return ret;
}

// quick and dirty surface representation
if (mode==-3) {
	int l;
	float slx,sly,v0[3],v1[3],v2[3];

	if (!parent) {
		parent=copy();
		parent->setParent(NULL);
//		error("Warning: parent allocated\n",ERR_WARN);
	}
	setRAlign(alt,az,phi);
	setTAlign(0,0,0);
	rotateAndTranslate();

	ret=new EMData;
	ret->setSize(nx,ny,1);
	dat=ret->getData();
	din=getDataRO();
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			for (k=nz-2; k>0; k--) {
				l=i+j*nx+k*nx*ny;
				if (din[l]>thr) break;
			}
			if (k==0) dat[i+j*nx]=-1000;
			else dat[i+j*nx]=nz-k-(thr-din[l+nx*ny])/(din[l]-din[l+nx*ny]);
		}
	}

	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i++) {
			k=i+j*nx;
			slx=dat[k+1]+dat[k+nx+1]-dat[k]-dat[k+nx];
			sly=dat[k+nx]+dat[k+nx+1]-dat[k]-dat[k+1];
			if (fabs(slx)>50 || fabs(sly)>50 ||dat[k+1]==-1000|| j==ny-1 || i==nx-1) { dat[k]=2.0; continue; }
			v0[0]=1.0/hypot(1.0,slx); v0[1]=0; v0[2]=slx/hypot(1.0,slx);
			v1[0]=0; v1[1]=1.0/hypot(1.0,sly); v1[2]=sly/hypot(1.0,sly);
			v2[0]=v0[1]*v1[2]-v0[2]*v1[1];
			v2[1]=v0[2]*v1[0]-v0[0]*v1[2];
			v2[2]=v0[0]*v1[1]-v0[1]*v1[0];
			dat[k]=-v2[0]-v2[1]+v2[2];
		}
	}

	doneData();
	ret->doneData();
	ret->update();
	return ret;
}

return NULL;
}

EMData * EMData::project3dOpt(float alt,float az,float phi,int ri,int dcx,int dcy,int dcz)
{
    //   EMData *ret;
    if(!vol_table)
    {
	vol_table = new VolTable;
	vol_table->setSize(nx,ny,nz);
	vol_table->setVolTable(this,dcx,dcy,dcz,ri);
    }
    return vol_table->project3d(alt,az,phi);
  
}

//EMData * EMData::project3dOpt(float alt,float az,float phi,char *fsp,float ri,int dcx,int dcy,int dcz)
//{

    // VolTable* dataTable;
//  Euler eu_Ang;
//  int i, j, k, nn;
//  float xb, yb;
//  int cx, cy, cz;
//  int iqx,iqy, iloc;
//  float dipx, dipy, dipy1m, dipx1m;
//  float *dm;
//  float *dat;
//  float *din;
//  float trueData;
//  int *pInt;
//  int sx,sy,sz;
//  EMData* ret;
//  if(!vol_table)
//  {
//
//      vol_table = new VolTable;
	 //	 dataTable = (VolTable*)vol_table;
//	if(fsp)
//{
//    vol_table->setTableFromMRC(fsp,ri);
//}
//else
//{
//    vol_table->setSize(nx, ny, nz);
//    vol_table->setVolTable(this,dcx, dcy, dcz, ri);
//}
//  }
    //   else
    //vol_table = (VolTable*)vol_table;
//  cx = vol_table->getCXSize();
//  cy = vol_table->getCYSize();
//  cz = vol_table->getCZSize();
//  sx = vol_table->getXSize();
//  sy = vol_table->getYSize();
//  sz = vol_table->getZSize();
//  eu_Ang.setAngle(alt, az, phi);
//  ret = new EMData;
//  dm = eu_Ang.mat();
//  ret->setSize(sx,sy,1);
//  ret->zero();
//  dat = ret->getData();
//  din = vol_table->getData();
//  nn = vol_table->getNoOfRow();
//  for(i = 0; i<nn; i++)
//  {
//pInt = vol_table->getPterAt(i);
//xb = (pInt[0]-cx)*dm[0] +(pInt[1]-cy)*dm[1] + (pInt[2]-cz)*dm[2] +cx;
//yb = (pInt[0]-cx)*dm[3] +(pInt[1]-cy)*dm[4] + (pInt[2]-cz)*dm[5] +cy;
//for(k = pInt[3]-1; k < pInt[4]; k++)
//{
//    iqx =(int) floor(xb);
//    iqy =(int) floor(yb);
//    trueData = din[k];
//    dipx = (xb - iqx);
//    dipy = (yb - iqy) * trueData;
//    dipy1m = (trueData - dipy);
//    dipx1m = (1.0 - dipx);
//    iloc = iqy*sx + iqx;
//    dat[iloc] = dat[iloc] + dipx1m*dipy1m;
//    dat[iloc+1] = dat[iloc+1] + dipx*dipy1m;
//    dat[iloc+sx] = dat[iloc+sx] + dipx1m*dipy;
//    dat[iloc+sx+1] = dat[iloc+sx+1]+ dipx*dipy;
//    xb = xb + dm[0];
//    yb = yb + dm[3];
//}
//
//  }
//  doneData();
//  ret->doneData();
//  ret->update();
//  return ret;
//  }*/
    
    


// Note that once this has been allocated, it is used
// globally and never freed.
float *gimx = NULL;

// initialize global gaussian interpolation matrix
void initGIMX() {
int i,j,k;
float r;

gimx=(float*)malloc(sizeof(float)*1000000);
for(k=0; k<100; k++) {
	for(j=0; j<100; j++) {
		for(i=0; i<100; i++) {
			r=hypot3(i/39.0,j/39.0,k/39.0);
			gimx[i+j*100+k*10000]=exp(-r/I5G);
		}
	}
}

}

// This will cut a plane out of a complex volume and return a
// trilinear interpolated image. This is used to generate
// projections of a real space volume
EMData *EMData::fftSlice(float alt,float az,float phi,int mode) {
float xx,yy,zz;
float t,u,v;
EMData *ret;
float *dat,mx[6],cc,r;
int i,j,k,ii,jj,kk,i0,j0,k0,x,y;

if (!isComplex() || nz!=ny || nx!=ny+2) { error("Cannot fftSlice this image",ERR_ERROR); return NULL; }
if (mode==5 && !gimx) initGIMX();

ap2ri();			// make sure we're in real/imaginary space
getDataRO();

ret=new EMData();
ret->setSize(nx,ny,1);
ret->setComplex(1);
ret->setRI(1);
dat=ret->getData();

mx[0]=(cos(phi)*cos(az)-cos(alt)*sin(az)*sin(phi));
mx[1]=-(sin(phi)*cos(az)+cos(alt)*sin(az)*cos(phi));
mx[2]=(cos(phi)*sin(az)+cos(alt)*cos(az)*sin(phi));
mx[3]=(-sin(phi)*sin(az)+cos(alt)*cos(az)*cos(phi));
mx[4]=sin(alt)*sin(phi);
mx[5]=sin(alt)*cos(phi);

for (y=0; y<ny; y++) {
	for (x=0; x<nx/2; x++) {
		ii=x*2+y*nx;
		if (hypot(x,y-ny/2)>=ny/2-1) { dat[ii]=0; dat[ii+1]=0; continue; }
	
		xx=x*mx[0]+(y-ny/2)*mx[1];
		yy=x*mx[2]+(y-ny/2)*mx[3];
		zz=x*mx[4]+(y-ny/2)*mx[5];

		// complex conjugate
		if (xx<0) { xx=-xx; yy=-yy; zz=-zz; cc=-1; } else cc=1;
		yy+=ny/2;
		zz+=nz/2;

		// gaussian interpolation, must be corrected in real space
		interpFT3d(xx,yy,zz,dat+ii,mode);
		dat[ii+1]*=cc;
	}
}

doneData();

ret->doneData();
ret->update();

return ret;
}


// This will interpolate a point in a 3d complex image with various
// quality levels.
// NOTE: no error checking is done because this routine is called a LOT
void EMData::interpFT3d(float x,float y,float z,float *dat,int mode) {
int x0,y0,z0;
int mx0,my0,mz0,mmx,mmy,mmz;
float dx,dy,dz,r,n,g;
int i,j,k,ii;

if (!supp) { error("WARNING: no setup4Slice done",ERR_WARN); setup4Slice(); }
getDataRO();

dat[0]=dat[1]=0;

switch (mode) {
case 1:
		x0=2*(int)floor(x+.5);
		y0=(int)floor(y+.5);
		z0=(int)floor(z+.5);

		dat[0]=rdata[x0+y0*nx+z0*nx*ny];
		dat[1]=rdata[x0+y0*nx+z0*nx*ny+1];
		break;
case 2:
		x0=(int)floor(x);
		y0=(int)floor(y);
		z0=(int)floor(z);

		dx=x-x0;
		dy=y-y0;
		dz=z-z0;

		if (x0>nx-2 || y0>ny-1 || z0>nz-1) break;

		i=(int)(x0*2+y0*nx+z0*nx*ny);

		n=agauss(1.0,dx,dy,dz,I2G)+agauss(1.0,1.0-dx,dy,dz,I2G)+
			agauss(1.0,dx,1.0-dy,dz,I2G)+agauss(1.0,1.0-dx,1.0-dy,dz,I2G)+
			agauss(1.0,dx,dy,1.0-dz,I2G)+agauss(1.0,1.0-dx,dy,1.0-dz,I2G)+
			agauss(1.0,dx,1.0-dy,1.0-dz,I2G)+agauss(1.0,1.0-dx,1.0-dy,1.0-dz,I2G);

		dat[0]=agauss(rdata[i],dx,dy,dz,I2G)+agauss(rdata[i+2],1.0-dx,dy,dz,I2G)+
			agauss(rdata[i+nx],dx,1.0-dy,dz,I2G)+agauss(rdata[i+nx+2],1.0-dx,1.0-dy,dz,I2G)+
			agauss(rdata[i+nx*ny],dx,dy,1.0-dz,I2G)+agauss(rdata[i+2+nx*ny],1.0-dx,dy,1.0-dz,I2G)+
			agauss(rdata[i+nx+nx*ny],dx,1.0-dy,1.0-dz,I2G)+agauss(rdata[i+2+nx+nx*ny],1.0-dx,1.0-dy,1.0-dz,I2G)/n;

		i++;

		dat[1]=agauss(rdata[i],dx,dy,dz,I2G)+agauss(rdata[i+2],1.0-dx,dy,dz,I2G)+
			agauss(rdata[i+nx],dx,1.0-dy,dz,I2G)+agauss(rdata[i+nx+2],1.0-dx,1.0-dy,dz,I2G)+
			agauss(rdata[i+nx*ny],dx,dy,1.0-dz,I2G)+agauss(rdata[i+2+nx*ny],1.0-dx,dy,1.0-dz,I2G)+
			agauss(rdata[i+nx+nx*ny],dx,1.0-dy,1.0-dz,I2G)+agauss(rdata[i+2+nx+nx*ny],1.0-dx,1.0-dy,1.0-dz,I2G)/n;

/*		dat[0]=trilin(rdata[i],rdata[i+2],rdata[i+nx],rdata[i+nx+2],
			rdata[i+nx*ny],rdata[i+nx*ny+2],rdata[i+nx*ny+nx],rdata[i+nx*ny+nx+2],
			dx,dy,dz);

		dat[1]=trilin(rdata[i+1],rdata[i+2+1],rdata[i+nx+1],rdata[i+nx+2+1],
			rdata[i+nx*ny+1],rdata[i+nx*ny+3],rdata[i+nx*ny+nx+1],rdata[i+nx*ny+nx+3],
			dx,dy,dz);*/
		break;
case 3:
		x0=2*(int)floor(x+.5);
		y0=(int)floor(y+.5);
		z0=(int)floor(z+.5);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		// if too close to x origin, use supplemental array
		n=0;
		if (x0==0) {
			x0+=4;
			for (k=z0-1; k<=z0+1; k++) {
				for (j=y0-1; j<=y0+1; j++) {
					for (i=x0-2; i<=x0+2; i+=2) {
						r=hypot3((float)i/2-x-2.0,(float)j-y,(float)k-z);
						
						g=exp(-r/I3G);
						n+=g;
						dat[0]+=supp[i+j*12+k*12*ny]*g;
						dat[1]+=supp[i+j*12+k*12*ny+1]*g;
					}
				}
			}
		}
		else {
			for (k=z0-1; k<=z0+1; k++) {
				for (j=y0-1; j<=y0+1; j++) {
					for (i=x0-2; i<=x0+2; i+=2) {
						r=hypot3((float)i/2-x,(float)j-y,(float)k-z);

						g=exp(-r/I3G);
						n+=g;
						dat[0]+=rdata[i+j*nx+k*nx*ny]*g;
						dat[1]+=rdata[i+j*nx+k*nx*ny+1]*g;
					}
				}
			}
		}
		dat[0]/=n;
		dat[1]/=n;
//printf("%f %f %f %f %f,%f\n",x,y,z,n,dat[0],dat[1]);
		break;
case 4:			// better check this one sometime
		x0=2*(int)floor(x);
		y0=(int)floor(y);
		z0=(int)floor(z);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		// if too close to x origin, use supplemental array
		n=0;
		if (x0==0) {
			x0+=4;
			for (k=z0-1; k<=z0+2; k++) {
				for (j=y0-1; j<=y0+2; j++) {
					for (i=x0-2; i<=x0+4; i+=2) {
						r=hypot3((float)i/2-x-2.0,(float)j-y,(float)k-z);
						
						g=exp(-r/I4G);
						n+=g;
						dat[0]+=supp[i+j*12+k*12*ny]*g;
						dat[1]+=supp[i+j*12+k*12*ny+1]*g;
					}
				}
			}
		}
		else {
			for (k=z0-1; k<=z0+2; k++) {
				for (j=y0-1; j<=y0+2; j++) {
					for (i=x0-2; i<=x0+4; i+=2) {
						r=hypot3((float)i/2-x,(float)j-y,(float)k-z);

						g=exp(-r/I4G);
						n+=g;
						dat[0]+=rdata[i+j*nx+k*nx*ny]*g;
						dat[1]+=rdata[i+j*nx+k*nx*ny+1]*g;
					}
				}
			}
		}
		dat[0]/=n;
		dat[1]/=n;
//printf("%f %f %f %f %f,%f\n",x,y,z,n,dat[0],dat[1]);
		break;
case 5:
		x0=(int)floor(x+.5);
		y0=(int)floor(y+.5);
		z0=(int)floor(z+.5);

		mx0=-(int)floor((x-x0)*39.0+.5)-78;
		my0=-(int)floor((y-y0)*39.0+.5)-78;
		mz0=-(int)floor((z-z0)*39.0+.5)-78;
		x0*=2;
		
		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		// if too close to x origin, use supplemental array
		n=0;
		if (x0<4) {
			x0+=4;
			for (k=z0-2,mmz=mz0; k<=z0+2; k++,mmz+=39) {
				for (j=y0-2,mmy=my0; j<=y0+2; j++,mmy+=39) {
					for (i=x0-4,mmx=mx0; i<=x0+4; i+=2,mmx+=39) {
						g=gimx[abs(mmx)+abs(mmy)*100+abs(mmz)*10000];

						n+=g;
						dat[0]+=supp[i+j*12+k*12*ny]*g;
						dat[1]+=supp[i+j*12+k*12*ny+1]*g;
					}
				}
			}
		}
		else {
			for (k=z0-2,mmz=mz0; k<=z0+2; k++,mmz+=39) {
				for (j=y0-2,mmy=my0; j<=y0+2; j++,mmy+=39) {
					for (i=x0-4,mmx=mx0; i<=x0+4; i+=2,mmx+=39) {
						ii=i+j*nx+k*nx*ny;
						g=gimx[abs(mmx)+abs(mmy)*100+abs(mmz)*10000];

						n+=g;
						dat[0]+=rdata[ii]*g;
						dat[1]+=rdata[ii+1]*g;
					}
				}
			}
		}
		dat[0]/=n;
		dat[1]/=n;
//printf("%f %f %f %f %f,%f\n",x,y,z,n,dat[0],dat[1]);
		break;
case 6:	
		x0=2*(int)floor(x+.5);
		y0=(int)floor(y+.5);
		z0=(int)floor(z+.5);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		// if too close to x origin, use supplemental array
		n=0;
		if (x0<4) {
			x0+=4;
			for (k=z0-2; k<=z0+2; k++) {
				for (j=y0-2; j<=y0+2; j++) {
					for (i=x0-4; i<=x0+4; i+=2) {
						r=hypot3((float)i/2-x-2.0,(float)j-y,(float)k-z);
						
						g=exp(-r/I5G);
						n+=g;
						dat[0]+=supp[i+j*12+k*12*ny]*g;
						dat[1]+=supp[i+j*12+k*12*ny+1]*g;
					}
				}
			}
		}
		else {
			for (k=z0-2; k<=z0+2; k++) {
				for (j=y0-2; j<=y0+2; j++) {
					for (i=x0-4; i<=x0+4; i+=2) {
						r=hypot3((float)i/2-x,(float)j-y,(float)k-z);

						g=exp(-r/I5G);
						n+=g;
						dat[0]+=rdata[i+j*nx+k*nx*ny]*g;
						dat[1]+=rdata[i+j*nx+k*nx*ny+1]*g;
					}
					
				}
			}
		}
		dat[0]/=n;
		dat[1]/=n;
//printf("%f %f %f %f %f,%f\n",x,y,z,n,dat[0],dat[1]);
		break;
// hypergeometric kernel, FT of exp(-x^4 * 1.8)
case 7:
		x0=2*(int)floor(x+.5);
		y0=(int)floor(y+.5);
		z0=(int)floor(z+.5);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		// if too close to x origin, use supplemental array
		n=0;
		if (x0<4) {
			x0+=4;
			for (k=z0-2; k<=z0+2; k++) {
				for (j=y0-2; j<=y0+2; j++) {
					for (i=x0-4; i<=x0+4; i+=2) {
						r=sqrt(hypot3((float)i/2-x-2.0,(float)j-y,(float)k-z));
						
						g=hyperg(r);
						n+=g;
						dat[0]+=supp[i+j*12+k*12*ny]*g;
						dat[1]+=supp[i+j*12+k*12*ny+1]*g;
					}
				}
			}
		}
		else {
			for (k=z0-2; k<=z0+2; k++) {
				for (j=y0-2; j<=y0+2; j++) {
					for (i=x0-4; i<=x0+4; i+=2) {
						r=sqrt(hypot3((float)i/2-x,(float)j-y,(float)k-z));

						g=hyperg(r);
						n+=g;
						dat[0]+=rdata[i+j*nx+k*nx*ny]*g;
						dat[1]+=rdata[i+j*nx+k*nx*ny+1]*g;
					}
					
				}
			}
		}
		dat[0]/=n;
		dat[1]/=n;
//printf("%f %f %f %f %f,%f\n",x,y,z,n,dat[0],dat[1]);
		break;
}

doneData();
}

// this initializes a new image to receive 'insertSlice' operations
// for fourier volume inversion. If the image has a parent, it will also
// be affected. WARNING: will use ~6*size^3 bytes of memory.
void EMData::setup4IS(int size)
{
int i;
float f;

setSize(size+2,size,size);
setComplex(1);
setRI(1);
zero();

// The small random phase initialization caused some normalization problems
// initialize with small random-phase number so phase residuals don't look good when empty
//getData();
//for (i=0; i<nx*ny*nz; i+=2) {
//	f=frand(0,2.0*PI);
//	rdata[i]=1.0e-10*sin(f);
//	rdata[i+1]=1.0e-10*cos(f);
//}
//doneData();

if (!parent) parent=new EMData();
parent->setSize(size+2,size,size);
parent->zero();
}


// When interpolating in fourier space there is a little problem when we get
// close to x=0, since f(-x,-y,-z)=f(x,y,z)*
// So this makes a little array that allows for up to +-2 point interpolation
// all the way to the origin in x
void EMData::setup4Slice(int redo) {
int x,y,z,yy,zz;
int i;

if (!isComplex() || nz==1) return;
if (supp) {
	if (redo) free(supp);
	else return;
}
supp=(float *)calloc(12,ny*nz*sizeof(float));
getDataRO();

for (z=0; z<nz; z++) {
	for (y=0; y<ny; y++) {
		for (x=0; x<8; x++) {
			supp[(x+4)+y*12+z*12*ny]=rdata[x+y*nx+z*nx*ny];
		}
	}
}

for (z=1,zz=nz-1; z<nz; z++,zz--) {
	for (y=1,yy=ny-1; y<ny; y++,yy--) {
		supp[  y*12+z*12*ny]= rdata[4+yy*nx+zz*nx*ny];
		supp[1+y*12+z*12*ny]=-rdata[5+yy*nx+zz*nx*ny];
		supp[2+y*12+z*12*ny]= rdata[2+yy*nx+zz*nx*ny];
		supp[3+y*12+z*12*ny]=-rdata[3+yy*nx+zz*nx*ny];
	}
}

//for (i=0; i<12; i++) printf("%f ",supp[i+(ny/2-1)*12+z*6*ny]);
//printf("\n");

doneData();
}

// caclulates normalization and phase residual for a slice in an already existing volume
// phase residual is calculated using only the inner 1/2 of the fourier sphere
float EMData::normSlice(EMData *slice,float alt,float az,float phi,float *phaseres)
{
int x0,y0,z0;
float xx,yy,zz,pr,prn,p1,p2;
EMData *ret;
float *dat,*norm,mx[6],cc;
float dt[2],d,r,rn,gg,rad,thr=0;
int y,x,i;

if (!isComplex() || !slice->isComplex() || !parent) return 0;

thr=parent->Sigma()*2.0; // Only consider regions with significant 'filled in' amplitude
slice->ap2ri();
getDataRO();
norm=parent->getData();
dat=slice->getData();

//printf("* %f %f %f\n",parent->Min(),parent->Max(),parent->Sigma());
//i=0;
//for (x=0; x<nx*ny*nz; x+=2) if (norm[x]>parent->Sigma()*2.0) i++;
//printf("*** %d\n",i);


mx[0]=(cos(phi)*cos(az)-cos(alt)*sin(az)*sin(phi));
mx[1]=-(sin(phi)*cos(az)+cos(alt)*sin(az)*cos(phi));
mx[2]=(cos(phi)*sin(az)+cos(alt)*cos(az)*sin(phi));
mx[3]=(-sin(phi)*sin(az)+cos(alt)*cos(az)*cos(phi));
mx[4]=sin(alt)*sin(phi);
mx[5]=sin(alt)*cos(phi);

r=rn=0;		// normalization
pr=prn=0;	// phase residual
// find points that are already inserted and
for (y=0; y<ny; y++) {
	for (x=0; x<nx/2; x++) {
		rad=hypot(x,y-ny/2);
		if (rad>=ny/2-1||rad<2.0) continue;

		xx=x*mx[0]+(y-ny/2)*mx[1];
		yy=x*mx[2]+(y-ny/2)*mx[3];
		zz=x*mx[4]+(y-ny/2)*mx[5];
		// complex conjugate
		if (xx<0) { xx=-xx; yy=-yy; zz=-zz; cc=-1.0; } else cc=1.0;
		yy+=ny/2;
		zz+=nz/2;


		x0=2*(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);

//		d=hypot3((float)x0/2-xx,(float)y0-yy,(float)z0-zz);
//		gg=exp(-d/I5G);

		i=x0+y0*nx+z0*nx*ny;
		if (fabs(norm[i])<thr) continue;
//		if (rdata[i]==0 || rdata[i+1]==0) continue;
//		if (norm[i]==0) continue;
/*		dt[1]=hypot(rdata[i],rdata[i+1]);
		dt[0]=dat[x*2+y*nx]/rdata[i]+cc*dat[x*2+1+y*nx]/rdata[i+1];
		r+=norm[i]*norm[i]*dt[1]*dt[0];
		rn+=norm[i]*dt[1];*/
		dt[0]=hypot(rdata[i],rdata[i+1]);
		dt[1]=hypot(dat[x*2+y*nx],dat[x*2+1+y*nx]);
//		r+=norm[i]*dt[1];
//		rn+=dt[0];
//		r+=SQR(norm[i])*dt[1]/dt[0];
//		rn+=norm[i];
//		r+=norm[i]*dt[1]/dt[0];
//		rn+=norm[i];
		r+=norm[i]*dt[1];
		rn+=norm[i]*dt[0];
		
/*		if (rad>3.0 && norm[i]>0) {
			pr+=fabs(dt[0]/norm[i]-dt[1])*norm[i];
			prn+=norm[i];
		} */

		if (rdata[i+1]==0 && rdata[i]==0) p1=0;
		else p1=atan2(cc*rdata[i+1],rdata[i]);
		if (dat[x*2+1+y*nx]==0 && dat[x*2+y*nx]==0) p2=0;
		else p2=atan2(dat[x*2+1+y*nx],dat[x*2+y*nx]);

		// phase residual using only inner 1/2 of fourier space
//		if (rad<ny/4 && rad>3.0) {
		if (rad>3.0) {
			pr+=asub(p1,p2)*dt[0]*dt[1]*norm[i];
			prn+=dt[0]*dt[1]*norm[i];
		}
	}
}
//printf("normslice %f   %f   %f\n",r,rn,r/rn);

if (phaseres) *phaseres= (prn==0)?0:pr/prn;

// ok now normalize
if (rn!=0) {
	r=r/rn;
//	for (i=0; i<nx*ny; i++) dat[i]/=r; 
//	printf("Slice norm %f (%f) \n",r,rn);
}
else r=1.0;

doneData();
parent->doneData();
slice->doneData();
slice->update();

return r;
}


// this inserts a complex slice into an image that has been setup4IS()
// NOTE:  ONLY MODE 5 & 6 DO A COMPLETE JOB (the others are close, just a little issue near the x origin)
void EMData::insertSlice(EMData *slice,float alt,float az,float phi,int mode, float weight) {
int x0,y0,z0;
int mx0,my0,mz0,mmx,mmy,mmz;
float xx,yy,zz;
float t,u,v,d,r,gg;
EMData *ret;
float *dat,*norm,mx[6],cc;
int i,j,k,l,ll,ii,x,y,off[8],rl;
float dt[2],g[8];

if (!isComplex() || !slice->isComplex() || !parent) return;

// This is a big matrix for speedy interpolation. Hopefully
// it's accurate enough...
// really should take advantage of symmetry here...
if (mode==5 && !gimx) initGIMX();

if (mode==2) {
	off[0]=0;
	off[1]=2;
	off[2]=nx;
	off[3]=nx+2;
	off[4]=nx*ny;
	off[5]=nx*ny+2;
	off[6]=nx*ny+nx;
	off[7]=nx*ny+nx+2;
}

getDataRO();
norm=parent->getData();
dat=slice->getData();

mx[0]=(cos(phi)*cos(az)-cos(alt)*sin(az)*sin(phi));
mx[1]=-(sin(phi)*cos(az)+cos(alt)*sin(az)*cos(phi));
mx[2]=(cos(phi)*sin(az)+cos(alt)*cos(az)*sin(phi));
mx[3]=(-sin(phi)*sin(az)+cos(alt)*cos(az)*cos(phi));
mx[4]=sin(alt)*sin(phi);
mx[5]=sin(alt)*cos(phi);

rl=(ny/2-1)*(ny/2-1);

for (y=0; y<ny; y++) {
	for (x=0; x<nx/2; x++) {
		if ((x*x+SQR(y-ny/2))>=rl) continue;

		xx=x*mx[0]+(y-ny/2)*mx[1];
		yy=x*mx[2]+(y-ny/2)*mx[3];
		zz=x*mx[4]+(y-ny/2)*mx[5];
		// complex conjugate
		if (xx<0) { xx=-xx; yy=-yy; zz=-zz; cc=-1.0; } else cc=1.0;
		yy+=ny/2;
		zz+=nz/2;

		dt[0]=dat[x*2+y*nx];
		dt[1]=cc*dat[x*2+1+y*nx];

switch (mode) {
	case 1:
		x0=2*(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);
		rdata[x0+y0*nx+z0*nx*ny]+=weight*dt[0];
		rdata[x0+y0*nx+z0*nx*ny+1]+=weight*dt[1];
		norm[x0+y0*nx+z0*nx*ny]+=weight;
		break;
	case 2:
		x0=(int)floor(xx);
		y0=(int)floor(yy);
		z0=(int)floor(zz);

		dx=xx-x0;
		dy=yy-y0;
		dz=zz-z0;

		if (x0>nx-2 || y0>ny-1 || z0>nz-1) break;

		i=(int)(x0*2+y0*nx+z0*nx*ny);

		g[0]=agauss(1.0,dx,dy,dz,I2G);
		g[1]=agauss(1.0,1.0-dx,dy,dz,I2G);
		g[2]=agauss(1.0,dx,1.0-dy,dz,I2G);
		g[3]=agauss(1.0,1.0-dx,1.0-dy,dz,I2G);
		g[4]=agauss(1.0,dx,dy,1.0-dz,I2G);
		g[5]=agauss(1.0,1.0-dx,dy,1.0-dz,I2G);
		g[6]=agauss(1.0,dx,1.0-dy,1.0-dz,I2G);
		g[7]=agauss(1.0,1.0-dx,1.0-dy,1.0-dz,I2G);

		for (j=0; j<8; j++) {
			k=i+off[j];
			rdata[k]+=weight*dt[0]*g[j];
			rdata[k+1]+=weight*dt[1]*g[j];
			norm[k]+=weight*g[j];
			norm[k+1]+=weight*g[j]*dt[0]*dt[0]*dt[1]*dt[1];
		}

		break;
	case 3:
		x0=2*(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		// if too close to x origin, use supplemental array
		if (x0==0) l=x0; else l=x0-2;
		for (k=z0-1; k<=z0+1; k++) {
			for (j=y0-1; j<=y0+1; j++) {
				for (i=l; i<=x0+2; i+=2) {
					r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
					gg=exp(-r/I3G);

					rdata[i+j*nx+k*nx*ny]+=weight*gg*dt[0];
					rdata[i+j*nx+k*nx*ny+1]+=weight*gg*dt[1];
					norm[i+j*nx+k*nx*ny]+=weight*gg;
				}
			}
		}
		break;
	case 4:			// better check this one sometime
		x0=2*(int)floor(xx);
		y0=(int)floor(yy);
		z0=(int)floor(zz);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		if (x0==0) l=x0; else l=x0-2;
		for (k=z0-1; k<=z0+2; k++) {
			for (j=y0-1; j<=y0+2; j++) {
				for (i=l; i<=x0+4; i+=2) {
					r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
					gg=exp(-r/I4G);

					rdata[i+j*nx+k*nx*ny]+=weight*gg*dt[0];
					rdata[i+j*nx+k*nx*ny+1]+=weight*gg*dt[1];
					norm[i+j*nx+k*nx*ny]+=weight*gg;
				}
			}
		}
		break;
	case 5:
	
		x0=(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);

		mx0=-(int)floor((xx-x0)*39.0+.5)-78;
		my0=-(int)floor((yy-y0)*39.0+.5)-78;
		mz0=-(int)floor((zz-z0)*39.0+.5)-78;
		x0*=2;
//printf("%3d %3d %3d   %f %f %f\n",mx0,my0,mz0,xx,yy,zz);
		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		if (x0==0) { l=0; mx0+=78; }
		else if (x0==2) { l=0; mx0+=39; }
		else l=x0-4;
		for (k=z0-2,mmz=mz0; k<=z0+2; k++,mmz+=39) {
			for (j=y0-2,mmy=my0; j<=y0+2; j++,mmy+=39) {
				for (i=l,mmx=mx0; i<=x0+4; i+=2,mmx+=39) {
					ii=i+j*nx+k*nx*ny;
					gg=weight*gimx[abs(mmx)+abs(mmy)*100+abs(mmz)*10000];

//					r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
//	if (mx0%39!=0)				printf("%d %d %d\t%f %f\t%f\n",mmx,mmy,mmz,gg,weight*exp(-r/I5G),gg-weight*exp(-r/I5G));
					
					rdata[ii]+=gg*dt[0];
					rdata[ii+1]+=gg*dt[1];
					norm[ii]+=gg;
				}
			}
		}

		// when we're too close to the origin we need to deal with the mirrored points
		if (x0<=2) {
			xx=-xx;
			yy=-(yy-ny/2)+ny/2;
			zz=-(zz-nz/2)+nz/2;
			x0=(int)floor(xx+.5);
			y0=(int)floor(yy+.5);
			z0=(int)floor(zz+.5);
			mx0=-(int)floor((xx-x0)*39.0+.5);
			x0*=2.0;

			if (y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

			for (k=z0-2,mmz=mz0; k<=z0+2; k++,mmz+=39) {
				for (j=y0-2,mmy=my0; j<=y0+2; j++,mmy+=39) {
					for (i=0,mmx=mx0; i<=x0+4; i+=2,mmx+=39) {
						ii=i+j*nx+k*nx*ny;
						gg=weight*gimx[abs(mmx)+abs(mmy)*100+abs(mmz)*10000];
	
						rdata[ii]+=gg*dt[0];
						rdata[ii+1]-=gg*dt[1];	// note the -, complex conj.
						norm[ii]+=gg;
					}
				}
			}
		}
		break;
// mode 6 is now mode 5 without the fast interpolation
	case 6:

		x0=2*(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		if (x0<=2) l=0;
		else l=x0-4;
		for (k=z0-2; k<=z0+2; k++) {
			for (j=y0-2; j<=y0+2; j++) {
				for (i=l; i<=x0+4; i+=2) {
					ii=i+j*nx+k*nx*ny;
					r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
					gg=weight*exp(-r/I5G);

					rdata[ii]+=gg*dt[0];
					rdata[ii+1]+=gg*dt[1];
					norm[ii]+=gg;
				}
			}
		}
		
		// when we're too close to the origin we need to deal with the mirrored points
		if (x0<=2) {
			xx=-xx;
			yy=-(yy-ny/2)+ny/2;
			zz=-(zz-nz/2)+nz/2;
			x0=2*(int)floor(xx+.5);
			y0=(int)floor(yy+.5);
			z0=(int)floor(zz+.5);

			if (y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

			for (k=z0-2; k<=z0+2; k++) {
				for (j=y0-2; j<=y0+2; j++) {
					for (i=0; i<=x0+4; i+=2) {
						ii=i+j*nx+k*nx*ny;
						r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
						gg=weight*exp(-r/I5G);
	
						rdata[ii]+=gg*dt[0];
						rdata[ii+1]-=gg*dt[1];	// note the -, complex conj.
						norm[ii]+=gg;
					}
				}
			}
		}
		break;
// mode 7 is 5x5x5 with a hypergeometric correction factor
	case 7:

		x0=2*(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		if (x0<=2) l=0;
		else l=x0-4;
		for (k=z0-2; k<=z0+2; k++) {
			for (j=y0-2; j<=y0+2; j++) {
				for (i=l; i<=x0+4; i+=2) {
					ii=i+j*nx+k*nx*ny;
					r=sqrt(hypot3((float)i/2-xx,(float)j-yy,(float)k-zz));
					gg=weight*hyperg(r);

					rdata[ii]+=gg*dt[0];
					rdata[ii+1]+=gg*dt[1];
					norm[ii]+=gg;
				}
			}
		}
		
		// when we're too close to the origin we need to deal with the mirrored points
		if (x0<=2) {
			xx=-xx;
			yy=-(yy-ny/2)+ny/2;
			zz=-(zz-nz/2)+nz/2;
			x0=2*(int)floor(xx+.5);
			y0=(int)floor(yy+.5);
			z0=(int)floor(zz+.5);

			if (y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

			for (k=z0-2; k<=z0+2; k++) {
				for (j=y0-2; j<=y0+2; j++) {
					for (i=0; i<=x0+4; i+=2) {
						ii=i+j*nx+k*nx*ny;
						r=sqrt(hypot3((float)i/2-xx,(float)j-yy,(float)k-zz));
						gg=weight*hyperg(r);
	
						rdata[ii]+=gg*dt[0];
						rdata[ii+1]-=gg*dt[1];	// note the -, complex conj.
						norm[ii]+=gg;
					}
				}
			}
		}
		break;
	}

	}
}

doneData();
parent->doneData();
slice->doneData();
slice->update();

return;
}

// this inserts a complex slice into an image that has been setup4IS()
// This version assumes slices are already Wiener filtered based on the SNR curve
// provided. Original Wiener filter is removed, and weight array constructed ST
// a new Wiener fileter can be applied to the results in 3D
// since slices are usually padded before FFT padratio contains a radial scaling factor
// to use on SNR, ie - an 80x80 image padded to 120x120 would have padratio=1.5
void EMData::insertSliceWF(EMData *slice,float alt,float az,float phi,int mode, float *SNR, float padratio, int deWF, int snr_ndim) {
int x0,y0,z0;
int mx0,my0,mz0,mmx,mmy,mmz;
float xx,yy,zz;
float t,u,v,d,r,gg,weight;
EMData *ret;
float *dat,*norm,mx[6],cc;
int i,j,k,l,ll,ii,x,y,off[8],rl;
float dt[2],g[8];

if (!isComplex() || !slice->isComplex() || !parent) return;

// This is a big matrix for speedy interpolation. Hopefully
// it's accurate enough...
// really should take advantage of symmetry here...
if (mode==5 && !gimx) initGIMX();

if (mode==2) {
	off[0]=0;
	off[1]=2;
	off[2]=nx;
	off[3]=nx+2;
	off[4]=nx*ny;
	off[5]=nx*ny+2;
	off[6]=nx*ny+nx;
	off[7]=nx*ny+nx+2;
}

getDataRO();
norm=parent->getData();
dat=slice->getData();

mx[0]=(cos(phi)*cos(az)-cos(alt)*sin(az)*sin(phi));
mx[1]=-(sin(phi)*cos(az)+cos(alt)*sin(az)*cos(phi));
mx[2]=(cos(phi)*sin(az)+cos(alt)*cos(az)*sin(phi));
mx[3]=(-sin(phi)*sin(az)+cos(alt)*cos(az)*cos(phi));
mx[4]=sin(alt)*sin(phi);
mx[5]=sin(alt)*cos(phi);

rl=(ny/2-1)*(ny/2-1);

for (y=0; y<ny; y++) {
	for (x=0; x<nx/2; x++) {
		if ((x*x+SQR(y-ny/2))>=rl) continue;

		if (snr_ndim == 2) {
			weight = SNR[y*nx+2*x];	// treat it as the same size as the image itself
			//weight=1.0;
		}
		else if (snr_ndim == 1){
			int r=round(hypot((float)x,(float)y-ny/2)*CTFOS/padratio);
			if (r>=CTFOS*ny/2) r=CTFOS*ny/2-1;
			weight=SNR[r];
		}
		else
			weight=1.0;

		xx=x*mx[0]+(y-ny/2)*mx[1];
		yy=x*mx[2]+(y-ny/2)*mx[3];
		zz=x*mx[4]+(y-ny/2)*mx[5];
		// complex conjugate
		if (xx<0) { xx=-xx; yy=-yy; zz=-zz; cc=-1.0; } else cc=1.0;
		yy+=ny/2;
		zz+=nz/2;

		if(deWF) {
			dt[0]=dat[x*2+y*nx]*(1.0+1.0/weight);	// remove old Wiener filter
			dt[1]=cc*dat[x*2+1+y*nx]*(1.0+1.0/weight);
		}
		else {
			dt[0]=dat[x*2+y*nx];
			dt[1]=cc*dat[x*2+1+y*nx];
		}

switch (mode) {
	case 1:
		x0=2*(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);
		rdata[x0+y0*nx+z0*nx*ny]+=weight*dt[0];
		rdata[x0+y0*nx+z0*nx*ny+1]+=weight*dt[1];
		norm[x0+y0*nx+z0*nx*ny]+=weight;
		break;
	case 2:
		x0=(int)floor(xx);
		y0=(int)floor(yy);
		z0=(int)floor(zz);

		dx=xx-x0;
		dy=yy-y0;
		dz=zz-z0;

		weight/=pow(I2G*PI,1.5);

		if (x0>nx-2 || y0>ny-1 || z0>nz-1) break;

		i=(int)(x0*2+y0*nx+z0*nx*ny);

		g[0]=agauss(1.0,dx,dy,dz,I2G);
		g[1]=agauss(1.0,1.0-dx,dy,dz,I2G);
		g[2]=agauss(1.0,dx,1.0-dy,dz,I2G);
		g[3]=agauss(1.0,1.0-dx,1.0-dy,dz,I2G);
		g[4]=agauss(1.0,dx,dy,1.0-dz,I2G);
		g[5]=agauss(1.0,1.0-dx,dy,1.0-dz,I2G);
		g[6]=agauss(1.0,dx,1.0-dy,1.0-dz,I2G);
		g[7]=agauss(1.0,1.0-dx,1.0-dy,1.0-dz,I2G);

		for (j=0; j<8; j++) {
			k=i+off[j];
			rdata[k]+=weight*dt[0]*g[j];
			rdata[k+1]+=weight*dt[1]*g[j];
			norm[k]+=weight*g[j];
			norm[k+1]+=weight*g[j]*dt[0]*dt[0]*dt[1]*dt[1];
		}

		break;
	case 3:
		x0=2*(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);

		weight/=pow(I3G*PI,1.5);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		// if too close to x origin, use supplemental array
		if (x0==0) l=x0; else l=x0-2;
		for (k=z0-1; k<=z0+1; k++) {
			for (j=y0-1; j<=y0+1; j++) {
				for (i=l; i<=x0+2; i+=2) {
					r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
					gg=exp(-r/I3G);

					rdata[i+j*nx+k*nx*ny]+=weight*gg*dt[0];
					rdata[i+j*nx+k*nx*ny+1]+=weight*gg*dt[1];
					norm[i+j*nx+k*nx*ny]+=weight*gg;
				}
			}
		}
		break;
	case 4:			// better check this one sometime
		x0=2*(int)floor(xx);
		y0=(int)floor(yy);
		z0=(int)floor(zz);

		weight/=pow(I4G*PI,1.5);
		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		if (x0==0) l=x0; else l=x0-2;
		for (k=z0-1; k<=z0+2; k++) {
			for (j=y0-1; j<=y0+2; j++) {
				for (i=l; i<=x0+4; i+=2) {
					r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
					gg=exp(-r/I4G);

					rdata[i+j*nx+k*nx*ny]+=weight*gg*dt[0];
					rdata[i+j*nx+k*nx*ny+1]+=weight*gg*dt[1];
					norm[i+j*nx+k*nx*ny]+=weight*gg;
				}
			}
		}
		break;
	case 5:
	
		x0=(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);

		weight/=pow(I5G*PI,1.5);

		mx0=-(int)floor((xx-x0)*39.0+.5)-78;
		my0=-(int)floor((yy-y0)*39.0+.5)-78;
		mz0=-(int)floor((zz-z0)*39.0+.5)-78;
		x0*=2;
//printf("%3d %3d %3d   %f %f %f\n",mx0,my0,mz0,xx,yy,zz);
		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		if (x0==0) { l=0; mx0+=78; }
		else if (x0==2) { l=0; mx0+=39; }
		else l=x0-4;
		for (k=z0-2,mmz=mz0; k<=z0+2; k++,mmz+=39) {
			for (j=y0-2,mmy=my0; j<=y0+2; j++,mmy+=39) {
				for (i=l,mmx=mx0; i<=x0+4; i+=2,mmx+=39) {
					ii=i+j*nx+k*nx*ny;
					gg=weight*gimx[abs(mmx)+abs(mmy)*100+abs(mmz)*10000];

//					r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
//	if (mx0%39!=0)				printf("%d %d %d\t%f %f\t%f\n",mmx,mmy,mmz,gg,weight*exp(-r/I5G),gg-weight*exp(-r/I5G));
					
					rdata[ii]+=gg*dt[0];
					rdata[ii+1]+=gg*dt[1];
					norm[ii]+=gg;
				}
			}
		}

		// when we're too close to the origin we need to deal with the mirrored points
		if (x0<=2) {
			xx=-xx;
			yy=-(yy-ny/2)+ny/2;
			zz=-(zz-nz/2)+nz/2;
			x0=(int)floor(xx+.5);
			y0=(int)floor(yy+.5);
			z0=(int)floor(zz+.5);
			mx0=-(int)floor((xx-x0)*39.0+.5);
			x0*=2.0;

			if (y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

			for (k=z0-2,mmz=mz0; k<=z0+2; k++,mmz+=39) {
				for (j=y0-2,mmy=my0; j<=y0+2; j++,mmy+=39) {
					for (i=0,mmx=mx0; i<=x0+4; i+=2,mmx+=39) {
						ii=i+j*nx+k*nx*ny;
						gg=weight*gimx[abs(mmx)+abs(mmy)*100+abs(mmz)*10000];

						rdata[ii]+=gg*dt[0];
						rdata[ii+1]-=gg*dt[1];	// note the -, complex conj.
						norm[ii]+=gg;
					}
				}
			}
		}
		break;
// mode 6 is now mode 5 without the fast interpolation
	case 6:
	
		x0=2*(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);

		weight/=pow(I5G*PI,1.5);
		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		if (x0<=2) l=0;
		else l=x0-4;
		for (k=z0-2; k<=z0+2; k++) {
			for (j=y0-2; j<=y0+2; j++) {
				for (i=l; i<=x0+4; i+=2) {
					ii=i+j*nx+k*nx*ny;
					r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
					gg=weight*exp(-r/I5G);

					rdata[ii]+=gg*dt[0];
					rdata[ii+1]+=gg*dt[1];
					norm[ii]+=gg;
				}
			}
		}
		
		// when we're too close to the origin we need to deal with the mirrored points
		if (x0<=2) {
			xx=-xx;
			yy=-(yy-ny/2)+ny/2;
			zz=-(zz-nz/2)+nz/2;
			x0=2*(int)floor(xx+.5);
			y0=(int)floor(yy+.5);
			z0=(int)floor(zz+.5);

			if (y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

			for (k=z0-2; k<=z0+2; k++) {
				for (j=y0-2; j<=y0+2; j++) {
					for (i=0; i<=x0+4; i+=2) {
						ii=i+j*nx+k*nx*ny;
						r=hypot3((float)i/2-xx,(float)j-yy,(float)k-zz);
						gg=weight*exp(-r/I5G);
	
						rdata[ii]+=gg*dt[0];
						rdata[ii+1]-=gg*dt[1];	// note the -, complex conj.
						norm[ii]+=gg;
					}
				}
			}
		}
		break;
// mode 7 is 5x5x5 with a hypergeometric correction factor
	case 7:

		x0=2*(int)floor(xx+.5);
		y0=(int)floor(yy+.5);
		z0=(int)floor(zz+.5);

		// there is a 2 pixel zero zone around the edge of fourier space
		if (x0>=nx-4 || y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

		if (x0<=2) l=0;
		else l=x0-4;
		for (k=z0-2; k<=z0+2; k++) {
			for (j=y0-2; j<=y0+2; j++) {
				for (i=l; i<=x0+4; i+=2) {
					ii=i+j*nx+k*nx*ny;
					r=sqrt(hypot3((float)i/2-xx,(float)j-yy,(float)k-zz));
					gg=weight*hyperg(r);

					rdata[ii]+=gg*dt[0];
					rdata[ii+1]+=gg*dt[1];
					norm[ii]+=gg;
				}
			}
		}
		
		// when we're too close to the origin we need to deal with the mirrored points
		if (x0<=2) {
			xx=-xx;
			yy=-(yy-ny/2)+ny/2;
			zz=-(zz-nz/2)+nz/2;
			x0=2*(int)floor(xx+.5);
			y0=(int)floor(yy+.5);
			z0=(int)floor(zz+.5);

			if (y0>ny-3 || z0>nz-3 || y0<2 || z0<2) break;

			for (k=z0-2; k<=z0+2; k++) {
				for (j=y0-2; j<=y0+2; j++) {
					for (i=0; i<=x0+4; i+=2) {
						ii=i+j*nx+k*nx*ny;
						r=sqrt(hypot3((float)i/2-xx,(float)j-yy,(float)k-zz));
						gg=weight*hyperg(r);
	
						rdata[ii]+=gg*dt[0];
						rdata[ii+1]-=gg*dt[1];	// note the -, complex conj.
						norm[ii]+=gg;
					}
				}
			}
		}
		break;
	}

	}
}

doneData();
parent->doneData();
slice->doneData();
slice->update();

return;
}

// once all of the slices have been added with insertSlice(), this routine
// does a normalization do an IFT to get the reconstructed volume
// dlog unused
void EMData::doneSlice(int dlog) {
int i;
float *norm;
float d;
EMData *ret;

getData();
norm=parent->getData();

// This normalizes the data such that intensities of the class-averages
// are preserved on average (only works with mode2 insertslice)
if (dlog) {
	for (i=0; i<nx*ny*nz; i+=2) {
		d=norm[i];
		if (d==0) rdata[i]=rdata[i+1]=0;
		else {
			float in=norm[i+1]/norm[i];
			float cin=hypot2(rdata[i],rdata[i+1]);
			rdata[i]*=sqrt(in/cin);
			rdata[i+1]*=sqrt(in/cin);
		}
	}
} else {
// regular normalization
	for (i=0; i<nx*ny*nz; i+=2) {
		d=norm[i];
		if (d==0) rdata[i]=rdata[i+1]=0;
		else {
			rdata[i]/=d;
			rdata[i+1]/=d;
		}
	}
}

doneData();
parent->doneData();

}

// once all of the slices have been added with insertSlice(), this routine
// does a normalization do an IFT to get the reconstructed volume
// dlog unused
void EMData::doneSliceWF(int dlog) {
int i;
float *norm;
float d;
EMData *ret;

getData();
norm=parent->getData();

for (i=0; i<nx*ny*nz; i+=2) {
	d=norm[i];
	if (d==0) rdata[i]=rdata[i+1]=0;
	else {
		rdata[i]/=d*(1.0+1.0/d);	// reapply Wiener filter with new weights
		rdata[i+1]/=d*(1.0+1.0/d);	// and divide out the weighting factor
	}
}

doneData();
parent->doneData();

}


// does an IFT and corrects for trilinear sampling
EMData *EMData::iftSlice() {
float x,y,z,r;
int i;
EMData *ret;
float *rd;

ret=doIFT();
/*ret->toCorner();

rd=ret->getData();

i=0;
for (z=-.5; z<=.5; z+=1.0/(nz-1.0)) {
	for (y=-.5; y<.5; y+=1.0/(ny-1.0)) {
		for (x=-.5; x<.5; x+=1.0/(nx-3.0),i++) { 
			rd[i]/=((1.0-cos(2.0*PI*x))*(1.0-cos(2.0*PI*y))*(1.0-cos(2.0*PI*z))/
				(4*PI*PI*x*x*y*y*z*z));
		}
	}
}

ret->doneData();
ret->toCorner();*/
return ret;
}

// This implements 3D reconstruction via back-projection. This is a basic slice insertion routine.
// Any desired filtering must be performed outside the context of this routine.
void EMData::insertSlice3DReal(EMData *slice,float weight) {
EMData tmp;
float *dp,*dv;
int i;

if (slice->xSize() != slice->ySize() || slice->xSize() != zSize()) {
	error("Improper size in insertSlice3DReal()",ERR_CRIT);
	return;
}
slice=slice->copy();
slice->multConst(weight);

tmp.setSize(nx,ny,nz);

// Project the slice along the z-axis
dp=slice->getDataRO();
dv=tmp.getData();
for (i=0; i<nx; i++) {
	memcpy(&dv[nx*nx*i],dp,nx*nx*sizeof(float));
}
slice->doneData();
tmp.doneData();

tmp.setRAlign(-slice->alt(),-slice->phi(),-slice->az());
tmp.setTAlign(0,0,0);
tmp.rotateAndTranslate();

add(&tmp);
delete slice;
}

void EMData::add(float d) {
int i;

if (isComplex()) ap2ri();

getData();
for (i=0; i<nx*ny*nz; i++) rdata[i]+=d;
doneData();
update();
}

void EMData::BilateralFilter(float Sigma1, float Sigma2, int ITER, int  LocalWidth){
// This is modified from code of Zeyun Yu  <zeyun@cs.utexas.edu>
// This routine is to do Bilateral Filtering on 3D volume data
// The input data type is FLOAT array OrgImg,  the storing order is as follows:
// The data for first slice is written first, then second slice, until the last slice.
// Within each slice the data for the first row is stored first, then the second row....
// And the filtered data will be stored in output FLOAT array NewImg 
// What each parameter means?
// width is the number of columns in each slice
// height is the number of rows   in each slice
// slicenum is the number of slices 
// Bilateral Filtering does Nonlinear weighted averaging filtering within a certain window,
// the window size is determined as (2*LocalWidth+1)^3
// So if you set LocalWidth as 1, you will do weighted average of 3*3*3 voxels around the 
// the voxel of interest
// Sigma1 means how large the voxel has impact on its neighbors in spatial domain.
// The larger the Sigma1, the more blurry the resulting image
// Sigma2 means how large the voxel has impact on its in range domain
// The larger the Sigma2, the more blurry the resulting image
// ITER means how many times you want apply this filtering method on your data
// Further questions or comments about this routine shall be directed to qwu@ticam.utexas.edu
// Original function interface
//void BilaFilter(float *OrgImg, float *NewImg, int width, int height,int slicenum,int ITER, float Sigma1, float Sigma2, int  LocalWidth)

if(LocalWidth<Sigma1) fprintf(stderr,"Warning: localwidth(=%d) should be larger than sigma1=(%f)\n",LocalWidth,Sigma1);
Sigma1*=Sigma1;

if(Sigma()>Sigma2) fprintf(stderr,"Warning: image sigma(=%f) should be smaller than sigma2=(%f)\n",Sigma(),Sigma2);
Sigma2*=Sigma2;

if(nz==1){	// 2D case
	int width=nx, height=ny; 

	int i,j,m,n;

	float tempfloat1,tempfloat2,tempfloat3;
	int   index1,index2,index;
	int   Iter;
	int   tempint1,tempint3;

	tempint1=width;
	tempint3=width+2*LocalWidth;

	float* mask=(float*)calloc((2*LocalWidth+1)*(2*LocalWidth+1),sizeof(float));
	float* OrgImg=(float*)calloc((2*LocalWidth+width)*(2*LocalWidth+height),sizeof(float));
	float* NewImg=getData();

	for(m=-(LocalWidth);m<=LocalWidth;m++)
		for(n=-(LocalWidth);n<=LocalWidth;n++) {
       	   index=(m+LocalWidth)*(2*LocalWidth+1)+(n+LocalWidth);
       	   mask[index]=exp((float)(-(m*m+n*n)/Sigma1/2.0));
  	}

	//printf("entering bilateral filtering process \n");

	Iter=0;
	while(Iter<ITER) {
		for(i=0;i<height;i++)
    		for(j=0;j<width;j++) {
	    		index1=(i+LocalWidth)*tempint3+(j+LocalWidth);
				index2=i*tempint1+j;
        		OrgImg[index1]=NewImg[index2];      
   		}

		// Mirror Padding
		for(i=0;i<height;i++){	
			for(j=0;j<LocalWidth;j++) OrgImg[(i+LocalWidth)*tempint3+(j)]=OrgImg[(i+LocalWidth)*tempint3+(2*LocalWidth-j)];
			for(j=0;j<LocalWidth;j++) OrgImg[(i+LocalWidth)*tempint3+(j+width+LocalWidth)]=OrgImg[(i+LocalWidth)*tempint3+(width+LocalWidth-j-2)];
   		}
		for(i=0;i<LocalWidth;i++){	
			for(j=0;j<(width+2*LocalWidth);j++) OrgImg[i*tempint3+j]=OrgImg[(2*LocalWidth-i)*tempint3+j];
			for(j=0;j<(width+2*LocalWidth);j++) OrgImg[(i+height+LocalWidth)*tempint3+j]=OrgImg[(height+LocalWidth-2-i)*tempint3+j];
		}

		//printf("finish mirror padding process \n");
		//now mirror padding have been done

		for(i=0;i<height;i++){
			//printf("now processing the %d th row \n",i);
			for(j=0;j<width;j++){
				tempfloat1=0.0; tempfloat2=0.0;
				for(m=-(LocalWidth);m<=LocalWidth;m++)
					for(n=-(LocalWidth);n<=LocalWidth;n++){
						index =(m+LocalWidth)*(2*LocalWidth+1)+(n+LocalWidth);
						index1=(i+LocalWidth)*tempint3+(j+LocalWidth);
						index2=(i+LocalWidth+m)*tempint3+(j+LocalWidth+n);
						tempfloat3=(OrgImg[index1]-OrgImg[index2])*(OrgImg[index1]-OrgImg[index2]);

						tempfloat3=mask[index]*(1.0/(1+tempfloat3/Sigma2));	// Lorentz kernel
						//tempfloat3=mask[index]*exp(tempfloat3/Sigma2/(-2.0));	// Guassian kernel
						tempfloat1+=tempfloat3;

						tempfloat2+=tempfloat3*OrgImg[(i+LocalWidth+m)*tempint3+(j+LocalWidth+n)];     
        			}
				NewImg[i*width+j]=tempfloat2/tempfloat1;
			}
		}
   		Iter++;
    }

    //printf("have finished %d  th iteration\n ",Iter);
	doneData();
	free(mask); 
	free(OrgImg);
	// end of BilaFilter routine
} else { // 3D case
	int width=nx, height=ny, slicenum=nz; 

	int i,j,k,m,n,p,q;

	float tempfloat1,tempfloat2,tempfloat3;
	int   index1,index2,index;
	int   Iter;
	int   tempint1,tempint2,tempint3,tempint4;

	tempint1=width;
	tempint2=width*height;
	tempint3=width+2*LocalWidth;
	tempint4=(width+2*LocalWidth)*(height+2*LocalWidth);

	float* mask=(float*)calloc((2*LocalWidth+1)*(2*LocalWidth+1)*(2*LocalWidth+1),sizeof(float));
	float* OrgImg=(float*)calloc((2*LocalWidth+width)*(2*LocalWidth+height)*(2*LocalWidth+slicenum),sizeof(float));
	float* NewImg=getData();

	for(p=-(LocalWidth);p<=LocalWidth;p++)
		for(m=-(LocalWidth);m<=LocalWidth;m++)
			for(n=-(LocalWidth);n<=LocalWidth;n++) {
        	   index=(p+LocalWidth)*(2*LocalWidth+1)*(2*LocalWidth+1)+(m+LocalWidth)*(2*LocalWidth+1)+(n+LocalWidth);
        	   mask[index]=exp((float)(-(m*m+n*n+p*p)/Sigma1/2.0));
        	}

	//printf("entering bilateral filtering process \n");

	Iter=0;
	while(Iter<ITER) {
	   for(k=0;k<slicenum;k++)
    	  for(i=0;i<height;i++)
        	 for(j=0;j<width;j++) {
	    		index1=(k+LocalWidth)*tempint4+(i+LocalWidth)*tempint3+(j+LocalWidth);
				index2=k*tempint2+i*tempint1+j;

	        	OrgImg[index1]=NewImg[index2];      
        	}

		// Mirror Padding
		for(k=0;k<slicenum;k++){	
			for(i=0;i<height;i++){	
				for(j=0;j<LocalWidth;j++) OrgImg[(k+LocalWidth)*tempint4+(i+LocalWidth)*tempint3+(j)]=OrgImg[(k+LocalWidth)*tempint4+(i+LocalWidth)*tempint3+(2*LocalWidth-j)];
				for(j=0;j<LocalWidth;j++) OrgImg[(k+LocalWidth)*tempint4+(i+LocalWidth)*tempint3+(j+width+LocalWidth)]=OrgImg[(k+LocalWidth)*tempint4+(i+LocalWidth)*tempint3+(width+LocalWidth-j-2)];
	    	}

			for(i=0;i<LocalWidth;i++){	
				for(j=0;j<(width+2*LocalWidth);j++) OrgImg[(k+LocalWidth)*tempint4+i*tempint3+j]=OrgImg[(k+LocalWidth)*tempint4+(2*LocalWidth-i)*tempint3+j];
				for(j=0;j<(width+2*LocalWidth);j++) OrgImg[(k+LocalWidth)*tempint4+(i+height+LocalWidth)*tempint3+j]=OrgImg[(k+LocalWidth)*tempint4+(height+LocalWidth-2-i)*tempint3+j];
			}
		}// end of k

		for(k=0;k<LocalWidth;k++){
			for(i=0;i<height+2*LocalWidth;i++)
				for(j=0;j<width+2*LocalWidth;j++){
					index1=k*tempint4+i*tempint3+j;
					index2=(2*LocalWidth-k)*tempint4+i*tempint3+j;
					OrgImg[index1]=OrgImg[index2];
				}
			for(i=0;i<height+2*LocalWidth;i++)	
				for(j=0;j<width+2*LocalWidth;j++){
		    		index1=(k+slicenum+LocalWidth)*tempint4+i*tempint3+j;
					index2=(slicenum+LocalWidth-2-k)*tempint4+i*tempint3+j;
					OrgImg[index1]=OrgImg[index2];
            	}
    	  }// end of loop k

		//printf("finish mirror padding process \n");
		//now mirror padding have been done

		for(k=0;k<slicenum;k++){
			//printf("now processing the %d th slice \n",k);
			for(i=0;i<height;i++)
				for(j=0;j<width;j++){
					tempfloat1=0.0; tempfloat2=0.0;
					for(p=-(LocalWidth);p<=LocalWidth;p++)
						for(m=-(LocalWidth);m<=LocalWidth;m++)
							for(n=-(LocalWidth);n<=LocalWidth;n++){
								index=(p+LocalWidth)*(2*LocalWidth+1)*(2*LocalWidth+1)+(m+LocalWidth)*(2*LocalWidth+1)+(n+LocalWidth);
								index1=(k+LocalWidth)*tempint4+(i+LocalWidth)*tempint3+(j+LocalWidth);
								index2=(k+LocalWidth+p)*tempint4+(i+LocalWidth+m)*tempint3+(j+LocalWidth+n);
								tempfloat3=(OrgImg[index1]-OrgImg[index2])*(OrgImg[index1]-OrgImg[index2]);

								tempfloat3=mask[index]*(1.0/(1+tempfloat3/Sigma2));	// Lorentz kernel
								//tempfloat3=mask[index]*exp(tempfloat3/Sigma2/(-2.0));	// Guassian kernel
								tempfloat1+=tempfloat3;

								tempfloat2+=tempfloat3*OrgImg[(k+LocalWidth+p)*tempint4+(i+LocalWidth+m)*tempint3+(j+LocalWidth+n)];     
        					}
					NewImg[k*height*width+i*width+j]=tempfloat2/tempfloat1;
				}
    	}

    	Iter++;
    	//printf("have finished %d  th iteration\n ",Iter);
	}  // Iteration loop is over
	doneData();
	free(mask); 
	free(OrgImg);
	// end of BilaFilter routine
}
update();
}


// apply CTF curve to the REAL image
void EMData::applyCTF(float* ctfcurve) {
if(!ctfcurve) {
	ctfcurve = ctfCurve();
	if(!ctfcurve)error("The image does not have CTF parameters set\n",ERR_EXIT);
}

edgeNormalize();

EMData *d2;
if(zSize()==1) d2=clip(-xSize()/2,-ySize()/2,xSize()*2,ySize()*2);
else d2=clip(-xSize()/2,-ySize()/2,-zSize()/2,xSize()*2,ySize()*2,zSize()*2);

EMData *d3=d2->doFFT();
d2->gimmeFFT();
delete d2;
d3->applyRadFn(CTFOS*ySize()/2,0,2.0/CTFOS,ctfcurve,0);
d2=d3->doIFT();
d3->gimmeFFT();
delete d3;

EMData *d;
if(zSize()==1) d=d2->clip(d2->xSize()/4,d2->ySize()/4,d2->xSize()/2,d2->ySize()/2);
else d=d2->clip(d2->xSize()/4,d2->ySize()/4,d2->zSize()/4,d2->xSize()/2,d2->ySize()/2,d2->zSize()/2);

delete d2;
d->setParent(NULL);
if(!ctfcurve) free(ctfcurve);

// now swap the data pointers for the result and the original, this
// is dangerous, but a lot faster than doing a copy  :^)
float *myd=getData();
d->getData();
rdata=d->rdata;
d->rdata=myd;
d->doneData();
delete d;
doneData();
update();
}
